// $Author: jmartin $
// $Date: 2000-02-05 02:18:34 $
// $Source: /root/cvs/Rigi/rigiedit/UT/src/pmdb.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>     // for varargs stuff

#include "sexps.h"	// for reading and writing s-expressions
#include "pmdb.h"
#include "avlnew.h"	// data structure for PMDB graphs (webs)

extern void Error( char* ... );
extern int debug;
extern int iflag;		// for building inverse "in" and "isa" links

static AvlSet srcnames = NULL;
static AvlSet nodetypes = NULL;

static AvlSet relntypes = NULL;

static AvlSet srcweb = NULL;	// "Give me object <blah>"
static AvlSet arcweb = NULL;	// "Give me the call graph"
static AvlSet dstweb = NULL;	// "What nodes are instanceof (in) Module?"

static FILE *avlfile = NULL;	// For writing out set contents

static int nsrcs = 0;
static int nrelntypes = 0;
static int nnodetypes = 0;

static const int ARCSIZE = 128;
static const int SRCSIZE = 256;
static const int DSTSIZE = 1024;

static char arcbuff[ARCSIZE], srcbuff[SRCSIZE], dstbuff[DSTSIZE];

static int fscans( FILE *fp, char *buff, int size );
static void DumpSet( AvlSet, FILE *fp );
static void PrintSet( AvlSet, FILE *fp );
static int FreeWeb( AvlSet* );

int LoadRSF( const char *fn )
{
    // open file
    FILE *fp = fn ? fopen( fn,"r" ) : stdin;
    if (fp==NULL) { Error("couldn't open RSF file %s",fn ); return 0; }

    AvlSet *x;

    nsrcs = 0;
    int j;
    for (j=0;;j++ )
    {
	// Read in a tuple

	if (0==fscans( fp, arcbuff, ARCSIZE )) break;
	if (0==fscans( fp, srcbuff, SRCSIZE )) break;
        fscans( fp, dstbuff, DSTSIZE );

	if ( NULL==AvlSetInfo( &relntypes, arcbuff, INVALID ) ) nrelntypes++;
	if ( NULL==AvlSetInfo( &srcnames, srcbuff, INVALID ) ) nsrcs++;

	// Convert RSF domain info to pmdb-style info

	if (!strcmp(arcbuff,"type")) strcpy(arcbuff,"@in");
	else if (!strcmp(arcbuff,"isa")) strcpy(arcbuff,"@isa");

	// Insert in our data structures: sets within sets within sets

	x = &srcweb;

	x = (AvlSet*)AvlSetInfo( x, srcbuff, NULL );
	x = (AvlSet*)AvlSetInfo( x, arcbuff, NULL );
	x = (AvlSet*)AvlSetInfo( x, dstbuff, NULL );

	// Now, put in inverse "in" and "isa" links

	if (!iflag) continue;
	if (!strcmp(arcbuff,"@in")) strcpy(arcbuff,"@in-1");
	else if (!strcmp(arcbuff,"@isa")) strcpy(arcbuff,"@isa-1");
	else continue;

	x = &srcweb;

        x = (AvlSet*)AvlSetInfo( x, dstbuff, NULL );
        x = (AvlSet*)AvlSetInfo( x, arcbuff, NULL );
        x = (AvlSet*)AvlSetInfo( x, srcbuff, NULL );

    }
    Error( "file %s, %d RSF tuples, %d distinct srcnames read",
	fn?fn:"<stdin>",j,nsrcs );

    int mem, charmem, nrecs;
    AvlMem( &mem, &charmem, &nrecs );
    Error( "AVL memory usage: mem=%d bytes, charmem=%d, nrecs=%d\n",
	mem,charmem,nrecs );

    //i = FreeSet( &srcnames );
    //Error("%d freed from srcnames",i);

    //i = FreeSet( &relntypes );
    //Error("%d freed from relntypes",i);

    if ( fp && fp!=stdin ) fclose( fp );

    //Error("Here be srcweb (from parsing RSF):");
    //DumpSet( srcweb,stdout );

    return 1;
}



    static AvlSet *srcw = &srcweb;	// main set
    static AvlSet *arcw = NULL;		// set in a set
    static AvlSet *dstw = NULL;		// set in a set in a set

    static AvlSet attrs = NULL;
    static AvlName lastdstname = NULL;
    static const char *lastname;

    static int attrs_callback( AvlName s )
    {
	dstw = (AvlSet*)AvlSetInfo( arcw, s, NULL );
	AvlSetInfo( dstw, lastdstname, NULL );
	lastname = AvlLastName();
	return 0;
    }

    static void sexp_parse_callback( char *s, int d, int i, int top )
    {
	// With this callback, we be writing stuff to the PMDB
	// For now, we are only concerned with the srcweb.

	// Affected statics:
	//	srcweb, srcnames, nodetypes, relntypes
	//	nsrcs, nnodetypes, nrelntypes

	static int narcs = 0;
	static char builtin_attr[8]; // "in" or "isa", and inverses

	switch (d)
	{
	case 0:

	    if (i==0)
	    {
	        // A src object has been read
	        if ( NULL==AvlSetInfo( &srcnames, s, INVALID ) ) nsrcs++;
		if (iflag) strcpy( srcbuff, s );

	        arcw = (AvlSet*)AvlSetInfo( srcw, s, NULL );
	    }
	    else
	    {
		// A nodetype name (Token, SClass, ... ) has been read
        	//if ( NULL==AvlSetInfo( &nodetypes, s, INVALID ) )
		    //nnodetypes++;

		//dstw = (AvlSet*)AvlSetInfo( arcw, "@in", NULL );
		//(AvlSet*)AvlSetInfo( dstw, s, NULL );
	    }
	    break;

	case 1:

	    // either the "in" list or the "isa" list
	    strcpy( builtin_attr, (top==2)?"@in":"@isa" );
	    
	    if (top==2)
	    if ( NULL==AvlSetInfo( &nodetypes, s, INVALID ) ) nnodetypes++;

	    dstw = (AvlSet*)AvlSetInfo( arcw, builtin_attr, NULL );
	    (AvlSet*)AvlSetInfo( dstw, s, NULL );

	    // Put in the inverse relation, if desired
	    if (iflag)
	    {
		strcat( builtin_attr, "-1" );
        	AvlSet *x = &srcweb;
        	x = (AvlSet*)AvlSetInfo( x, s, NULL );
        	x = (AvlSet*)AvlSetInfo( x, builtin_attr, NULL );
        	AvlSetInfo( x, srcbuff, NULL ); 
	    }
	    break;

	case 3:

	    // An arctype name has been read
	    if (NULL==AvlSetInfo( &relntypes, s, INVALID ))
		nrelntypes++;

	    // In case of multiple attribute categories, we must build
	    // them up in a set
	    if (i==0) FreeSet( &attrs );
	  AvlSetInfo( &attrs, s, NULL );

	  //dstw = (AvlSet*)AvlSetInfo( arcw, s, NULL );
	    break;

	case 4:
	case 5: // Note: d==5 implies a HACK syntax for attrs in a Sequence

	    if (i>1)
	    {
		Error("syntax error, string is %s",s );
		break;
	    }

	    if (i==1)
	    {
		// This is a Class-level syntax variant (not Token, we hope!)
		dstw = (AvlSet*)AvlSetInfo( arcw, lastname, NULL );
		AvlSetInfo( dstw, s, NULL );
		break;
	    }

	    // A dstname has been read. Now we retrieve those attr categories
	    lastdstname = s;
	    AvlIterate( attrs, attrs_callback );
/*
*/

	  //(AvlSet*)AvlSetInfo( dstw, s, NULL );

	    // Keep the name s, in case there is a class-level syntax
	    // variant coming up immediately
	  //lastname = AvlLastName();

	    narcs++;
	    if (debug) if (narcs%1000==0)
		printf("%d RSF tuples\n",narcs);
	    break;

	default:
	    ;
	}
    }


int LoadSexp( const char *fn )
{
    // open file
    FILE *fp = fn ? fopen( fn,"r" ) : stdin;
    if (fp==NULL) { Error("couldn't open s_exp file \"%s\"",fn ); return 0; }

    nsrcs = 0;
    parse_sexps( fp, sexp_parse_callback );
    fclose(fp);

    if (debug)
    {
	printf("# nodetypes\n");
	PrintSet( nodetypes,stdout );
    }
    int i = FreeSet( &relntypes );

    Error( "file %s, %d relation types read", fn?fn:"<stdin>", i );
    Error( "%d distinct object%s read from %s",nsrcs,nsrcs==1?"":"s",fn );

    int mem, charmem, nrecs;
    AvlMem( &mem, &charmem, &nrecs );
    Error( "AVL memory usage: mem=%d bytes, charmem=%d, nrecs=%d\n",
	mem,charmem,nrecs );

    //printf("Here be srcweb (from parsing sexps):\n");
    //DumpSet( srcweb,stdout );

    return 1;
}

    static int save_callback( const char *s )
    {
        const char *p = MakeSexp(s);
        fputs(p,avlfile);
        fputs("\n",avlfile);
        return 0;
    }

int SaveSexp( const char *fn )
{
    // open file
    FILE *fp = fn ? fopen( fn,"w" ) : stdout;
    if (fp==NULL) { Error("couldn't open s_exp file \"%s\"",fn ); return 0; }

    // For each name in the srcweb, we write out its sexp
    avlfile = fp;
    ForAllObjects( save_callback );

    //DumpSet( srcweb, fp );
    if (fp!=stdout) fclose(fp);
    return 1;
}

void DumpNames() { PrintDictionary(); }

void EmptyPMDB()
{
    // Must free all localglobals associated with the "active" DB
    int i = FreeSet( &srcnames );
    int j = FreeSet( &nodetypes );
    int k = FreeSet( &relntypes );

    nsrcs = nnodetypes = nrelntypes = 0;

    int a = FreeWeb( &srcweb );
    FreeWeb( &arcweb );
    FreeWeb( &dstweb );

    Error("%d objects, %d srcnames, %d nodetypes, %d relntypes freed",
	a, i,j,k);
}

const void ParseSexp( const char *s ) { parse_sexp( s, sexp_parse_callback ); }


    static int write_callback( AvlName s, AvlInfo )
    {
	write_sexp(s);
	write_sexp(" ");
	return 0;
    }

    static int dst_its;
    static int tml_flag;

    static int dst_callback( AvlName s, AvlInfo )
    {
        // Make nice indentation, if NOT to be written to a TML message
	if (!(dst_its%4)) if (!tml_flag) write_sexp("\n   ");

	write_sexp(" (");
	write_sexp( (char*)s );
	write_sexp(")");

	dst_its++;
	return 0;
    }

    static int arc_callback( AvlName s, AvlInfo dst_set )
    {
	// Nothing to print out if no dst values
	if (!dst_set) return 0;

	// Forget about "in" and "isa"; already put in inlist
	if (!strcmp(s,"@in")) return 0;
	if (!strcmp(s,"@isa")) return 0;

	// Forget about inverse "in" and "isa"; only used internally
	if (!strcmp(s,"@in-1")) return 0;
	if (!strcmp(s,"@isa-1")) return 0;

        // Make nice indentation, if NOT to be written to a TML message
	if (!tml_flag) write_sexp("\n  ");

	write_sexp(" ( (");
	write_sexp( (char*)s );
	write_sexp(") (");

	dst_its = 0;
	AvlIterate( (AvlSet)dst_set, dst_callback );

	write_sexp("))");
	return 0;
    }


// Convert info inside the PMDB data structures to an s-expr,
// conforming to the grammar for NetworkObjects

const char *MakeSexp( const char *s, int flag )
{
    //printf("Making a sexp for %s ...\n",s);

    if (srcweb==NULL)
    {
	Error("error, no active DB!");
	return NULL;
    }
    tml_flag = flag;
    AvlSet *arcset, *typeset, *isa_set;

    arcset = (AvlSet*)AvlSetInfo( &srcweb, s, PROBEONLY );
    if (!arcset) return NULL;

    // Must extract type(s) of this object
    typeset = (AvlSet*)AvlSetInfo( arcset, "@in", PROBEONLY );
    isa_set = (AvlSet*)AvlSetInfo( arcset, "@isa", PROBEONLY );

    reset_sexp();
    write_sexp("( ");
    write_sexp(s);
    write_sexp(" Token (");

    if (typeset) AvlIterate( *typeset, write_callback );
    else write_sexp("unknown");

    write_sexp(") (");
    if (isa_set) AvlIterate( *isa_set, write_callback );
    write_sexp(") (");

    AvlIterate( *arcset, arc_callback );

    write_sexp(" ) )");

    return get_sexp();
}

// **************************************************************


static int fscans( FILE *fp, char *buff, int size )
{
    // Load buff with next string in the file, skipping comments
    register int c,r,i,x,y;

    for ( i=0;;)
    {
	c = getc(fp);
	if (c=='#') i = 1; else if (c=='\n') i = 0;
	else if (isspace(c)) continue; else if (i==0) break;
    }

    if (c=='"') { x=1; c=getc(fp); } else x=0;
    r = size-1;
    for ( i=y=0; ; )
    {
	if (c==EOF) break;
	if (x)
	{
	    if (!y) { if (c=='"') break; }
	    y = c=='\\';
	}
	else if (isspace(c)) break;
	if (i<r) buff[i++] = c;
	c = getc(fp);
    }
    buff[i] = '\0';
    if (i==r) Error("fscans(), buffer overflow, contents are\n%s",buff);
    return i;
}


    static int avllevel = 0;
 
    static int avl_print_callback( AvlName s,AvlInfo a )
    {
	int d = avllevel;

	if (d==0) // src name 
	{
		fputs("\n( ",avlfile);
		fputs(s,avlfile);
		fputs(" Token () () (\n",avlfile);
	}
	if (d==1) // arc name
	{
		fputs("  ( ( ",avlfile);
		fputs(s,avlfile);
		fputs(" ) ( ",avlfile);
	}
	if (d==2) // dst name
	{
		fputs("(",avlfile);
		fputs(s,avlfile);
		fputs(")",avlfile);
	}

	if (a) // recurse
	{
	    avllevel++;
	    AvlIterate( (AvlSet)a, avl_print_callback );
	    avllevel--;
	}
	if (d==0) fputs("))\n",avlfile);
	if (d==1) fputs("))\n",avlfile);
	if (d==2);

	return 0;
    }

    static int print( AvlName s, AvlInfo )
    {
	fputs(s,avlfile);
	//if (i!=0 && i!=INVALID)
	    //{ fputs( "\t", avlfile); fputs( (char*)i,avlfile ); }
	fputs("\n",avlfile);
	return 0;
    }

    static void free_callback( AvlName, AvlInfo *i )
    {
	if (i) FreeSet( (AvlSet*)i, free_callback );
    }

// Recursively free child sets

static int FreeWeb( AvlSet *web ) { return FreeSet( web, free_callback ); }

static void DumpSet( AvlSet s, FILE *fp )
{
    avlfile = fp;
    avllevel = 0;
    AvlIterate( s, avl_print_callback );
    fputs("\n",fp);
}

static void PrintSet( AvlSet s, FILE *fp )
{
    avlfile = fp;
    avllevel = 0;
    AvlIterate( s,print );
    fputs("\n",fp);
}

void ForAllObjects( int (*f)(const char*) )
{
    // Call f on the name of each object in the DB
    AvlIterate( srcweb, f );
}

void ForEachObject( const char *n, const char *a, int (*f)(const char*) )
{
    AvlSet *arcset, *dstset;

    arcset = (AvlSet*)AvlSetInfo( &srcweb, n, PROBEONLY );
    if (!arcset) return;

    dstset = (AvlSet*)AvlSetInfo( arcset, a, PROBEONLY );
    if (!dstset) return;
    if (!*dstset) Error("error in ForEachObject()");

    AvlIterate( *dstset, f );
}

// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
// Revision 1.1  1997/12/05 01:57:18  jmartin
// Initial revision
//
// Revision 1.8  1996/10/19 19:03:46  jmartin
// Changes for Linux Port
//
//   Revision 1.7  1995/10/29  19:00:57  mwhitney
//   make inverse in and isa links
//
//   Revision 1.6  1995/10/27  15:54:10  mwhitney
//   many changes
//
//   Revision 1.5  1995/05/10  22:37:19  mwhitney
//   minor
//
//   Revision 1.4  1995/02/15  19:10:50  mwhitney
//   extra parm for parse callback function
//    .
//
//   Revision 1.3  1994/10/28  16:56:06  mwhitney
//   many many changes! Panic for CASCON94.
//
//   Revision 1.2  1994/08/29  23:13:03  mwhitney
//   Many changes
//
//   Revision 1.1  1994/08/17  20:48:07  mwhitney
//   Initial revision
//
//
