// $Author: jmartin $
// $Date: 2000-02-05 02:18:35 $
// $Source: /root/cvs/Rigi/rigiedit/UT/src/tds.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------


// EXAMPLE Telos Data Server accepting messages sent specifically to
// the client or to the tool class TELOS_DATA_SERVER.

// This file was originally TDS.C, supplied by UofT as an example
// Telos Data Server. It is now to be connected to the pseudo-repository
// PMDB. Similar Rigi-side code resides in ut.C.

#include <stdlib.h>
#include <unistd.h>	// MJW: for usleep()

#include "tml.H"	// the TML library
#include "pmdb.h"	// the PMDB stuff

static const int USLEEPINTERVAL = 10000;
static int MAXOBJECTS = 40;

static void ProcessAsk( Request* );
static void ProcessTell( Request* );
static void ProcessCommand( Request* );

static void SendResponse( Request*, const char *err );

static int AddObject( const char* );
static void FlushResponse( int lastflag );

void Error( char *format ... );

static char *progname;
static char *server = NULL;
static char *port = NULL;

static Request *current_request;
static Response *current_response = NULL;


int debug = 0;
int iflag = 0;		// build inverse links for "in" and "isa"
int nflag = 0;		// do not connect to server
int oflag = 0;		// write the internal sets to stdout
int rflag = 0;		// read RSF, not NetworkObject s-exprs
int pflag = 0;		// port
int sflag = 0;		// server host
int wflag = 0;		// print out the AVL word dictionary

void PrintOptions()
{
    printf( "\
	d	debug\n\
	i	build inverse in and isa links\n\
	h	help\n\
	n	no server\n\
	o	write sets to stdout\n\
	p <n>	port <num>\n\
	r	read RSF, not s-exprs\n\
	s <n>	server <name>\n\
	t	test (no storage except main sets; debug)\n\
	w	print out the set of words (symbols)\n\
" );
}

int main( int argc, char *argv[] )
{
    progname = *argv;

    // process arguments

    while (--argc)

        if (**++argv=='-') while (*++*argv) switch (**argv)
	{
            case 'd':	debug = 1;	break;
            case 'i':	iflag = 1;	break;
	    case 'h':	PrintOptions();	exit(0);
            case 'n':	nflag = 1;	break;
            case 'o':	oflag = 1;	break;
            case 'p':	pflag = 1;	break;
            case 'r':	rflag = 1;	break;
            case 's':	sflag = 1;	break;
            case 'w':	wflag = 1;	break;
	    default :	Error( "error, bad option -%c",**argv );
        }
        else
	{
	    if (sflag) { server = *argv; sflag = 0; }
	    else if (pflag) { port = *argv; pflag = 0; }
	    else
		if (rflag) LoadRSF(*argv); else LoadSexp(*argv);
	}

    if (wflag) DumpNames();

    if (oflag) SaveSexp(NULL);
    if (nflag) exit(0);


    // Here's all the TML stuff, largely from an old example client.C program

    TMLinit();

    // set the current client toolClass to the Telos data server
    CLIENT->setToolClass(TELOS_DATA_SERVER);

    server = server ? server : getenv( "RIGIDBHOST" );
    if (server && *server)
    {
	Error( "setting SERVER host id to \"%s\"", server );
	String str(server);
	SERVER->setHostId( &str );
    }
    port = port ? port : getenv( "RIGIDBPORT" );
    if (port && *port)
    {
	Error( "setting SERVER port number to %s",port );
	SERVER->setPortNumber( atoi(port) );
    }

    Command *C; Boolean r;

    C = new Command(MS_OPEN);
    r = C->sendCommand();
    delete C;
    if (!r)
    {
	String *str = SERVER->getHostId();
	Error("error, can't connect to server on host %s at port %d",
		str->StringToChar(),SERVER->getPortNumber());
	exit(1);
    }

    // accept message intended for the tool TELOS_DATA_SERVER

    CLIENT->addDomain(TELOS_DATA_SERVER);

    // This command accepts the userIds, ToolClasses, or Contexts (listed in
    // the client's DomainList) as the client's domain of interest.
    // By default, domain of interest is the user id of the client.
    // If this command is not sent, then the client cannot not
    // receive any messages.

    C = new Command(MS_ACCEPT_DOMAINS);
    r = C->sendCommand();
    delete C;
    if (!r)
    {
        Error("error, can't command server to accept domains" );
        exit(1);
    }

    CLIENT->print();
    SERVER->print();

    // Now we just sit around in a loop, waiting to receive messages
    // from various tools, and forward responses to these messages to
    // appropriate tools (so far, we only receive messages of class REQUEST,
    // and they may be either "ASK" or "TELL")

    for ( int i=0; ; i++ )
    {
#ifdef __EMX__
	_sleep2(USLEEPINTERVAL/1000);
#else
	usleep(USLEEPINTERVAL);
#endif

	NewMessageStatus status = TMBgetNewMessage();

	if ( status != NO_NEW_MESSAGE && status != NEW_REQUEST )
	{
	    Error( "error, unexpected message %d received", status );
	    return 1;
	}

	switch (status)
	{
	    case NO_NEW_MESSAGE:
		break;
	    case NEW_REQUEST:
	    {
		// A request arrived and can be accessed via REQUEST_QUEUE

		Request* request = REQUEST_QUEUE->front();
	        REQUEST_QUEUE->leave();

		tml_Method *method = request->getMethod();

		if ( (*method)=="ASK" )
		    ProcessAsk( request );
		else if ( (*method)=="TELL" )
		    ProcessTell( request );
		else if ( (*method)=="COMMAND" )
		    ProcessCommand( request );
		else
		    Error("error, unrecognized method??");
	        delete request;
	        break;
	     }
	 case UNRECOGNIZABLE_MESSAGE:
	 case NEW_RESPONSE:
	 case NEW_MONOLOGUE:
	 default:
	      {
		Error("error, bad return code (status is %d)",status);
		break;
	      }
	}

    }
    return 0;
}

static void ProcessAsk( Request *request )
{
    // First, find out what info we be requesting. It should
    // be an ASK request.

    ArgumentList* al = request->getArgumentList();
    ArgumentString *as;
    char *attrname = NULL;
    int i, flag=0;

    current_request = request;

    for ( i=0; as = al->getArgument(i); i++ )
    {
	const char *s = as->StringToChar();

	// These are commands unique to pmdb server
	if (!strcmp(s,"@all"))
	    ForAllObjects( AddObject );

	// These ask commands are available in the telos server

	else if (!strcmp(s,"getObject"))	flag = 1;

	else if (!strcmp(s,"getInstances"))	attrname = "@in-1";
	//else if (!strcmp(s,"getAllInstances"))

	else if (!strcmp(s,"getAncestors"))	attrname = "@in";
	//else if (!strcmp(s,"getAllAncestors"))

	else if (!strcmp(s,"getChildren"))	attrname = "@isa-1";
	//else if (!strcmp(s,"getAllChildren"))

	else if (!strcmp(s,"getParents"))	attrname = "@isa";
	//else if (!strcmp(s,"getAllParents"))

	else if (flag==1)

	    AddObject(s);
	
	else if (attrname)

	    ForEachObject( s, attrname, AddObject );

	else
	{
	    Error( "error, can't process ASK for \"%s\"",s );
	    break;
	}
    }
    Error("received ASK request with %d argument%s",i,i==1?"":"s");
    FlushResponse( 1 );
}

static void ProcessTell( Request *request )
{
    // Here, the server processes a TELL Request, sent from a
    // client (eg. Rigi) to the database server (which be PMDB).
    // A Response is expected.

    ArgumentList* al = request->getArgumentList();
    ArgumentString *as;

    static int j=0; int i;
    j++;

    for ( i=0; as = al->getArgument(i); i++ )
    {
        const char *s = as->StringToChar();

        if(debug){printf("\nParsing %s\n",s);fflush(stdout);}
        ParseSexp(s);
    }
    Error("received TELL request %d with %d argument%s",j,i,i==1?"":"s");

    // Be sending a confirming Response message
    SendResponse( request, NULL );
}

static void ProcessCommand( Request *request )
{
    // Process a COMMAND Request, send a Response

    ArgumentString *as;
    int i=0, k, r=1;
    static int j = 0;

    j++;
    ArgumentList* al = request->getArgumentList();
    k = al->getSize();
    Error("received COMMAND request %d with %d argument%s", j,k,k==1?"":"s");

    for ( k=0; as = al->getArgument(k); k++ )
    {
        const char *s = as->StringToChar();

	if (!strcmp(s,"load")) i = 1;
	else if (!strcmp(s,"save")) i = 2;
	else if (!strcmp(s,"max")) i = 3;
	else if (!strcmp(s,"debug"))
	{
            debug = !debug;
            Error("debug flag set %s",debug?"on":"off");
	}
	else if (!strcmp(s,"empty")) EmptyPMDB();

	else if (i==1) { r=LoadSexp(s); i=0; }
	else if (i==2) { r=SaveSexp(s); i=0; }
	else if (i==3)
	{
            int m = MAXOBJECTS; MAXOBJECTS = atoi(s);
            Error("objects-per-message set to %d from %d",MAXOBJECTS,m);
	    i=0;
	}
	else
	{
	    Error("unrecognizable command \"%s\"",s);
	    r=0;
	}
    }
    if (k!=al->getSize()) Error("error, problem in ProcessCommand()");

    // Special case, for writing DB to stdout
    if ( i==2 ) r=SaveSexp( NULL );

    // Be sending a confirming Response message
    char *errormessage = NULL;
    if (r==0) errormessage = "couldn't process COMMAND";
    SendResponse( request, errormessage );
}

static int AddObject( const char *s )
{
    // For the given key string, compile its s-exp into the
    // current response message. If the message is "filled up",
    // send it.

    static int i = 0;
    int tml_flag = 1;

    // Make an s-expression suitable for a TML message
    const char *p = MakeSexp( s,tml_flag );

    if (!p)
    {
	Error("error, couldn't make an s-expression for %s!",s);
	return 0;
    }

    if (!current_response)
    {
	current_response = new Response( current_request );
	i = 0;
    }
    i++;

    ReturnValueString *rv;

    rv = new ReturnValueString(p);
    current_response->addReturnValue(rv);
    delete rv;

    if ( i==MAXOBJECTS ) FlushResponse( 0 );
    return 0;
}

static void FlushResponse( int lastflag )
{
    static int j = 0;

    if (!current_response)
    {
	if (!lastflag) return;
	current_response = new Response( current_request );
    }
    if (!lastflag)
    {
	// Indicate to client that more info is coming
	ReturnValueString *rv = new ReturnValueString("@continue");
	current_response->addReturnValue(rv);
	delete rv;
    }
    SendResponse( NULL,NULL );
}

static void SendResponse( Request *request, const char *errormessage )
{
    static int count=0;
    count++;

    if (!current_response) current_response = new Response( request );

    if (errormessage)
    {
	ErrorMessageString *e = new ErrorMessageString(errormessage);
        current_response->addErrorMessage(e);
        current_response->setCompletionStatus( ERROR );
    }
    else current_response->setCompletionStatus( SUCCESS );

    int n = current_response->getReturnValues()->getSize();
    int i = current_response->getErrorMessages()->getSize();
    Error("sent %sresponse #%d with %d value%s and %d errormessage%s",
	errormessage?"ERROR ":"", count, n,n==1?"":"s", i,i==1?"":"s");

    Boolean r = current_response->sendResponse();
    if (!r) Error("error, couldn't send response!");
    delete current_response;
    current_response = NULL;
}

void Error( char *format ... )
{
    va_list ap;
    va_start( ap, format );
    fprintf( stderr,"%s: ",progname );
    if (format) vfprintf( stderr, format, ap );
    fprintf( stderr,"\n" );
    va_end(ap);
}

// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
// Revision 1.2  1999/06/07 17:35:03  jmartin
// changes for compilation under OS/2
//
// Revision 1.1  1997/12/05 01:57:18  jmartin
// Initial revision
//
//   Revision 1.6  1995/10/29  20:36:23  mwhitney
//   changes to allow COMMAND Requests
//
//   Revision 1.5  1995/10/27  17:36:09  mwhitney
//   many changes
//
//   Revision 1.4  1995/01/12  23:23:45  mwhitney
//   many changes
//
//   Revision 1.3  1994/10/26  21:14:17  mwhitney
//   many many changes! Panic for CASCON94.
//
//   Revision 1.2  1994/09/12  22:16:33  mwhitney
//   Many changes
//
//   Revision 1.1  1994/08/17  20:48:07  mwhitney
//   Initial revision
//
//
