#ifndef ERROR_H
#define ERROR_H

#include <iostream.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
extern "C" {
#include "api.h"
}

#include "GenericList.h"
#include "String.H"

// ErrorMessageString is used to return an error message if request is
// not completed successfully.

class ErrorMessageString : public Generic, public  String 
{
  public:
    ErrorMessageString(String s)       : String(s) {};
    ErrorMessageString(const char* s)  : String(s) {};
    ErrorMessageString()               : String()  {};
};

// ErrorMessageList is used to return error messages if a request is
// not completed successfully.

class ErrorMessageList : public GenericList
{
public:
  ErrorMessageList(): GenericList() {};
  ErrorMessageList(t_sexp sexp): GenericList() {
      int n = MBlength(sexp);
      ErrorMessageString* arg;
      char* tmp;
//      ErrorMessageString* errMsgStr = new ErrorMessageString;
      for (int i = 0; i < n; i++) { 
	  tmp = MBprint_Cstring(MBnth(sexp, i));
          arg = new ErrorMessageString(tmp);
//	  *errMsgStr += tmp;
	  (*MBfree_func)(tmp);
              addItem(arg);
      }
//      addItem(errMsgStr);
  };

    // read only routines
    ErrorMessageString* getErrorMessage(int i) 
                            { return (ErrorMessageString*)getItem(i); };
    int findErrorMessage(ErrorMessageString* s) { return findItem(s); };

    void print (ostream& outStream = cout)
      { outStream << "(";
        int n = getSize();
        if (n > 0)
	  { for (int i = 0; i < n - 1; i++)
                outStream << getErrorMessage(i) << ", ";
            outStream << getErrorMessage(n-1);
	   };
        outStream << ")";
          
      };

    // write access routines
    void addErrorMessage(ErrorMessageString* s)    { addItem(s);      };
    void removeErrorMessage(ErrorMessageString* s) { removeItem(s);   };
    void clearErrorMessageList()                   { clear();         };
    ErrorMessageString* removeErrorMessage(int i) 
                       { return (ErrorMessageString*)removeItem(i);   };
};


#define TMB_NULL                  300
#define TMB_CannotConnect         301
#define TMB_CannotDisconnect      302
#define TMB_CannotClear           303
#define TMB_CannotRegisterId      304
#define TMB_CannotAccept          305
#define TMB_CannotSend            306
#define TMB_Disconnected          307
#define TMB_UnrecognizableMessage 308
class ErrorType
{  
    int type;
    String* errorBuffer;
    String* messageBuffer;
    Boolean errorFound;

  public:
    ErrorType()
      { type = TMB_NULL;
        errorBuffer = new String();
        messageBuffer = new String();
        errorFound = FALSE;
      }

    ~ErrorType()
      { delete(errorBuffer);
        delete(messageBuffer);
      }

    Boolean found()                { return errorFound; };
    int getErrorType()             { return type;       };
    String* getErrorBuffer()       { return errorBuffer;     };
    String* getMessageBuffer()     { return messageBuffer;    };

    void clearErrorMessage() 
      { type = TMB_NULL;
        errorFound = FALSE;
        *errorBuffer = "";
        *messageBuffer = "";
      };

    String* generateStandardErrorBuffer(int t) 
      { String* ret;
        switch (t) {
          case TMB_NULL:
               ret = new String();
               break;
          case TMB_CannotConnect:
               ret = new String("Cannot connect to message server.");
               break;
          case TMB_CannotDisconnect:
               ret = new String("Cannot disconnect from message server");
               break;
          case TMB_CannotClear:
               ret = new String("Cannot clear domain list at message server");
               break;
          case TMB_CannotRegisterId:
               ret = new String("Cannot register userid with message server");
               break;
          case TMB_CannotAccept:
               ret = new String(
               "Message server did not accept client's domains of interest.");
               break;
          case TMB_CannotSend:
               ret = new String(
               "Message server did not send client's message.");
               break;
          case TMB_Disconnected:
               ret = new String("Client disconnected from message server.");
               break;
          case TMB_UnrecognizableMessage:
               ret = new String("Message format is unrecognizeable.");
               break;
        }
        return ret;
      };
  

    String* generateStandardErrorBuffer() 
      { return generateStandardErrorBuffer(type); };


    void setError(int e)
      { errorFound = TRUE;
        type = e;
        *errorBuffer = "";
        *messageBuffer = "";
      }

    void setErrorBuffer  (char* buf) { errorBuffer = new String(buf);   }
    void setMessageBuffer(char* msg) { messageBuffer = new String(msg); }
};

extern ErrorType *TMB_ERROR;

#endif
