// file:        os.C
// created by:  David Lauzon, University of Toronto
// date:        March 29, 1994.
// version:     0.0
// last update: March 29, 1994.



// This module is an interface to the operating system. It contains system
// dependent functions.
//
// getClientLogin() --> String*
//    The function getClientLogin returns the login name of the current client.
//
// getClientHostId() --> String*
//    The function getClientHostId returns the host id of the current client.
//
// getClientPId() --> String*
//    The function getClientPId returns the process id of the current client.
//
// getClientPNumber() --> int
//    The function getClientPNumber returns the host id (int) of the 
//    current client.
//
// getUniqueTimeString() --> String*
//    The function getUniqueTimeString returns a unique string of the form:
//                    seconds.microsecond
//    This is used to uniquely identify messages.



#include <sys/errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <limits.h>
#include <sys/time.h>

#include "os.H"

#define _BSD 
#define MAX_LOGIN_NAME_SIZE      256
#define MAX_HOST_NAME_SIZE       256
#define MAX_PROCESS_ID_SIZE      256
#define MAX_UNIQUE_TIME_SIZE     256
#define MAX_PROCESS_ID_NUMBER  99999

#ifndef __linux__
extern "C" int gethostname(char*, int);
#endif


// getClientLogin() --> String*
// The function getClientLogin returns the login name of the current client.

String* getClientLogin() 
  { char userName[MAX_LOGIN_NAME_SIZE+1];
    struct passwd *pw;
    uid_t uid;

    if ((pw = getpwuid(getuid())) == (passwd*)NULL)
      cerr << "SYSTEM ERROR: unable to access login name.\n";
    else
      strncpy(userName, pw->pw_name, MAX_LOGIN_NAME_SIZE);
    return (new String(userName));
  };



// getClientHostId() --> String*
// The function getClientHostId returns the host id of the current client.

String* getClientHostId()
  { int i;
    char hostName[MAX_HOST_NAME_SIZE+1];
    if ((i = gethostname(hostName, MAX_HOST_NAME_SIZE)) == -1)
      cerr << "SYSTEM ERROR: unable to access host id.\n";
    return (new String(hostName));
  };



// getClientPId() --> String*
// The function getClientPId returns the process id of the current client.

String* getClientPId()   
  { int i; 
    char processId[MAX_PROCESS_ID_SIZE+1];
    if ((i = getpid()) > MAX_PROCESS_ID_NUMBER)
      cerr << "SYSTEM ERROR: unable to access process id.\n";
    sprintf(processId, "%d", i);
    return (new String(processId));
  };



// getClientPNumber() --> int
// The function getClientPNumber returns the host id (int) of the 
// current client.

int getClientPNumber()   
  { int i; 
    if ((i = getpid()) > MAX_PROCESS_ID_NUMBER)
      cerr << "SYSTEM ERROR: unable to access process id.\n";
    return i;
  };



// getUniqueTimeString() --> String*
// The function getUniqueTimeString returns a unique string of the form:
//                    seconds.microsecond
// This is used to uniquely identify messages.

String* getUniqueTimeString()   
  { int err;
    struct timeval t, *tp;
    struct timezone tz, *tzp;
    tp = &t;
    tzp = &tz;

    char processId[MAX_UNIQUE_TIME_SIZE+1];
    if ((err = gettimeofday(tp, tzp)) != 0)
      cerr << "SYSTEM ERROR: unable to access system time. \n";
    sprintf(processId, ".%d.%d", tp->tv_sec, tp->tv_usec);
    return (new String(processId));
  };




