// $Author: bcorrie $
// $Date: 1997-09-27 17:45:59 $
// $Source: /root/cvs/Rigi/rigiedit/adt/bitarray.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#undef FDEBUG
#ifndef DEBUGTYPEH
#include "ccdebug.h"
#endif

#ifndef BITARRAY_H
#include "bitarray.h"
#endif

static int BitIds = 0;
static int const BYTE_LENGTH = 8;
static int const char_length = sizeof(char) * BYTE_LENGTH;

BitArray::BitArray(unsigned int sze) {
	DebugOff();
	register i;
	register char *p1;

	/*
	 * Size is dependent on the size of a char (normally 8 bits). The
	 * number of chars requried is the number of bits divided by the size
	 * of the char.  Add one char to pick up stray bits when size is not
	 * modulo char size.
	 */
	if(sze % char_length) {
		size = (sze/char_length) + 1;
	}
	else {
		size = (sze/char_length);
 	}
	ValueDebug("Creating bit array of size ",size,BitArray::BitArray);
	id = BitIds++;

	ptr = new char[size];
	for(i = size,p1 = ptr;--i >= 0;) {
		*p1 = 0;
	}

}

BitArray::BitArray(const BitArray& other) {
	register int i = other.size;
	register char *p1,*p2;

	ptr  = new char[other.size];
	for(p1 = ptr,p2 = other.ptr; --i >= 0;) {
		*p1++ = *p2++;
	}
}

BitArray::~BitArray() {
	delete[] ptr;
}

int BitArray::FindEmptyBit() {
	DebugOff();

	register int  p;
	register int  i;
	register char res;

	char* src;
	int result;

	/*
	 * Find first byte with a zero bit.
	 */
	for(src = ptr,i = 0;*src == ~0 && i < size;src++,i++);
	if(i == size) return -1; // full
	ValueDebug("working with byte = ",i,BitArray::FindEmptyBit);

	/*
	 * Find zero bit
	 */
	res = *src;
	for(p = char_length - 1;((res>>p) & ~(~0<<1)) && (p > 0); p--) {
		    ValueDebug("p = ",p,BitArray::FindEmptyBit);
	}	
	
	ValueDebug("Bit = ",p,BitArray::FindEmptyBit);

	/*
	 * Set bit.
	 */
	*src = *src | (1 << p);
	result = (i<< 3) + char_length - p - 1;

	ValueDebug("Returning Value of = ",result,BitArray::FindEmptyBit);

	return(result); 
}

void BitArray::ClearBit(int off) {
	DebugOff();
	int forward;
	int shift;
	int i;
	char *src;
	
	forward = off / char_length;
	if(forward > size) {
		MsgDebug("Bit outside array.",BitArray::ClearBit);
		return;
	}
	/*
	 * Find correct char
	 */
	for(i = 0,src = ptr;i< forward;i++,src++);
	ValueDebug("Missing ",forward,BitArray::ClearBit);

	/*
	 * Clear bit
	 */
	shift = (char_length - 1) - (off % char_length);
	*src = *src & ~(1 <<  shift);

	ValueDebug("Shift = ",shift,BitArray::ClearBit);

}

RigiBool BitArray::operator==(const BitArray& a) const {
	if(size != a.size) return RigiFalse;

	register int i = a.size;
	register char *p1,*p2;

	for(p1 = ptr,p2 = a.ptr; --i >= 0;) {
		if(*p1++ != *p2++) return RigiFalse;
	}
	return RigiTrue;
}	

void BitArray::operator=(const BitArray& a) {
	register int i = a.size;
	register char *p1,*p2;
	if(ptr != NULL) delete[] ptr;
	ptr  = new char[a.size];
	for(p1 = ptr,p2 = a.ptr; --i >= 0;) {
		*p1++ = *p2++;
	}
}
 
int BitArray::Printout(int) const {
	register int i = size;
	register char* p1;

	for(p1 = ptr; --i >= 0;p1++) {
		cerr << hex << (0x0ff&(*p1));
	}
	cerr << '\n';
	return 0;
}

unsigned int BitArray::Hash() const {
	register unsigned int h;
	register char*	  s = ptr;
	if(*s == NULL) return 0;
	h = ((int)*s++);
	while(*s++) h^=(((int)*s) & 0xff);

	return h;

}


RigiBool BitArray::isEqual(void* a) const {
	return ((RigiBool)(*this==*(BitArray*)a));
}






// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
//   Revision 1.2  1995/06/12  22:06:09  jmcdanie
//   PC port
//
//   Revision 1.1  1994/06/30  20:09:03  bcorrie
//   Initial revision
//
//

