// $Author: jmartin $
// $Date: 2000-02-05 02:18:28 $
// $Source: /root/cvs/Rigi/rigiedit/avl.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

//
//	avl.c: AVL treesets
//
#ifndef __STDIO_H
#include	<stdio.h>
#endif
#ifndef __STDLIB_H
#include	<stdlib.h>	// for malloc (MUST be new/delete free)
#endif
#ifndef __CTYPE_H
#include	<ctype.h>
#endif
#ifndef __STRING_H
#include	<string.h>
#endif

#ifndef _avlh
#include	"avl.h"
#endif
#ifndef _CCHEADER_H_
#include	"CCheader.h"	// for RigiBool
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

const int INVALID	= -1;

static const int EVEN	= 0;
static const int LEFT	= 1;
static const int RIGHT	= 2;

static const int MAXSETS = 32; // HACK!

typedef int (*AvlCompFunc)(void*a,void*b);	// comparison function type

struct avlrecord
{
    int info;
    int lean;
    char *name;
    struct avlrecord *left;
    struct avlrecord *right;
};
typedef struct avlrecord Record;

/*
// for 4-word records (with lean and info overlapping)
#define	set_lean(x,l)	(x)->info = (((x)->info>>2)<<2) | (l) 
#define	get_lean(x)	((x)->info & 0x3 )
#define	set_info(x,i)	(x)->info = ((x)->info & 0x3) | ((i)<<2)
#define	get_info(x)	((x)->info >> 2 )
*/
#define	set_lean(x,l)	(x)->lean = (l)
#define	get_lean(x)	((x)->lean)
#define	set_info(x,i)	(x)->info = (i)
#define	get_info(x)	((x)->info)

#define CHECKSET(x,p)	do{if((x)<=0||(x)>newsetindex){\
                        mayday("AVL error: invalid set %d in %s()",\
                        x,p);return 0;}}while(0)

static char	types[MAXSETS];
static int	sizes[MAXSETS];
static Record*	trees[MAXSETS];
static Record*	newrecord = NULL;
static Record*	NewRecord();
static int	FreeRecord( Record* );
static char*	newstring( char* );

static int newsetindex=0;

static int info;
static int returnvalue;
static int currset;
static char *string;

static int debug=0;

static int currtype;
static int probeonly;
static int intree = 0;

static const int KEY_ON_STRING	= 0;
static const int KEY_ON_INFO	= 1;

static int AVL = 1;

// either store a new record, or just find it

static int Store( Record **tree )
{
    // static vars: string, info, currtype

    Record *n = *tree;
    int j;

    if (n==NULL)
    {
        returnvalue = INVALID;
        if (probeonly) return 0;

        *tree = n = newrecord = NewRecord();
        sizes[currset]++;
        if (info!=INVALID) set_info(n,info);
        n->name = (currtype==2) ? string : newstring( string );

        string = n->name;

        if(debug)if(currtype==2)
        printf("new record 0x%x, info is %d, returnvalue is %d, size now %d\n",
            string,info,returnvalue, sizes[currset] );
        else
        printf("new record %s, info is %d, returnvalue is %d, size now %d\n",
            string,info,returnvalue, sizes[currset] );

        // this tree just got deeper
        return 1;
    }

    // comparison
    if (currtype==1) j = info - get_info(n);
    else if (currtype==2) j = (string==n->name)?0:((string<n->name)?-1:1);
    else {
        if (!string) mayday("null string in AvlStore()");
        if (!n->name) mayday("null name in AvlStore()");
        j = strcmp( string, n->name );
    }

    if (j==0) // found the appropriate record
    {
        returnvalue = get_info(n);
        if (info!=INVALID) set_info(n,info);
        string = n->name;

        if(debug)if(currtype==2)
        printf("found record 0x%x, info is %d, returnvalue is %d\n",
        string,info,returnvalue);
        else
        printf("found record %s, info is %d, returnvalue is %d\n",
        string,info,returnvalue);

        return 0;
    }
    else if (j<0) // recur down the tree
    {
        j = Store( &(n->left) ) ? LEFT : EVEN;
    }
    else
    {
        j = Store( &(n->right) ) ? RIGHT : EVEN;
    }
    if (j==0 || AVL==0) return 0;


    if (j!=get_lean(n))
    {
        // not a pivot - just set the lean factor
        if (get_lean(n)!=EVEN) set_lean(n,EVEN); else set_lean(n,j);

        // no change to nodes above if this subtree became EVEN
        return get_lean(n) ? 1 : 0;
    }

    // a rotation is required
    if (j==RIGHT && get_lean(n->right)==RIGHT)
    {
        set_lean(n,EVEN); set_lean(n->right,EVEN);

        *tree = (n->right);
        n->right = (n->right)->left;
        (*tree)->left = n;
        return 0;
    }
    else if (j==LEFT && get_lean(n->left)==LEFT)
    {
        set_lean(n,EVEN); set_lean(n->left,EVEN);

        *tree = (n->left);
        n->left = (n->left)->right;
        (*tree)->right = n;
        return 0;
    }
    else if (j==RIGHT && get_lean(n->right)==LEFT)
    {
        (*tree) = (n->right)->left;
        (n->right)->left = (*tree)->right;
        (*tree)->right = n->right;
        n->right = (*tree)->left;
        (*tree)->left = n;

        set_lean(n,get_lean(*tree)!=RIGHT ? EVEN : LEFT );
        set_lean((*tree)->right, get_lean(*tree)!=LEFT ? EVEN : RIGHT );
        set_lean(*tree,EVEN);
        return 0;
    }
    else if (j==LEFT && get_lean(n->left)==RIGHT)
    {
        (*tree) = (n->left)->right;
        (n->left)->right = (*tree)->left;
        (*tree)->left = n->left;
        n->left = (*tree)->right;
        (*tree)->right = n;

        set_lean(n,get_lean(*tree)!=LEFT ? EVEN : RIGHT);
        set_lean((*tree)->left,get_lean(*tree)!=RIGHT ? EVEN : LEFT);
        set_lean(*tree,EVEN);
        return 0;
    }
    else mayday("error: inconsistent state in AVL tree-set!");

    return 0;
}

int SetInfo( Avlset set, char *s, int i )
{
    CHECKSET(set,"SetInfo");
if (!s) mayday("null string in SetInfo()" );
    string = s; info = i;
    if (currset!=set) newrecord = NULL;
    currset = set;
    currtype = types[set];

    if (currtype==KEY_ON_STRING)
    if (newrecord) if (!strcmp(s,newrecord->name))
    {
        if(debug)printf("shortcut find %s, info =%d, prev=%d\n",
        s,i,get_info(newrecord));

        int r = get_info(newrecord);
        if (i!=INVALID) set_info(newrecord,i);
        return r;
    }

    if ( i==-1) probeonly = 1;

    if (intree) fprintf(stderr,"error in AVL, interleaved traversal!\n");
    intree = 1;
    Store( trees+set );
    intree = 0;
    probeonly = 0;

    return returnvalue;
}

static int maxlevel;
static int numset;
static int unbalanced;

static int Print( Record **tree, int level )
{
    Record *n = *tree; int l,r;

    if (n==NULL) return 0;

    ++ numset;
    if (level>maxlevel) maxlevel = level;

    l = 1 + Print( &(n->left), level+1 );

    if (currtype==2)
    printf("%2d%c 0x%x %d\n", level,
        get_lean(n)==EVEN?'E':(get_lean(n)==LEFT?'L':'R'),
        n->name, get_info(n) );
    else
    printf("%2d%c %4d \"%s\"\n", level,
        get_lean(n)==EVEN?'E':(get_lean(n)==LEFT?'L':'R'),
        get_info(n), n->name );

    r = 1 + Print( &(n->right), level+1 );

    if ((r-l)>1 || (r-l)<-1) unbalanced = 1;
    return l>r ? l : r;
}

int PrintSet( Avlset set )
{
    int d;

    CHECKSET(set,"PrintSet");
    numset = maxlevel = unbalanced = 0;

    d = Print( trees+set, 0 );
    if (unbalanced)
        mayday( "error: AVL sets not working properly" );
    if (maxlevel!=d-1)
        mayday( "error: PrintSet() not working (d=%d)!",d );
    printf( "(total=%d maxdepth=%d)\n",numset,maxlevel );
    return numset;
}

Avlset NewSet() { return NewSet(0); }

Avlset NewSet( int type )
{
    if (newsetindex>=MAXSETS)
    {
        mayday("avlset error: overflow! (due to poor programming)");
        return INVALID;
    }
    ++newsetindex; 
    trees[newsetindex] = NULL;
    types[newsetindex] = type;
    sizes[newsetindex] = 0;
    if(debug)printf("NewSet(), set=%d, type=%d\n",newsetindex,type);
    return newsetindex;
}

int SetSize( Avlset set )
{
    CHECKSET(set,"SetSize");
    return sizes[set];
}

// MEMORY ALLOCATION

static Record *rlist = NULL;
static int mem = 0;
static int charmem = 0;

static Record *NewRecord()
{
    static int maxindex = 64;
    static int index = maxindex;
    static const int maxblock = 512;
    static Record *array = NULL;

    Record *r;

    if (rlist) {
        r = rlist; rlist = r->right;
    } else if (index<maxindex) {
        r = array + index++;
    } else {
        if ( maxindex < maxblock ) maxindex<<=1;
        array = (Record*)malloc(maxindex*sizeof(Record));
        mem += maxindex*sizeof(Record);
        index = 0;
        r = array + index++;
    }
    r->name = NULL; r->left = r->right = NULL;
    r->info = r->lean = 0;
    return r;
}

static int FreeRecord( Record *r )
{
    //r->info = 0; r->lean = 0; r->name = NULL; r->left = NULL;
    r->right = rlist; rlist = r;
    return 1;
}

static int FreeRecords( Record **tree )
{
    Record *n = *tree;
    if (n==NULL) return 0;
    int r = FreeRecords( &(n->left) );
    r += FreeRecords( &(n->right) );
    return r + FreeRecord(n);
}

static char *newstring( register char *s )
{
    static int SIZE = 512;
    static char* index = NULL;
    static char* maxindex = NULL;

    register int len = 1;
    register char *i = s;

    if (!i) return NULL;
    if (*i) do { len++; i++; } while (*i);

    i = index;
    if (i+len>maxindex)
    {
        i = (char*)malloc(SIZE);
        charmem += SIZE;
        maxindex = i + SIZE;
    }
    if (*s) do { *i++ = *s++; } while (*s);
    *i++ = 0;
    index = i;
    return i-len;
}

int FreeSet( Avlset set )
{
    CHECKSET(set,"FreeSet");
    if (set==newsetindex) newsetindex--;
    newrecord = NULL;
    int r = FreeRecords( trees+set );
    return r;
}

int AvlSet( int i ) { return AVL = i; }

int AvlMin( int d )
{
    if (d==0) return 1;
    else if (d==1) return 2;
    else return 1 + AvlMin(d-1) + AvlMin(d-2);
}

int GetAvlMem() { return mem; }
int GetAvlCharMem() { return charmem; }

