// $Author: jmartin $
// $Date: 2002-09-27 00:27:42 $
// $Source: /root/cvs/Rigi/rigiedit/cl_base_window.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef BASEWINDOW_H
#define BASEWINDOW_H

#ifndef CHARTYPEH
#include "adt/chartype.h"     // For CharType
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"      // For Point
#endif

// Enumerated type that determines the placement of text relative to the
// position given to the text rendering routines.

typedef enum {
    T_TOPCENTER, T_BOTCENTER,
    T_BOTLEFT, T_TOPLEFT,
    T_BOTRIGHT, T_TOPRIGHT,
    T_CENTER
} TextLocation;

#define Above 0             // X definitions (x.h) used by Restack
#define Below 1

// Window class that is generic to all windowing systems. This class should
// be clompletely independent of the windowing system that implements the
// actual windowing functionality.

class BaseWindow
{
public:
    BaseWindow(char* type);
    BaseWindow(BaseWindow&);
    virtual ~BaseWindow();

    virtual int    OfType(char*);
    virtual char*  GetType();
    virtual int    GetId();
    virtual char*  GetTitle() = 0;
    virtual void   SetTitle(char*) = 0;

    // Set normal and wait cursors for the generic window
    virtual void   NormalCursor() = 0;
    virtual void   WaitCursor() = 0;

    // Message bar manipulation routines
    virtual void   SetMessage(char *text) = 0;
    virtual void   ResetMessage() = 0;

    // functions for retrieving/setting state info of any generic window
    virtual void   Restack(int order, BaseWindow *otherWin) = 0;
    virtual void   DestroyChildWindows() {};
    virtual void   Hide() = 0;
    virtual void   Show() = 0;
    virtual void   Location( int& x, int& y ) = 0;
    virtual void   SetLocation( int x, int y ) = 0;
    virtual void   SetGeometry( int x, int y, int w, int h) = 0;

    virtual char*  GetFont() = 0;
    virtual void   SetFont( char* ) = 0;

    // Returns RigiTrue if the window is iconified
    virtual int Iconified() = 0;
    virtual void SetIconified( int ) = 0;

    // Scrollbar and coordinate translation

    virtual void    SetScrollbarStartX( int )  {}
    virtual void    SetScrollbarStartY( int )  {}
    virtual int     ScrollbarStartX()          { return 0; }
    virtual int     ScrollbarStartY()          { return 0; }
    virtual Point2d Window2Canvas( Point2d p ) { return p; }
    virtual Point2d Canvas2Window( Point2d p ) { return p; }
    virtual Point   Window2Canvas( Point p )   { return p; }
    virtual Point   Canvas2Window( Point p )   { return p; }

    // Methods to find out infomration about the drawind area of the window
    virtual int     Depth() = 0;
    virtual void    FrameSize( unsigned short& w, unsigned short& h ) = 0;
    virtual void    SetFrameSize( int w, int h ) = 0; 
    virtual void    GetDrawSize( int*w, int*h ) = 0;

    // Basic window drawing operations
    virtual void Clear() = 0;
    virtual void DrawIcon( Point p, int size, int type, int active ) = 0;
    virtual void DrawBox( Point sp, Point dp, int thick, int type, int dpyOnly ) = 0;
    virtual void DrawLine( Point sp, Point dp, int width, int style ) = 0;
    virtual void DrawString( char* str, Point p ) = 0;
    virtual void DrawString( char* str, Point p, TextLocation loc ) = 0;
    virtual void TextExtent( char *str, int *w, int *h, int *a) = 0;

    // Drawing, batching operations

    virtual int  Drawing() = 0;
    virtual void Drawing( int ) = 0;
    virtual void DrawSynch( RigiBool ) = 0;
    virtual void BunchRect( int ) = 0;
    virtual void BunchLine( int ) = 0;
    virtual void SetComplexity( int n ) { numDrawObj = n; redrawCnt = 1; }
    virtual unsigned GetComplexity()    { return redrawCnt? numDrawObj/redrawCnt : numDrawObj; }

    virtual void Batch( int ) = 0;
    virtual void BackStore( int ) = 0;
    virtual int BackStore() = 0;
    virtual void Restore( int x, int y, int w, int h ) = 0;
    virtual void RestoreWindow() = 0;
    virtual void FreeStore() = 0;


    // Query shift key status
    virtual int IsShifted()             { return 0; }

private:
    char     *type;         // The type of the derived class, unknown to this
    int      id;            // Unique id for this window

protected:
    unsigned numDrawObj;    // cummulative # of rects, lines, captions
    unsigned redrawCnt;     // times BackStore function called
};

#endif

