#ifndef FILTERSET_H
#define FILTERSET_H
#ifndef FLAG_H
#include "cl_flag.h"
#endif

// Class that maintains arc and nodetype filter flags
// in a bit vector.

class FilterSet {
    private:
        BitSet* toBeApplied;    // filters set but not applied
        BitSet* filterSet;      // filters applied and in effect
        int filterSetSize;      // number of different filter types

    public:
        FilterSet(const int numFlags);
        ~FilterSet();

        int  GetFilter(const int filterId) const;
        void SetFilter(const int filterId);
        void ClearFilter(const int filterId);
        void ClearAll();
        void ApplyFilters();
        int FilterSetSize() const;
        FilterSet & operator=(const FilterSet &fs);
};

// inline member definitions
inline FilterSet::FilterSet (const int setSize) {
    // Initialize flags representing state of filters
    filterSet = new BitSet(setSize);
    toBeApplied = new BitSet(setSize);
    filterSetSize = setSize;
}

inline FilterSet::~FilterSet() {
    delete filterSet;
    delete toBeApplied;
    toBeApplied = filterSet = NULL;
}

inline int  FilterSet::GetFilter(const int filterId) const {
    return (filterId < filterSetSize)? filterSet->GetBit(filterId) : 0;
}

inline void FilterSet::SetFilter(const int filterId) {
    if (filterId >= filterSetSize) {
        filterSetSize = filterId + 1;
        filterSet->SetSize(filterSetSize);
        toBeApplied->SetSize(filterSetSize);
    }
    toBeApplied->SetBit(filterId);
}

inline void FilterSet::ClearFilter(const int filterId) {
    if (filterId >= filterSetSize) {
        filterSetSize = filterId + 1;
        filterSet->SetSize(filterSetSize);
        toBeApplied->SetSize(filterSetSize);
    }
    toBeApplied->ClearBit(filterId);
}

inline void FilterSet::ClearAll() {filterSet->ClearAll(); toBeApplied->ClearAll();}

inline void FilterSet::ApplyFilters() {*filterSet = *toBeApplied;}

inline int FilterSet::FilterSetSize() const {return filterSetSize;}

inline FilterSet & FilterSet::operator=(const FilterSet &fs) {
    *filterSet = *fs.filterSet;
    filterSetSize = fs.filterSetSize;
    toBeApplied->ClearAll();
    return *this;
}
#endif
