//
//

// $Author: jmcdanie $
// $Date: 1996-09-14 00:11:41 $
// $Source: /root/cvs/Rigi/rigiedit/cl_flag.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_flag.C,v 1.19 1996-09-14 00:11:41 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef FLAG_H
#include "cl_flag.h"
#endif
#ifndef GRAPHERR_H
#include "grapherr.h"
#endif

BitSet::BitSet() {
    numBits = 0;
    numElements = ElementAlloc;
    numBitsAlloc = numElements * BitsPerElement;
    elements = new unsigned int[ElementAlloc];
    memset(elements, 0, numElements * sizeof(int));
}

BitSet::BitSet(const int size) {

    numBits = size;
    if (size < 0) {
        numBits = 0;
        mayday( "BitSet::BitSet - Invalid number of bits (%d)", size );
        mayday( "BitSet::BitSet - Creating a BitSet with 0 bits");
    }

    // calculate the size of the vector.
    numElements = (size + BitsPerElement - 1) / BitsPerElement;
    if (numElements < ElementAlloc)
        numElements = ElementAlloc;
    numBitsAlloc = numElements * BitsPerElement;

    // allocate the vector and clear it.    
    elements = new unsigned int[numElements];
    memset(elements, 0, numElements * sizeof(int));
}

BitSet::~BitSet() {
    delete[] elements;
}

void BitSet::SetSize(const int size) {

    int newNumElements, newNumBitsAlloc;
    unsigned int *newElements;

    if (size < 0) {
        mayday("BitSet::SetSize - invalid number of bits (%d), size unchanged",
                size );
        return;
    }

    // just adjust the number of bits if the vector is 
    // already large enough to hold them.
    if (size >= numBits && size < numBitsAlloc) {
        numBits = size;
        return;
    }

    // calculate the number of vector elements that are
    // needed to contain the number of bits needed.
    newNumElements = (size + BitsPerElement - 1) / BitsPerElement;

    // if vector has become smaller, clear the
    // bits but don't deallocate the vector.
    if (size < numBits) {

        // initialize whole elements
        if (newNumElements < numElements) 
            memset(&elements[newNumElements], 0, 
                (numElements - newNumElements) * sizeof(int));

        if (newNumElements > 0) {

            // calculate number of bits to preserve in the 
            // last non-zero element.
            int i = (size % BitsPerElement);

            // build a mask by clearing low order bits
            // and calculating its complement.
            if (i) elements[newNumElements - 1] &= (~(-1 << i));
        }
        numBits = size;
        return;
    }
            
    // if the vector needs more elements, reallocate
    // the vector and copy the bits from the smaller
    // vector to the larger.
    newNumBitsAlloc = newNumElements * BitsPerElement;

    // allocate the new vector.
    newElements = new unsigned int[newNumElements];

    // copy the old vector into the new, larger vector.
    memcpy(newElements, elements, numElements * sizeof(int));

    // clear the extension.
    memset(&newElements[numElements], 0, 
        (newNumElements - numElements) * sizeof(int));

    // free the old vector's storage.
    delete[] elements;

    // swap in the new vector.
    numBits = size;
    numElements = newNumElements;
    numBitsAlloc = newNumBitsAlloc;
    elements = newElements;
}

void BitSet::SetBit(const int bit) {
    if (bit >= numBits || bit < 0) {
        mayday( "BitSet::SetBit() - bit %d not in BitSet of size %d", 
                bit, numBits);
        return;
    }
    int elementIdx = bit / BitsPerElement;
    int bitIdx = bit % BitsPerElement;
    elements[elementIdx] |= (1 << bitIdx);
}

void BitSet::ClearBit(const int bit) {
    if (bit >= numBits || bit < 0) {
        mayday( "BitSet::SetBit() - bit %d not in BitSet of size %d", 
                bit, numBits);
        return;
    }
    int elementIdx = bit/BitsPerElement;
    int bitIdx = bit % BitsPerElement;
    elements[elementIdx] &= ~(1 << bitIdx);
}

void BitSet::ClearAll() {
    if (numElements) memset(elements, 0, numElements * sizeof(int));
}

void BitSet::SetAll() {
    int numElementsInUse = (numBits + BitsPerElement - 1) / BitsPerElement;
    if (!numElementsInUse) return;

    memset(elements, -1, numElementsInUse * sizeof(int));

    // Clear the bits in the last element that
    // are not being used.

    // Calculate number of bits to preserve in the 
    // last non-zero element.
    int i = numBits % BitsPerElement;

     // build a mask by clearing low order bits
     // and calculating its complement.
     if (i) elements[numElementsInUse - 1] &= (~(-1 << i));
}

void BitSet::Print(char *str) const {
    Mess("%s... ",str) ;
    for(int i = 0; i < numBits; i++)
        if (GetBit(i))
            Mess("1... ");
        else
            Mess("0... ") ;
    Mess(NULL);
}

void BitSet::operator=( const BitSet& set ) {
    SetSize(set.numBits);
    memcpy(elements, set.elements, numElements * sizeof(int));
}

