/*
        File:    list.c - generic list (stack, queue) implementation module
        Author:  Hausi Muller
        Date:    May 21, 1989
        Project: Rigi V4.0
*/

// $Author: jmcdanie $
// $Date: 1996-05-26 18:13:01 $
// $Source: /root/cvs/Rigi/rigiedit/cl_list.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_list.C,v 1.14 1996-05-26 18:13:01 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef __STDIO_H
#include <stdio.h>
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef LIST_H
#include "cl_list.h"
#endif

static void Error( char*s ) { mayday("List class error: %s",s); }

void ItemList::Print() {
    for ( Item *s = head; s; s=s->next ) Mess("%d ... ",s->value);
    Mess(NULL);
}

void List::Print() {
    for ( SimpleItem *s = head; s; s=s->next ) Mess("%d ... ",s->value);
    Mess(NULL);
}

ItemList::ItemList() {
    head = NULL;
    tail = NULL;
    CurItem = NULL;
    NoItems = 0;
}

List::List() {
    head = NULL;
    CurItem = NULL;
    NoItems = 0;
}

ItemList::ItemList(ItemList& list) {
    Mess("ItemList copy constructor called");

    // Initialize the list

    head = NULL;
    tail = NULL;
    CurItem = NULL;
    NoItems = 0;

    // Make a copy of the list. We take advantage of the fact that
    // we are a friend of the Item class so that we can traverse the
    // list that we are given without changing the state of the
    // CurItem pointer.

    Item *item = list.head;
    while(item != NULL) {
        this->InsertLast(item->GetValue());
        item = item->next;
    }
}

List::List(List& list) {
    Mess("List copy constructor called");

    // Initialize the list

    head = NULL;
    CurItem = NULL;
    NoItems = 0;
    
    SimpleItem *item = list.head;
    while(item != NULL) {
        this->Insert(item->GetValue());
        item = item->next;
    }
    Reverse();
}

ItemList::~ItemList() {
    Item* s = head;
    while (s != NULL) {
        Item* doomed = s;
        s = s->next;
        delete doomed;
    }
}

List::~List() {
    SimpleItem* s = head;
    while (s != NULL) {
        SimpleItem* doomed = s;
        s = s->next;
        delete doomed;
    }
}

Info* ItemList::Insert( Info* value ) {
    Item* s = new Item(value);
    NoItems++;
    if (head == NULL) {
        head = s;
        tail = s;
    } else {
        s->next = head;
        head->prev = s;
        head = s;
    }
    return value;
}

Info* List::Insert( Info* value ) {
    SimpleItem* s = new SimpleItem(value);
    NoItems++;
    if (head == NULL) {
        head = s;
    } else {
        s->next = head;
        head = s;
    }
    return value;
}

Info* ItemList::InsertLast( Info* value ) {
    Item* s = new Item(value);
    NoItems++;
    if (tail == NULL) {
        head = s;
        tail = s;
    } else {
        s->prev = tail;
        tail->next = s;
        tail = s;
    }
    return value;
}

Info* ItemList::Remove( Info *i ) {
    Item *s;
    for ( s=head; s; s=s->next )
        if (s->value==i) break;
    if (s == NULL) return NULL;

    if (head == tail) {
        head = NULL;
        tail = NULL;
    } else if (s == head) {
        s->next->prev = NULL;
        head = s->next;
    } else if (s == tail) {
        s->prev->next = NULL;
        tail = s->prev;
    } else {
        s->next->prev = s->prev;
        s->prev->next = s->next;
    }
    if (CurItem == s) CurItem = NULL;
    NoItems--;
    delete s;
    return i;
}

Info* List::Remove( Info *i ) {
    SimpleItem *s, *p;
    for ( p=NULL,s=head; s; p=s,s=s->next )
        if (s->value==i) break;
    if (!s) return NULL;

    if (!p) head = s->next;
        else p->next = s->next;
    NoItems--;
    if (CurItem==s) CurItem = NULL;
    delete s;
    return i;
}

Info* ItemList::RemoveFirst() {
    if (head == NULL) return NULL;

    Info* val = head->value;
    Item* doomed = head;
    head = head->next;
    if (head == NULL) tail = NULL;
    NoItems--;
    delete doomed;
    return val;
}

Info* List::RemoveFirst() {
    if (head == NULL) return NULL;

    Info* val = head->value;
    SimpleItem* doomed = head;
    head = head->next;
    NoItems--;
    delete doomed;
    return val;
}

void ItemList::Empty() {
    while (RemoveFirst());
    if (NoItems!=0) Error("problem with Empty()");
}

void List::Empty() {
    while (RemoveFirst());
    if (NoItems!=0) Error("problem with Empty()");
}

void List::Reverse() {
    SimpleItem *newhead = NULL;
    SimpleItem *i = head;
    while ( i != NULL ) {
        SimpleItem *j = i->next;
        i->next = newhead;
        newhead = i;
        i = j;
    }
    head = newhead;
    CurItem = NULL;
}

Info* ItemList::RemoveLast() {
    if (tail == NULL) return NULL;

    Info* val = tail->value;
    Item* doomed = tail;
    tail = tail->prev;
    if (tail == NULL) head = NULL;
    NoItems--;
    delete doomed;
    return val;
}

Info* ItemList::Find( Info* val ) {
    for ( Item* s = head; s; s = s->next )
        if (s->value == val) return val;
    return NULL;
}

Info* List::Find( Info* val ) {
    for ( SimpleItem* s = head; s; s = s->next )
        if (s->value == val) return val;
    return NULL;
}

Info* ItemList::First() {
    return ( CurItem = head ) ? CurItem->value : NULL;
}

Info* List::First() {
    return ( CurItem = head ) ? CurItem->value : NULL;
}

Info* ItemList::Next() {
    if (!CurItem) return NULL;
    return ( CurItem = CurItem->next ) ? CurItem->value : NULL;
}

Info* List::Next() {
    if (!CurItem) return NULL;
    return ( CurItem = CurItem->next ) ? CurItem->value : NULL;
}

Info* ItemList::Last() {
    return ( CurItem = tail ) ? CurItem->value : NULL;
}

Info* ItemList::Prev() {
    if (!CurItem) return NULL;
    return ( CurItem = CurItem->prev ) ? CurItem->value : NULL;
}

void ItemList::Iterator( VarArgFunRef ifp ) {
    if (ifp == NULL) return;

    Item* item = this->head;
    while (item != NULL) {
        Info* ip = item->GetValue();
        (*ifp)(ip);
        item = item->next;
    }
}

void List::Iterator( VarArgFunRef ifp ) {
    if (ifp == NULL) return;

    SimpleItem* item = this->head;
    while (item != NULL) {
        Info* ip = item->GetValue();
        (*ifp)(ip);
        item = item->next;
    }
}


// SimpleList members

SimpleList::SimpleList(SimpleList& list) {
    Mess("SimpleList copy constructor called");

    head = NULL;
    SimpleItem *item = list.head;
    while (item != NULL) {
        Insert( item->GetValue() );
        item = item->next;
    }
    Reverse();
}

SimpleList::~SimpleList() {
    SimpleItem* s = head;
    while (s != NULL) {
        SimpleItem* doomed = s;
        s = s->next;
        delete doomed;
    }
}

Info* SimpleList::Insert( Info* value ) {
    SimpleItem* s = new SimpleItem(value);
    s->next = head;
    head = s;
    return value;
}

Info* SimpleList::Remove( Info *i ) {
    SimpleItem *s, *p;
    for ( p=NULL,s=head; s; p=s,s=s->next )
        if (s->value==i) break;
    if (!s) return NULL;

    if (!p) head = s->next;
        else p->next = s->next;
    delete s;
    return i;
}

Info* SimpleList::RemoveFirst() {
    if (head == NULL) return NULL;

    Info* val = head->value;
    SimpleItem* doomed = head;
    head = head->next;
    delete doomed;
    return val;
}

void SimpleList::Empty() {
    while (RemoveFirst());
}

void SimpleList::Reverse() {
    SimpleItem *newhead = NULL;
    SimpleItem *i = head;
    while ( i != NULL ) {
        SimpleItem *j = i->next;
        i->next = newhead;
        newhead = i;
        i = j;
    }
    head = newhead;
}

Info* SimpleList::Find( Info* val ) {
    for ( SimpleItem* s = head; s; s = s->next )
        if (s->value == val) return val;
    return NULL;
}

void SimpleList::Print() {
    for ( SimpleItem *s = head; s; s=s->next ) Mess("%d ... ",s->value);
    Mess(NULL);
}

void SimpleList::Iterator( VarArgFunRef ifp ) {
    if (!ifp) return;
    for ( SimpleItem* item = head; item; item = item->next )
        (*ifp)(item->GetValue());
}

