//
// File:    list.h -- generic list (stack, queue) definition module
// Author:  Hausi Muller
// Date:    May 21, 1989
// Project: Rigi V4.0

// $Author: jmcdanie $
// $Date: 1996-03-21 02:39:39 $
// $Source: /root/cvs/Rigi/rigiedit/cl_list.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef LIST_H
#define LIST_H

#ifndef _CCHEADER_H_
#include "CCheader.h"
#endif
typedef void(*VarArgFunRef)(...);
typedef int (*IntFunction)(...);
typedef void Info;

class Item
{
friend class ItemList;
    Item* next;
    Item* prev;
    Info* value;

public:
    Item(Info* val)   { value = val; next = prev = NULL; }
    ~Item()           { }
    Info* GetValue()  { return value; }
};

class SimpleItem
{
friend class List;
friend class SimpleList;
    SimpleItem *next;
    Info* value;

public:
    SimpleItem(Info* val)  { value = val; next = NULL; }
    ~SimpleItem()          { }
    Info* GetValue()       { return value; }
};

class ItemList
{
    // doubly-linked list
    // storage is currently allocated on an item by item basis
    // this needs to be changed to a fixed number of nodes by
    // fixed number of nodes allocation scheme.

    Item* head;
    Item* tail;
    Item* CurItem;
    int NoItems;

public:
    ItemList();
    ItemList(ItemList&);
    ~ItemList();

    Info* Insert(Info*);	// insert at one end of the list
    Info* InsertLast(Info*);    // insert at the other end of the list
    Info* Remove(Info*);
    Info* RemoveFirst();
    Info* RemoveLast();
    void  Empty();		// Remove all items in the list
    void  Reverse();
    Info* GetFirst()	{ return head ? head->value : NULL; }
    Info* GetLast()	{ return tail ? tail->value : NULL; }
    Info* Find(Info*);
    void  Print();

    RigiBool IsEmpty()	{ return (RigiBool)(head == NULL); }
    Info* First();	// start traversing the list in one direction
    Info* Next();	// continue traversing the list in one direction
    Info* Last();	// start traversing the list in the other direction
    Info* Prev();	// continue traversing the list in the other direction
    int GetNoItems()	{ return NoItems; }
    void Iterator(VarArgFunRef);
};

class List
{
    // singly-linked list
    // storage is currently allocated on an item by item basis
    // this needs to be changed to a fixed number of nodes by
    // fixed number of nodes allocation scheme.

    SimpleItem* head;
    SimpleItem* CurItem;
    int NoItems;

public:
    List();
    List(List&);
    ~List();

    Info* Insert(Info*);
    Info* Remove(Info*);
    Info* RemoveFirst();
    void  Empty();
    void  Reverse();
    Info* GetFirst()    { return head ? head->value : NULL; }
    Info* Find(Info*);
    void  Print();

    RigiBool IsEmpty()	{ return (RigiBool)(head == NULL); }
    Info* First();
    Info* Next();
    int GetNoItems()	{ return NoItems; }
    void Iterator(VarArgFunRef);
};

class SimpleList
{
    // minimal singly-linked list

    SimpleItem* head;

public:
    SimpleList()         { head = NULL; }
    SimpleList(SimpleList&);
    ~SimpleList();

    Info* Insert(Info*);
    Info* Remove(Info*);
    Info* RemoveFirst();
    void  Empty();
    void  Reverse();
    Info* Find(Info*);
    void  Print();

    RigiBool IsEmpty()	{ return (RigiBool)(head == NULL); }
    void Iterator(VarArgFunRef);
};

class Stack : private List
{
public:
    Stack() {}
    ~Stack() {}
    void Push(Info* val) { Insert(val); }
    Info* Pop() { return RemoveFirst(); }
    Info* Top() { return GetFirst(); }
    RigiBool IsEmpty() { return List::IsEmpty(); }
    void Empty() { List::Empty(); }
    int Height() { return GetNoItems(); }
};

class Queue : private ItemList
{
public:
    Queue() {}
    ~Queue() {}
    void Enqueue(Info* val) { Insert(val); }
    Info* Dequeue() { return RemoveLast(); }
    Info* Front() { return GetLast(); }
    RigiBool IsEmpty() { return ItemList::IsEmpty(); }
    void Empty() { ItemList::Empty(); }
    int Length() { return GetNoItems(); }
};

#endif

