//
// File:    rigiarc.c
// Author:  Brian Corrie
// Date:    January 16, 1991
// Project: Rigi V4.0
//

// $Author: jmcdanie $
// $Date: 1996-05-25 01:19:04 $
// $Source: /root/cvs/Rigi/rigiedit/cl_rigiarc.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_rigiarc.C,v 1.26 1996-05-25 01:19:04 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef ACTIVEARC_H
#include "cl_activearc.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

#define	uns	unsigned int


static ActiveArc *GetActive( GraphArc *a ) {
    if (!a) mayday("NULL arc in GetActive()");
    if (!a->GetActiveArc()) {
        ActiveArc* aa = new ActiveArc();
        a->SetActiveArc( aa );
    }
    return (ActiveArc*)(a->GetActiveArc());
}

static int narcs=0;

RigiArc::RigiArc( GraphArc* a, RigiWindow* win ) {
    narcs++;
    arc = a;
    window = win;

    ActiveArc* aa = GetActive( a );
    aa->AddArc( this );
}

RigiArc::~RigiArc() {
    narcs--;
    if (this==NULL) mayday("NULL this in RigiArc::~RigiArc()!");

    // Remove refs from the RigiNode src and dst lists
    RigiNode *src=GetSrc(), *dst=GetDst();
    if ( src ) src->RemoveArc( this );
    if ( dst ) if ( dst!=src ) dst->RemoveArc( this );

    // Remove all references to this arc from its active arcs list

    if (arc) ((ActiveArc*)(arc->GetActiveArc()))->RemoveArc( this );
}

GraphArc *RigiArc::SetNull() {
    // Need a method for telling a rigiarc that it is defunct.

    window = NULL;
    return arc;
}

int GetNoRigiArcs() { return narcs; }

RigiNode *RigiArc::GetSrc() {
    if (!window) return NULL;
    return window->FindRigiNode( GetArc()->GetSrc() );
}

RigiNode *RigiArc::GetDst() {
    if (!window) return NULL;
    return window->FindRigiNode( GetArc()->GetDst() );
}

RigiNode *RigiArc::GetOther( RigiNode *rn ) {
    if (!window) return NULL;
    RigiNode *src = GetSrc();
    return rn==src ? GetDst() : src;
}

int RigiArc::isVisible() {
    // Check filters of itself, its nodes, and its window
    return window->CheckArc( this );
}

void RigiArc::Draw() {
    if ( this==NULL || arc==NULL || window==NULL )
        mayday( "NULL reference in RigiArc::Draw" );

    // Check arc against its rigiwindow's filters

    if (!isVisible()) return;

    // Draw the arc in its window

    window->DrawArc( this );
    return;
}

