// $Author: jmcdanie $
// $Date: 1996-09-11 19:19:07 $
// $Source: /root/cvs/Rigi/rigiedit/cl_rigiarchive.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------
//	File:	cl_rigiarchive.C
//	Date:	May 17, 1994
//		(Taken from cl_rigiwindow.C)
//

#ifndef GLOBAL_H
#include "global.h"		// for the editor
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

#ifndef RIGIERR_H
#include "rigierr.h"		// error handling stuff
#endif

#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif

// Archive method implementations

int RigiWindow::WriteArchive( FILE* fp )
{
    int i, x,y;
    AvlPos *P;
    RigiWinType wintype = GetType();

    // Note: The first line of the archived window specifies info that is
    // required to create (through the constructor) a RigiWindow. This info
    // contains the window type, the window depth, and the size of the node
    // and arc flags. The second line contains the number of parent nodes for
    // the window, and their GRAS ids. All of this info needs to be read to
    // call the constructor for RigiWindows. You can then use the ReadArchive
    // method to read in the rest of the archived RigiWindow information.

    // Print the window tag, window type, and window depth

    fprintf( fp,"\n%s %d %d","RIGIWINDOW",(int)wintype,Depth() );

    // Print the size of the node and arc flags. Specifies the number of nodes
    // and arcs that the window used to have. This should be checked with the
    // current state of the system when a node is unarchived.

    fprintf( fp,"    %d %d\n", nodeFilters->FilterSetSize(), arcFilters->FilterSetSize());

    // Number of parent nodes, and their ids
    fprintf(fp,"%d   ",NumParents());

    Mess("Parents for window (%s) are:", Window()->GetTitle() );
    for ( GraphNode* np=FirstParent(P); np; np=NextParent(P) )
    {
        fprintf(fp,"%d ",(int)(np->GetId()));
        if (!GutilNodeExists(np->GetId())) Mess("      doesn't exist");
        Mess("    Node %d, name = %s", np->GetId(), np->GetLabel() );
    }
    fputc( '\n',fp );

    if (wintype == WIN_GENERAL) {
        for ( i=0; i<inarctypes.Size(); i++ ) {
            fputc( inarctypes.GetBit(i)?'1':'0',fp );
        }
        fputc( '\n',fp );
        for ( i=0; i<outarctypes.Size(); i++ ) {
            fputc( outarctypes.GetBit(i)?'1':'0',fp );
        }
        fputc( '\n',fp );
        fprintf(fp, "%d\n", this->inclusive);
    }

    // Print out the state of the node and arc filter flags
    for (i = 0; i < nodeFilters->FilterSetSize(); i++ )
        fputc( nodeFilters->GetFilter(i)? '0':'1',fp );
    for (i = 0; i < arcFilters->FilterSetSize(); i++ )
        fputc( arcFilters->GetFilter(i)? '0':'1',fp );
    fputc( '\n',fp );

    // Print out the window scale, the windows location in world coordinates.
    // MJW HACK! New yscale also written out, but the corresponding read
    // must be backward compatible

    Point2d p = GetLocation(); 
    fprintf( fp,"XS=%g YS=%g   %d %d %d",GetXScale(),GetYScale(),p.x,p.y,0 );

    // Print out the size and location of the window on the screen
    unsigned short w,h;
    BaseWindow* bwin = window;
    bwin->FrameSize(w,h);

    // The following variables are used to convert between ushorts and ints
    x = w; y = h;
    fprintf( fp,"   %d %d",x,y );

    // Print out window location and if the window is iconified or not 
    bwin->Location( x,y ); 
    fprintf( fp,"   %d %d %d\n", x,y,bwin->Iconified() );

    // Print out the current scrollbar position
    fprintf( fp, "   %d %d\n",bwin->ScrollbarStartX(),bwin->ScrollbarStartY());

    // Print out the control panel state
    fprintf( fp, "   %d %d %d %d %f %f\n", 
                editor->GetOverviewDepth(), 
                editor->GetProjectionDepth(),
                editor->GetGridSize(), 
                ((editor->GetReportGeneration()==REPORT_SIMPLE)?1:0),
                editor->GetHighThreshold(),
                editor->GetLowThreshold() );

    // Print out the number of nodes in the window, and their states
    fprintf( fp,"%d\n",NodeNumber() );
    for ( RigiNode* rn=FirstNode(P); rn; rn=NextNode(P) )
    {
        fprintf( fp,"%d", rn->GetNode()->GetId() );
        p = rn->GetPosition();
        fprintf( fp,"    %d %d %d",p.x,p.y,0 );
        fprintf( fp,"    %d    %d    %d    %d    %d\n",rn->GetFilter(),
                (int)(rn->HideOutArcs()),(int)(rn->HideInArcs()),
                (int)(rn->HideName()),(int)(rn->DrawWithIcon()) );
    }
    return 1;
}

int RigiWindow::ReadArchive( FILE* fp )
{
    int result,index,numnodes,grasid;
    char inchar;
    float winxscale, winyscale;
    int i,j, w,h, x,y,z;

    // Saving control panel in save view
    int ovdepth, projdepth, grid, eisimple;
    float thigh, tlow;

    char buff[80];

    if (!fscanf( fp,"%s",buff )) mayday("can't read node/arc flags!");

    i = arcFilters->FilterSetSize() + nodeFilters->FilterSetSize();
    j = strlen(buff);
    if ( i != j )
    {
        //warn("inconsistent number of view nodes+arcs (%d vs %d)", i,j);
    }

    // Read the state of the node and arc filter flags.
    // We don't want to set any flags out of range though!

    for (i=0; i < nodeFilters->FilterSetSize() && i < j; i++ ) {
        inchar = buff[i];
        if (inchar == '1')
            nodeFilters->ClearFilter(i);
        else if (inchar == '0')
            nodeFilters->SetFilter(i);
        else {
            mayday("ReadArchive: Invalid node filter flag");
            return 0;
        }
    }
    nodeFilters->ApplyFilters();

    index = nodeFilters->FilterSetSize();
    for (i = 0; i < arcFilters->FilterSetSize() && index < j; i++, index++ ) {
        inchar = buff[index];
        if (inchar == '1')
            arcFilters->ClearFilter(i);
        else if (inchar == '0')
            arcFilters->SetFilter(i);
        else {
            mayday("ReadArchive: Invalid arc filter flag");
            return 0;
        }
    }
    arcFilters->ApplyFilters();

    // Read the window scale. HACK! Must read yscale if available

    result = fscanf( fp," XS=%f YS=%f", &winxscale, &winyscale );
    if (result==2) {
        SetXScale(winxscale);
        SetYScale(winyscale);
    } else {
        result = fscanf( fp,"%f",&winxscale );
        if (result != 1 || winxscale < 0) {
            mayday("ReadArchive: Invalid window scale");
            return 0;
        } else {
            SetXScale(winxscale);
            SetYScale(winxscale);
        }
    }

    // Read the window world location
    if ( 3 != fscanf(fp,"%d %d %d",&x,&y,&z) ) {
        mayday("ReadArchive: Could not read world location");
        return 0;
    }
    Point2d p;
    p.x = x; p.y = y;
    SetLocation(p); 

    // Read the window screen size
    if ( 2 != fscanf(fp,"%d %d",&w,&h) ) {
        mayday("ReadArchive: Could not read window size");
        return 0;
    }
    BaseWindow* bwin = window;

    // Read the windows screen location and if it is iconified or not
    
    int iconified = 0;
    result = fscanf( fp, "%d %d %d", &x,&y,&iconified );
    if (result != 3 && result != 2) {
        // could be an old file, before icon information was saved therefore
        // I would only have two numbers here - accept this
        mayday("ReadArchive: Could not read window location");
        return 0;
    }

    bwin->SetGeometry( x, y, w, h ); 
    bwin->SetIconified( iconified );


    // Read the scrollbar position
    if ( 2 != fscanf(fp, "%d %d", &x, &y ) ) {
        mayday("ReadArchive: Could not read the scrollbar positions");
        return 0;
    }
    bwin->SetScrollbarStartX(x);
    bwin->SetScrollbarStartY(y);

    result = fscanf( fp, "%d %d %d %d %f %f", 
        &ovdepth, &projdepth, &grid, &eisimple, &thigh, &tlow );

    if (result != 6) {
        mayday("ReadArchive: Could not read the control panel state");
        return 0;
    }

    // Read the number of nodes in the window, and their states
    if (fscanf(fp,"%d",&numnodes) != 1) {
        mayday("ReadArchive: Could not read node list size");
        return 0;
    }

    int lostanode = 0;
    for (i = 0; i < numnodes; i++) {
        int f,oa,ia,name,draw;

        // Read in the gras id for the node. Then try to find it in the list
        // of riginodes for this window. If we do not find it report the error.
        // Even if we do not find the RigiNode, we still must read in the info
        // so that the file structure does not become corrupt.

        fscanf(fp,"%d",&grasid);
        AvlPos *P;
        RigiNode* rn = FirstNode(P);
        while (rn && rn->GetNode()->GetId() != grasid) rn = NextNode(P);
        if (rn == NULL)
        {
        if (!lostanode)
            mayday("Warning - Could not set state on RigiNode %d",grasid);
        lostanode = 1;
        }

        int z;
        if (fscanf(fp,"%d %d %d",&(p.x),&(p.y),&z) != 3)
        {
            mayday("ReadArchive: Could not read node position");
            return 0;
        }
            
        if (fscanf(fp,"%d %d %d %d %d",&f,&oa,&ia,&name,&draw) != 5)
        {
            mayday("ReadArchive: Could not read node state");
            return 0;
        }

        // If we found the RigiNode, set the information read
        if (rn != NULL)
        {
            rn->SetPosition(p);
            rn->SetFilter(f);
            rn->SetHideOutArcs((RigiBool)oa);
            rn->SetHideInArcs((RigiBool)ia);
            rn->SetHideName((RigiBool)name);
            rn->SetDrawWithIcon((RigiBool)draw);
        }
    }
    return 1;
}

