//
// File:    riginode.h
// Author:  Brian Corrie
// Date:    January 16, 1991
// Project: Rigi V4.0
//

// $Author: jmcdanie $
// $Date: 1996-09-11 19:19:09 $
// $Source: /root/cvs/Rigi/rigiedit/cl_riginode.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef RIGINODE_H
#define RIGINODE_H

//#ifndef OBJECTH
//#include "adt/object.h"         // Base class for RigiNode
//#endif

#ifndef LIST_H
#include "cl_list.h"            // Base class for RigiNodeList
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"        // Point2d, ArcDirection
#endif

#ifndef RIGIARCSET_H
#include "cl_rigiarcset.h"      // RigiArcSet
#endif
#ifndef _CCHEADER_H_
#include "CCheader.h"           // RigiBool
#endif

class RigiWindow;
class RigiArc;
class AvlPos;

class RigiNodeList : private List
{
public:

    // Constructor and destructor for this class.

    RigiNodeList() {}
    ~RigiNodeList() {}

    // Routines provided by this class.

    void      Insert( RigiNode*n )  { List::Insert(n); }
    RigiNode* RemoveFirst()         { return (RigiNode*)List::RemoveFirst(); }

    // MJW  These next two should removed! Still needed in ActiveNode though.
    RigiNode* Find( RigiNode*n )    { return (RigiNode*)List::Find(n); }
    RigiNode* Remove( RigiNode*n )  { return (RigiNode*)List::Remove(n); }

    RigiNode* First()      { return (RigiNode *)List::First(); }
    RigiNode* Next()       { return (RigiNode *)List::Next(); }
    void      Empty()      { List::Empty(); }
    int       GetNoItems() { return List::GetNoItems(); }

    void HorizSort();     
    void VertSort();     
    void DiagSort();     
    void GridSort();     
};

class RigiNode //: private RigiObject
{
private:

    RigiWindow*  window;     // Rigiwindow associated with this riginode
    GraphNode*   node;       // Pointer to the actual graphnode
    Point2d      position;   // Local location in the Rigiwindow

    RigiArcSet   srcarcs;    // arcs going out (this node is src)
    RigiArcSet   dstarcs;    // arcs coming in (this node is dst)

    // Flags: hideinarcs, hideoutarcs, hidename, drawicon
    unsigned short flags;
    unsigned short nodesize;

public:

    RigiNode( GraphNode*, RigiWindow* );
    ~RigiNode();

    // Functions to access the elements of this class.

    GraphNode*   GetNode() const              { return node; }
    RigiWindow*  GetWindow() const            { return window; }
    int          GetState() const;

    void         SetPosition(Point2d p)       { position = p; }
    Point2d      GetPosition() const          { return position; }

    void         SetNodeSize( int n)          { nodesize = n; }
    int          NodeSize() const             { return nodesize; }

    void         SetFilter( int f )           { flags&=~1; flags|=(f&1); }
    int          GetFilter() const            { return flags&1; }

    void         GetExtents( Point2d& minp, Point2d& maxp ) const;


    void     SetHideOutArcs(RigiBool b)	 { flags&=~2; flags|=(b<<1); }
    RigiBool HideOutArcs() const         { return flags&2?RigiTrue:RigiFalse; }

    void     SetHideInArcs(RigiBool b)   { flags&=~4; flags|=(b<<2); }
    RigiBool HideInArcs() const	         { return flags&4?RigiTrue:RigiFalse; }

    void     SetHideName(RigiBool b)     { flags&=~8; flags|=(b<<3); }
    RigiBool HideName() const            { return flags&8?RigiTrue:RigiFalse; }

    void     SetDrawWithIcon(RigiBool b) { flags&=~16; flags|=(b<<4); }
    RigiBool DrawWithIcon() const        { return flags&16?RigiTrue:RigiFalse;}

    // Arc list operations

    void	InsertArc( RigiArc* );
    RigiArc*	RemoveArc( RigiArc* );
    RigiArc*	FindArc( RigiNode* );
    RigiArc*	FirstArc( AvlPos*&, ArcDirection=ARC_INOUT );
    RigiArc*	NextArc( AvlPos*&, ArcDirection=ARC_INOUT );
    int		ArcNum( ArcDirection=ARC_INOUT );

    // Node movement and drawing methods

    RigiBool    isUnique();
    RigiBool    isVisible();
    void	Move( Point2d );
    void	Draw();
    
};

#endif
