//
// $Author: jmcdanie $
// $Date: 1996-10-26 00:17:12 $
// $Source: /root/cvs/Rigi/rigiedit/cl_rigistate.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_rigistate.C,v 1.65 1996-10-26 00:17:12 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"
#endif
#ifndef GRAPHERR_H
#include "grapherr.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"		// for RinitGetDomain() - MJW HACK
#endif

#ifndef GRAPHNODELIST_H
#include "cl_graphnodelist.h"
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif
#ifndef GRAPHARCSET_H
#include "cl_grapharcset.h"
#endif
#ifndef RIGINODESET_H
#include "cl_riginodeset.h"
#endif
#ifndef GRAPHNAMESET_H
#include "cl_graphnameset.h"	// MJW: for access-by-name
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef ACTIVEARC_H
#include "cl_activearc.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef EDIT_H
#include "edit.h"
#endif

// Default initial values for the control items

static const int DEF_OVERDEPTH = 5;
static const int DEF_HIERDEPTH = 10;
static const int DEF_SCALE = 150;
static const int DEF_GRIDSIZE = 100;
static const float DEF_THRESH = 0.0F;

EditorState::EditorState()
{
    // ALL initialization done in Init(). No good if Init() called TWICE!
}

void EditorState::Init()
{
    InitObjId();
    globalnodeset = new NodeSet;
    globalarcset  = new ArcSet;
    globalwinlist = new RigiWindowList;
    globalnameset = new NodeNameSet;

    selectnodes = new NodeSet;
    lastselectednode = NULL;
    filterlist = new RigiNodeList;

    selectarc = NULL;
    over_depth = DEPTH_INFINITE;
    proj_depth = DEPTH_INFINITE;
    hier_depth = DEF_HIERDEPTH;
    scale = DEF_SCALE;
    grid_size = DEF_GRIDSIZE;
    high_threshold = DEF_THRESH;
    low_threshold = DEF_THRESH;

    display_anno = RigiTrue;
    display_message = RigiTrue;
    display_command = RigiTrue;
    display_menu = RigiTrue;

    scale_function = SCALE_WINDOW;
    report_type = REPORT_VERBOSE;
    inherit_filters = RigiTrue;
    move_all = RigiFalse;

    arc_type = ANY_ARC;
    level_arc = 0;
    composite_arc = 0;
    collapse_node = 0;
}

EditorState::~EditorState()
{
    delete globalnodeset;
    delete globalarcset;
    delete globalwinlist;
    delete globalnameset;
    delete selectnodes;
    delete filterlist;
}

int            EditorState::GetOverviewDepth() const { return over_depth; }
int            EditorState::GetProjectionDepth() const { return proj_depth; }
int            EditorState::GetHierarchyDepth() const { return hier_depth; }
RigiScale      EditorState::GetScaleType() const { return scale_function; }
int            EditorState::GetScaleFactor() const { return scale; }
RigiConstraint EditorState::GetConstraintType() const { return constraint; }
RigiBool       EditorState::GetDisplayAnnotation() const { return display_anno;}
RigiBool       EditorState::GetDisplayMessage() const { return display_message;}
RigiBool       EditorState::GetDisplayCommand() const { return display_command;}
RigiBool       EditorState::GetDisplayMenu() const { return display_menu; }
int            EditorState::GetGridSize() const { return grid_size; }
float          EditorState::GetHighThreshold() const { return high_threshold; }
float          EditorState::GetLowThreshold() const { return low_threshold; }
RigiReport     EditorState::GetReportGeneration() const { return report_type; }
int            EditorState::GetArcType() const { return arc_type; }
RigiBool       EditorState::GetInheritFilters() const { return inherit_filters;}
RigiBool       EditorState::GetMoveUnison() const { return move_all;}


void EditorState::SetOverviewDepth(int v) { over_depth = v; }
void EditorState::SetProjectionDepth(int v) { proj_depth = v; }
void EditorState::SetHierarchyDepth(int v) { hier_depth = v; }
void EditorState::SetScaleType(RigiScale v) { scale_function = v; }
void EditorState::SetScaleFactor(int v) { scale = v; }
void EditorState::SetConstraintType(RigiConstraint v) { constraint = v; }
void EditorState::SetGridSize(int v) { grid_size = v; }
void EditorState::SetHighThreshold(float v) { high_threshold = v; }
void EditorState::SetLowThreshold(float v) { low_threshold = v; }
void EditorState::SetInheritFilters(RigiBool v) { inherit_filters = v; }
void EditorState::SetMoveUnison(RigiBool v) { move_all = v; }
void EditorState::SetReportGeneration(RigiReport v) { report_type = v; }
void EditorState::SetDisplayAnnotation(RigiBool v) { display_anno = v; }

void EditorState::SetArcType( int v ) { 
    arc_type = v; 

    // Changed arc type, so need to update all of the windows' message bars
    // This crappola shouldn't be in here!!  MJW

    if (!RinitGetDomain()) return;
    char *str = RinitGetDomain()->ArcNameFromType( v );
    for ( RigiWindow* rw = FirstWindow(); rw; rw = NextWindow() )
        rw->UpdateArcMessage( str );
}

void EditorState::SetDisplayMessage(RigiBool v) { 
    display_message = v; 
}

void EditorState::SetDisplayCommand(RigiBool v) { 
    display_command = v; 
}

void EditorState::SetDisplayMenu(RigiBool v) { 
    display_menu = v; 
}

// Arc set methods

unsigned char EditorState::InArcSet( int arctype ) const {
    // Make sure the arc type is in the appropriate range before setting

    if (arctype < 0 || arctype >= arc_set.Size())
        return RigiError(ALERT, "Invalid arc type (%d)", arctype );
    return arc_set.GetBit( arctype );
}

unsigned char EditorState::InArcSet( char *arctypename ) const {
    // Translate the type name into a type number and make sure its valid.

    int arctype = -1;
    if (arctypename == NULL) 
        return RigiError(ALERT, "Null arc type specified");

    else arctype = RinitGetDomain()->ArcTypeFromName( arctypename );
    if (arctype < 0 || arctype >= arc_set.Size())
        return RigiError(ALERT, "Invalid arc type (%s) specified", arctypename );

    // Return the required value

    return arc_set.GetBit( arctype );
}

void EditorState::SetArcSet( int arctype, unsigned char value ) {
    // Make sure the arc type is in the appropriate range

    if (arctype < 0 || arctype >= arc_set.Size()) {
        RigiError(ALERT, "Invalid arc type (%d) specified", arctype);
        return;
    }

    // Set the appropriate bit in the arc set

    if (value) arc_set.SetBit( arctype );
    else       arc_set.ClearBit( arctype );
}

void EditorState::SetArcSet( char *arctypename, unsigned char value ) {
    // Translate the type name into a type number and make sure its valid.

    int arctype = -1;
    if (arctypename == NULL) {
        RigiError(ALERT, "Null arc type in arc set specified");
        return;
    }
    arctype = RinitGetDomain()->ArcTypeFromName( arctypename );
    if (arctype < 0 || arctype >= arc_set.Size()) {
        RigiError(ALERT, "Invalid arc type (%s) specified", arctypename );
        return;
    }

    // Set the appropriate bit in the arc set.

    if (value) arc_set.SetBit( arctype );
    else       arc_set.ClearBit( arctype );
}

void EditorState::ClearArcSet() {
    if (arc_set.Size() != RinitGetDomain()->NumArcTypes())
        arc_set.SetSize(RinitGetDomain()->NumArcTypes());
    arc_set.ClearAll();
}

void EditorState::SetAllArcSet() {
    arc_set.SetAll();
}


// Selection processing

GraphNode* EditorState::FirstSelectedNode( AvlPos*& P )
    { return selectnodes->First(P); }
GraphNode* EditorState::NextSelectedNode( AvlPos*& P )
    { return selectnodes->Next(P); }
int EditorState::NumSelectedNodes()
    {return selectnodes->Size();}

GraphNode* EditorState::LastSelectedNode() { return lastselectednode; }

void EditorState::SelectNode( RigiNode* rn ) {
    // don't select filtered nodes
    if ( !rn->GetWindow()->CheckNode(rn) ) return;
    SelectNode( rn->GetNode() );
}
void EditorState::SelectNode( GraphNode *n ) {
    if (n == NULL) {
        RigiError(ALERT, "NULL node pointer" );
        return;
    }

    // If the node is already in the selectnodes, don't add it again.
    if (selectnodes->Find(n) != NULL) return;

    lastselectednode = n;
    selectnodes->Insert(n);

    // Draw each RigiNode and change its selection state before redrawing
    ActiveNode* an = (ActiveNode*)(n->GetActiveNode());
    if (an==NULL) return;

    RigiNode *cn;
    for ( cn=an->First(); cn; cn=an->Next() ) cn->Draw();
    an->SetState(ACTIVE);
    for ( cn=an->First(); cn; cn=an->Next() ) cn->Draw();
}

// Select all the nodes that are in the rectangle defined by p1,p2
// Deselect any nodes previously selected in the same rectangle

void EditorState::InvertSelectionNodes( RigiWindow *window, Point p1, Point p2 ) {
    Point minp,maxp;

    if ( window == NULL ) {
        RigiError(ERRORFLAG, "NULL rigi window" );
        return;
    }

    // Get the extents of the box.

    minp = p1;
    if ( p2.x < minp.x ) minp.x = p2.x;
    if ( p2.y < minp.y ) minp.y = p2.y;
    if ( p2.z < minp.z ) minp.z = p2.z;
    maxp = p1;
    if ( p2.x > maxp.x ) maxp.x = p2.x;
    if ( p2.y > maxp.y ) maxp.y = p2.y;
    if ( p2.z > maxp.z ) maxp.z = p2.z;

    // Check all nodes in the window.

    AvlPos *P;
    for ( RigiNode *rn=window->FirstNode(P); rn; rn=window->NextNode(P) )
    {
        Point rmaxp, rminp;
        window->NodeExtents( rn, &rminp, &rmaxp );

        int typeord = rn->GetNode()->GetType();
    
        // If the node extents overlap the box extents,
        // select/deselect the node
        if (rminp.x <= maxp.x && rmaxp.x >= minp.x
            && rminp.y <= maxp.y && rmaxp.y >= minp.y
            && !window->GetNodeTypeFilters()->GetFilter(typeord)
            && !rn->GetFilter()) {
            GraphNode *gn = rn->GetNode();
            ActiveNode* an = (ActiveNode*)(gn->GetActiveNode());
            if (an->GetState()==NORMAL)
                SelectNode(gn);
            else
                DeselectNode(gn);
        }
    }
}


// MJW  static (local) stuff to aid in MoveNodes()

static NodeSet *moveset;
static RigiWindow *rwin;
static int newx, newy;
static int move;

static GraphNode *DrawOrMove( GraphNode *n ) {
    ActiveNode* an = rwin ? NULL : (ActiveNode*)(n->GetActiveNode());
    RigiNode *rn = rwin ? rwin->FindRigiNode(n) : (an?an->First():NULL);
    for ( ; rn; rn = an ? an->Next() : NULL )

        if (move)  // adjust the riginodes' positions
        {
            Point2d p = rn->GetPosition();
            p.x += newx; p.y += newy;
            rn->Move( p );
        }
        else      // erase or draw all riginodes and rigiarcs
        {
            rn->Draw();
            AvlPos *P; RigiArc *ra;
            for ( ra = rn->FirstArc(P); ra; ra = rn->NextArc(P) )
            {
                RigiNode *on = ra->GetOther(rn);
                if ( !moveset->Find(on->GetNode()) || on>=rn ) ra->Draw();
            }
        }
    return NULL;
} 

void EditorState::DrawNodes( NodeSet *set ) {
    moveset = set; newx = 0; newy = 0;
    rwin = move_all ? NULL : RutilActiveWindow();
    move = 0; set->Iterator( DrawOrMove );
}

void EditorState::MoveNodes( NodeSet *set, int x, int y ) {
    if (x || y) {
        moveset = set; newx = x; newy = y;
        rwin = move_all ? NULL : RutilActiveWindow();
        move = 1; set->Iterator( DrawOrMove );
    }
}

// Deselect the given RigiNode

void EditorState::DeselectNode( RigiNode *rn ) { DeselectNode(rn->GetNode()); }
void EditorState::DeselectNode( GraphNode *n ) {
    if ( n==NULL ) {
        RigiError(ALERT, "NULL graph node" );
        return;
    }

    ActiveNode* an = (ActiveNode*)(n->GetActiveNode());
    if (an && an->GetState()==NORMAL) return;

    selectnodes->Remove( n );
    if (n==lastselectednode) lastselectednode = NULL;

    if (!an) return;

    RigiNode *cn;
    for ( cn=an->First(); cn; cn=an->Next() ) cn->Draw();
    an->SetState(NORMAL);
    for ( cn=an->First(); cn; cn=an->Next() ) cn->Draw();
}

// Select the given RigiArc

void EditorState::SelectArc( RigiArc* rap ) {
    rap->Draw();
    selectarc = rap;
    rap->Draw();
}

void EditorState::SelectArc( GraphArc* ap ) {
    RigiWindow *rw = RutilActiveWindow();
    if (!rw) return;
    RigiArc* rap = rw->FindRigiArc( ap );
    if (!rap) {
        RigiError(ALERT, "Arc not found in window");
        return;
    }
    
    rap->Draw();
    selectarc = rap;
    rap->Draw();
}

// Deselect the currently selected RigiArc

void EditorState::DeselectArc() {
    RigiArc *ra = selectarc;
    if ( ra == NULL ) return;

    ra->Draw();
    selectarc = NULL;
    ra->Draw();
}

// Deselect a graph arc

void EditorState::DeselectArc( GraphArc* ap ) {
    RigiWindow *rw = RutilActiveWindow();
    if (!rw) return;
    RigiArc* rap = rw->FindRigiArc( ap );
    if (!rap) {
        RigiError(ALERT, "Arc not found in window");
        return;
    }
    
    rap->Draw();
    selectarc = NULL;
    rap->Draw();
}

// Return a pointer to the selected arc

RigiArc* EditorState::GetSelectedArc() {
    return selectarc;
}

// Return a copy of the list of selected nodes. Uses the list iterator of
// the itemlist class and the function below to make the copy.

    static NodeList *local_nl;
    static GraphNode *addfunc( GraphNode*n ) { local_nl->Insert(n); return NULL; }

void EditorState::GetSelectionList( NodeList& list ) {
    local_nl = &list;
    local_nl->Empty();
    selectnodes->Iterator( addfunc );
}

    static RigiNodeList *local_rnl;
    static RigiWindow *local_rw;
    static GraphNode *rn_addfunc( GraphNode*n )
    {
        RigiNode *rn;
        if (n) if (rn = local_rw->FindRigiNode(n)) local_rnl->Insert(rn);
        return NULL;
    }

void EditorState::GetSelectionList( RigiNodeList& rnl, RigiWindow *rw ) {
    local_rnl = &rnl;
    local_rw = rw;
    selectnodes->Iterator( rn_addfunc );
}

// Deselect every node in the select list

void EditorState::EmptySelection() {
    AvlPos *P;
    for (GraphNode *n=selectnodes->First(P); n; n=selectnodes->Next(P)) {
        ActiveNode* an = (ActiveNode*)(n->GetActiveNode());
        if (!an || an->GetState()==NORMAL) continue;

        RigiNode *cn;
        for ( cn=an->First(); cn; cn=an->Next() ) cn->Draw();
        an->SetState(NORMAL);
        for ( cn=an->First(); cn; cn=an->Next() ) cn->Draw();
    }
    selectnodes->Empty();
    lastselectednode = NULL;
}

// Invert the selection list for the current window.

void EditorState::InvertSelection() {
    RigiNodeList newlist;
    RigiNode *rn;
    RigiWindow *rw = RutilActiveWindow();

    AvlPos *P;
    for ( rn=rw->FirstNode(P); rn; rn=rw->NextNode(P) )
        if (!selectnodes->Find( rn->GetNode() )) newlist.Insert( rn );

    EmptySelection();

    for ( rn=newlist.First(); rn; rn=newlist.Next() ) SelectNode( rn );
}

// Filter processing

RigiNode* EditorState::FirstFilteredNode()
    { return filterlist->First(); }
RigiNode* EditorState::NextFilteredNode()
    { return filterlist->Next(); }

int EditorState::NumFilteredNodes()
    { return filterlist->GetNoItems(); }

// Reset the filter list

void EditorState::ResetFilter() {
    RigiNode* rnp = filterlist->First();
    while( rnp != NULL )
    {
        rnp->SetFilter(RigiFalse);
        filterlist->Remove( rnp );
        rnp = filterlist->First();
    }
}

void EditorState::ResetFilter( RigiNode *rn ) {
    rn->SetFilter(RigiFalse);
    filterlist->Remove( rn );
}

void EditorState::ResetFilter( RigiWindow* rw ) {
    RigiNode *rn; RigiNodeList newlist;

    for ( rn=filterlist->First(); rn; rn=filterlist->First() )
    {
        filterlist->Remove( rn );
        if ( rn->GetWindow() == rw ) rn->SetFilter(RigiFalse);
            else newlist.Insert(rn);
    }
    for ( rn=newlist.First(); rn; rn=newlist.Next() )
        filterlist->Insert(rn);
}

// Set the filter list to contain all the nodes in the select list

void EditorState::SetFilter(RigiNodeList& list) {
    RigiNode* rnp = list.First();
    while( rnp != NULL )
    {
        SetFilter(rnp);
        rnp = list.Next();
    }
}

// P13
void EditorState::SetFilter(RigiNodeList& list, RigiWindow* rwin) {
    RigiNode* rnp = list.First();
    while( rnp != NULL )
    {
        if ( rnp->GetWindow() == rwin )
            SetFilter(rnp);
        rnp = list.Next();
    }
}

void EditorState::SetFilter(RigiNode* rnp) {
    filterlist->Insert(rnp);
    rnp->SetFilter(RigiTrue);
}

// State Node and Arc bookkeeping functions

static int narcs = 0;
static int nnodes = 0;

RigiBool EditorState::AddNode(GraphNode *n) {
    nnodes++;
    globalnameset->Insert(n);
    return globalnodeset->Add(n) ? RigiTrue : RigiFalse;
}

RigiBool EditorState::RemoveNode( GraphNode *n ) {
    nnodes--;
    if (!globalnameset->Remove(n))
        return RigiError(ERRORFLAG,
            "Unable to delete node name (%s)", (n)? n->GetLabel() : "<null>" );

    return globalnodeset->Remove( n->GetId() ) ? RigiTrue : RigiFalse;
}

RigiBool EditorState::ChangeNodeName( GraphNode *n, char *newname ) {
    globalnameset->Remove(n);

    if (newname == NULL) 
        n->SetLabel("");
    else
        n->SetLabel(newname);
    globalnameset->Insert(n);

    return RigiTrue;
}

int EditorState::PrintNodes() {
    globalnodeset->Print();
    return 0;
}

int EditorState::PrintNames( char *name ) {
    if (!name) { globalnameset->Print(); return 0; }

    AvlPos *P; GraphNode *n;
    for ( n = globalnameset->First(P,name); n; n=globalnameset->Next(P) )
        printf("%s\n",n->GetLabel());
    return 0;
}

    static GraphNode *WDN(GraphNode*n) {
        //WindowDeleteNode(n);
        ActiveNode* an = (ActiveNode*)n->GetActiveNode();
        if (an) delete an;
        delete n;
        return NULL;
    }

int EditorState::EmptyNodes() {
    int i = globalnameset->Size();
    int j = globalnodeset->Size();
    if ( i != j ) {
        RigiError(ERRORFLAG, "Node sets are corrupt");
        return 0;
    }

    globalnodeset->Iterator( WDN );

    globalnameset->Empty();
    nnodes -= globalnodeset->Empty();
    return j;
}

GraphNode* EditorState::FirstNode(AvlPos*&P) { return globalnodeset->First(P); }
GraphNode* EditorState::NextNode(AvlPos*&P)  { return globalnodeset->Next(P); }
int EditorState::NumNodes()  { return globalnodeset->Size(); }

GraphNode* EditorState::FindNode(const NodeId n) {
    GraphNode *clipnode = RigiClipNode();
    return (n == clipnode->GetId())? clipnode : globalnodeset->Find(n);
}

GraphNode* EditorState::FindNode( char *name ) {
    GraphNode *clipnode = RigiClipNode();
    return (!strcmp(name, clipnode->GetLabel()))? clipnode : globalnameset->Find(name);
}

int EditorState::FindNodes( char *name, NodeList& nl ) {
    return 0;
}

RigiBool EditorState::AddArc(GraphArc& n) {
    narcs++;
    return globalarcset->Add(&n) ? RigiTrue : RigiFalse;
}

GraphArc* EditorState::RemoveArc(const ArcId grasid) {
    narcs--;
    return globalarcset->Remove(grasid);
}

int GetNoEdArcs() { return narcs; }
int GetNoEdNodes() { return nnodes; }

GraphArc* EditorState::FindArc(const ArcId grasid) {
    return globalarcset->Find(grasid);
}

int EditorState::PrintArcs() {
    globalarcset->Print();
    return 0;
}

    static GraphArc *WDA(GraphArc*a)
    {
        //WindowDeleteArc(a);
        ActiveArc* aa = (ActiveArc*)a->GetActiveArc();
        if (aa) delete aa;
//printf("    be deleting arc %d\n",a);
        delete a;
        return NULL;
    }

int EditorState::EmptyArcs() {
    globalarcset->Iterator( WDA );
    narcs -= globalarcset->Empty();
    return 0;
}

// State window bookkeeping and iterator functions

void EditorState::AddWindow(RigiWindow* rwp) {
    globalwinlist->Insert(rwp);
}

RigiWindow* EditorState::RemoveWindow(RigiWindow* rwp) {
    if (rwp == RutilActiveWindow())
        RutilSetActiveWindow(RigiBaseWindow());
    return globalwinlist->Remove(rwp);
}

RigiWindow* EditorState::FindWindow(RigiWindow* rwp) {
    return globalwinlist->Find(rwp);
}

RigiWindow* EditorState::FirstWindow() {
    return globalwinlist->First();
}

RigiWindow* EditorState::NextWindow() {
    return globalwinlist->Next();
}

void EditorState::RigiDisplay() {
    for ( RigiWindow *rw = FirstWindow(); rw; rw = NextWindow() )
        rw->RigiDisplay();
}

        static GraphArc *SetNull( GraphArc *a )
        { a->SetSrc(NULL); a->SetDst(NULL); return NULL; }

void EditorState::Reset() {
    EmptySelection();
    DeselectArc();
    ResetFilter();

    // HACK
    // MJW assume that the arcs and nodes need no cleaning up;
    // just empty those lists, deleting their contents

    // Tell the arcs that they don't have to delete their refs in nodes
    globalarcset->Iterator( SetNull );

    EmptyNodes();
    EmptyArcs();
    InitObjId();
}

void EditorState::CloseWindow(RigiWindow *rw) {
    ResetFilter(rw);
    DeselectArc();
    RutilSetActiveWindow( RigiBaseWindow() );
    // hide the window to make the destruction seem faster
    rw->Window()->Hide();
    globalwinlist->Remove(rw);
    if (rw->GetType() == WIN_SCRATCH)
        RigiDeleteClipBoard();
    else
        delete rw;
}

void EditorState::CloseAll() {
    RigiWindowList winlist(*globalwinlist);
    RigiWindow *baseWindow = RigiBaseWindow();
    EmptySelection();
    DeselectArc();
    ResetFilter();

    // Make the base window the active one; do it first so that
    // The previous active window's title can be safely changed
    RutilSetActiveWindow(baseWindow);

    // Now, close all the rigiwindows except the base window
    for (RigiWindow* rw = winlist.First(); rw != NULL; rw = winlist.Next()) {
        if (rw == baseWindow) 
            // destroy any panels associated with the base window
            rw->DestroyChildWindows();

        else
            CloseWindow(rw);
    }

    // reset the active window and update the display 
    RutilSetActiveWindow(baseWindow);
}

void EditorState::CloseBaseWindow() {
    CloseAll();
    globalwinlist->Empty();
    delete RigiBaseWindow();
}

// State archiving functions

int EditorState::WriteArchive(FILE* fp) {
    // Write a tag to identify this section of the view

    fprintf(fp,"GLOBAL\n");

    // Write out the number of selected nodes, and their GRAS ids

    AvlPos *P;
    fprintf(fp,"%d   ",NumSelectedNodes());
    for ( GraphNode* n=FirstSelectedNode(P); n; n=NextSelectedNode(P) ) 
        fprintf(fp,"%d ",n->GetId());
    fprintf(fp,"\n");

    // Write out the GRAS id of the selected arc, if it is valid

    RigiArc* rap = GetSelectedArc();
    fprintf(fp,"%d\n", rap?rap->GetArc()->GetId():(-1) );

    // Write out the list of filtered nodes

    fprintf(fp,"%d   ",NumFilteredNodes());
    RigiNode *rnp = FirstFilteredNode();
    while(rnp != NULL) {
        fprintf(fp,"%d ",rnp->GetNode()->GetId());
        rnp = NextFilteredNode();
    }
    fprintf(fp,"\n");

    return 1;
}

int EditorState::ReadArchive(FILE* fp) {
    char str[80];
    int numnodes;
    int grasid;
    int i;

    // Read the GLOBAL tag to identify this section of the view

    if (fscanf(fp,"%s",str) != 1 || strcmp(str,"GLOBAL") != 0)
    {
        warn("Error - Corrupt GLOBAL section");
        return 0;
    }

    // Read the number of selected nodes, and their GRAS ids. Then see if that
    // node exists in the system. If so, find a RigiNode that represents it
    // and select that RigiNode.

    if (fscanf(fp,"%d",&numnodes) != 1)
    {
        warn("Error - Corrupt selection section");
        return 0;
    }
    for ( i=0; i<numnodes; i++ )
    {
        if (fscanf(fp,"%d",&grasid) != 1)
        {
            warn("Error - Corrupt selection list");
            return 0;
        }
        GraphNode* n = GutilNodeExists(grasid);
        if (n) SelectNode(n);
        else warn("Warning: Selected node doesn't exist, ignored!");
    }

    // Read the id of the selected arc, if it is valid
    fscanf(fp,"%d",&grasid);
    GraphArc* ap = GutilArcExists(grasid);
    if (ap)
    {
        // MJW don't understand this code; removing because of new activearc
        //ActiveArc* aap = (ActiveArc*)ap->GetActiveArc();
        //RigiArc* rap = aap->First();
        //if (!rap)
            //warn("Warning - Selected arc doesn't exist, ignored!");
        //else SelectArc(rap);
    }

    // Read the list of filtered nodes

    if (fscanf(fp,"%d",&numnodes) != 1)
    {
        warn("Error - Corrupt filter section");
        return 0;
    }
    for( i=0; i<numnodes; i++ )
    {
        if (fscanf(fp,"%d",&grasid) != 1)
        {
            warn("Error - Corrupt filter list");
            return 0;
        }
        GraphNode* np = GutilNodeExists(grasid);
        if (np != NULL)
        {
            ActiveNode* an = (ActiveNode*)np->GetActiveNode();
            RigiNode* rnp = an ? an->First() : NULL;
            if (rnp == NULL)
            warn("Warning - Selected node doesn't exist, ignored!");
            else SetFilter(rnp);
        }
        else warn("Warning: Selected node doesn't exist, ignored!");
    }
    return 1;
}

RigiWindow* EditorState::GetWindowByNumber(int windowNumber) {
    RigiWindow* rw = FirstWindow();
    while ( rw != NULL) {
        if (rw->GetWindowNumber() == windowNumber) return rw;
        rw = editor->NextWindow();
    }
    return NULL;
}

