
// File:      rigiwindow.c
// Author:    Brian Corrie
// Date:      January 16, 1991
// Project:   Rigi V4.0
//

// $Author: kienle $
// $Date: 2002-03-01 02:12:38 $
// $Source: /root/cvs/Rigi/rigiedit/cl_rigiwindow.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_rigiwindow.C,v 1.127 2002-03-01 02:12:38 kienle Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <math.h>

#ifndef _CCHEADER_H_
#include "CCheader.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"
#endif
#include "cl_tk_window.h"
#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef ACTIVEARC_H
#include "cl_activearc.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif

#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif
#ifndef GRAPHARCSET_H
#include "cl_grapharcset.h"
#endif
#ifndef RIGINODESET_H
#include "cl_riginodeset.h"
#endif

#ifndef GRAPH_H
#include "cl_graph.h"		// all the low level graph stuff.
#endif

#ifndef RIGIERR_H
#include "rigierr.h"		// error handling stuff
#endif
#ifndef TIDY_H
#include "tidy.h"		// interface to Lynns tidy stuff
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef WINUTIL_H
#include "win_util.h"		// for BIG, SMALL, THICK
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif

#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif
#ifndef FILENAME_H
#include "adt/filename.h"
#endif
#ifndef CSTUFF_H
#include "cstuff.h"         // for WindowSetCursor
#endif
#ifndef GROUP_H
#include "group.h"
#endif

static int id = 0;

// Definitions that define the maximum limits for the world coordinate space
// and the size of the windows that look in on the world coordinate system.

// Size of root window frame (adjusted to compensate for message bar)
static const int ROOTSIZEX = 500;
static const int ROOTSIZEY = 400;

// Size of generic window frame
static const int FRAMESIZEX = 500;
static const int FRAMESIZEY = 400;

// Adjustable canvas dimensions to suit differing memory capabilities
static Coordinate WINSIZEX = 1280;
static Coordinate WINSIZEY = 1024;

// Conversion factors between World and NDC coordinate spaces.
static const Coordinate NDC2WORLDX = 10240;
static const Coordinate NDC2WORLDY = 10240;
static const Coordinate NDC2WORLDZ = 10240;

static const NDC_Coord WORLD2NDCX = 1.0F/NDC2WORLDX;
static const NDC_Coord WORLD2NDCY = 1.0F/NDC2WORLDY;
static const NDC_Coord WORLD2NDCZ = 1.0F/NDC2WORLDZ;

// defines an infinite depth projection (externally visible - blech!)
const int DEPTH_INFINITE = -1;

// define the maximum and minimum size of a node.
const int MAXNODESIZE = 64;
const int MINNODESIZE = 6;

// define the max distance from an arc for selection
const int MAXARCDISTANCE = 3;

// define the positioning of the nw corner node
const int BORDER = 16;

void RigiWindow::CenterScrollbars(int x, int y ) {
    int cw, ch;
    int ww, wh;
    Point2d p;
    RutilGetMaxCanvasSize(cw, ch);
    window->GetDrawSize(&ww, &wh);

    // calculate pixel offset of window origin from canvas origin
    p.x = (ww < cw)? (int)((float)((cw - ww)) / 2.0) : 0;
    p.y = (wh < ch)? (int)((float)((ch - wh)) / 2.0) : 0;
    window->SetScrollbarStartX(p.x);
    window->SetScrollbarStartY(p.y);

    // calculate canvas origin in world coordinates.
    canOrig.x = x - (int)(p.x / xscale);
    canOrig.y = y - (int)(p.y / yscale);
}

void RigiWindow::DrawAll() {
    if ( !this || !window ) mayday( "NULL self or window in DrawAll" );

    RigiNode *rn; RigiArc *ra; AvlPos *P, *Q;

    for ( rn=nodeset->First(P); rn; rn=nodeset->Next(P) ) {
        rn->Draw();
        for ( ra=rn->FirstArc(Q,ARC_OUT); ra; ra=rn->NextArc(Q,ARC_OUT) )
            ra->Draw();
    }
}

// Perform required stuff for changing the Domain

GraphNode* RigiWindow::GetNode() {
    return nlist ? nlist->First() : NULL;
}

void RigiWindow::SetNode( GraphNode *n ) {
    if (!n) { mayday("set node of window to NULL!"); return; }
    //node = n;

    nlist->Empty();
    nlist->Add( n );
    RigiDomain *domain = RinitGetDomain();
    int nn = domain->NumNodeTypes();
    int na = domain->NumArcTypes();
    Mess("setting window's node to %s, with %d node and %d arc types",
        n->GetLabel(), nn, na );

    nodeFilters->ClearAll();
    arcFilters->ClearAll();
    inarctypes.SetSize( na );
    outarctypes.SetSize( na );
    inarctypes.SetBit( editor->LevelArc() );
    outarctypes.SetBit( editor->LevelArc() );

    UpdateLabel();
}

int RigiWindow::NumParents() {
    return nlist->Size();
}

GraphNode* RigiWindow::FirstParent( AvlPos*&P ) {
    return nlist ? nlist->First(P) : NULL;
}

GraphNode* RigiWindow::NextParent( AvlPos*&P ) {
    return nlist ? nlist->Next(P) : NULL;
}

GraphNode* RigiWindow::NodeIsParent(NodeId nodeId) {
    return nlist->Find(nodeId);
}

void RigiWindow::RemoveParent(NodeId nodeId) {
    nlist->Remove(nodeId);
}

// find the upper-left corner around visible nodes in the window.
// flag > 0 means only the SELECTED nodes (rather than all).

void RigiWindow::GetNodeInWindow(Point *origin) {
    int numNodes = 0;
    Point2d p;
    RigiNode *rn; RigiNodeList slist;
    Point minp, maxp;
    int f0, f0prev, f1, f1prev;

    origin->x = origin->y = 0;
    editor->GetSelectionList(slist, this);
    int numSel = slist.GetNoItems();
    GetWorldExtents(&minp, &maxp, numSel);

    AvlPos *P;
    rn = (numSel)? slist.First() : nodeset->First(P);
    while (rn!=NULL) {
        if (CheckNode(rn)) {
            p = rn->GetPosition();

            // sum of deltax and deltay
            f0 = p.x -minp.x + p.y - minp.y;

            // difference of deltax and deltay
            f1 = abs(p.x - minp.x - p.y + minp.y);
            if (!numNodes || (f0 <= f0prev && f1 <= f1prev)) {
                    origin->x = p.x; origin->y = p.y;
                    f0prev = f0; f1prev = f1;
                    numNodes++;
            }
        }
        rn = (numSel)?  slist.Next() : nodeset->Next(P);
    }
}

// find the box around visible nodes in the window.
// flag > 0 means only the SELECTED nodes (rather than all).

void RigiWindow::GetWorldExtents( Point *minp, Point *maxp, int flag ) {
    Point2d p;
    int i = 1;
    RigiNode *rn; RigiNodeList slist;

    if ( this==NULL || minp==NULL || maxp==NULL )
        mayday( "NULL parameter in GetWorldExtents" );

    if (flag) editor->GetSelectionList( slist, this );

    AvlPos *P;
    rn = flag ? slist.First() : nodeset->First(P);
    while ( rn!=NULL ) {
        if (CheckNode(rn)) {
            p = rn->GetPosition();
            if ( i || p.x < minp->x ) minp->x = p.x;
            if ( i || p.y < minp->y ) minp->y = p.y;

            if ( i || p.x > maxp->x ) maxp->x = p.x;
            if ( i || p.y > maxp->y ) maxp->y = p.y;
            i = 0;
        }
        rn = flag ? slist.Next() : nodeset->Next(P);
    }
    if (i) {
        minp->x = minp->y = minp->z = 0;
        maxp->x = maxp->y = maxp->z = 0;
    }
}

// Coordinate conversion methods

Point RigiWindow::NDC2World( NDC_Point ndcp ) {
    Point p;
    if (this==NULL) mayday("NULL self reference in RigiWindow::NDC2World");

    p.x = (Coordinate)(ndcp.x * NDC2WORLDX);
    p.y = (Coordinate)(ndcp.y * NDC2WORLDY);
    p.z = (Coordinate)(ndcp.z * NDC2WORLDZ);
    return p;
}

Point RigiWindow::NDC2World( NDC_Point2d ndcp ) {
    Point p;
    if (this==NULL) mayday("NULL self reference in RigiWindow::NDC2World");

    p.x = (Coordinate)(ndcp.x * NDC2WORLDX);
    p.y = (Coordinate)(ndcp.y * NDC2WORLDY);
    p.z = 0;
    return p;
}

Point2d RigiWindow::NDC2World2d( NDC_Point2d ndcp ) {
    Point2d p;
    if (this==NULL) mayday("NULL self reference in RigiWindow::NDC2World");

    p.x = (Coordinate)(ndcp.x * NDC2WORLDX);
    p.y = (Coordinate)(ndcp.y * NDC2WORLDY);
    return p;
}

NDC_Point RigiWindow::World2NDC( Point p ) {
    NDC_Point ndcp;
    if (this==NULL) mayday("NULL self reference in RigiWindow::World2NDC");

    ndcp.x = (NDC_Coord)(p.x * WORLD2NDCX);
    ndcp.y = (NDC_Coord)(p.y * WORLD2NDCY);
    ndcp.z = (NDC_Coord)(p.z * WORLD2NDCZ);
    return ndcp;
}

NDC_Point2d RigiWindow::World2NDC2d( Point p ) {
    NDC_Point2d ndcp;
    if (this==NULL) mayday("NULL self reference in RigiWindow::World2NDC2d");

    ndcp.x = (NDC_Coord)(p.x * WORLD2NDCX);
    ndcp.y = (NDC_Coord)(p.y * WORLD2NDCY);
    return ndcp;
}

NDC_Point2d RigiWindow::World2NDC2d( Point2d p ) {
    Point p1; p1.x = p.x; p1.y = p.y;
    return World2NDC2d(p1);
}

Point RigiWindow::World2Canvas( Point p ) {
    if (this == NULL) mayday("NULL self in RigiWindow::World2Canvas");

    p.x = (int)((p.x - canOrig.x)*xscale);
    p.y = (int)((p.y - canOrig.y)*yscale);

    return p;
}

Point RigiWindow::World2Canvas( Point2d pos ) {
    Point p; p.x = pos.x; p.y = pos.y; p.z = 0;
    return World2Canvas(p);
}

Point RigiWindow::Canvas2World( Point p ) {
    if ( this == NULL )
    mayday("NULL self reference in RigiWindow::Canvas2World");

    p.x = (int)(p.x/xscale);
    p.y = (int)(p.y/yscale);
    p.z = (int)(p.z/xscale);

    // Add the window location to convert to world coordinates.

    p.x += canOrig.x;
    p.y += canOrig.y;
    return p;
}

// Scaling methods

void RigiWindow::ScaleToBox( Point p1, Point p2 ) {
    const int TEXTHEIGHT = 13;	// HACK!
    const int NODESIZE = MAXNODESIZE;	// HACK!

    int w,h;
    float scalex = 1.0F, scaley = 1.0F;
    window->GetDrawSize( &w, &h );
//  Mess("Scale to box (%d,%d) - (%d,%d), w=%d h=%d",
//      p1.x, p1.y, p2.x, p2.y, w,h);

    int xdiff = p1.x - p2.x;
    int ydiff = p1.y - p2.y;

    // If the box has 0 area, then don't perform any scaling. We do have to
    // move the window posiiton though.

    if (ydiff || xdiff) {

        xdiff = ((xdiff<0) ? -xdiff : xdiff);
        ydiff = ((ydiff<0) ? -ydiff : ydiff);

        xdiff += NODESIZE;
        ydiff += NODESIZE;

        if (xdiff) scalex = ((float)w-BORDER*2)/(float)xdiff;
        if (ydiff) scaley = ((float)h-BORDER*2-TEXTHEIGHT)/(float)ydiff;

        if ( scalex>1.0F || scalex<=0.0F ) scalex = 1.0F;
        SetXScale( scalex );

        if ( scaley>1.0F || scaley<=0.0F ) scaley = 1.0F;
        SetYScale( scaley );
    }

    // now that scales are calculated, we could be in trouble if
    // they are quite different ... don't worry for now

    p1.x -= (int)(BORDER/scalex);
    p1.y -= (int)(BORDER/scaley);

    canOrig.x = p1.x; canOrig.y = p1.y;
    CenterScrollbars(p1.x, p1.y);
}

void RigiWindow::ScaleToView() {
    Point minp, maxp;
    GetWorldExtents(&minp, &maxp, 0);
    ScaleToBox(minp, maxp);
}

void RigiWindow::ScaleSelect() {
    Point minp, maxp;
    GetWorldExtents(&minp, &maxp, 1 );
    ScaleToBox(minp, maxp);
}

void RigiWindow::ScaleFactor() {
    Point minp;

    // get the upper left corner of graph
    GetNodeInWindow(&minp);
    SetXScale(GetXScale()*(editor->GetScaleFactor()/100.0F)) ;
    SetYScale(GetYScale()*(editor->GetScaleFactor()/100.0F)) ;
    minp.x -= (int)((float)BORDER/GetXScale()); 
    minp.y -= (int)((float)BORDER/GetYScale());
    canOrig.x = minp.x; canOrig.y = minp.y;
    CenterScrollbars(minp.x, minp.y);
}        

void RigiWindow::ScaleUnscale() {
    Point minp;

    // get the upper left corner of graph
    GetNodeInWindow(&minp);
    SetXScale(1.0F); SetYScale(1.0F);
    minp.x -= BORDER; minp.y -= BORDER;
    canOrig.x = minp.x; canOrig.y = minp.y;
    CenterScrollbars(minp.x, minp.y);
}

//---------------------------------------------------------------------------
// Initialize an Overview window
//---------------------------------------------------------------------------

// The depth is set in the RigiWindow::InitOverview function and used here
// to determine how far down the tree we should go.

static int ovdepth;
static int ovstart;

// Some static lists used by various Init* functions

static NodeSet visited;

static void AddOverviewNodes( RigiNode *src, RigiWindow *rw )
{
    visited.Insert(src->GetNode());

    // For each level arc out, be adding that arc and its dest
    // (if that dest not be visited)

    GraphNode *n = src->GetNode();
    AvlPos *P;
    GraphArc *arc = n->FirstAdj( P, outLevArcs );
    while (arc) {
        GraphNode* ndst = arc->GetDst();
        if (!visited.Find(ndst)) {

            RigiNode* dst = new RigiNode( ndst,rw );
            rw->InsertNode(dst);

            RigiArc *ra = new RigiArc( arc,rw );
            src->InsertArc(ra);
            dst->InsertArc(ra);

            AddOverviewNodes( dst, rw );
        }
        arc = n->NextAdj( P, outLevArcs );
    }
}

void RigiWindow::InitOverview( GraphNode* n )
{
    if ( !n || !this )
        mayday("NULL reference in RigiWindow::InitOverview");

    // Do a depth first search on the level arcs, adding all the nodes
    // and arcs that we see to their respective lists.

    RigiNode *rn = new RigiNode( n,this );
    InsertNode(rn);

    visited.Empty();
    visited.Insert(n);

    AddOverviewNodes( rn, this );

    // Compute nice location for the nodes.

/*
    int size = rn->NodeSize();
    int d = ovdepth==DEPTH_INFINITE ? ovdepth : ovstart+ovdepth;
    Mess("Ovdepth = %i, Ovstart = %i", ovdepth, ovstart);
    TidyUpForest(n,d,size,size);
*/

    // Set the riginodes to the tidy positions.
    extern int Tidy(RigiNode *, int);
    rn = FindRigiNode(n);
    Tidy(rn, 0);
    
    AvlPos *P;
    for ( rn = nodeset->First(P); rn; rn = nodeset->Next(P) ) {
        rn->SetHideName(RigiTrue);
        rn->SetDrawWithIcon(RigiFalse);
    }
}

//---------------------------------------------------------------------------
// Initialize a Scratch window
//---------------------------------------------------------------------------

void RigiWindow::InitScratch( GraphNode* n ) {
    ArcSet arcset;
    RigiNodeList nodelist;
    int numNodesLaidOut = 0;

    // Paranoid error checking

    if ( n == NULL ) return;
    if (this==NULL)
        mayday("NULL self reference in RigiWindow::InitScratch");

    // Traverse all of the level arcs from node n, and add each destination
    // node to the list of nodes that are contained in this window.

    int num = 0;
    AvlPos *P;
    GraphArc* arc = n->FirstAdj( P, outLevArcs );
    while ( arc != NULL ) {
        // Get the dst of the level arc and create a new riginode

        GraphNode* dst = arc->GetDst();
        RigiNode* rn = new RigiNode(dst,this);

        // Insert the node in the node list

        nodeset->Insert(rn);
        num++;

        // Get the next level arc

        arc = n->NextAdj( P, outLevArcs );
    }

    // Get the arcs for this window and while we are at it, lets set the
    // position of the RigiNodes.

    for ( RigiNode* rn = nodeset->First(P); rn; rn = nodeset->Next(P) ) {
        // Get the node associated with this RigiNode

        GraphNode *n = rn->GetNode();

        // Get the NDC position of the node and transform to local coords.

        NDC_Point2d ndcp = n->GetNDCPosition();
        Point2d p = NDC2World2d( ndcp );
        rn->SetPosition( p );
        nodelist.Insert(rn);

        // keep tally to determine if a layout is needed.
        if (p.x || p.y) numNodesLaidOut++;
        
        // For each normal arc in this node

        AvlPos *Q;
        GraphArc *a = n->FirstAdj( Q, outArcs );
        while ( a != NULL ) {
            // Don't add the arc to the window if it already exists, or if the
            // node does not belong to window being processed. If it doesn't
            // already exist, create a rigiarc for it, and insert into the
            // riginodes' arclists

            GraphNode *dst = a->GetDst();
            RigiNode *rdst = FindRigiNode(dst);
            if ( rdst && !arcset.Find( a->GetId() ) ) {
                arcset.Add( a );
                RigiArc* ra = new RigiArc( a,this );
                rn->InsertArc( ra );

                // If the arcs source and destination are the same then it is
                // a recursive arc and the arc should not be added to the
                // RigiNodes list again.

                if (ra->GetArc()->GetSrc()->GetId() !=
                    ra->GetArc()->GetDst()->GetId() ) rdst->InsertArc( ra );
            }

        a = n->NextAdj( Q, outArcs );
        }
    }

    // do a grid layout if needed.
    if (nodelist.GetNoItems() > 1 && !numNodesLaidOut)
        GridNodes(this, nodelist, 0, 0, editor->GetGridSize());
}

//---------------------------------------------------------------------------
// Initialize a Projection window
//---------------------------------------------------------------------------

static NodeSet projectnodes;
static NodeSet windownodes;

static void
ProjectCheck( GraphNode* node, RigiWindow* win, int maxdepth, int level ) {
    // Check to see if we have exceeded the maximum depth.

    if ((maxdepth != DEPTH_INFINITE) && (level > maxdepth)) return;

    // Add this node to the list of visited nodes.

    projectnodes.Add( node );

    // Get the first Level GraphArc leaving this node. If there are none, it
    // is a leaf node and we need to add it to the RigiWindow. We also
    // add the node to the window if it is at level maxlevel. If there
    // are Level Arcs, we recurse on them to get to the leaves.

    AvlPos *P;
    GraphArc* ap = node->FirstAdj( P, outLevArcs );
    if ( ap == NULL || level == maxdepth) {
        // Add the node to the set of nodes contained in this window.

        if (windownodes.Find(node) == NULL) {
            windownodes.Add( node );

            // Create the RigiNode for this node and add it to the windows list

            RigiNode* rn = new RigiNode( node,win );
            win->InsertNode(rn);
        }

    } else {
        // For each level arc, get the destination. If we have not visited
        // this node then recurse down that branch of the tree. If we have
        // been here before, skip this node and go on to the next one.

        while ( ap ) {
            GraphNode* dst = ap->GetDst();
            if (!projectnodes.Find(dst) )
                ProjectCheck(dst, win, maxdepth, level+1);
            ap = node->NextAdj(P, outLevArcs);
        }
    }
}

static void InitGeneralNodes( RigiWindow* rw, GraphNode* np,
ArcFlags& arctypes, ArcDirection arcdir,
int inclusive, int project, int level, NodeSet& visited ) {
    ArcDirection dir = arcdir;
    // Check to see that we haven't recursed too far

    if ( project != -1 && level > project ) return;

    // If the node is already in the list, we have been here before and
    // don't need to repeat the process.

    if ( visited.Find(np) ) return;
    visited.Add( np );

    // Check to see how many bits are set in the ArcFlags

    int ntypes = arctypes.Size();
    int type, setbits;
    GraphArc *levarc = NULL;

    for ( type=0,setbits=0; type<ntypes; type++ ) {
        if (!arctypes.GetBit(type)) continue;

        AvlPos *P; // memory leak!
        setbits++;
        if (!levarc) levarc = np->FirstArc( P, type, arcdir );
    }

    // Add the current node to the window if required. We add the node if the
    // inclusive flag is set, if the level we are at is equal to the projection
    // level for the window, and if the window is a projection of infinite
    // depth and the node is a leaf node.

    // Check to see if it is an infinite projection and a leaf node.

    int leafnode = (project==-1 && levarc==NULL && setbits>0);

    // Add the node to the window if required and it is not already
    // in the window.

    RigiNode *rn = rw->FindRigiNode(np);
    if (!rn) if ( inclusive || leafnode || level == project ) {
        RigiNode* rn = new RigiNode( np, rw );
        rw->InsertNode(rn);
    }

    // Traverse all the arcs of the specified types from the node and
    // recursively perform this function until the value of level is greater
    // than project.

    for ( type=0; type<ntypes; type++ ) {
        if (!arctypes.GetBit(type)) continue;

        AvlPos *P;
        GraphArc* ap = np->FirstArc( P, type, dir);
        while ( ap != NULL ) {
            // Get the type of the arc and check to see if that type's bit is
            // set in the arctypes flag. If so, we traverse that arc and add
            // the destination node to the window.

            int type = ap->GetType();
            if ( arctypes.GetBit( type ) ) {
                // Get the destination of the arc, create a new riginode, and
                // add the node to the windows nodelist. We add the node to
                // a NodeSet as well to make searching fast.

                GraphNode* dst = NULL;
                if ( arcdir == ARC_OUT )
                    dst = ap->GetDst();
                else if (arcdir == ARC_IN )
                    dst = ap->GetSrc();
                else {
                    if (ap->GetDst() == np)
                        dst = ap->GetSrc();
                    else if (ap->GetSrc() == np)
                        dst = ap->GetDst();
                    else mayday("Could not get dst for arc in InitGeneral");
                }

                if ( rw->FindRigiNode(dst)==NULL ) {
                    InitGeneralNodes( rw, dst, arctypes, dir, inclusive,
                        project, level+1, visited );
                }
            }
            ap = np->NextArc( P, type, dir );
        }
    }
}

void RigiWindow::InitGeneral( NodeSet& list,
ArcFlags& intypes, ArcFlags& outtypes, int incl, int proj ) {
    ArcDirection arcout = ARC_OUT;
    NodeSet visited;
    ArcSet arcset;
    GraphNode *n;
    AvlPos *P;
    RigiNodeList nodelist;
    int numNodesLaidOut = 0;

    // Save the instance variables for the window except the list of parents,
    // which is done at the next step.

    inarctypes = intypes;
    outarctypes = outtypes;
    depth = proj;
    inclusive = incl;

    // Traverse all the nodes in the list and add them to the window as
    // required.

    for ( n = list.First(P); n; n = list.Next(P) ) {
        nlist->Add( n );
        InitGeneralNodes( this, n, intypes, ARC_IN, incl, proj,
                        0, visited );
        visited.Empty();
        InitGeneralNodes( this, n, outtypes, ARC_OUT, incl, proj,
                        0, visited );
    }

    // Get the arcs for this window and while we are at it, lets set the
    // position of the RigiNodes.

    for ( RigiNode* rn = FirstNode(P); rn; rn = NextNode(P) ) {
        GraphNode *n = rn->GetNode();

        // Get the NDC position of the node and transform to local coords.

        NDC_Point2d ndcp = n->GetNDCPosition();
        Point2d p = NDC2World2d( ndcp );
        rn->SetPosition( p );
        nodelist.Insert(rn);

        // keep tally to determine if a layout is needed.
        if (p.x || p.y) numNodesLaidOut++;
        
        // For each arc leaving this node

        AvlPos *Q;
        GraphArc *a = n->FirstArc( Q, ANY_ARC, arcout );
        while ( a ) {
            // Don't add the arc to the window if it already exists, or if the
            // node does not belong to the window being processed. If it doesn't
            // already exist, create a rigi arc for it, insert it into the nodes
            // and the windows arc list.

            GraphNode *dst = a->GetDst();
            RigiNode *rdst = FindRigiNode(dst);
            if ( rdst && !arcset.Find( a->GetId() ) ) {
                arcset.Add( a );
                RigiArc* ra = new RigiArc( a, this );
                rn->InsertArc( ra );

                // If the arcs source and destination are the same then it is
                // a recursive arc and the arc should not be added to the
                // RigiNodes list again.

                if (ra->GetArc()->GetSrc()->GetId() !=
                    ra->GetArc()->GetDst()->GetId() ) rdst->InsertArc( ra );
            }
            a = n->NextArc( Q, ANY_ARC, arcout );
        }
    }

    // do a grid layout if needed.
    if (nodelist.GetNoItems() > 1 && !numNodesLaidOut)
        GridNodes(this, nodelist, 0, 0, editor->GetGridSize());
}

void RigiWindow::InitProjection( NodeSet& nodes ) {
    ArcSet arcset;
    GraphNode *n;
    AvlPos *P;
    RigiNodeList nodelist;
    int numNodesLaidOut = 0;

    if ( this == NULL )
        mayday("NULL self reference in RigiWindow::InitProjection");

    // empty the nodeset that saves the nodes that are
    // contained in this window. 

    windownodes.Empty();

    // Traverse all of the level arcs from each node in the node list to
    // the leaf nodes at that tree, and add each leaf node to the window.
    // Do my own depth first traversal so I don't do more work than I need
    // to.

    for ( n = nodes.First(P); n; n = nodes.Next(P) ) {
        // Empty the nodeset that marks the nodes that we have visited.
        projectnodes.Empty();
        ProjectCheck(n, this, depth, 0);
    }

    // Traverse the list of RigiNodes and get all of the normal arcs that
    // have a destination node in this window. This is done FAST by looking
    // up the destination RigiNode's id in the nodeset created by
    // ProjectCheck()

    RigiNode* rn;
    for ( rn = nodeset->First(P); rn; rn = nodeset->Next(P) ) {
        // Get the node associated with this RigiNode, and set its location
        // with respect to the current World location and the Window location.

        GraphNode *n = rn->GetNode();
        NDC_Point2d ndcp = n->GetNDCPosition();
        Point2d p = NDC2World2d( ndcp );
        rn->SetPosition( p );
        nodelist.Insert(rn);

        // keep tally to determine if a layout is needed.
        if (p.x || p.y) numNodesLaidOut++;
        

        // Traverse the list of outgoing normal arcs and add it to the window
        // if it's destination is in the window and it has not already been
        // added to the window.

        GraphArc *a; AvlPos *Q;
        for ( a=n->FirstAdj( Q,outArcs ); a; a=n->NextAdj( Q,outArcs ) ) {	
            // Get the destination node and see if it is in this window. If
            // so, add the arc to the window if it has not already been added.

            GraphNode* dst = a->GetDst();
            int founddst = windownodes.Find( dst ) ? 1 : 0;

            // Check to see if we have already added this one, if not add it
            // to both the window and the arcset.

            if ( founddst && !arcset.Find( a->GetId() ) ) {
                arcset.Add( a );
                RigiArc* ra = new RigiArc( a,this );
                rn->InsertArc( ra );

                ActiveNode* anp = (ActiveNode*)dst->GetActiveNode();
                RigiNode* currnp = anp->First();
                while ( currnp != NULL && currnp->GetWindow() != this ) { 
                    currnp = anp->Next();
                }
                if ( currnp == NULL ) {
                    mayday("Couldn't find node in this window");
                    mayday("Using self as dst for the RigiArc");
                } else if (ra->GetArc()->GetSrc()->GetId() !=
                    ra->GetArc()->GetDst()->GetId()) {
                    currnp->InsertArc(ra);
                }
            }
        }
    }
    // do a grid layout if needed.
    if (nodelist.GetNoItems() > 1 && !numNodesLaidOut)
        GridNodes(this, nodelist, 0, 0, editor->GetGridSize());
}

//---------------------------------------------------------------------------
// Constructors for RigiWindow.
//---------------------------------------------------------------------------

// Constructor for the general RigiWindow.
// (1)
RigiWindow::RigiWindow(NodeSet& list, ArcFlags& intypes, ArcFlags& outtypes,
int inclusive, int project, int x, int y, char *label ) {

    windownum = ++id;                                  
    windowname.Assign( "General" );
    conditionstring = NULL;

    nodeset = new RigiNodeSet;
    nlist = new NodeSet;

    // Add the list of nodes that are the parents of this window to the
    // node list and add the first node as the immediate parent.

    GraphNode* n = NULL;
    GraphNode *np; AvlPos *P;
    for ( np = list.First(P); np; np = list.Next(P) ) {
        nlist->Add( np );
        if ( !n ) n = np;
    }
    if ( n == NULL ) mayday("Empty list in RigiWindow::RigiWindow");

    // Save the instance info and set up the windows for this RigiWindow.

    RigiDomain *domain = RinitGetDomain();
    InitInstance( WIN_GENERAL, domain->NumNodeTypes(), domain->NumArcTypes());
    InitWindow( n, x, y );

    // Initialize the windows RigiNodes and RigiArcs

    InitGeneral( *nlist, intypes, outtypes, inclusive, project );

    // Scale the window and update its label.

    Mess("Scale to View");
    ScaleToView();
    UpdateLabel( label );
}

// (2)
RigiWindow::RigiWindow( NodeSet& list, ArcFlags& arctypes,
ArcDirection arcdir, int inclusive, int project, int x, int y, char *label ) {
    ArcFlags intypes, outtypes;
    windownum = ++id;
    windowname.Assign( "General" );
    conditionstring = NULL;

    nodeset = new RigiNodeSet;
    nlist = new NodeSet;

    // Add the list of nodes that are the parents of this window to the
    // node list and add the first node as the immediate parent.

    GraphNode* n = NULL;
    GraphNode *np; AvlPos *P;
    for ( np = list.First(P); np; np = list.Next(P) ) {
        nlist->Add( np );
        if ( !n ) n = np;
    }
    if ( n == NULL ) mayday("Empty list in RigiWindow::RigiWindow");

    // Save the instance info and set up the windows for this RigiWindow.

    RigiDomain *domain = RinitGetDomain();
    InitInstance( WIN_GENERAL, domain->NumNodeTypes(), domain->NumArcTypes() );
    InitWindow( n, x, y );

    // Initialize the windows RigiNodes and RigiArcs based on the type.

    intypes.SetSize( arctypes.Size() );
    intypes.ClearAll();
    outtypes.SetSize( arctypes.Size() );
    outtypes.ClearAll();
    if (arcdir == ARC_IN || arcdir == ARC_INOUT) intypes = arctypes;
    if (arcdir == ARC_OUT || arcdir == ARC_INOUT) outtypes = arctypes;
    InitGeneral( *nlist, intypes, outtypes, inclusive, project );

    // Scale the window and update its label.

    ScaleToView();
    UpdateLabel( label );
}

// Looks like the projection or overview window
// except for views, the overview window is constructed with
// RigiWindow(Graphnode *, RigiWinType, ...) (below).
// (3)
RigiWindow::RigiWindow( NodeSet& list, RigiWinType type, int depth,
int numnodes, int numarcs, int x, int y ) {
    windownum = ++id;
    if (type == WIN_OVERVIEW) 
        windowname.Assign("Overview");
    else
        windowname.Assign("Projection");
    conditionstring = NULL;

    nodeset = new RigiNodeSet;
    nlist = new NodeSet;

    // Do some paranoid error checking
    if ( this == NULL ) mayday("NULL self in RigiWindow::RigiWindow");

    // Get the first node in the list

    GraphNode* n = NULL;
    GraphNode *np; AvlPos *P;
    for ( np = list.First(P); np; np = list.Next(P) ) {
        nlist->Add( np );
        if (!n) n = np;
    }
    if ( n == NULL ) mayday("Empty list in RigiWindow::RigiWindow");

    // Save the instance info and set up the windows for this RigiWindow.
    InitInstance( type, numnodes, numarcs );
    SetDepth( depth );
    InitWindow( n,x,y );

    // Initialize the windows RigiNodes and RigiArcs based on the type.
    InitType( list, type );

    // Scale the window and update its label.
    ScaleToView();
    UpdateLabel();
}

// Looks like the project or overview window
// (4)
RigiWindow::RigiWindow( GraphNode* n, RigiWinType type, int depth,
int numnodes, int numarcs, int x, int y ) {
    windownum = ++id;
    windowname.Assign( "Overview" );
    conditionstring = NULL;

    nodeset = new RigiNodeSet;
    nlist = new NodeSet;

    // Do some paranoid error checking

    if ( this == NULL ) mayday("NULL self in RigiWindow::RigiWindow");
    if ( n == NULL ) mayday( "NULL GraphNode in RigiWindow Constructor" );
    if ( type == WIN_PROJECT ) {
        mayday("RigiWindow::RigiWindow - ");
        mayday("Should use other constructor for PROJECTIONS");
    }

    // Save the instance info and set up the windows for this RigiWindow.
    InitInstance( type, numnodes, numarcs );
    SetDepth( depth );
    InitWindow( n,x,y );

    // Initialize the window's RigiNodes and RigiArcs based on the type, as
    // well as storing the list of nodes that create this window.

    nlist->Add( n );

    NodeSet list;
    list.Add( n );
    InitType( list, type );

    // Scale the window and update its label.
    Mess("scale an OVERVIEW window");
    ScaleToView();
    UpdateLabel();
}

// this is for scratch
// (5)
RigiWindow::RigiWindow( GraphNode* n, RigiWinType type, int numnodes,
int numarcs, int x, int y) {
    windownum = ++id;
    conditionstring = NULL;

    nodeset = new RigiNodeSet;
    nlist = new NodeSet;

    if (type == WIN_SCRATCH) windowname.Assign( "Scratch" );
    else if (type == WIN_GENERAL) windowname.Assign( "General" );
    else mayday("Error: Should not call this constructor with this type");

    // Do some paranoid error checking
    if ( this == NULL ) mayday("NULL self in RigiWindow::RigiWindow");
    //if ( n == NULL ) mayday( "NULL GraphNode in RigiWindow Constructor" );
    if ( type == WIN_PROJECT ) {
        mayday("RigiWindow::RigiWindow - ");
        mayday("Should use other constructor for PROJECTIONS");
    }

    // Save the instance info and set up the windows for this RigiWindow.
    InitInstance( type, numnodes, numarcs );
    SetDepth( DEPTH_INFINITE );
    InitWindow( n,x,y );

    // Initialize the windows RigiNodes and RigiArcs based on the type, as
    // well as storing the list of nodes that create this window.

    if (n) nlist->Add( n );
    NodeSet list;
    if (n) list.Add( n );
    InitType( list, type );

    // Scale the window and update its label.
    ScaleToView();
    UpdateLabel();
}

//---------------------------------------------------------------------------
// Initialize the instance variables, the window, and type specific stuff.
//---------------------------------------------------------------------------

void RigiWindow::InitInstance( RigiWinType t, int nnode, int narcs ) {
    type = t;
    nodeFilters = new FilterSet(nnode);
    arcFilters = new FilterSet(narcs);
    xscale = yscale = 1.0F;
    canOrig.x = canOrig.y = 0;
}

void RigiWindow::InitWindow( GraphNode* n, int xloc, int yloc ) {
    int w=0,h=0;

    RutilGetMaxCanvasSize( WINSIZEX, WINSIZEY );
    RutilGetDefFrameSize( w, h );

    // If this is the basenode, create the window as the base frame. If not,
    // get the window from the node pointer n, and use its frame as the parent
    // frame for this window.

    RutilGetMaxCanvasSize( WINSIZEX, WINSIZEY );
    window = new TKWindow( NULL, 0,0, WINSIZEX,WINSIZEY, xloc,yloc );

    UpdateFilterMessage();
    UpdateArcMessage(RinitGetDomain()->ArcNameFromType( editor->GetArcType()));
}

void RigiWindow::InitType( NodeSet& nlist, RigiWinType type ) {
    GraphNode* n = NULL;
    GraphNode *np;
    AvlPos *P;
    int num = 0;

    for ( np = nlist.First(P); np; np = nlist.Next(P) ) {
        n = np;
        num++;
    }

    switch ( type ) {
    case WIN_OVERVIEW:
        if (num <= 1 || RigiError(ALERT, "One node used as root for OVERVIEW window"))
            InitOverview( n );
        break;
    case WIN_PROJECT:
        InitProjection( nlist );
        break;
    case WIN_SCRATCH:
        if (num <= 1 || RigiError(ALERT, "One node used as root for SCRATCH window"))
            InitScratch( n );
        break;
    case WIN_GENERAL:
        {
        // Should not normally be called through InitType. This way
        // we just set up a bunch of default stuff.
        ArcFlags intypes, outtypes;
        outtypes.SetBit( editor->LevelArc() );
        InitGeneral( nlist, intypes, outtypes, 0, 1 );
        break;
        }
    default:
        RigiError(ERRORFLAG, "Unknown type in RigiWindow::RigiWindow()");
        break;
    }
    UpdateFilterMessage();
}

// Destructor

RigiWindow::~RigiWindow() {
    Mess("RigiWindow::~RigiWindow %d", windownum);
    if (this==NULL) {
        RigiError(ERRORFLAG, "NULL RigiWindow");
        return;
    }

    // Delete all of the RigiArcs and RigiNodes associated with this window
    Empty();
    delete nlist;
    delete nodeset;

    delete nodeFilters; 
    delete arcFilters; 

    Mess("Deleting window");
    delete window;
    Mess("Done Destructor");
}

void RigiWindow::DestroyChildWindows() {
    window->DestroyChildWindows();
}

// UpdateLabel. When first arg non-null, update typename. When second
// arg non-null, update the "conditionstring" (eg. "ACTIVE" or not).
// Update the scale percentages regardless.

void RigiWindow::UpdateLabel( char *newlabel, char *condition ) {
    char label[MAXTITLE + 1];

    int xscale = (int)(100*GetXScale());
    int yscale = (int)(100*GetYScale());

    // Assign the name to the instance variable
    if (newlabel) windowname.Assign( newlabel );

    const char *name = GetNode() ? GetNode()->GetLabel() : "<no node>";

    CharType str(windowname.string());
    sprintf(label, " - %d ", windownum);
    str.Append(label);

    switch (GetType()) {
    case WIN_PROJECT:
    case WIN_GENERAL:
        {
        int count = 0;
        AvlPos *P;
        for ( GraphNode *n = nlist->First(P); n; n = nlist->Next(P) ) {
            if (count) str.Append("/");
            str.Append(n->GetLabel());
            count++;
        }
        break;
        }
    case WIN_OVERVIEW:
    case WIN_SCRATCH:
        str.Append(name);
        break;
    default:
        mayday("Unknown type in RigiWindow::UpdateLabel()");
        break;
    }

    if (condition) {
        if (conditionstring) free(conditionstring);
        conditionstring = strdup(condition);
    }

    // calculate the length of str to copy into the label string.
    int i = (conditionstring == NULL)? 0 : strlen(conditionstring) + 2;
    RigiBool ovf = RigiFalse;
    if (i + str.Length() > MAXTITLE) {
        i = MAXTITLE - i - 3;
        ovf = RigiTrue;
    } else {
        i = str.Length();
    }
    if (i < 0) i = 0;
    strncpy(label, str.string(), i);
    label[i] = 0;
    if (ovf) strcat(label, "...");
    if (conditionstring != NULL) {
        strcat(label, "  ");
        strncat(label, conditionstring, MAXTITLE - strlen(label));
    }

    window->SetTitle(label);
}

void RigiWindow::Empty() {
    // Delete all of the riginodes associated with this window
    // Delete the rigiarcs along the way

    Mess("deleting %d riginodes",nodeset->Size());
    RigiNode *rn; RigiArc *ra; AvlPos *P, *Q;

    for ( rn = nodeset->First(P); rn; rn = nodeset->Next(P) ) {
        for ( ra=rn->FirstArc(Q,ARC_OUT); ra; ra=rn->NextArc(Q,ARC_OUT) ) {
            ra->SetNull();   // Mark rigiarc as defunct
            delete ra;       // delete without worrying about nodes' lists
        }
        delete rn;
    }
    nodeset->Empty();
}

void RigiWindow::Update() {
    Empty();
    // Reinitialize the rigi window stuff.
    switch ( type ) {
    case WIN_OVERVIEW:
        InitOverview( GetNode() );
        break;
    case WIN_PROJECT:
        InitProjection( *nlist );
        break;
    case WIN_SCRATCH: InitScratch( GetNode() );
        break;
    case WIN_GENERAL:
        InitGeneral( *nlist, inarctypes, outarctypes, inclusive, depth );
        break;
    default:
        break;
    }

    if (editor->GetScaleType() == SCALE_WINDOW) {
        Point p;
        p.x = p.y = p.z = 0;
        RigiDisplay(0, p, p);
    } else {
        RigiDisplay();
    }
}

//----------------------------------
// Hacks for efficient drawing
//----------------------------------

void RigiWindow::Dimensions( int*x, int*y, int*w, int*h ) {
    if (w||h) window->GetDrawSize( w,h );
    if (x) *x = canOrig.x;
    if (y) *y = canOrig.y;
}


// B23: new rigidisplay which gets the current scaling function
// before displaying the screen
// I'm not certain that scaling should be independent of window
// type -- for example, is an overview window scalable by factor?  JGM.

void RigiWindow::RigiDisplay( int boxflag, Point p1, Point p2 ) {
    switch (editor->GetScaleType()) {
    case SCALE_NONE:
        break;
    case SCALE_WINDOW:
        ScaleToView();
        break;
    case SCALE_BOX:
        if (boxflag) ScaleToBox( p1, p2 );
        break;
    case SCALE_ONE:
        SetXScale( 1.0F );
        SetYScale( 1.0F );
        break;
    case SCALE_FACTOR:
        SetXScale( GetXScale() * editor->GetScaleFactor() / 100.0F);
        SetYScale( GetYScale() * editor->GetScaleFactor() / 100.0F);
        break;
    default:
        break;
    }
    RigiDisplay();
}

void RigiWindow::RigiDisplay() {
    Mess("RigiDisplay");

    // If the window has no drawing complexity, then set it
    if (!window->GetComplexity())
        window->SetComplexity(ShowedNodeNumber() * 2);

    // if Drawing is disabled, quit right away
    if (window->Drawing())
      return;

    WindowSetCursor( CURSOR_WAIT );
    Batch( RigiTrue );
    BunchNode( RigiTrue );
    BunchArc( RigiTrue );
    Clear();
    DrawAll();
    BunchArc( RigiFalse );
    BunchNode( RigiFalse );
    Batch( RigiFalse );
    WindowSetCursor( CURSOR_NORMAL );
}

void RigiWindow::DrawStore() {
    // If the window has no drawing complexity, then set it
    if (!window->GetComplexity())
        window->SetComplexity(ShowedNodeNumber() * 2);

    // if Drawing is disabled, quit right away
    if (window->Drawing())
      return;

    WindowSetCursor( CURSOR_WAIT );
    BackStore( RigiTrue ); 
    BunchNode( RigiTrue );
    BunchArc( RigiTrue );
    Clear();
    DrawAll();
    BunchArc( RigiFalse );
    BunchNode( RigiFalse );
    BackStore( RigiFalse ); 
    WindowSetCursor( CURSOR_NORMAL );
}

// Return the size of the RigiNode given in cavas coordinate space. This
// is not trivial because of the limit placed on the size of the nodes when
// they are drawn onto the canvas.

int RigiWindow::WindowNodeExtents( RigiNode *rnp, Point *minp, Point *maxp ) {
    RigiWindow *rw = rnp->GetWindow();
    if (rw != this)
        mayday("RigiWindow::WindowNodeExtents - window!=this, node = %s",
                rnp->GetNode()->GetLabel());

    // Get the opposite corners of the node based on the node's size. These
    // locations are in canvas coordinates and have been scaled by the
    // cavas's scale factors.

    int window_size = MAXNODESIZE;

    int xsize = (int)(window_size * xscale); 
    window_size = (int)(window_size * yscale);
    window_size = (window_size<xsize) ? window_size : xsize;

    // Make the node's size the minimum of the x and y sizes after scaling
    // to keep the node square. Also make sure the node size is between
    // the minimum and maximum node sizes.

    window_size = (window_size>MAXNODESIZE) ? MAXNODESIZE : window_size;
    window_size = (window_size<MINNODESIZE) ? MINNODESIZE : window_size;

    // update the node's size
    rnp->SetNodeSize( window_size );

    // Calculate the maximum point of the node based on the new size

    Point2d p = rnp->GetPosition();
    *minp = rw->World2Canvas(p);
    maxp->x = minp->x + window_size;
    maxp->y = minp->y + window_size;

    // Return the node's size on the canvas

    return window_size;
}

// Return the size of the RigiNode given in world coordinate space. This
// is not trivial because of the limit placed on the size of the nodes when
// they are drawn on the canvas.

void RigiWindow::NodeExtents( RigiNode *rnp, Point *minp, Point *maxp ) {
    RigiWindow *rw = rnp->GetWindow();
    if (rw != this)
        mayday("RigiWindow::NodeExtents - nodes window != this");

    // Calculate the node extents in canvas coordinates

    WindowNodeExtents( rnp, minp, maxp );

    // Convert the window coordinates into world coordinates

    *minp = Canvas2World( *minp );
    *maxp = Canvas2World( *maxp );
}

void RigiWindow::DrawNode( RigiNode *rnp) {
    // Check to see if doing redundant drawing 

    int active = (rnp->GetState()==ACTIVE);
    int name = (rnp->HideName()==RigiFalse);

    // return if not drawing this type of node in this window
    if ( !CheckNode( rnp ) ) return;

    // Get some info about this node
    GraphNode* node = rnp->GetNode();
    int type = node->GetType();

    // Get the opposite corners of the node and the node's size. These
    // locations are in window coordinates and have been scaled by the
    // window's scale factors.

    Point window_loc, window_loc2;
    int window_size = WindowNodeExtents( rnp, &window_loc, &window_loc2 );

    // Use this for normal hintless erase/draw
    window->DrawBox( window_loc, window_loc2, (active?BIG:SMALL), type, 0 );

    // Draw the node's name if desired
    if (name) {
        // Get the node label and return if it does not exist

        char* str = node->GetLabel();
        if (!str) return;


        // Calculate point on node where the text centre should be in window
        // coordinate space (ie. bottom/middle of the node).

        Point text_loc;
        text_loc.x = window_loc.x + (int)(window_size/2); 
        text_loc.y = window_loc.y + window_size; 

        // Draw the string in the window, giving text_loc as the
        // TOP CENTER of the string

        window->DrawString( str, text_loc, T_TOPCENTER );
    }
}

void RigiWindow::DrawIcon( Point p, int size, int type, int active ) {
    // Transform the world location to a local window location
    Point l = World2Canvas(p);
    p.x += size;
    p.y += size;
    p.z += size;
    p = World2Canvas(p);

    // Scale the icon size to minimum of dimensional scales
    size = p.x - l.x;
    int y = p.y - l.y;
    size = (size<y) ? size : y;
    size = (size>MAXNODESIZE) ? MAXNODESIZE : size;

    // Draw the icon
    window->DrawIcon( l, size, type, active );
}

void RigiWindow::DrawBox( Point sp, Point dp, int thick, int type, int dpyOnly ) {
    // Transform the world location to a local window location
    Point s = World2Canvas(sp);
    Point d = World2Canvas(dp);

    // Draw the box
    window->DrawBox( s, d, thick, type, dpyOnly );
}

void RigiWindow::DrawSquare( Point p, int size, int thick, int type ) {
    Point d;
    // Transform the world location to a local window location
    p = World2Canvas(p);

    // Ensure that box is square in the generic window
    // A non-positive size means to interpret as pixels, no scaling
    size = size<=0 ? -size :
        (int)(size * ((xscale<yscale) ? xscale : yscale ));
    d.x = p.x + size;
    d.y = p.y + size;

    // Draw the box
    window->DrawBox( p, d, thick, type, 0 );
}

Point RigiWindow::InArcPoint( RigiNode *rnp ) {
    // Calculate the node extents in the window coordinate system

    Point window_loc, window_loc2;
    int window_size = WindowNodeExtents( rnp, &window_loc, &window_loc2 );

    // Calculate the incoming arc point and translate that back to world
    // coordinates.

    window_loc.x += (int)(window_size/2);
    window_loc = Canvas2World(window_loc);
    return window_loc;
}

Point RigiWindow::OutArcPoint( RigiNode *rnp ) {
    // Calculate the node extents in the window coordinate system

    Point window_loc, window_loc2;
    int window_size = WindowNodeExtents( rnp, &window_loc, &window_loc2 );

    // Calculate point on node where the text centre should be in window
    // coordinate space (ie. bottom/middle of the node).

    Point text_loc;
    text_loc.x = window_loc.x + (int)(window_size/2);
    text_loc.y = window_loc.y + window_size;

    // If not hiding the node's name, calculate offset for text as well as
    // for the node. The calculation is done in Window space, since the
    // text is not scalable. The extra yoffset/4 factor is to take care of
    // characters that have letters descending below the text's baseline.

    if (!rnp->HideName()) {
        int yoffset;
        window->TextExtent( "Aj", NULL, NULL, &yoffset );
        yoffset += yoffset/4;
        text_loc.y += yoffset;
    }

    // Transform the Window coordinates back to the World coordinate system
    // and return the calculated location

    text_loc = Canvas2World(text_loc);
    return text_loc;
}

void RigiWindow::DrawArc( RigiArc *rap ) {
    RigiNode *src = rap->GetSrc();
    RigiNode *dst = rap->GetDst();
    int style;

    // Get the position for the in and out arc points for the destination and
    // source nodes respectively

    Point sp = OutArcPoint( src );
    Point dp = InArcPoint( dst );

    // Translate the source and destination positions to window coordinates

    sp = World2Canvas(sp);
    dp = World2Canvas(dp);

    // Draw the arc, depending on its type and its consistency state

    int width = (rap->GetArc()->GetConsistent() ? SMALL : THICK);

    if ( rap==editor->GetSelectedArc() ) width = BIG;	// HACK!
    style = GetArcType( rap );

extern int XMJW;
    if (XMJW==3) {
        float cx, cy;
        float dx, dy;
        dx = dp.x - (float)sp.x;
        dy = dp.y - (float)sp.y;

        cx = (float)sp.x + 0.5F * (float)(dp.x - sp.x);
        cy = (float)sp.y + 0.5F * (float)(dp.y - sp.y);
        Point p, cp;
        float distx, disty, d;
        p.x = (int)(cx - (dy));
        p.y = (int)(cy + (dx));
        cp.x = (int)(cx);
        cp.y = (int)(cy);
        distx = (float)(p.x - cp.x);
        disty = (float)(p.y - cp.y);
        d = (float)sqrt(distx * distx + disty * disty);
        d = 10.0F/d;
        if (d > 10.0) {
            d = d / 10.0F;
            dx /= d;
            dy /= d;
        }
        p.x = (int)(cx - (dy));
        p.y = (int)(cy + (dx));
        cp.x = (int)(cx);
        cp.y = (int)(cy);
        window->DrawLine( sp, p, width, style);
        window->DrawLine( p, dp, width, style);
    } else {
        window->DrawLine( sp, dp, width, style);
    }
}

void RigiWindow::DrawLine( Point s, Point d, int width, int style ) {
    // Transform the world location to the local window space
    s = World2Canvas(s);
    d = World2Canvas(d);

    // Draw the line
    window->DrawLine( s, d, width, style );
}

// Draws string str, with the its bounding box top centre at point p.
// It does NOT use the bottom left corner of the bounding box as most
// text rendering algorithms do, so the location of the text is not
// skewed by any transformations, since the text is not scalable.

void RigiWindow::DrawString( char* str, Point p ) {
    // Get the offsets for the size of the text
    //TextExtent( str, &xoffset, NULL, &yoffset );

    // Transform the point to window coordinates
    p = World2Canvas(p);

    // Compute the bottom left hand corner of the bounding box of the text,
    //p.x -= xoffset/2;
    //p.y += yoffset;

    // Draw the string in the window
    window->DrawString( str, p );
}

RigiArc* RigiWindow::FindArc(Point loc) {
    if (this == NULL) mayday("NULL self in RigiWindow::FindArc");

    // scale the point to pixel coordinates
    loc = World2Canvas(loc);

    // Scan the list of RigiArcs for one close to loc, if any.
    // use the closest distance from a point to a line.

    RigiNode *rn; RigiArc *ra; AvlPos *P, *Q;
    for (rn=nodeset->First(P); rn; rn=nodeset->Next(P)) {
        for (ra = rn->FirstArc(Q,ARC_OUT); ra; ra = rn->NextArc(Q,ARC_OUT)) {
            if (!CheckArc(ra)) continue;
            Point s, d, tl, br;

            s = OutArcPoint(ra->GetSrc());
            d = InArcPoint(ra->GetDst());

            // the source and destination
            // to pixel coordinates
            s = World2Canvas(s);
            d = World2Canvas(d);

            // calculate the bounding box of the arc
            tl.x = (s.x < d.x)? s.x : d.x;
            tl.y = (s.y < d.y)? s.y : d.y;
            br.x = (s.x > d.x)? s.x : d.x;
            br.y = (s.y > d.y)? s.y : d.y;
            tl.x -= MAXARCDISTANCE;
            tl.y -= MAXARCDISTANCE;
            br.x += MAXARCDISTANCE;
            br.y += MAXARCDISTANCE;

            // check the basic ranges of the coordinates
            // point must be inside the bounding square enclosing the arc
            if (loc.x < tl.x || loc.y < tl.y ) continue;
            if (loc.x > br.x || loc.y > br.y ) continue;

            // check the distance of the line to the point    
            double m = (s.x == d.x)? 2.0E+10 : ((double)s.y - d.y)/(s.x - d.x);
            double b = (double)s.y - m * s.x;
            double distance = ((double)loc.y - m * loc.x - b) / pow((m * m + 1.0), 0.5);
            if (distance < 0.0) distance = - distance;
            if (distance <= MAXARCDISTANCE) return ra;
        }
    }
    return NULL;
}

int RigiWindow::ArcNumber( int type ) {
    int number = 0;
    RigiNode *rn; RigiArc *ra; AvlPos *P, *Q;

    for ( rn=nodeset->First(P); rn; rn=nodeset->Next(P) )
    for ( ra=rn->FirstArc(Q,ARC_OUT); ra; ra=rn->NextArc(Q,ARC_OUT) )
        if ( ra->GetArc()->GetType() == type  ) number++;
    return number;
}


// MJW HACK! Compute total_arcs as a side effect.
// Counting by type and filter condition needs work.

int RigiWindow::ShowedArcNumber() {
    int i=0,j=0; AvlPos *P, *Q; RigiNode *rn; RigiArc *ra;

    for ( rn=nodeset->First(P); rn; rn=nodeset->Next(P) )
    for ( ra=rn->FirstArc(Q,ARC_OUT); ra; ra = rn->NextArc(Q,ARC_OUT) )
        { j++; if ( CheckArc(ra) ) i++; }
    total_arcs = j;
    return i;
}

int RigiWindow::ShowedArcNumber( int type )
{
    int i=0,j=0; AvlPos *P, *Q; RigiNode *rn; RigiArc *ra;

    for ( rn=nodeset->First(P); rn; rn=nodeset->Next(P) )
    for ( ra=rn->FirstArc(Q,ARC_OUT); ra; ra = rn->NextArc(Q,ARC_OUT) )
        { j++; if ( ra->GetArc()->GetType()==type && CheckArc(ra) ) i++; }
    total_arcs = j;
    return i;
}

int RigiWindow::GetArcType( RigiArc *ra ) {
  if( ra->GetArc()->GetType() != editor->CompositeArc() ) {
    return( ra->GetArc()->GetType() );
  } else {
    int i, iFound = 0, iType;
    
    for( i=0; i < ra->GetArc()->NumContributorIds(); i++ ) {
      int iTmp_type = editor->FindArc( ra->GetArc()->GetContributorId( i ) )->GetType();

      if( !arcFilters->GetFilter( iTmp_type ) ) {
	if( iFound
	    && iTmp_type != iType ) {
	  // Different contributing type found, just call this arc Composite.
	  return( editor->CompositeArc() );
	}
	iFound = 1;
	iType = editor->FindArc(ra->GetArc()->GetContributorId( i ) )->GetType();
      }
    }
    return( iType );
  }
}

int RigiWindow::CheckArc( RigiArc *ra ) {
    // Check filters of arc itself, and its src and dst, to see if drawable

    if ( GetType() != WIN_OVERVIEW && 
        arcFilters->GetFilter((ra->GetArc()->GetType()))) return RigiFalse;

    if( ra->GetArc()->GetType() == editor->CompositeArc() ) {
      int i, iFound = 0;

      for( i=0; i < ra->GetArc()->NumContributorIds(); i++ ) {
	if( !arcFilters->GetFilter(editor->FindArc( ra->GetArc()->GetContributorId( i ) )->GetType()) ) {
	  iFound = 1;
	  break;
	}
      }
      if( !iFound ) {
	return( RigiFalse );
      }
    }

    RigiNode* src = ra->GetSrc();
    RigiNode* dst = ra->GetDst();

    if ( src->GetFilter() || dst->GetFilter() ) return RigiFalse;
    if ( src->HideOutArcs() || dst->HideInArcs() ) return RigiFalse;

    int srctype = src->GetNode()->GetType();
    int dsttype = dst->GetNode()->GetType();

    if (nodeFilters->GetFilter(srctype)) return RigiFalse;
    if (nodeFilters->GetFilter(dsttype)) return RigiFalse;
    return RigiTrue;
}

RigiNode* RigiWindow::FindNode( Point loc ) {
    int x = loc.x;
    int y = loc.y;

    if (this == NULL) mayday("NULL self in RigiWindow::FindNode");

    // scan list of RigiNodes for one at loc, if any

    AvlPos *P;
    for ( RigiNode* rn = nodeset->First(P); rn; rn = nodeset->Next(P) ) {
        Point2d p = rn->GetPosition();
        int size = rn->NodeSize();

        // Get the opposite corners of the node based on the node's size. These
        // locations are in window coordinates and have been scaled by the
        // windows scale factors.

        Point win_min = World2Canvas(p);
        p.x += size;
        p.y += size;
        Point win_max = World2Canvas(p);

        // Make the nodes size the minimum of the x and y sizes after scaling
        // to keep the node square. Also make sure the node size is between
        // the minimum and maximum node sizes.

        int window_size = win_max.x - win_min.x;
        int ydiff = win_max.y - win_min.y;
        window_size = (window_size<ydiff) ? window_size : ydiff;
        window_size = (window_size>MAXNODESIZE) ? MAXNODESIZE : window_size;
        window_size = (window_size<MINNODESIZE) ? MINNODESIZE : window_size;

        // Compute new max size based on window_size

        win_max.x = win_min.x + window_size;
        win_max.y = win_min.y + window_size;
        win_max.z = win_min.z + window_size;

        // Convert both min and max back to world coordinates

        win_min = Canvas2World( win_min );
        win_max = Canvas2World( win_max );

        if ( x >= win_min.x && x <= win_max.x &&
            y >= win_min.y && y <= win_max.y && CheckNode( rn ) ) {
            return rn;
        }
    }
    return NULL;
}

int RigiWindow::CheckNode( RigiNode *n ) {
    int nodeord = n->GetNode()->GetType();
    return !nodeFilters->GetFilter(nodeord) && !n->GetFilter();
return !n->GetFilter();
}

int RigiWindow::NodeNumber( int type ) {
    RigiNode *rn; int i=0; AvlPos *P;
    for ( rn=nodeset->First(P); rn; rn=nodeset->Next(P) )
        if ( rn->GetNode()->GetType() == type ) i++;
    return i;
}

int RigiWindow::ShowedNodeNumber() {
    RigiNode *rn; int i=0; AvlPos *P;
    for ( rn=nodeset->First(P); rn; rn=nodeset->Next(P) )
        if ( CheckNode( rn ) ) i++;
    return i;
}

int RigiWindow::ShowedNodeNumber( int type ) {
    RigiNode *rn; int i=0; AvlPos *P;
    for ( rn=nodeset->First(P); rn; rn=nodeset->Next(P) )
        if ( rn->GetNode()->GetType() == type )
        if ( CheckNode( rn ) ) i++;
    return i;
}

void RigiWindow::SetNumFilteredNodes( int num ) {
    filtered_nodes = num;
}

int RigiWindow::GetNumFilteredNodes() {
    return filtered_nodes;
}

void RigiWindow::SetNumFilteredArcs( int num ) {
    filtered_arcs = num;
}

int RigiWindow::GetNumFilteredArcs() {
    return filtered_arcs;
}


////////////////////////////////////
// Update the message bar routines
////////////////////////////////////

void RigiWindow::UpdateFilterMessage() {
    // Update the filter info for this window, and the message bar

    // MJW Yikes! O(n) crappola - HACK!
    SetNumFilteredNodes( nodeset->Size() - ShowedNodeNumber() );
    filtered_arcs = ShowedArcNumber(); //Ak! Compute total as side effect
    filtered_arcs = total_arcs - filtered_arcs;

    // Build the message

    CharType msg;
    char str[80];
    msg.Assign(arc_message.string());
    sprintf(str,"    Filtered: %d node%s, %d arc%s",
                filtered_nodes,filtered_nodes==1?"":"s",
                filtered_arcs,filtered_arcs==1?"":"s" );

    filter_message.Assign(str);
    msg.Append(str);
    window->SetMessage(msg.string());
}

void RigiWindow::UpdateArcMessage( char *str ) {
    CharType msg;

    arc_message.Assign("ArcType: ");
    arc_message.Append(str);

    msg.Assign(arc_message.string());
    msg.Append(" ");
    msg.Append(filter_message.string());
    window->SetMessage(msg.string());
}

int       RigiWindow::NodeNumber()               { return nodeset->Size(); }
int       RigiWindow::InsertNode( RigiNode *rn ) { return nodeset->Insert(rn); }
RigiNode* RigiWindow::RemoveNode( RigiNode *rn ) { return nodeset->Remove(rn); }
RigiNode* RigiWindow::FindNode( RigiNode *rn )   { return nodeset->Find(rn); }
RigiNode* RigiWindow::FirstNode( AvlPos*&P )     { return nodeset->First(P); }
RigiNode* RigiWindow::NextNode( AvlPos*&P )      { return nodeset->Next(P); }

RigiNode *RigiWindow::FindRigiNode( GraphNode *n ) {
    if (!n) return NULL;
    ActiveNode *an = (ActiveNode*)(n->GetActiveNode());
    if (!an) return NULL;
    return an->FindRigiNode(this);
}

RigiArc *RigiWindow::FindRigiArc( GraphArc *n ) {
    if (!n) return NULL;
    ActiveArc *an = (ActiveArc*)(n->GetActiveArc());
    if (!an) return NULL;
    return an->FindRigiArc(this);
}

void RigiWindow::CreateRigiArcs( RigiNode *rn ) {
    RigiNodeList rnl; rnl.Insert(rn); CreateRigiArcs( &rnl );
}

void RigiWindow::CreateRigiArcs( RigiNodeList *rnl ) {
    ArcSet arcs;
    RigiNode *rn;
    for ( rn = rnl->First(); rn; rn = rnl->Next() ) {
        GraphArc *a;
        AvlPos *P;
        GraphNode *n = rn->GetNode();
        for ( a=n->FirstAdj(P,inArcs); a; a=n->NextAdj(P,inArcs) ) {
            GraphNode *src = a->GetSrc();
            RigiNode *rn1 = FindRigiNode(src);

            if (rn1) if (!arcs.Find(a->GetId())) {
                arcs.Add( a );
                RigiArc* ra = new RigiArc( a,this );

                ra->Draw();
                rn->InsertArc( ra );

                // do it once for self (recursive) arcs
                if (rn != rn1) rn1->InsertArc( ra );
            }
        }
        for ( a=n->FirstAdj(P,outArcs); a; a=n->NextAdj(P,outArcs) ) {
                GraphNode *src = a->GetDst();
            RigiNode *rn1 = FindRigiNode(src);
    
            if (rn1) if (!arcs.Find(a->GetId()))
                {
                arcs.Add( a );
                RigiArc* ra = new RigiArc( a,this );

                ra->Draw();
                rn->InsertArc( ra );

                // do it once for self (recursive) arcs
                // (a redundant check here, I believe)
                if (rn != rn1) rn1->InsertArc( ra );
            }
        }
    }
}

void RigiWindow::InheritTypeFilters(const RigiWindow *baseWin) {
    *nodeFilters = *baseWin->nodeFilters;
    *arcFilters = *baseWin->arcFilters;
    UpdateFilterMessage();
}
