//
// $Author: jmartin $
// $Date: 2003-02-13 13:41:19 $
// $Source: /root/cvs/Rigi/rigiedit/cl_tk_window.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_tk_window.C,v 1.59 2003-02-13 13:41:19 jmartin Exp $";

//------------------------------------------------------------------------

#include <stream.h>
#include <math.h>
#include <string.h>

#ifdef __OS2__
#define INCL_WININPUT
#include <os2.h>
extern "C" {
#include <X11/Xlib.h>
}
#else
#if defined(__WINDOWS__) || defined(__WIN32__)
#include <windows.h>
#endif
#endif

#include <tcl.h>
#include <tk.h>

// tkInt.h has a List struct that conflicts with cl_list.h's List class
#define List ___TkList
#include <tkInt.h>
#undef List

#include "cl_tk_window.h"   // Include the class definition
#include "cl_domain.h"
#include "rcl_commands.h"
#include "rigi_init.h"
#include "tk_colormap.h"

#include "global.h"		    // globals from rigi.c
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"	 	// interface for the error functions
#endif
#ifndef LIST_H
#include "cl_list.h"        // for List class
#endif
#ifndef ENVIRON_H
#include "environ.h"
#endif
#include "win_util.h"       // For BIG and SMALL drawing

// Backing store tuning parameters
static const unsigned minBSReqmnts = 20;    // # of draw objs needed to request BS
static RigiBool localServer = RigiTrue;     // server and client on same machine

// Static variables that get set up when the first window is set up

static unsigned long foreground, background;
static GC gcdefault;
static Tk_Font font = 0;             
static char *fontName = NULL;            
static int noDraw = 0;                      // disable all drawing
static List tkWinList;                      // list of all TKWindow instances
static Tk_Cursor waitCursor = NULL;            // wait cursor handle
static Tk_Cursor arrowCursor = NULL;           // arrow cursor handle
static RigiBool cursorStateWait = RigiFalse; // current state of cursor
static Display *dpy;                        // display pointer

// keycodes (for motif)
#ifndef __WINDOWS__
static int KeyCodeLChar, KeyCodeLBit;
static int KeyCodeRChar, KeyCodeRBit;
#endif

#ifdef __WIN32__

static void XDrawRectangle(Display* dpy, Window window, GC gcdefault, int x, int y, int width, int height) {
  XDrawLine( dpy, window, gcdefault, x, y, x + width, y );
  XDrawLine( dpy, window, gcdefault, x + width, y, x + width, y + height );
  XDrawLine( dpy, window, gcdefault, x, y + height, x + width, y + height );
  XDrawLine( dpy, window, gcdefault, x, y + height, x, y );
}

static void XFillRectangle(Display* dpy, Window window, GC gcdefault, int x, int y, int width, int height) {
  for (int i = 0; i < height; i++) 
    XDrawLine( dpy, window, gcdefault, x, y + i, x + width, y + i);
}

#endif

// Deletes any event handler that matches the standard canvas event masks
// Returns RigiTrue if successful.

// apparantly, if this isn't done, tk overwrites the contents of the window
// with what it thinks should be in it. Since Rigi is doing it's own drawing
// (?!?), this is not what we want. (Johannes Martin, 11/2002)

static RigiBool DeleteCanvasEventHandler(Tk_Window hWnd) {
    TkWindow *winPtr = (TkWindow *)hWnd;
    const unsigned long EventMask = ExposureMask|StructureNotifyMask|FocusChangeMask;
    TkEventHandler *procPtr = winPtr->handlerList;
    for (; procPtr != NULL; procPtr = procPtr->nextPtr) {
            if (procPtr->mask == EventMask) {
                    Tk_DeleteEventHandler(hWnd, procPtr->mask, procPtr->proc, procPtr->clientData);
                    return RigiTrue;
            }
    }
    return RigiFalse;
}

static RigiBool ResetBlackCanvas() {
    RigiColor canvasCol;
    WindowForegroundColor(&canvasCol);
    if (canvasCol.pixel) {
        XSetForeground(dpy, gcdefault, canvasCol.pixel);
        return RigiTrue;
    }
    return RigiFalse;
}


// windw and winh give outside dimensions of window
TKWindow::TKWindow( BaseWindow* /*parent*/,
    int windw, int windh, int draww, int drawh, int locx, int locy ) :
    BaseWindow("tk")
{
    static int tkwindow_init = 0;
    int id = GetId();
    int fw, fh, cw, ch;

    char winname[80];
    char cmd[160];
    char name[80];

    iconified = RigiFalse;
    drawBatch = RigiFalse;
    drawWindow = RigiFalse;
    drawRealTime = RigiTrue;
    winOrig.x = winOrig.y = 0;
    InitExtent();
    sprintf(winname, ".window%d",id);
    RutilGetDefFrameSize(fw, fh);
    RutilGetDefWindowSize(cw, ch);
    canw = draww;
    canh = drawh;
    if (!windw) windw = fw;
    if (!windh) windh = fh;
    cw += windw - fw;
    ch += windh - fh;
    sprintf(cmd, "canvas_build %s %d \"\" %d %d %d %d %d %d",
            winname, id, cw, ch, draww, drawh, locx, locy);
    exec_command( cmd );
    tkwin = Tk_NameToWindow(rcl_interp, winname, Tk_MainWindow(rcl_interp));
    if (tkwin == NULL) mayday("NULL base window for tk window");
    Tk_MapWindow(tkwin);

    Tk_Window xscrollwin, yscrollwin;
    sprintf(name, "%s.frame.canvas", Tk_PathName(tkwin));
    canvaswin = Tk_NameToWindow(rcl_interp, name, tkwin);
    sprintf(name, "%s.frame.xscroll", Tk_PathName(tkwin));
    xscrollwin = Tk_NameToWindow(rcl_interp, name, tkwin);
    sprintf(name, "%s.frame.yscroll", Tk_PathName(tkwin));
    yscrollwin = Tk_NameToWindow(rcl_interp, name, tkwin);
    if (canvaswin == NULL || xscrollwin == NULL || yscrollwin == NULL) {
        mayday("Null canvaswin || xscrollwin || yscrollwin in DrawBox");
        return;
    }

    if (!DeleteCanvasEventHandler(canvaswin))
      mayday("default canvas event handler is not deleted");
    Tk_MapWindow(canvaswin);

    if (!Tk_IsMapped(canvaswin)) Mess("canvas is not mapped");

    // Get the X related stuff from the TK window

    Window window = Tk_WindowId(canvaswin);

    // If this is the first window, initialize the global GC info

    if (!tkwindow_init) {
        XGCValues gcv;
        unsigned long mask;

        dpy = Tk_Display(canvaswin);

        // Get colour resources of the canvas to store in a GC

        RigiColor col;
        WindowForegroundColor(&col);
        foreground = col.pixel;
        WindowBackgroundColor(&col);
        background = col.pixel;

        // Set up the GC for drawing lines
        gcv.background = background;
        gcv.foreground = foreground;
        gcv.function = GXxor;
        gcv.line_style = LineSolid;
        mask = GCForeground|GCBackground|GCLineStyle|GCFunction;
        gcdefault = Tk_GetGC(canvaswin, mask, &gcv);

        // Set up the GC for drawing strings


        waitCursor = Tk_GetCursor(rcl_interp,tkwin,"watch");
#ifdef __WINDOWS__
        arrowCursor = Tk_GetCursor(rcl_interp,tkwin,"IDC_ARROW");
#else
        arrowCursor = Tk_GetCursor(rcl_interp,tkwin,"arrow");
#if (!defined(__OS2__) && !defined(__WIN32__))
        // get the shift key codes for the IsShifted function
        KeyCodeLBit = XKeysymToKeycode(dpy, XK_Shift_L);
        KeyCodeLChar = KeyCodeLBit/8;
        KeyCodeLBit %= 8;
        KeyCodeRBit = XKeysymToKeycode(dpy, XK_Shift_R);
        KeyCodeRChar = KeyCodeRBit/8;
        KeyCodeRBit %= 8;
#endif
#endif

        char *fn = EnvGet("GRAPHFONT");
        if (fn == NULL) SetFont("-*-helvetica-medium-r-normal--*-120-*-*-*-*-*-*");
        else SetFont(fn);

        // Set the flag so we don't do this again

        tkwindow_init = 1;

    }

    // Prepare a backing store object for each window.
    // The backing store must be mapped (associated) with
    // a particular window before it can be used.
    pBackStore = new BckStore(dpy, window,
                draww, drawh, Tk_Depth(canvaswin), NULL);

    if (pBackStore == NULL)
        // touble in  River City - crash time
        mayday("Unable to construct BckStore object");

    // Save some state information and set the windows location

    winx = locx;
    winy = locy;
    winw = windw;
    winh = windh;
    tkWinList.Insert((Info*)this);

    Clear();

    tkcanw = Tk_Width(canvaswin);
    tkcanh = Tk_Height(canvaswin);

    Mess("TKWindow - done window construction");
}


TKWindow::~TKWindow() {
    CharType cmd;

    if (pBackStore != NULL) delete pBackStore;
    cmd.Assign("bind ");
    cmd.Append(Tk_PathName(tkwin));
    cmd.Append(" <Destroy> {}");
    exec_command(cmd.string());

    cmd.Assign("bind ");
    cmd.Append(Tk_PathName(tkwin));
    cmd.Append(" <Expose> {}");
    exec_command(cmd.string());

    Tk_DestroyWindow(tkwin);
    tkWinList.Remove((Info*)this);
}

void TKWindow::DestroyChildWindows() {
    CharType cmd;

    cmd.Assign("canvas_destroy_children ");
    cmd.Append(Tk_PathName(tkwin));
    exec_command(cmd.string());
}

void TKWindow::Hide() {
    CharType cmd;
    cmd.Assign("wm withdraw ");
    cmd.Append(Tk_PathName(tkwin));
    exec_command(cmd.string());
}

void TKWindow::Show()
{
    // Mess("TK - Show (iconified = %d)", iconified);
    CharType path(Tk_PathName(tkwin));
    // set the window geometry by calling the window manager
    char *geo = new char[strlen(path.string())+70];
    sprintf(geo, "wm geometry %s %dx%d+%d+%d",path.string(), winw,winh,winx,winy);
    exec_command(geo);
    delete[] geo;
    CharType cmd;
    if (!iconified) {
        // deiconify the window if it is not iconified
        CharType cmd("wm deiconify ");
        cmd.Append(path.string());
        exec_command(cmd.string());
    }
}

char* TKWindow::GetTitle()
{
    return title.string();
} // GetTitle

void TKWindow::SetTitle( char* str )
{
    // Save the title

    title.Assign(str);

    // Send the command to the interpreter

    CharType cmd;
    cmd.Assign("wm title ");
    cmd.Append(Tk_PathName(tkwin));
    cmd.Append(" {");
    cmd.Append(str);
    cmd.Append("}");
    exec_command(cmd.string());
}

int TKWindow::Iconified()
{
    return iconified;
}

void TKWindow::SetIconified( int value )
{
    if (value != iconified) {
        iconified = value;
        if (iconified) {
            CharType cmd;
            cmd.Assign("wm iconify ");
            cmd.Append(Tk_PathName(tkwin));
            exec_command(cmd.string());
        } else {
            Show(); // set the geometry and deiconify
        }
    }
}

// Change the windows cursor
void TKWindow::NormalCursor()
{
    cursorStateWait = RigiFalse;
#ifdef __WINDOWS__
    SetCursor(arrowCursor); // bug in XDefineCursor
#else
    Tk_DefineCursor(tkwin, arrowCursor);
#endif
}

void TKWindow::WaitCursor()
{
    cursorStateWait = RigiTrue;
#ifdef __WINDOWS__
    SetCursor(waitCursor); // bug in XDefineCursor
#else
    Tk_DefineCursor(tkwin, waitCursor);
#endif
}

// Get the windows absolute location on the workstation screen
void TKWindow::Location( int& x, int& y )
{
    x = winx;
    y = winy;
}

// Set the windows absolute location on the workstation screen
void TKWindow::SetLocation( int x, int y )
{
    SetGeometry(x, y, winw, winh);
}

void TKWindow::SetFont( char* fName ) {
  Mess("SetFont <%s>", fName);
  if (font != 0)
    Tk_FreeFont(font);
  font = Tk_GetFont(NULL, canvaswin, fName);
  if (font == NULL) 
    mayday("Unable to find specified canvas font %s.", fName);
  else {
    if (fontName)
      free(fontName);
    fontName = strdup(fName);
#if defined(__WINDOWS__) || defined(__WIN32__)
    XSetFont(dpy, gcdefault, (Font) font); 
#endif
  }
}

char *TKWindow::GetFont() { return fontName; }

// Set the windows absolute location on the workstation screen
// Set the dimensions of the window simulataneously
void TKWindow::SetGeometry( int x, int y, int w, int h )
{
    if (x != Tk_X(tkwin) || y != Tk_Y(tkwin)
        || w != Tk_Width(tkwin) || h != Tk_Height(tkwin)) {

        // keep track of current canvas size because its the
        // only way to determine their validity in GetDrawSize.
        tkcanw = Tk_Width(canvaswin);
        tkcanh = Tk_Height(canvaswin);
        if (!iconified) {
	  // the use of Tk_MoveResizeWindow causes problems with tk 8.x
	  // wm geometry does the job.
	  //	  Tk_MoveResizeWindow(tkwin, x, y, w, h);
	  CharType path(Tk_PathName(tkwin));
	  char *geo = new char[strlen(path.string())+70];
	  sprintf(geo, "wm state %s normal", path.string());
	  exec_command(geo);
	  sprintf(geo, "wm geometry %s %dx%d+%d+%d",path.string(), winw,winh,winx,winy);
	  exec_command(geo);
	  delete[] geo;
	}
        winx = x;
        winy = y;
        winw = w;
        winh = h;
    }
}

void TKWindow::SetScrollbarStartX(int scrollStart) {
    int w, h;
    char ratioStr[20];

    GetDrawSize(&w, &h);
    if (scrollStart + w >= canw) scrollStart = canw - w;
    if (scrollStart < 0) scrollStart = 0;

    sprintf(ratioStr, " %f", (double)scrollStart/canw);
    CharType cmd(Tk_PathName(tkwin));
    cmd.Append(".frame.xscroll set");
    cmd.Append(ratioStr);
    sprintf(ratioStr, " %f", (double)(scrollStart + w) / canw);
    cmd.Append(ratioStr);
    exec_command(cmd.string());
    winOrig.x = scrollStart;
}

void TKWindow::SetScrollbarStartY(int scrollStart) {
    int w, h;
    char ratioStr[20];

    GetDrawSize(&w, &h);
    if (scrollStart + h >= canh) scrollStart = canh - h;
    if (scrollStart < 0) scrollStart = 0;

    sprintf(ratioStr, " %f", (double)scrollStart/canh);
    CharType cmd(Tk_PathName(tkwin));
    cmd.Append(".frame.yscroll set");
    cmd.Append(ratioStr);
    sprintf(ratioStr, " %f", (double)(scrollStart + h) / canh);
    cmd.Append(ratioStr);
    exec_command(cmd.string());
    winOrig.y = scrollStart;
}

int TKWindow::Depth() {
    return Tk_Depth(canvaswin);
}

void TKWindow::SetFrameSize( int w, int h ) {
    SetGeometry(winx, winy, w, h);
} // SetFrameSize

void TKWindow::FrameSize( unsigned short& w, unsigned short& h ) {
    // Mess("TK - FrameSize");
    w = winw;
    h = winh;
} // FrameSize

void TKWindow::GetDrawSize( int *w, int *h ) {
    int ww, wh, fw, fh;

    // the window dimensions are derived from two sources
    // 'cause Tk often seems to forget to update its
    // state variables.

    RutilGetDefWindowSize(ww, wh);
    RutilGetDefFrameSize(fw, fh);
    *w = Tk_Width(canvaswin);
    *h = Tk_Height(canvaswin);
    if ((*w == tkcanw && *h == tkcanh) || *w < 2 || *h < 2) {
        // go with the dimensions last requested
        *w = winw - (fw - ww);
        *h = winh - (fh - wh);
    }

    // following is a correction - Tk does not know the proper
    // canvas size before the window is constructed...
    *w += 15;
    *h += 15;

    // Mess("GetDrawSize (%d %d) (%d %d)",*w,*h,
    // Tk_Width(canvaswin), Tk_Height(canvaswin));
}

void TKWindow::Clear()
{
    Window window;
    Point2d p;

    XSetForeground(dpy, gcdefault, foreground);
    XSetFunction(dpy, gcdefault, GXcopy);
    if (drawRealTime) {
        // clear the actual window
        window = Tk_WindowId(canvaswin);
        XFillRectangle(dpy, window, gcdefault, 0, 0, canw, canh);
    }

    // if the window has not been mapped before, maybe
    // it can now...
    CheckMappings();

    if (pBackStore->IsMapped()) {

        // clear the pixmap
        window = pBackStore->GetPixmap();
        XFillRectangle(dpy, window, gcdefault, 0, 0, canw, canh);

        // If the window had no backstore previously,
        // then it will have to be completely restored (not just
        // a simple object redraw).
        if (!drawRealTime) {
            // We do not want to send a bigger pixmap than is
            // necessary to the window to clear it.  Translate
            // the window coords to canvas coords.
            int w, h;
                GetDrawSize( &w, &h );
            p.x = p.y = 0;
            p = Window2Canvas( p );
            UpdateExtent(p.x, p.y, w, h);
        }
    }
    XSetFunction( dpy, gcdefault, GXxor );
}

void TKWindow::DrawIcon( Point p, int size, int type, int active )
{
    if (noDraw) return;
    Point p2;
    p2.x = p.x + size;
    p2.y = p.y + size;
    p2.z = p.z + size;
    if (drawWindow) numDrawObj++;
    if (active) DrawBox(p, p2, BIG, type, 0);
    else        DrawBox(p, p2, SMALL, type, 0);
}

void TKWindow::DrawBox( Point sp, Point dp, int width, int type, int dpyOnly )
{
    if (noDraw) return;
    if (!Tk_IsMapped(canvaswin)) Mess("canvas is not mapped");
    Window window = pBackStore->GetPixmap();

    // Check for boundary cases and make sure sp is less than dp.

    int t, w, h;
    if (sp.x==dp.x) return;
    else if (sp.x > dp.x) { t = sp.x; sp.x = dp.x; dp.x = t; }

    if (sp.y==dp.y) return;
    else if (sp.y > dp.y) { t = sp.y; sp.y = dp.y; dp.y = t; }

    w = dp.x - sp.x;
    h = dp.y - sp.y;

    // Draw the box. If the width is less than 2, then draw the box as a
    // solid box.

    unsigned long fg = rigi_colormap->GetNodeColor(type)->pixel;
    if (width<2)
        XSetLineAttributes( dpy, gcdefault, width, LineSolid, 0, 2 );
    XSetFunction( dpy, gcdefault, GXxor );

    if (!dpyOnly && pBackStore->IsMapped()) {
        window = pBackStore->GetPixmap();
        UpdateExtent(sp.x, sp.y, w, h);
        if (ResetBlackCanvas()) {
            if (width >= 2)
                XFillRectangle( dpy, window, gcdefault, sp.x, sp.y, w, h);
            else
                XDrawRectangle( dpy, window, gcdefault,sp.x, sp.y, w-1, h-1);
        }
        if (type >= 0) XSetForeground( dpy, gcdefault, fg );
        if (width >= 2)
            XFillRectangle( dpy, window, gcdefault, sp.x, sp.y, w, h);
        else
            XDrawRectangle( dpy, window, gcdefault, sp.x, sp.y, w-1, h-1);
    }
    if (dpyOnly || drawRealTime) {
        window = Tk_WindowId(canvaswin);
        sp = Canvas2Window(sp);     // draw in window coordinates
        if (ResetBlackCanvas()) {
            if (width >= 2)
                XFillRectangle( dpy, window, gcdefault, sp.x, sp.y, w, h);
            else
                XDrawRectangle( dpy, window, gcdefault,sp.x, sp.y, w-1, h-1);
        }
        if (type >= 0) XSetForeground( dpy, gcdefault, fg );
        if (width >= 2)
            XFillRectangle( dpy, window, gcdefault, sp.x, sp.y, w, h);
        else
            XDrawRectangle( dpy, window, gcdefault,sp.x, sp.y, w-1, h-1);
    }
}

void TKWindow::DrawLine( Point s, Point d, int width, int type )
{
    if (noDraw) return;
    if (width <= 0) width = 1;
    Window window;
    unsigned long fg = rigi_colormap->GetArcColor(type)->pixel;
    XSetLineAttributes( dpy, gcdefault, width, LineSolid, 0, 2 );

    if (drawWindow) numDrawObj++;
    if (pBackStore->IsMapped()) {
        int x, y, w, h;
        x = (s.x < d.x)? s.x : d.x;
        y = (s.y < d.y)? s.y : d.y;
        w = d.x - s.x;
        h = d.y - s.y;
        if (w < 0) w = -w;
        if (h < 0) h = -h;
        width = ((width + 1) >> 1) << 1;
        w += width;
        h += width;
        x -= (width >> 1);
        y -= (width >> 1);
        window = pBackStore->GetPixmap();
        UpdateExtent(x, y, w, h);
        if (ResetBlackCanvas())
            XDrawLine( dpy, window, gcdefault, s.x, s.y, d.x, d.y );
        XSetForeground(dpy, gcdefault, fg);
        XDrawLine( dpy, window, gcdefault, s.x, s.y, d.x, d.y );
        if (localServer && drawRealTime) {
            Restore(x, y, w, h);
            return;
        }
    }
    if (drawRealTime) {
        window = Tk_WindowId(canvaswin);
        s = Canvas2Window(s);   // draw in window coordinates
        d = Canvas2Window(d);
        if (ResetBlackCanvas())
            XDrawLine( dpy, window, gcdefault, s.x, s.y, d.x, d.y );
        XSetForeground(dpy, gcdefault, fg);
        XDrawLine( dpy, window, gcdefault, s.x, s.y, d.x, d.y );
    }
}

void TKWindow::DrawString( char *str, Point p )
{
    if (noDraw) return;
    Point tmpp;
    int text_width, text_ascent, text_height;
    Window window = pBackStore->GetPixmap();
    unsigned long fg = rigi_colormap->GetStringColor(0)->pixel;

    // Convert the text location given as the top center of the text into a
    // location at the bottom left corner of the string for writing
    // with X windows.

    TextExtent( str, &text_width, &text_height, &text_ascent );
    tmpp.x = p.x - text_width / 2;
    tmpp.y = p.y + text_ascent;

    XSetFunction( dpy, gcdefault, GXxor );
    //    Mess("gc->function = %d\n", gcdefault->function);
    if (drawWindow) numDrawObj++;
    if (pBackStore->IsMapped()) {
        window = pBackStore->GetPixmap();
        UpdateExtent(tmpp.x, tmpp.y, text_width, text_height);
        if (ResetBlackCanvas())
	  Tk_DrawChars(dpy, window, gcdefault, font, 
		       str, strlen(str), tmpp.x, tmpp.y);

        XSetForeground( dpy, gcdefault, fg );
	Tk_DrawChars(dpy, window, gcdefault, font, 
		     str, strlen(str), tmpp.x, tmpp.y);
        if (localServer && drawRealTime) {
            Restore(tmpp.x, p.y, text_width, text_height);
            return;
        }
    }
    if (drawRealTime) {
        window = Tk_WindowId(canvaswin);
        tmpp = Canvas2Window(tmpp);
        if (ResetBlackCanvas())
	  Tk_DrawChars(dpy, window, gcdefault, font, 
		       str, strlen(str), tmpp.x, tmpp.y);
        XSetForeground( dpy, gcdefault, fg );
	Tk_DrawChars(dpy, window, gcdefault, font, 
		     str, strlen(str), tmpp.x, tmpp.y);
    }
}


void TKWindow::DrawString( char *str, Point p, TextLocation loc)
{
    if (noDraw) return;
    int text_width, text_ascent, text_height;

    // Convert the text location given, based on the loc flag, into a
    // location at the bottom left corner of the string for writing
    // with X windows.

    TextExtent( str, &text_width, &text_height, &text_ascent );
    switch(loc) {
    case T_BOTCENTER:
        p.y += text_ascent;
        break;
    case T_TOPLEFT:
        p.x -= text_width / 2;
        break;
    case T_BOTLEFT:
        p.x -= text_width / 2;
        p.y += text_ascent;
        break;
    case T_TOPRIGHT:
        p.x += text_width / 2;
        break;
    case T_BOTRIGHT:
        p.x += text_width / 2;
        p.y += text_ascent;
        break;
    case T_CENTER:
        p.y += text_ascent/2;
        break;
    default:
        break;
    }
    DrawString(str,p);

}

// Get the pixel size of the given text string

void TKWindow::TextExtent( char *str, int *w, int *h, int *a)
{
  if (w != NULL)
    *w = Tk_TextWidth(font, str, strlen(str));
  if (h != NULL || a != NULL) {
    Tk_FontMetrics fontMetrics;
    Tk_GetFontMetrics(font, &fontMetrics);
    if (h != NULL) 
      *h = fontMetrics.ascent + fontMetrics.descent;
    if (a != NULL) 
      *a = fontMetrics.ascent;
  }
}

void TKWindow::SetMessage( char *text )
{
    CharType msg;
    msg.Assign(Tk_PathName(tkwin));
    msg.Append(".frame.text configure -text {");
    msg.Append(text);
    msg.Append("}");
    exec_command(msg.string());
}

void TKWindow::ResetMessage()
{
    SetMessage("");
}

void TKWindow::Restack(int order, BaseWindow *otherWin) {
    if (otherWin != NULL)
        Tk_RestackWindow(tkwin, order, ((TKWindow*)otherWin)->tkwin);
    else
        Tk_RestackWindow(tkwin, order, NULL);
}

// ----------------------------------------------------------
// draw control functions follow ...
// ----------------------------------------------------------

int TKWindow::Drawing() { return noDraw!=0; }

void TKWindow::Drawing( int i )
{
    noDraw += (i)? -1 : 1;
    if (noDraw < 0) {
        warn("unbalanced Drawing() commands");
        noDraw = 0;
    }
}

void TKWindow::DrawSynch( RigiBool i ) {
//  Enable/disable batching in all windows
    TKWindow *p;
    for (p = (TKWindow*)tkWinList.First(); p != NULL;
        p = (TKWindow*)tkWinList.Next()) {
        p->Batch(i);
    }
}

void TKWindow::BunchRect( RigiBool /*i*/ ) {}
void TKWindow::BunchLine( RigiBool /*i*/ ) {}

void TKWindow::Batch( RigiBool i )
{
    //  Draw to backing store only (TRUE), then flush to display (FALSE)

    drawWindow = drawBatch = i;
    drawRealTime = RigiTrue;
    if (pBackStore->IsMapped() && localServer) {
        drawRealTime = !drawBatch;
        if (drawRealTime && dsExtent.w && dsExtent.h && localServer) {

            // restore dirty part of screen
            Restore(dsExtent.x, dsExtent.y, dsExtent.w, dsExtent.h);

        }
    }
    if (!drawRealTime) InitExtent();
    if (!drawWindow) redrawCnt++;
}

void TKWindow::BackStore( RigiBool i )
{
    drawBatch = i;
    drawRealTime = RigiTrue;
    if (pBackStore->IsMapped() && localServer) drawRealTime = !drawBatch;
    if (!drawRealTime) InitExtent();
}

int TKWindow::BackStore() { return pBackStore->IsMapped(); }

void TKWindow::Restore( int x, int y, int w, int h )
{
    if (pBackStore->IsMapped()) {
        int ww, wh;
        Point2d dsWin,dsCan;

        GetDrawSize(&ww, &wh);
        UpdateExtent(x, y, w, h);
        Window window = Tk_WindowId(canvaswin);

        // translate destination to window coordinates

        dsWin.x = dsExtent.x; dsWin.y = dsExtent.y;
        dsWin = Canvas2Window(dsWin);

        // window coords cannot be negative
        if (dsWin.x < 0) {
            dsWin.x = 0;
            dsCan = Window2Canvas( dsWin );
            dsExtent.w -= (dsCan.x - dsExtent.x);
            dsExtent.x = dsCan.x;
        }

        if (dsWin.y < 0) {
            dsWin.y = 0;
            dsCan = Window2Canvas( dsWin );
            dsExtent.h -= (dsCan.y - dsExtent.y);
            dsExtent.y = dsCan.y;
        }


        if (dsWin.x + dsExtent.w > ww) dsExtent.w = ww - dsWin.x;
        if (dsWin.y + dsExtent.h > wh) dsExtent.h = wh - dsWin.y;

        if (dsExtent.w > 0 && dsExtent.h > 0) {
            // write backing store to the visible window

            XSetFunction( dpy, gcdefault, GXcopy );

            // Mess("XCopyArea %d %d %d %d -> %d %d",dsExtent.x, dsExtent.y,
            // dsExtent.w, dsExtent.h, dsWin.x, dsWin.y);
            XCopyArea( dpy, pBackStore->GetPixmap(), window, gcdefault,
                dsExtent.x, dsExtent.y, dsExtent.w, dsExtent.h,
                dsWin.x, dsWin.y );
            XSetFunction( dpy, gcdefault, GXxor );
            InitExtent();
        }
    }
}

void TKWindow::RestoreWindow()
{
    int w, h;
    GetDrawSize(&w, &h);
    Restore( winOrig.x, winOrig.y, w, h);
}


void TKWindow::FreeStore()
{
    if (pBackStore != NULL) pBackStore->Unmap();
    drawRealTime = RigiTrue;
    InitExtent();
}

// -------------------------------------------------------------------
// private functions follow
// -------------------------------------------------------------------

void TKWindow::InitExtent() {
    dsExtent.x = dsExtent.y = 0x7fffffff;
    dsExtent.w = dsExtent.h = 0;
}

void TKWindow::UpdateExtent(int x, int y, int w, int h) {
    int ex, ey;

    if (w < 0) {
        x += w;
        w = -w;
    }
    if (h < 0) {
        y += h;
        h = -h;
    }
    if (dsExtent.w == 0 || dsExtent.h == 0) {
        dsExtent.x = x;
        dsExtent.y = y;
        dsExtent.w = w;
        dsExtent.h = h;
        return;
    }
    ex = (x < dsExtent.x)? x : dsExtent.x;
    ey = (y < dsExtent.y)? y : dsExtent.y;
    dsExtent.w = (dsExtent.x + dsExtent.w < x + w)?
        x + w - ex : dsExtent.x + dsExtent.w - ex;
    dsExtent.h = (dsExtent.y + dsExtent.h < y + h)?
        y + h - ey : dsExtent.y + dsExtent.h - ey;
    dsExtent.x = ex;
    dsExtent.y = ey;
}

void TKWindow::CheckMappings() {

    int n = pBackStore->NumFreeBackStore();
    unsigned maxNumPerRedraw = numDrawObj;
    if (redrawCnt) maxNumPerRedraw /= redrawCnt;

    // following two lines cause the average objects per draw
    // to be weighted 50% toward history.
    numDrawObj = maxNumPerRedraw;
    redrawCnt = 1;

//    Mess("CheckMappings Ismapped %d maxNumPerRedraw %d",
//        pBackStore->IsMapped(),maxNumPerRedraw);

    // There may be too few objects in the window
    // to justify it being mapped to a backing store.
    // Alternatively, there may be no backing stores
    // to allocate.

    if (!pBackStore->IsMapped() &&
        (maxNumPerRedraw < minBSReqmnts || n < 1)) return;

    // If there is no contention for backstores,
    // simply allocate one if needed and return.

    if (n > 1) {
        if (!pBackStore->IsMapped()) {
            pBackStore->Map();
            if (pBackStore->IsMapped()) drawRealTime = !drawBatch;
        }
        return;
    }

    // Decide whether or not to deallocate a backing store,
    // (if there are none free),
    // or whether to allocate a backing store,
    // (if one is free).

    // Check for most needy window.

    TKWindow *p,*mostNeedy;
    mostNeedy = this;
        for (p = (TKWindow*)tkWinList.First(); p != NULL;
        p = (TKWindow*)tkWinList.Next()) {
        if (this != p && !p->pBackStore->IsMapped()) {
            unsigned numPerRedraw = p->numDrawObj;
            if (p->redrawCnt) numPerRedraw /= p->redrawCnt;
            if (maxNumPerRedraw < numPerRedraw) {
                mostNeedy = p;
                maxNumPerRedraw = numPerRedraw;
            }
        }
    }

    // decide whether or not to allocate or free the backing store

    if (pBackStore->IsMapped()) {
        if (mostNeedy != this) {
            // This window has a back store but is not most needy of one.
            // Release the backing store ...
            pBackStore->Unmap();
            drawRealTime = RigiTrue;
            drawBatch = RigiFalse;
        }
    } else if (mostNeedy == this) {
        // This window is most needy of the backing store.
        // Allocate the backing store ...
        pBackStore->Map();
        if (pBackStore->IsMapped()) drawRealTime = !drawBatch;
    }

}

int TKWindow::IsShifted() {
#if defined (__WINDOWS__) || defined(__WIN32__)
    BYTE keyState[256];
    GetKeyboardState((LPBYTE)&keyState);
    return ((keyState[VK_SHIFT] & 0x80) > 0);
#elif defined (__OS2__)
    return (WinGetKeyState(HWND_DESKTOP, VK_SHIFT) & 0x8000) > 0;
#else
    if (dpy == NULL) return 0;      // no window yet defined (call is too early)
    char keys[32];

    XQueryKeymap(dpy, keys);

    return ((keys[KeyCodeLChar] >> KeyCodeLBit) & 1)
        || ((keys[KeyCodeRChar] >> KeyCodeRBit) & 1);
#endif
}
