//
//

// $Author: jmcdanie $
// $Date: 1996-05-25 01:19:04 $
// $Source: /root/cvs/Rigi/rigiedit/client.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: client.C,v 1.18 1996-05-25 01:19:04 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif

#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef CLIENT_H
#include "client.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

#ifndef HASHNODELIST_H
#include "cl_hashnodelist.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif

const int RIGI_CLIENT = 0;
const int RIGI_SUPPLIER = 1;

// Get the neighbours or common clients and common suppliers of the given node
// based on the thresholds provided. See the algorithm in Hausi's paper
// ``Discovering and Reconstructing Subsystem Structures using Graph Filters''
// for details on the implementation.

static void GetNeighbours( RigiNode* rn,
                    float thcc, float tlcc, float thcs, float tlcs,
                    ArcType arctype, HashNodeList &winnodes, int use_winnodes,
                    HashNodeList& commonclient, HashNodeList& commonsupplier) {
    GraphNode* x = rn->GetNode();
    int n;

    // Get the predecessor and successor lists for node x.

    HashNodeList succx,predx;
    x->GetSuccessor( succx,outArcs,arctype );
    x->GetPredecessor( predx,inArcs,arctype );

    if ( use_winnodes ) {
        // get intersections
        succx & winnodes;
        predx & winnodes;
    }

    // For each node y in the succesor list get its common client set, and
    // add the common client nodes to the hashlist.

    IterDO( succx,GraphNode*,y ) {
        HashNodeList predy;
        y->GetPredecessor( predy,inArcs,arctype );
        if ( use_winnodes ) {
            predy & winnodes;
        }
        IterDO( predy,GraphNode*,z ) {
            if ( z->GetId() != x->GetId() ) {
                HashNodeList succz;
                z->GetSuccessor( succz,outArcs,arctype );
                if ( tlcc < 0.0 ) {
                    int num = succz.NumElements();
                    if (num == (int)thcc) {
                        succz & succx;
                        n = succz.NumElements();
                        if ( n == thcc ) {
                            commonclient.Add( *z );
                        }
                    }
                } else {
                    succz & succx;
                    n = succz.NumElements();
                    if ( n >= tlcc && ( thcc < 0.0 || n <= thcc ) ) {
                        commonclient.Add( *z );
                    }
                }
            }
        }
        ODretI;
    }
    ODretI;

    // For each node y in the predecessor list get its common supplier set,
    // and add the common supplier nodes to the hashlist.

    IterDO( predx,GraphNode*,y ) {
        HashNodeList succy;
        y->GetSuccessor( succy,outArcs,arctype );
        if ( use_winnodes ) {
            succy & winnodes;
        }

        IterDO( succy,GraphNode*,z ) {
            if ( z->GetId() != x->GetId() ) {
                HashNodeList predz;
                z->GetPredecessor( predz,inArcs,arctype );
                if ( tlcs < 0.0 ) {
                    int num = predz.NumElements();
                    if (num == (int)thcs) {
                        predz & predx;
                        n = predz.NumElements();
                        if ( n == thcs ) {
                            commonsupplier.Add( *z );
                        }
                    }
                } else {
                    predz & predx;
                    n = predz.NumElements();
                    if ( n >= tlcs && ( thcs < 0.0 || n <= thcs ) ) {
                        commonsupplier.Add( *z );
                    }
                }
            }
        }
        ODretI;
    }
    ODretI;
}

// Select all node in this window that have a common client or a common
// supplier based on the thresholds thcc, tlcc, thcs and tlcs.

void CommonCSSelect( RigiWindow* win,
                    float thcc, float tlcc, float thcs, float tlcs,
                    ArcType arctype, int use_winnodes )
{
    HashNodeList allcommon;
    HashNodeList winnodes;
    RigiNode* rn; AvlPos *P;

    // Get a hash list of the nodes in this window

    if ( use_winnodes )
    for ( rn = win->FirstNode(P); rn; rn = win->NextNode(P) )
        winnodes.Add(*(rn->GetNode()));

    // For each Riginode in the window, get its neighbours. Union both the
    // common client and the common supplier sets together to form the
    // Union over all nodes in this window that have a common client/supplier.

    for ( rn = win->FirstNode(P); rn; rn = win->NextNode(P) )
    {
        HashNodeList cc,cs;

        GetNeighbours( rn, thcc, tlcc, thcs, tlcs, arctype,
                        winnodes, use_winnodes, cc, cs );
        allcommon | cc;
        allcommon | cs;
    }

    // Select each node in the allcommon set that is in the current window.

    for ( rn = win->FirstNode(P); rn; rn = win->NextNode(P) )
        if ( allcommon.Find(rn->GetNode()->GetId()))
            editor->SelectNode( rn->GetNode() );
}

// Select the Common Client/Supplier component based on the type of
// neighbour desired in TYPE.

void CommonCSComponent( RigiWindow *win, RigiNodeList& slist,
                        float thcc, float tlcc, float thcs, float tlcs,
                        ArcType arctype, int type, int use_winnodes ) {
    HashNodeList winnodes;
    RigiNode *rn; AvlPos *P;

//    Mess("Here I am in CommonCSComponent in client.C!");
    // Want at least one node.

    if (slist.GetNoItems() < 1 ) {
        RigiError(ERRORFLAG, "At least one node must be selected");
        return;
    }

    // Get a hash list of the nodes in this window

    if ( use_winnodes )
        for ( rn = win->FirstNode(P); rn; rn = win->NextNode(P) )
            winnodes.Add(*(rn->GetNode()));

    // Process each of the nodes in the list

    for ( RigiNode* comprn=slist.First(); comprn; comprn=slist.Next() ) {
        // Get the neighbours.

        HashNodeList cc,cs;
        GetNeighbours( comprn, thcc, tlcc, thcs, tlcs, arctype,
                        winnodes, use_winnodes, cc, cs);

        // Select the nodes in the window that are of the correct type.

        for ( rn = win->FirstNode(P); rn; rn = win->NextNode(P) )
        {
            if (type==RIGI_CLIENT && cc.Find(rn->GetNode()->GetId()))
                editor->SelectNode( rn );
            if (type==RIGI_SUPPLIER && cs.Find(rn->GetNode()->GetId()))
                editor->SelectNode( rn );
        }
    }
}

