//
// $Author: jmartin $
// $Date: 2000-02-05 02:18:30 $
// $Source: /root/cvs/Rigi/rigiedit/editutil.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: editutil.C,v 1.44 2000-02-05 02:18:30 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef DUPLIST_H
#include "cl_duplist.h"
#endif

#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef GRAPHARCLIST_H
#include "cl_grapharclist.h"
#endif
#ifndef GRAPHNODELIST_H
#include "cl_graphnodelist.h"
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef ACTIVEARC_H
#include "cl_activearc.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef WINUTIL_H
#include "win_util.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

//---------------------------------------------------------------------------
// External and local functions and objects used by this file
//---------------------------------------------------------------------------

// Data structure used for determining what is deleted when the clipboard
// is emptied.

struct sets {
    NodeSet treenodes;
    NodeSet cleannodes;
    NodeSet visited;
};

// Forward declarations of functions in this file

extern void edit_disconnect_node( GraphNode* parent, NodeList& list );
extern void edit_delete_rigi_nodes( GraphNode* n );

//---------------------------------------------------------------------------
// Static routines that are used by graph iterators to process nodes and arcs
//---------------------------------------------------------------------------

// Add a node to the locally defined delete list

static void edit_add_node( GraphNode* n, void *set )
{
    ((NodeSet*)set)->Add(n);
}

// Delete all Level arcs associated with this node.

static void edit_prune_node( GraphNode* n, NodeSet *cutlist )
{
    ArcList arclist;
    GraphArc* a; AvlPos *P;

    for ( a = n->FirstAdj( P,inLevArcs ); a; a = n->NextAdj( P,inLevArcs ) )
        if ( !cutlist->Find( a->GetSrc() ) ) arclist.Insert( a );

    for ( a = n->FirstAdj( P,outLevArcs ); a; a = n->NextAdj( P,outLevArcs ) )
        if ( !cutlist->Find( a->GetDst() ) ) arclist.Insert( a );

    // Now, get rid of the arcs

    for ( a = arclist.First(); a; a = arclist.Next() )
        GutilDeleteArc(a);
}

// Delete all Rigi/Window stuff associated with this node and prune all
// incoming and outgoing Level Arcs from it.

static void edit_update_node( GraphNode* n, void *set )
{
    edit_delete_rigi_nodes( n );
    edit_prune_node( n, (NodeSet*)set );
}

// Get a list of nodes that are rooted by the list of nodes that are given
// in treelist. Add each node to the treelist

static void edit_add_tree_node( GraphNode* n, void* val )
{
    NodeSet *treeset = (NodeSet*)val;
    treeset->Add( n );
}

static void edit_duplicate_node( GraphNode* node, DupList* newnodes )
{
    // Create a new node and copy its contents.

    GraphNode* np = new GraphNode( node->GetLabel() );
    node->Copy( np );
    Mess( "edit_duplicate_node: Duplicating node %s", node->GetLabel() );

    // Create a matching ActiveNode for the node.

    ActiveNode* anp = new ActiveNode();
    np->SetActiveNode( anp );

    // Append the new node to the duplication list.

    newnodes->Append( np,node,node->GetId() );
}

// Count the number of nodes in the subtree given to edit_count_nodes().

static void edit_count_node( GraphNode* node, int* count )
{
    if (node) (*count)++;
}

static void edit_set_level( GraphNode *node )
{
    GraphArc* levarc = node->FirstArc( inLevArcs ); // memory leak!!
    if ( levarc == NULL ) mayday("Couldn't find a levelarc?????" );
    node->SetLevel();
}

static int edit_process_node( GraphNode *n, void *val )
{
    NodeSet *visited = (NodeSet*)val;
    return !visited->Find( n );
}

// Add a node to the set if it does not contain more than one
// level arc incoming.

static void edit_add_clean_list( GraphNode* n, void* val )
{
    struct sets *l = (struct sets*)val;
    l->visited.Add(n);
}

static int edit_process_clean_node( GraphNode *n, void *val ) {
    int clean_it = 1;
    ArcDirection arcin = ARC_IN;
    struct sets *l = (struct sets *)val;

    if ( l->visited.Find( n ) ) return 0;

    // If the node has a level arc that is not within the set of tree nodes
    // then don't add it to the clean list. NOTE: There is a special case
    // if the incoming level arc originates from the clipnode. The clipnode
    // is not in the forest, but any nodes connected to it should be deleted
    // if the clipnode is the only node they are attached to by level arcs.

    int lt = editor->LevelArc();

    GraphArc *a; AvlPos *P;
    for ( a=n->FirstArc(P, lt, arcin); a; a=n->NextArc(P, lt, arcin) )
    {
        GraphNode *src = a->GetSrc();
        if ( l->treenodes.Find( src ) )
        {
            Mess( "clean: **** Adding node %s to clean list", n->GetLabel() );
            clean_it = 1;
        } else {
            clean_it = 0;
            break;
        }
    }

    // If we don't add it to the clean list, we return 0 to tell the tree
    // traversal to prune this branch from its traversal. If we do add it
    // to the clean list then we continue traversing that brach by returnig 1

    if ( clean_it ) l->cleannodes.Add( n );
    return clean_it;
}

//---------------------------------------------------------------------------
// Editing functions required by edit.C
//---------------------------------------------------------------------------

// Delete all Rigi/Window related stuff that has anything to do with this node

void edit_delete_rigi_nodes( GraphNode* n )
{
    // Get the active node (if it exists) for this node,
    // to trace all its RigiNodes

    //Mess( "-- Deleting rigi stuff for node %s", n->GetLabel() );
    ActiveNode* anp = (ActiveNode*)n->GetActiveNode();
    if (!anp) return;

    // For each RigiNode, delete all its RigiArcs, then itself.

    for ( RigiNode* rnp = anp->First(); rnp; rnp = anp->First() )
    {
        //Mess( "-- Deleting rigi stuff for a window" );
        RigiArc *ra; AvlPos *P;
        for ( ra = rnp->FirstArc(P); ra; ra = rnp->FirstArc(P) )
        {
            // Erase the RigiArc
            ra->Draw();

            // Delete it (that will take care of nodes and activearc)
            delete ra;
        }

        // Erase the RigiNode.
        rnp->Draw();

        // Remove the RigiNode from the ActiveNode list.
        anp->Remove( rnp );

        // Remove the RigiNode from its windows node set
        RigiWindow* rwp = rnp->GetWindow();
        rwp->RemoveNode( rnp );

        // Delete the actual RigiNode.
        delete rnp;
    }
}

// Get all of the nodes in the forest of trees rooted by the nodes in
// the list rootlist. The nodes in the forest are returned in the
// treeset

static void edit_get_tree_nodes( NodeList &rootlist, NodeSet &treeset ) {
    for (GraphNode *n = rootlist.First(); n != NULL; n=rootlist.Next())
        n->TreeBFS( editor->LevelArc(), ARC_OUT,
            (VarArgFunRef)edit_add_tree_node, (void*)(&treeset),
            NULL, NULL,
            (IntFunction)edit_process_node, (void*)(&treeset) );
}

// Clean up the hierarchies rooted at the nodes given in the NodeList.
// but only the ones that are not shared by another part of the graph

void edit_update_list_to_cut( NodeList& list )
{
    for ( GraphNode* n=list.First(); n; n=list.Next() )
        n->TreeBFS( outLevArcs, 0, (VarArgFunRef)edit_delete_rigi_nodes, NULL,
                    NULL, NULL, NULL, NULL );
}

// Get the list of nodes that are rooted at these nodes

void edit_get_ids( NodeList& list, NodeSet& ids )
{
    for ( GraphNode* n=list.First(); n; n=list.Next() )
        n->TreeDFS( outLevArcs,(VarArgFunRef)edit_add_node,&ids,NULL,NULL );
}

// Add the given list of nodes to the clipnode.

void edit_add_clipnode( NodeList& list )
{
    GraphNode* clipnode = RigiClipNode();

    // For each node in the list, add its hierarchy to the clipboard.

    int lt = editor->LevelArc();
    int i=0;
    for ( GraphNode* np = list.First(); np; np = list.Next(),i++ )
    {
        // Create a level arc between this node and the clipboard base node.

        GraphArc* ap = new GraphArc( clipnode, np, lt );
        clipnode->InsertArc( ap );
        np->InsertArc( ap );

        GutilAddArc(ap);
    }
    Mess( "%d node%s added to the clipboard",i,i==1?"":"s" );
}

// Clean up the hierarchies rooted at the nodes given in the NodeList.

void edit_update_list( NodeList& list, NodeSet& ids ) {
    for ( GraphNode* n=list.First(); n; n=list.Next() )
        n->TreeDFS(outLevArcs,(VarArgFunRef)edit_update_node,
            &ids,NULL,NULL);
}

void edit_delete_all( NodeList& list ) {
    if (list.GetNoItems() == 0) return;
    struct sets l;
    GraphNode *n; AvlPos *P;

    // Get a list of all of the nodes beneath the given node list that
    // have no incoming level arcs from outside the subgraph defined by
    // the forest rooted by the nodes in the nodelist.

    // First, get a list of all nodes in the forest

    edit_get_tree_nodes( list, l.treenodes );

    // Disconnect the nodes from the clipnode

    edit_disconnect_node( RigiClipNode(), list );

    // Determine which nodes need to be "cleaned"

    for ( n = list.First(); n; n = list.Next() )
        n->TreeDFS( editor->LevelArc(), ARC_OUT,
            (VarArgFunRef)(&(edit_add_clean_list)), &l, NULL, NULL,
            (IntFunction)edit_process_clean_node, (void*)(&l));

    // Delete each of the nodes compiled in the previous step. First we
    // delete the Rigi related stuff and then the underlying graphs stuff.

    int i=0;
    for ( n = l.cleannodes.First(P); n; n = l.cleannodes.Next(P),i++ )
    {
        edit_delete_rigi_nodes( n );

        GraphArc* a;
        while ((a = n->FirstArc()) != NULL && GutilDeleteArc(a)) ;

        GutilDeleteNode(n);
    }

    Mess( "delete_all: %d node%s deleted",i,i==1?"":"s" );
}

// Get all destination nodes of all level arcs that leave the clipnode.
// These are the roots of the trees to be cleaned up.

void edit_get_cliplist( NodeList& list ) {
    ArcDirection arcout = ARC_OUT;
    int lt = editor->LevelArc();
    GraphNode *cn = RigiClipNode();
    GraphArc* a; AvlPos *P; int i=0;

    for (a = cn->FirstArc(P, lt, arcout); a != NULL; a = cn->NextArc(P, lt, arcout),i++ )
        list.Insert(a->GetDst());

    Mess("%d clipnode child%s added to cliplist",i,i==1?"":"ren" );
}

void edit_disconnect_node( GraphNode* n, NodeList& list ) {
    if (n == NULL) return;
    ArcDirection arcout = ARC_OUT;
    // Disconnect the list of nodes from the given parent node.

    ArcList dellist;
    NodeSet nodes;

    GraphNode *p;
    for ( p = list.First(); p; p = list.Next() ) nodes.Add(p);

    // For each levelarc child of the parent, add to arc delete list
    // iff that child is in the set of nodes.

    int lt = editor->LevelArc();
    GraphArc *a; AvlPos *P;

    for ( a = n->FirstArc(P, lt, arcout); a; a = n->NextArc(P, lt, arcout))
        if (nodes.Find( a->GetDst() )) dellist.Insert(a);

    // Delete each of the arcs found in the previous step.

    for ( a = dellist.First(); a; a = dellist.Next() )
        GutilDeleteArc(a);
}

int edit_count_nodes( NodeList& list )
{
    int count = 0;
    for ( GraphNode* np=list.First(); np; np=list.Next() )
        np->TreeDFS(outLevArcs,(VarArgFunRef)(&(edit_count_node)),&count,
        NULL,NULL);
    return count;
}

void edit_duplicate_nodes( NodeList& list, NodeSet& copynodes,
                            NodeList& duplist )
{
    int num = edit_count_nodes( list );
    if (num == 0) return;

    // Create a list of new and old node pairs, based on old nodes ID

    DupList* newnodes = new DupList( num );
    GraphNode* n;
    for (n = list.First(); n; n = list.Next() )
        n->TreeDFS( outLevArcs,(VarArgFunRef)(&(edit_duplicate_node)),
            newnodes,NULL,NULL );

    // Set up duplist so it has the roots of the hierarchy that were copied

    for ( n = list.First(); n; n = list.Next() )
    {
        GraphNode* newnode = newnodes->FindNewNode( n->GetId() );
        duplist.Insert( newnode );
    }

    // Traverse the list of newnodes, and copy all the arcs (both internal
    // and external)

    int i;
    ArcList newarclist;
    for( i=0; i<newnodes->Number(); i++ )
    {
        DupPair* pair = newnodes->GetPair( i );
        GraphNode* oldnode = pair->oldnode;
        GraphNode* newnode = pair->newnode;

        // At each node, copy all the outgoing arcs. Since we hit each node
        // that we copy exactly once, we get all the internal arcs that join
        // two nodes in the copylist. We only set the src at this stage, and 
        // save the dst as the old arc, and retraverse the list when all the 
        // nodes have been created, setting the dst at the second stage. At 
        // the second stage, we check to see if the destination is external or
        // not, and connect the appropriate node (either a new node that is
        // external to the copy or an old node that is external).

        AdjList type = outArcs;
        while( type == outArcs || type == outLevArcs )
        {
            AvlPos *P;
            GraphArc* arc = oldnode->FirstAdj( P, type );
            while( arc != NULL )
            {
                // Get the src and dst of the arc

                GraphNode* src = arc->GetSrc();
                GraphNode* dst = arc->GetDst();

                // Create a new arc, copy the arc contents, and set the
                // source and destination. The source for the new arc is the
                // new node, but the destination is the old destination. We
                // need this to determine the destination later.

                Mess( "edit_duplicate_nodes: Creating an arc from %s to %s",
                        src->GetLabel(), dst->GetLabel() );
                GraphArc* newarc = new GraphArc( newnode, dst, arc->GetType() );

                // Insert the arc in the appropriate list

                newnode->InsertArc( newarc );

                // Keep track of the arc we made

                newarclist.Insert( newarc );

                // Get the next arc of this type in the list

                arc = oldnode->NextAdj( P, type );
            }

            // Determine the next type of arc to process based on the
            // current type.

            switch( type )
            {
                case outArcs: type = outLevArcs; break;
                case outLevArcs: type = invalidArcs; break;
                default: mayday("**** Invalid arc type"); arc = NULL; break;
            }
        }

        // Now we need to take care of all the arcs coming in from other nodes 
        // that are outside the delete list. We don't need to insert the new 
        // arc in the new arc list because we complete the arc at this stage 
        // (both src and dst are set).

        type = inArcs;
        while( type == inArcs || type == inLevArcs ) {
            AvlPos *P;
            ArcDirection arcinout = ARC_INOUT;
            GraphArc* arc = oldnode->FirstArc(P, type, arcinout);
            while( arc != NULL ) {
                // Get the src and dst of the incoming arc.

                GraphNode* src = arc->GetSrc();
                GraphNode* dst = arc->GetDst();

                if ( !copynodes.Find( src ) ) {
                    Mess( "edit_duplicate_nodes: EXTERNAL arc from %s to %s",
                                    src->GetLabel(),dst->GetLabel() );

                    // Create a new arc, copy the arc contents, and set the
                    // source and destination. The destination is the new
                    // new node while the src is the external node.

                    GraphArc*newarc=new GraphArc( src,newnode,arc->GetType() );
                    newnode->InsertArc( newarc );
                    src->InsertArc( newarc );

                    GutilAddArc(newarc);
                }
                arc = oldnode->NextAdj(P, type);
            }

            // Get the next type of incoming arc.

            switch( type )
            {
                case inArcs: type = inLevArcs; break;
                case inLevArcs: type = invalidArcs; break;
                default: mayday("**** Invalid arc type"); arc = NULL; break;
            }
        }

        // Add to the editor's set

        if (!editor->AddNode(newnode))
            mayday("graphNode already exists. HELP!!!!");
    }

    // Process all the new arcs we created and set the destinations

    GraphArc* arc = newarclist.First();
    while( arc != NULL )
    {
        // Get the dst for the arc.

        GraphNode* dst = arc->GetDst();
        Mess( "Setting the destination for arc %s to %s",
                        arc->GetSrc()->GetLabel(),dst->GetLabel() );

        // Check to see if the destination is included in the copy list. If
        // not we have an external arc. If it is in the copy list, then we
        // have an internal arc and the destination is a newnode rather than
        // a node that existed in the database before.

        if ( !copynodes.Find( dst ) )
        {
            Mess( "    External dst for arc, node = %s",dst->GetLabel());

            // Set the destination for the arc to be the external node DST and
            // insert the node into the appropriate list.

            arc->SetDst( dst );
            dst->InsertArc( arc );
        }
        else
        {
            // If we have an internal arc, we need to get the new version of the
            // node and use it as the destination. Then insert the arc in the
            // nodes appropriate arc list.

            Mess( "    Internal dst for arc, node = %s",dst->GetLabel());

            GraphNode* newdst = newnodes->FindNewNode( dst->GetId() );
            if (newdst == NULL)
                mayday("No newdst for this node in edit_duplicate_nodes");
            arc->SetDst( newdst );

            newdst->InsertArc( arc );
        }

        GutilAddArc(arc);

        arc = newarclist.Next();
    }
}

void edit_paste_nodes( NodeList& list, RigiWindow* window )
{
    GraphNode *np;
    RigiArcList newarcs;
    NodeSet nodes;

    // Make a set of the nodes in the current window

    AvlPos *P;
    for ( RigiNode* rn = window->FirstNode(P); rn; rn = window->NextNode(P) )
        nodes.Add( rn->GetNode() );

    // For each node in the list, move its hierarchy to the child of the
    // node represented by window.

    GraphNode* parent = window->GetNode();
    for ( np = list.First(); np; np = list.Next() )
    {
        // If this clip node already exists in the window, don't add it
        // again, as that implies that there is already a level arc between
        // it and the parent of the window.

        if ( nodes.Find(np) ) continue;


        np->SetLevel();

        // Create a level arc from the parent of this node.

        GraphArc* ap = new GraphArc( parent, np, editor->LevelArc() );
        parent->InsertArc( ap );
        np->InsertArc( ap );

        GutilAddArc(ap);

        // Create the RigiNode for this node and insert it into the window.

        RigiNode* rn = new RigiNode(np,window);
        Point2d p = window->NDC2World2d( np->GetNDCPosition() );
        rn->SetPosition( p );
        window->InsertNode( rn );

        // Add the node to the window list

        nodes.Add( np );

        // Add RigiArcs to the window if both the source and destination
        // node are inside the window.

        AvlPos *P;
        for ( ap=np->FirstAdj(P,outArcs); ap; ap=np->NextAdj(P,outArcs) )
            if ( nodes.Find( ap->GetSrc() ) && nodes.Find( ap->GetDst() ) )
            {
                RigiArc* rap = new RigiArc( ap,window );
                newarcs.Insert( rap );
                rn->InsertArc( rap );
            }
    }

    // For each arc added, set its destination

    for ( RigiArc* ra=newarcs.First(); ra; ra = newarcs.Next() )
    {
        GraphNode* dst = ra->GetArc()->GetDst();
        RigiNode *drnp = window->FindRigiNode( dst );
        if (drnp == NULL) Mess("**** NULL dest rigi node ****");
        else drnp->InsertArc( ra );
    }

    // I have no idea what this does. If you look at edit_set_level(), all it
    // does is set the level of the nodes. It is left here for historical
    // reasons (in other words I am afraid to take it out right now).

    for ( np = list.First(); np; np = list.Next() )
        np->TreeDFS(outLevArcs,(VarArgFunRef)(&(edit_set_level)),
        NULL,NULL,NULL);
}

void edit_get_comp_arcs( GraphNode* n, NodeSet& nlist,
                        ArcList& inlist, ArcList& outlist)
{
    // Check all the outgoing normal arcs

    GraphArc *a; AvlPos *P;
    for ( a = n->FirstAdj( P,outArcs ); a; a = n->NextAdj( P,outArcs ) )
    {
        GraphNode* dst = a->GetDst();

        // Check to see if the destination is external to the node set, if
        // so, add it to the out arc list.

        if ( !nlist.Find( dst ) ) outlist.Insert( a );
    }

    // Check all the incoming normal arcs

    for ( a = n->FirstAdj( P,inArcs ); a; a = n->NextAdj( P,inArcs ) )
    {
        GraphNode* src = a->GetSrc();

        // Check to see if the destination is external to the node set, if
        // so, add it to the in arc list.

        if ( !nlist.Find( src ) ) inlist.Insert( a );
    }
}

// Code to rebuild the contributor lists of all the composite arcs that
// originate from the tree rooted by the given node. Realistically, it
// should probably only be called with the root node so it rebuilds
// everything consistently. edit_check_composite_node recurses to the
// leaf nodes of the tree and from the bottom up rebuilds all of the
// contributor lists.

static void edit_check_composite_node( GraphNode *n, NodeSet *visited ) {
    ArcDirection arcout = ARC_OUT;
    // Check to see if we have visited this node. If so, return

    if ( visited->Find(n) ) return;
    visited->Add( n );

    // Recurse on all of my descendants

    int lt = editor->LevelArc();

    GraphArc *a; AvlPos *P;
    for ( a=n->FirstArc(P, lt, arcout); a; a=n->NextArc(P, lt, arcout) )
        edit_check_composite_node( a->GetDst(), visited );

    for ( a=n->FirstAdj( P,outArcs ); a; a=n->NextAdj( P,outArcs ) )
        if (a->GetType() == editor->CompositeArc())
            a->ComputeContributorList();
}

void edit_check_composite( GraphNode *n )
{
    NodeSet visited;

    // Call the recursive function that traverses the tree
    edit_check_composite_node( n, &visited );
}
