//
//

// $Author: jmcdanie $
// $Date: 1996-10-12 00:06:11 $
// $Source: /root/cvs/Rigi/rigiedit/exact.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: exact.C,v 1.35 1996-10-12 00:06:11 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#ifndef __STRING_H
#include <string.h>
#endif

#ifndef CHARARRAYH
#include "adt/chararray.h"
#endif
#ifndef GRAPHARCSET_H
#include "cl_grapharcset.h"
#endif
#ifndef HASHNODELIST_H
#include "cl_hashnodelist.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif

#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif
#ifndef TK_TEXTSW_H
#include "tk_textsw.h"
#endif

#ifndef RIGI_H
#include "rigi.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef TREELISTS_H
#include "treelists.h"
#endif
#ifndef EXACT_H
#include "exact.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif

// STRING1
#define BUFSIZE 800   // use a constant instead of 800!

// A class that contains a textual string, as well as a counter. This is
// necessary for the sort routines, as the count need to be kept with the
// text string.

class Counter
{
    private:
        char* name;
        int   count;

    public:
        Counter(char* str)
        {
            if (name != NULL) delete[] name;
            name = new char[strlen(str)+1];
            strcpy(name,str);
            count = 1;
        }
        Counter()
        {
            name = NULL;
            count = 0;
        }
        ~Counter()
        {
            if (name != NULL) delete[] name;
        }

        void  SetName(char* str)
        {
            if (name != NULL) delete[] name;
            name = new char[strlen(str)+1];
            strcpy(name,str);
            count = 1;
        }
        void Init()
        {
            name = NULL;
            count = 0;
        }

        char* Name()      {return name;}
        int   Count()     {return count;}
        void  Increment() {count++;}
};

// A global variable that determines whether exact interfaces should use
// all subtypes of arcs or not. exacttype determines which type to use
// if alltypes = RigiFalse

static RigiBool alltypes = RigiTrue;
static ArcType exacttype;

// Some local variables used for printing error messages etc.

static char buffer[BUFSIZE]; 
static char noname[] = " ";

// Return the parent of the target node that is closest to level LEVEL

static char* GetNodeNameAtLevel( int level, GraphNode* target ) {
    ArcDirection arcin = ARC_IN;
    if ( target == NULL ) {
        mayday("No target node specified");
        return noname;
    }

    int lt = editor->LevelArc();
    AvlPos *P; // memory leak!

    GraphArc* ap = target->FirstArc( P, lt, arcin);
    if ( ap == NULL ) {
        mayday("No incoming level arc to node %d", target->GetId());
        return noname;
    }
    GraphNode* curnode = ap->GetSrc();
    if ( curnode == NULL ) {
        mayday("No src found for arc id %d", ap->GetId());
        return noname;
    }
    int  curlevel = curnode->GetLevel();
    while( curnode != NULL && curlevel > level ) {
        ap = curnode->FirstArc(P, lt, arcin);
        if ( ap == NULL ) {
            mayday("No incoming level arc to node %d",
                curnode->GetId());
            return noname;
        }
        curnode = ap->GetSrc();
        if ( curnode == NULL ) {
            mayday("No src for arc %d", ap->GetId());
            return noname;
        }
        curlevel = curnode->GetLevel();
    }
    return (curnode == NULL)? noname : curnode->GetLabel();
}

void ArcExactInterface(GraphArc* ap)
{
    // Do some error checking

    if ( ap == NULL ) {
        RigiError(ERRORFLAG, "NULL arc in exact interface" );
        return;
    }
    if ( ap->GetType()!=editor->CompositeArc() ) {
        RigiError(ERRORFLAG, "Exact interfaces are only for composite arcs");
        return;
    }

    // Turn on text batching to speed up writing to a text window.
    TextStartReport();

    // Get the number of objects exchanged by this arc.

    RigiReport report_type = editor->GetReportGeneration();
    int i,gid,num = ap->NumContributorIds();
    GraphNode* compsrc = ap->GetSrc();
    GraphNode* compdst = ap->GetDst();
    // STRING1, check not over the size!
    if ( strlen( compsrc->GetLabel() ) + strlen( compdst->GetLabel() ) +
        sizeof(int)  + 50 >= BUFSIZE ) 
        Mess("Buffer too small in exact");
    else 	     
        sprintf( buffer,
            "Subsystem %s requires %d object%s from subsystem %s\n",
            compsrc->GetLabel(),num,num==1?"":"s",compdst->GetLabel() );
    if (!TextWriteStr(buffer)) {
        TextEndReport();
        mayday("Error writing to report");
        return;
    }

    // Print out the actual interfaces for the objects exchanged.

    for ( i=0; i<num; i++ ) {
        gid = ap->GetContributorId(i);
        GraphArc *ap = editor->FindArc(gid);
        if ( ap != NULL ) {
            GraphNode* dst = ap->GetDst();
            GraphNode* src = ap->GetSrc();
            char title[MAXPATHLEN];
            // STRING1
            if ( strlen( src->GetLabel() ) +
                strlen( GetNodeNameAtLevel( compsrc->GetLevel(), src) ) +
                5 > MAXPATHLEN )
                Mess("Title buffer too small in exact");
            else
                // STRING1
                sprintf(title,"%s (%s)",src->GetLabel(),
                    GetNodeNameAtLevel(compsrc->GetLevel(),src));
            if ( strlen( dst->GetLabel() ) +    
                strlen( GetNodeNameAtLevel(compdst->GetLevel(),dst) ) +
                40  > BUFSIZE )
                Mess("Buffer too small in exact");
            else
                sprintf( buffer,"    %-20s <-- %s (%s)\n",title,
                    dst->GetLabel(),GetNodeNameAtLevel(compdst->GetLevel(),dst));
        }
        // STRING1, this one is definitely less than BUFSIZE!
        else
            sprintf(buffer,"   Contributor not loaded yet??\n");
        if (report_type == REPORT_VERBOSE) {
            if (!TextWriteStr(buffer)) {
                TextEndReport();
                return;
            }
        }
    }
    TextEndReport();
    CharType cmd("w_texteditor ");
    cmd.Append(TextGetFileName());
    exec_command(cmd.string());
}

//------------------------------------------------------------------------------
// Operations for lists of strings. Used for sorting the output for the
// exact interfaces.
//------------------------------------------------------------------------------

// Clear the lists given, and reset the counters provided.

static void ClearLists(Counter namelist[], int& numcount,
                char** list,int& count)
{
    int i;

    if ( (count>0 && !list) || (numcount>=0 && !namelist) )
    {
        mayday("NULL list in ClearList");
        return;
    }

    for ( i=0; i<=numcount; i++ )
    {
        namelist[i].Init();
    }
    numcount = -1;

    for ( i=0; i<count; i++ )
    {
        if (list[i] != NULL) delete[] list[i];
        list[i] = NULL;
    }
    count = 0;
}

// Clear and delete the lists provided.

static void DeleteLists(Counter namelist[], int& numcount,
                char** list,int& count)
{
    ClearLists(namelist,numcount,list,count);
    delete[] namelist;
    delete[] list;
    list = NULL;
}

// Comparison function for two strings.

static int compare_func( const void *e1, const void *e2 )
{
    char *s1, *s2;
    s1 = *((char **)e1);
    s2 = *((char **)e2);
    return strcmp( s1, s2 );
}

// Comparison function for two counter objects. Sort the objects on the
// count rather on the text string.

static int compare2_func( const void *e1, const void *e2 )
{
    Counter* c1 = (Counter*)e1;
    Counter* c2 = (Counter*)e2;
    if (c1->Count() < c2->Count()) return 1;
    if (c1->Count() > c2->Count()) return -1;
    return strcmp(c1->Name(),c2->Name());
}

// Output the lists given for an exact interface report.

static int OutputLists(Counter namelist[], int& nnum, char* opstr,
                char** list, int& num)
{
    int i;

    // Sort the Counter list on the count field.

    if (nnum > 0) qsort((char*)namelist,nnum+1,sizeof(Counter), compare2_func);

    // Print out a report on the usage counts of the objects.

    for( i=0; i<=nnum; i++ )
    {
        if ( strlen( opstr ) + sizeof(int)  + 40  > BUFSIZE )
            mayday("exact.C, buffer not long enough");
        else
        {
            int j = namelist[i].Count();
            sprintf( buffer,"    %-20s %s %d object%s.\n",
                    namelist[i].Name(),opstr,j,j==1?"":"s");
        }
        if (!TextWriteStr(buffer))
        {
            DeleteLists(namelist,nnum,list,num);
            return 0;
        }
    }

    // Clean up the lists

    if (nnum >= 0)
    {
        if (!TextWriteStr("\n"))
        {
            DeleteLists(namelist,nnum,list,num);
            return 0;
        }
    }

    // Sort the list of strings alphabetically, and print out a report on
    // the individual objects and their exact dependecies.

    if (num > 0) qsort((char*)list,num,sizeof(char*),compare_func);
    for(i=0; i<num; i++)
    {
        if (!TextWriteStr(list[i]))
        {
            DeleteLists(namelist,nnum,list,num);
            return 0;
        }
    }
    return 1;
}

// Check the counter list for a name, and if it is found increment its count.
// If it is not found, add it to the list.

static void CheckCount(Counter namelist[],int& numcount,char* name)
{
    int i;

    for(i=0; i<=numcount; i++)
    {
        if (strcmp(namelist[i].Name(),name) == 0)
        {
            namelist[i].Increment();
            return;
        }
    }
    numcount++;  
    namelist[numcount].SetName(name); 
}

//----------------------------------------------------------------------------
// Node exact interface routines.
//----------------------------------------------------------------------------

void NodeInterface(GraphNode *np, ArcType type, RigiBool useall, ArcSet &require,
                    ArcSet &provide, ArcSet &internal)
{
    // Empty the hash lists that contain the children of this node,
    // the provisions, requirements, and internalizations of this node,
    // and the calls from/into this node and its children. We use the TreeList
    // module to create the children, depinto and depfrom lists.

    TreeInitLists();
    TreeSetType(type);
    if (useall) TreeUseAllTypes();

    provide.Empty();
    require.Empty();
    internal.Empty();

    // Now that the lists are empty, create them for the current node.

    np->TreeDFS( outLevArcs,(VarArgFunRef)(&(TreeMakeLists)),NULL,NULL,NULL );

    // Now that we have the lists, check all the calls from this node's
    // subtree. If the destination of one of these calls is inside the
    // subtree then that call has been internalized. If the destination
    // is outside the subtree, then the destination object is a
    // requirement of this subsystem.

    
    GraphArc *ap; AvlPos *P;
    for ( ap = depfrom.First(P); ap; ap = depfrom.Next(P) )
    {
        GraphNode* dst = ap->GetDst();
        if (children.Find(dst->GetId()) == NULL) require.Add(ap);
        else                                    internal.Add(ap);
    }

    // Now check all of the calls into this subtree. If the source
    // is outside the subtree, then the call determines an object
    // that this subsystem provides.

    for ( ap = depinto.First(P); ap; ap = depinto.Next(P) )
    {
        GraphNode* src = ap->GetSrc();
        if ( children.Find(src->GetId()) == NULL ) provide.Add( ap );
    }
}

// Perform an exact interface on the given node. Use the type in exacttype if
// alltypes is RigiFalse. These values either default or are set in the
// functions that use this function.

static void LocalNodeExactInterface( GraphNode* np )
{
    static ArcSet require,internal,provide;
    AvlPos *P;

    if ( np == NULL )
    {
        RigiError(ERRORFLAG, "NULL node in NodeExactInterface");
        return;
    }
    // Get the node interface.
    NodeInterface(np,exacttype,alltypes,require,provide,internal);

    RigiReport report_type = editor->GetReportGeneration();

    // Write the header for this subsystem.
    if ( strlen(np->GetLabel()) + 50 > BUFSIZE )
        mayday("Exact Interface, buffer not big enough");
    else
        sprintf(buffer,"Exact Interface for subsystem %s:\n\n",np->GetLabel());

    if (!TextWriteStr(buffer)) return;

    // Get the largest number of elements in either of the lists

    int n = provide.Size();
    if (n < require.Size()) n = require.Size();
    if (n < internal.Size()) n = internal.Size();

    // Create the lists needed for sorting

    Counter* namelist;
    // P8
    namelist = new Counter[ n ];
    int numcount = -1;
    char** list = new char*[n];
    int count = 0;

    // Get a simple or verbose description of the provisions. Make sure we free
    // up the name lists if the writes fail and we abort the operation.

    // STRING1, this one will be ok
    sprintf(buffer,"Provisions = %d\n",provide.Size());
    if (!TextWriteStr(buffer))
    {
        DeleteLists(namelist,numcount,list,count);
        mayday("LocalNodeExactInterface, couldn't write string");
        return;
    }
    if ( report_type == REPORT_VERBOSE )
    {
        // If it is not a simple request, build each of the text strings for
        // each of the arcs in the provide, require, and internal lists, sort
        // them, and output the report.

        if (!TextWriteStr("\n"))
        {
            DeleteLists(namelist,numcount,list,count);
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }
        for ( GraphArc *ap = provide.First(P); ap; ap = provide.Next(P) )
        {
            GraphNode *src = ap->GetSrc();
            GraphNode *dst = ap->GetDst();
            // STRING1
            if ( strlen( src->GetLabel() ) +
                strlen( GetNodeNameAtLevel( np->GetLevel(), src ) ) +
                40 > BUFSIZE )
                Mess("Buffer too small in exact");
            else
                sprintf( buffer,"    %-20s --> %s (%s)\n",
                        dst->GetLabel(),src->GetLabel(),
                        GetNodeNameAtLevel( np->GetLevel(),src ));
            list[count] = new char[strlen(buffer)+1];
            strcpy(list[count],buffer);
            CheckCount(namelist,numcount,dst->GetLabel());
            count++;
        }

        if (!OutputLists(namelist,numcount,"provides",list,count))
        {
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }
        ClearLists(namelist,numcount,list,count);
        if (!TextWriteStr("\n"))
        {
            DeleteLists(namelist,numcount,list,count);
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }
    }

    // Get a simple or verbose description of the requirements.

    // STRING1, this one will be big enough
    sprintf(buffer,"Requirements = %d\n",require.Size());
    if (!TextWriteStr(buffer))
    {
        DeleteLists(namelist,numcount,list,count);
        mayday("LocalNodeExactInterface, couldn't write string");
        return;
    }
    if ( report_type == REPORT_VERBOSE )
    {
        if (!TextWriteStr("\n"))
        {
            DeleteLists(namelist,numcount,list,count);
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }
        numcount = -1;
        count = 0;
        for ( GraphArc *ap = require.First(P); ap; ap = require.Next(P) )
        {
            GraphNode *src = ap->GetSrc();
            GraphNode *dst = ap->GetDst();
            // STRING1
            if ( strlen( dst->GetLabel() ) +
                strlen( GetNodeNameAtLevel( np->GetLevel(),dst) ) +
                40 > BUFSIZE )
                mayday("exact interface, buffer not big enough!");
            else  
                sprintf( buffer,"    %-20s <-- %s (%s)\n",
                        src->GetLabel(),dst->GetLabel(),
                        GetNodeNameAtLevel( np->GetLevel(),dst ));
            list[count] = new char[strlen(buffer)+1];
            strcpy(list[count],buffer);
            CheckCount(namelist,numcount,src->GetLabel()); 
            count++; 
        }

        if (!OutputLists(namelist,numcount,"requires",list,count))
        {
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }

        ClearLists(namelist,numcount,list,count);
        if (!TextWriteStr("\n"))
        {
            DeleteLists(namelist,numcount,list,count);
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }
    }

    // Get a simple or verbose description of the internalizations.

    // STRING1, this one should be big enough
    sprintf(buffer,"Internalizations = %d\n",internal.Size());
    if (!TextWriteStr(buffer))
    {
        DeleteLists(namelist,numcount,list,count);
        mayday("LocalNodeExactInterface, couldn't write string");
        return;
    }
    if ( report_type == REPORT_VERBOSE )
    {
        if (!TextWriteStr("\n"))
        {
            DeleteLists(namelist,numcount,list,count);
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }
        numcount = -1;
        count = 0;
        for ( GraphArc *ap = internal.First(P); ap; ap=internal.Next(P) )
        {
            GraphNode *src = ap->GetSrc();
            GraphNode *dst = ap->GetDst();
            // STRING1
            if ( strlen( dst->GetLabel() ) + 40 > BUFSIZE )
                Mess("Buffer too small in exact");
            else
                sprintf( buffer,"    %-20s <-- %s\n",
                    src->GetLabel(),dst->GetLabel());
            list[count] = new char[strlen(buffer)+1];
            strcpy(list[count],buffer);
            CheckCount(namelist,numcount,src->GetLabel()); 
            count++;
        }

        if (!OutputLists(namelist,numcount,"internalizes",list,count))
        {
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }

        ClearLists(namelist,numcount,list,count);
        if (!TextWriteStr("\n"))
        {
            DeleteLists(namelist,numcount,list,count);
            mayday("LocalNodeExactInterface, couldn't write string");
            return;
        }
    }

    DeleteLists(namelist,numcount,list,count);
}

void NodeExactInterface(GraphNode* np)
{
    alltypes = RigiTrue;

    // Turn on text batching to speed up display, then produce the exact
    // interface report, and turn off the batching.

    TextStartReport();
    LocalNodeExactInterface(np);
    TextEndReport();
    CharType cmd("w_texteditor ");
    cmd.Append(TextGetFileName());
    exec_command(cmd.string());
}

void NodeExactInterface( GraphNode* np,ArcType t )
{
    alltypes = RigiFalse;
    exacttype = t;

    // Turn on text batching to speed up display, then produce the exact
    // interface report, and turn off the batching.

    TextStartReport();
    LocalNodeExactInterface(np);
    TextEndReport();
    CharType cmd("w_texteditor ");
    cmd.Append(TextGetFileName());
    exec_command(cmd.string());
}

void NodeListExactInterface(RigiNodeList& list)
{
    alltypes = RigiTrue;
    RigiNode* rnp = list.First();

    // Turn on text batching to speed up display, then produce the exact
    // interface report, and turn off the batching.

    TextStartReport();
    while(rnp != NULL)
    {
        LocalNodeExactInterface(rnp->GetNode());
        rnp = list.Next();
    }
    TextEndReport();
    CharType cmd("w_texteditor ");
    cmd.Append(TextGetFileName());
    exec_command(cmd.string());
}

void NodeListExactInterface(RigiNodeList& list, ArcType t)
{
    alltypes = RigiFalse;
    exacttype = t;
    RigiNode* rnp = list.First();

    // Turn on text batching to speed up display, then produce the exact
    // interface report, and turn off the batching.

    TextStartReport();
    while(rnp != NULL)
    {
        LocalNodeExactInterface(rnp->GetNode());
        rnp = list.Next();
    }
    TextEndReport();
    CharType cmd("w_texteditor ");
    cmd.Append(TextGetFileName());
    exec_command(cmd.string());
}

