
// $Author: jmcdanie $
// $Date: 1996-09-14 00:11:41 $
// $Source: /root/cvs/Rigi/rigiedit/filterops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: filterops.C,v 1.35 1996-09-14 00:11:41 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef FILTEROPS_H
#include "filterops.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"		// for RinitGetDomain()
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"		// for RutilActiveWindow()
#endif
#include "win_util.h"		// for cursor stuff
#ifndef FILTERSET_H
#include "cl_filterset.h"
#endif

const int DRAWWITHBOX           = 0;
const int DRAWWITHICON          = 1;
const int HIDENAME              = 2;
const int SHOWNAME              = 3;
const int FILTER                = 4;
const int NOFILTER              = 5;
const int ARCTYPEFILTER         = 6;
const int NODETYPEFILTER        = 7;

const int APPLYTYPEFILTERS      = 8;
const int FIRSTFILTEREDNODE     = 9;
const int NEXTFILTEREDNODE      = 10;
const int NUMFILTEREDNODES      = 11;
const int ARCTYPEFILTERED       = 12;
const int NODETYPEFILTERED      = 13;

static void SetIntResult(int val, CharType &result) {
    char str[12];
    if (result.Length()) result.Append(" ");
    sprintf(str, "%d", val);
    result.Append(str);
}

static int NodesSelected() {
    int n = 0;
    if ((n = editor->NumSelectedNodes()) < 1)
        RigiError(ERRORFLAG, "At least one node must be selected");
    return n;
}

static RigiBool GetRigiWindow(char *argv, RigiWindow *&rw) {
    int winNum = 0;
    if (argv != NULL && *argv) winNum = atoi(argv);
    if (winNum) {
        if ((rw = editor->GetWindowByNumber(winNum)) != NULL) return RigiTrue;
        return RigiError(ERRORFLAG, "Incorrect window (%d) specified", winNum);
    }
    return ((rw = RutilActiveWindow()) == NULL)? RigiFalse : RigiTrue;
}

int FilterOp(int argc, char **argv, CharType &result) {
    GraphNode *n;
    RigiNode* rn;
    int r = 0, type;
    AvlPos *P;
    RigiDomain *domain;

    RigiWindow *rw, *activewin;


    // Set the cursor to the wait cursor

    WindowSetCursor(CURSOR_WAIT);

    switch (atoi(argv[1])) { 
    case FILTER:
    {
        if (!NodesSelected()) break;
        activewin = RutilActiveWindow();
        RigiNodeList slist;
        editor->GetSelectionList( slist, activewin );
        editor->SetFilter( slist, activewin );

        // MJW: what about selected nodes in other windows??
        activewin->UpdateFilterMessage();
        activewin->RigiDisplay();		// MJW HACK!
        break;
    }
    case NOFILTER:

        activewin = RutilActiveWindow();
        editor->ResetFilter( activewin );
        activewin->UpdateFilterMessage();
        activewin->RigiDisplay();		// MJW HACK!
        break;

    case HIDENAME:

        if (!NodesSelected()) break;
        activewin = RutilActiveWindow();
        n = editor->FirstSelectedNode(P);
        while (n){
            rn = activewin->FindRigiNode(n);
            if (rn) rn->SetHideName( RigiTrue );
            n = editor->NextSelectedNode(P);
        }
        // MJW, unfortunately because can't handle new arc posns
        activewin->RigiDisplay();
        break;

    case SHOWNAME:

        if (!NodesSelected()) break;
        activewin = RutilActiveWindow();
        n = editor->FirstSelectedNode(P);
        while (n) {
            rn = activewin->FindRigiNode(n);
            if (rn) rn->SetHideName( RigiFalse );
            n = editor->NextSelectedNode(P);
        }
        // MJW, unfortunately because can't handle new arc posns
        activewin->RigiDisplay();
        break;

    case DRAWWITHICON:

        if (!NodesSelected()) break;
        activewin = RutilActiveWindow();
        n = editor->FirstSelectedNode(P);
        while (n) {
            rn = activewin->FindRigiNode(n);
            if (rn) {
                rn->Draw();
                rn->SetDrawWithIcon( RigiTrue );
                rn->Draw();
            }
            n = editor->NextSelectedNode(P);
        }
        break;

    case DRAWWITHBOX: 

        if (!NodesSelected()) break;
        activewin = RutilActiveWindow();
        n = editor->FirstSelectedNode(P);
        while (n) {
            rn = activewin->FindRigiNode(n);
            if (rn) {
                rn->Draw();
                rn->SetDrawWithIcon( RigiFalse );
                rn->Draw();
            }
            n = editor->NextSelectedNode(P);
        }
        break;

    case ARCTYPEFILTER:
        if (!GetRigiWindow(argv[3], rw)) break;
        domain = RinitGetDomain();

        if (!strcmp(argv[2],"any")) {
            int j = domain->NumArcTypes();
            for ( int i=0; i<j; i++ )
                rw->GetArcTypeFilters()->SetFilter(i);
        } else if (!strcmp(argv[2],"none")) {
            int j = domain->NumArcTypes();
            for (int i = 0; i < j; i++)
                rw->GetArcTypeFilters()->ClearFilter(i);
        } else {
            type = domain->ArcTypeFromName(argv[2]);
            if (rw->GetArcTypeFilters()->GetFilter(type))
                rw->GetArcTypeFilters()->ClearFilter(type);
            else
                rw->GetArcTypeFilters()->SetFilter(type);
        }
        break;

    case NODETYPEFILTER:

        if (!GetRigiWindow(argv[3], rw)) break;
        domain = RinitGetDomain();
        if (!strcmp(argv[2],"any")) {
            int j = domain->NumNodeTypes();
            for ( int i=0; i<j; i++ )
                rw->GetNodeTypeFilters()->SetFilter(i);
        } else if (!strcmp(argv[2],"none")) {
            int j = domain->NumNodeTypes();
            for ( int i=0; i<j; i++ )
                rw->GetNodeTypeFilters()->ClearFilter(i);
        } else {
            type = domain->NodeTypeFromName( argv[2] );

            if (rw->GetNodeTypeFilters()->GetFilter(type))
                rw->GetNodeTypeFilters()->ClearFilter(type);
            else
                rw->GetNodeTypeFilters()->SetFilter(type);
        }
        break;

    case APPLYTYPEFILTERS:
        if (!GetRigiWindow(argv[3], rw)) break;
        domain = RinitGetDomain();

        if (!strcmp(argv[3], "arc") || !strcmp(argv[3], "any")) 
            rw->GetArcTypeFilters()->ApplyFilters();
        if (!strcmp(argv[3], "node") || !strcmp(argv[3], "any"))
            rw->GetNodeTypeFilters()->ApplyFilters();

        rw->UpdateFilterMessage();
        rw->RigiDisplay();
        break;

    // The following operations all return a node id

    case FIRSTFILTEREDNODE:

        rn = editor->FirstFilteredNode();
        if (rn) r = rn->GetNode()->GetId();
        break;

    case NEXTFILTEREDNODE:

        rn = editor->NextFilteredNode();
        if (rn) r = rn->GetNode()->GetId();
        break;

    case NUMFILTEREDNODES:

        r = editor->NumFilteredNodes();
        break;

    case ARCTYPEFILTERED:

        if (!GetRigiWindow(argv[3], rw)) break;
        domain = RinitGetDomain();
        type = domain->ArcTypeFromName(argv[2]);
        r = rw->GetArcTypeFilters()->GetFilter(type);
        break;

    case NODETYPEFILTERED:

        if (!GetRigiWindow(argv[3], rw)) break;
        domain = RinitGetDomain();
        type = domain->NodeTypeFromName(argv[2]);
        r = rw->GetNodeTypeFilters()->GetFilter(type);
        break;

    default:
        RigiError(ERRORFLAG, "Unknown filter operation (%s)", argv[1]);
        break;
    }

    WindowSetCursor(CURSOR_NORMAL);
    SetIntResult(r, result);
    return r;
}

