// $Author: jmcdanie $
// $Date: 1996-10-26 00:17:13 $
// $Source: /root/cvs/Rigi/rigiedit/graphutil.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#ifndef __CTYPE_H
#include <ctype.h>      // for isspace()
#endif
#ifndef __STDIO_H
#include <stdio.h>      // for type FILE
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGISTART_H
#include "cl_rigistate.h"
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef ACTIVEARC_H
#include "cl_activearc.h"
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif

#ifndef GRAPHERR_H
#include "grapherr.h"   
#endif

RigiBool GutilCheckList(RigiNodeList& list, RigiWindow* window) {
    for ( RigiNode* rn=list.First(); rn!=NULL; rn=list.Next() )
        if ( !window->FindNode( rn ) ) return RigiFalse;
    return RigiTrue;
}

GraphArc *GutilCreateArc(GraphNode *src, GraphNode *dst, int arctype) {
    if (!src || !dst) mayday("NULL src or dst in GutilCreate_arc");
    GraphArc* a = new GraphArc( src,dst, arctype );

    if ( arctype == editor->LevelArc() ) dst->SetLevel();

    src->InsertArc( a );
    dst->InsertArc( a );

    GutilAddArc(a);
    return a;
}

GraphNode* GutilCreateNode(GraphNode* parent, char* str, int type) {
    // Create the node

    GraphNode* n = new GraphNode( str );
    n->SetType(type);

    // Create the level arc between this node and its parent. This must be done
    // before the next step (setting the node's position) because SetLevel()
    // uses the level arc to determine the node's level.

    if (parent) {
        GutilCreateArc( parent,n,editor->LevelArc() );
        n->SetLevel();
    }

    // Add node to hash list

    GutilAddNode(n);
    return n;
}

GraphNode* GutilNodeExists(int id) { return editor->FindNode((NodeId)id); }
GraphArc* GutilArcExists(int id) { return editor->FindArc((ArcId)id); }

void GutilAddArc(GraphArc* ap) {
    // Add to the hash list. If the arc has not been created in the database
    // yet, an error may occur, as we are trying to add an arc with a hash id
    // that has not been received. If this is done more than once, we will be
    // adding more than one with the same hash id.
    if (!(editor->AddArc(*ap))) {
        GraphArc *a = editor->FindArc(ap->GetId());
        if (a != NULL) {
            mayday("GutilCreateArc: GraphArc from %s to %s already exists!",
                a->GetSrc()->GetLabel(), a->GetDst()->GetLabel());
        } else {
            mayday("GutilCreateArc: editor GraphArc not found!!!");
        }
        if (ap != NULL) {
            mayday("    duplicate id = %d, arc from %s to %s",
                ap->GetId(), ap->GetSrc()->GetLabel(), ap->GetDst()->GetLabel());
        }
    }
}

void GutilAddNode(GraphNode* np) {
    // Add to the hash list. If the node has not been created in the database
    // yet, an error may occur, as we are trying to add an node with a hash id
    // that has not been received. If this is done more than once, we will be
    // adding more than one with the same hash id, resulting in the error.

    if (!editor->AddNode(np))
        mayday("WindowCreateNode: Node already exists. HELP!");
}

RigiBool GutilDeleteArc(GraphArc* a) {
    RigiBool rtn = RigiTrue;

    if (a==NULL) mayday("NULL arc in WindowDeleteArc()");

    // Clean up anything that might point to the arc in the editor

    if (editor->RemoveArc( a->GetId() )) {
        editor->DeselectArc();
    } else {
        mayday("arc (%d) is missing from editor arc set",a,a->GetId());
        rtn = RigiFalse;
    }


    // Deleting the activearc will take care of any rigiarcs

    ActiveArc *aa = (ActiveArc*)(a->GetActiveArc());
    if (aa) delete aa;
    a->SetActiveArc(NULL);
    delete a;
    return rtn;
}

// Remove ALL the stuff in the editor associated with this node

void GutilDeleteNode(GraphNode* n) {
    //Mess("Deleting node %d (%d %s)",n,n->GetId(),n->GetLabel() );

    // Remove RigiWindows associated with this node
    GutilDeleteNodesWindows( n );

    // Remove the GraphNode from the editor's sets
    editor->DeselectNode( n );
    editor->RemoveNode( n );

    // The destructor will delete all of the RigiNodes associated with this
    // guy for us, but it will not delete the RigiNodes out of the lists that
    // it doesn't know about: the RigiWindow lists.
    // MJW: dead slow!!!

    ActiveNode *an = (ActiveNode*)(n->GetActiveNode());
    if (an) for ( RigiNode* rn = an->First(); rn; rn = an->Next() )
        rn->GetWindow()->RemoveNode( rn );

    // Delete the ActiveNode. NOTE: The destructor for the ActiveNode deletes
    // all of the RigiNodes that the ActiveNode represents.

    if (an) delete an;
    n->SetActiveNode(NULL);
    delete n;
}

void GutilDeleteNodesWindows(GraphNode* np) {
    // Compile a list of RigiWindows that are based on this node.

    RigiWindowList dellist ;
    RigiWindow* rw;
    for ( rw=editor->FirstWindow(); rw; rw=editor->NextWindow()) {
        // MJW HACK: do NOT kill the basewindow
        if ( rw==RigiBaseWindow() ) continue;

        // Delete only those windows where the node is the only
        // remaining parent.  If the node is in both the parent
        // and child window then this is a window opened on
        // a selection of nodes: this is a hack!.  In this case
        // remove the parent but don't delete the node itself.

        if (rw->NodeIsParent(np->GetId())) {
            rw->RemoveParent(np->GetId());
            if (!rw->NumParents())
                dellist.Insert(rw);
        }
    }

    // Delete the RigiWindows that are in the delete list.
    for (rw=dellist.First(); rw; rw=dellist.Next()) {
        editor->ResetFilter( rw );
        editor->RemoveWindow(rw);
        delete rw;
    }
}

void GutilChangeNodeName(GraphNode *n, char *name) {
    editor->RemoveNode(n); n->SetLabel(name); editor->AddNode(n);
}

void GutilChangeNodeId(GraphNode*n, int id) {
    editor->RemoveNode(n); n->SetId(id); editor->AddNode(n);
}

void GutilChangeArcType(GraphArc *a, int type) {
    ActiveArc *aa = (ActiveArc*)(a->GetActiveArc());

    // Erase all the old rigiarcs
    if (aa) aa->DrawAll();

    // Src and dst nodes' arcsets are sorted by type, so ...
    a->GetSrc()->RemoveArc(a);
    if (a->GetDst()!=a->GetSrc()) a->GetDst()->RemoveArc(a);
    a->SetType(type);
    a->GetSrc()->InsertArc(a);
    if (a->GetDst()!=a->GetSrc()) a->GetDst()->InsertArc(a);

    // Redraw the rigiarcs for the arc, with new type
    if (aa) aa->DrawAll();
}

RigiNode *GutilNewNode(RigiWindow *rw, char *name, int type, int x, int y) {
    // Find parent, and create a level arc between it and new node
    GraphNode *parent = rw->GetNode();
    if (!name) name = "new";
    GraphNode *np = GutilCreateNode( parent,name,type );

    // Create a new rigi node and insert it in the list for this window
    RigiNode* rn = new RigiNode( np,rw );

    // Calc where the window origin is wrt world coords.
    Point p;
    Point2d p2d;
    p.x = x; p.y = y; p.z = 0;
    p = rw->Canvas2World(rw->Window2Canvas(p));
    p2d.x = p.x; p2d.y = p.y;
    rn->SetPosition( p2d );

    if (rw->GetType() != WIN_OVERVIEW ) {
        // Update the NDC position of the source node

        NDC_Point2d ndcp = rw->World2NDC2d(p2d);
        rn->GetNode()->SetNDCPosition(ndcp);
    }

    rw->InsertNode( rn );
    rn->Draw();
    return rn;
}

// Function that creates a new arc based on the context of the type of
// window it is in and other various restrictions. Lots of hacks that our
// "general" editor shouldn't have. The function returns a pointer
// to the arc if the arc already exists, and NULL if the arc was created.

GraphArc *GutilNewArc( RigiWindow *rw, int arctype, RigiNode *sn1, RigiNode *sn2) {
    RigiDomain *domain = RinitGetDomain();

    // MAD Note:  I have already done the following checks in
    // win_event.C (this is where Editop with ARC_CREATE_BASE)
    // is called from in interactive mode, but I added this code
    // again, in case this call is made from a script.

    int wintype = rw->GetType();
    if ( wintype==WIN_SCRATCH ) return NULL;

    // use controlpanel type if input parameter is -1
    if ( wintype==WIN_OVERVIEW ) arctype = editor->LevelArc();
    if ( arctype== -1) arctype = editor->GetArcType();
    if ( arctype==ANY_ARC ) arctype = (editor->LevelArc() == 0) ? 1 : 0;

    // do not permit a second arc of the same type
    GraphNode* src = sn1->GetNode();
    GraphNode* dst = sn2->GetNode();

    if (!src || !dst) mayday("null src or dst in GutilNewArc!");

    // Not permitting multiarcs yet
    GraphArc *a;
    if ( a = src->FindArc( dst, outArcs ) ) return a;

    a = new GraphArc(src,dst,arctype);
    src->InsertArc( a );
    dst->InsertArc( a );

    GutilAddArc(a);

    // Make a new rigiarc for the arc
    RigiArc *rap = new RigiArc( a,rw );
    sn1->InsertArc( rap );
    if (sn2!=sn1) sn2->InsertArc( rap );
    rap->Draw();
    return NULL;
}

