
// $Author: jmartin $
// $Date: 2002-11-10 22:16:58 $
// $Source: /root/cvs/Rigi/rigiedit/nodeops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: nodeops.C,v 1.65 2002-11-10 22:16:58 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef RIGISTATE_H
#include "cl_rigistate.h"               // Global instance of editor's state
#endif

#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"         // Basic graph definitions
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Class definition for graph Nodes
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"           // Class definition for RigiDomain
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"       // Window menu state access functions
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef NODEOPS_H
#include "nodeops.h"
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif

#ifndef RIGI_H
#include "rigi.h"		// Basic control functions for the editor
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"		// Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"		// Utility control functions for the editor
#endif
#ifndef RIGIERR_H
#include "rigierr.h"		// Error reporting functions
#endif
#include "win_util.h"		// Window state access functions
#include "event.h"		// for EventUpdateNodeLabel()
#include "win_archierarchy.h"
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"	// Required to include rcl_interp.
#endif

const int COMARG = 0;
const int OPARG = 1;
const int IDARG = 2;
const int WIN = 2;
const int NODENAME = 2;
const int NODEID = 3;
const int TYPEARG = 3;
const int X_COORD = 3;
const int Y_COORD = 4;
const int DIRARG = 4;
const int FILTERARG = 5;
const int WINDOWARG = 6;
const int XFRAMEOFFSET = 20;
const int YFRAMEOFFSET = 20;

// -------------- utility functions ------------------------------

static  ArcDirection GetDirection(char *argv) {
    if (!strcmp(argv, "in")) return ARC_IN;
    if (!strcmp(argv, "out")) return ARC_OUT;
    if (!strcmp(argv, "any")) return ARC_INOUT;
    RigiError(ERRORFLAG, "Invalid arc direction (%s)", argv);
    return ARC_DIR_UNKNOWN;
}

static RigiBool ConstructArcSet(int argc, const char **argv, ArcFlags &arcset) {

    RigiDomain *domain = RinitGetDomain();
    arcset.ClearAll();
    for(int i = 0 ; i < argc ; i++) {
        ArcType type = domain->ArcTypeFromName((char*) argv[i]);
        if (type == ANY_ARC) {
            arcset.SetAll();
            break;
        } else if (type == -1) {
            if (!RigiError(ALERT, "Invalid arc type %s", argv[i]))
                return RigiFalse;
        } else
            arcset.SetBit(type);
    }
    return RigiTrue;
}

static RigiBool GetGraphNode(char *argv, GraphNode *&graphNode) {
    graphNode = editor->FindNode(atoi(argv));
    if (graphNode == NULL)
        return RigiError(ERRORFLAG, "Node %s not found", argv);
    return RigiTrue;
}

static RigiBool GetRigiWindow(char *argv, RigiWindow *&rw) {
    int winNum = atoi(argv);
    if (winNum) {
        if ((rw = editor->GetWindowByNumber(winNum)) != NULL) return RigiTrue;
        return RigiError(ERRORFLAG, "Incorrect window (%d) specified", winNum);
    }
    return ((rw = RutilActiveWindow()) == NULL)? RigiFalse : RigiTrue;
}

static RigiBool GetBaseWindow(RigiWindow *rw, BaseWindow *&basewin) {
    basewin = rw->Window();
    if (basewin == NULL)
        return RigiError(ERRORFLAG, "No base window");
    return RigiTrue;
}

static RigiBool NumArgsOK(int argc, int argsExpected) {
    if (argc != argsExpected)
        return RigiError(ERRORFLAG, "Wrong number of arguments (%d)", argc);
    return RigiTrue;
}

static void SetIntResult(int val, CharType &result) {
    char str[12];
    if (result.Length()) result.Append(" ");
    sprintf(str, "%d", val);
    result.Append(str);
}

    
static int OpenSingleNode(GraphNode *graphNode, CharType &result) {
    RigiWindow *rw, *activewin;
    BaseWindow *basewin;
    int x, y;

    if (!GetRigiWindow("", activewin)) return 0;
    if (!GetBaseWindow(activewin, basewin)) return 0;

    RigiDomain *domain = RinitGetDomain();

    WindowSetCursor(CURSOR_WAIT);
    basewin->Location(x, y); 
    x += XFRAMEOFFSET; y += YFRAMEOFFSET;

    Mess("Creating a window");

    // Put the node in a list for the constructor
    NodeSet nodes;
    nodes.Add(graphNode);

    // Set the bits for the types of arcs to traverse for the open
    // operation.

    ArcFlags types;
    Mess("After creating arc types");
    int level_type = editor->LevelArc();
    Mess("current type = %d", level_type);
    if (level_type == domain->ArcTypeFromName("any")) {
        types.SetAll();
        int tmptype = domain->ArcTypeFromName("refractive");
        if (tmptype != -1) types.ClearBit( tmptype );
    } else {
         types.SetBit(level_type);
    }

    // Create the new RigiWindow. Set the inclusive flag to 1 and the
    // number of levels to traverse to 1.

    rw = new RigiWindow(nodes, types, ARC_OUT, 0, 1, x, y);
    if (editor->GetInheritFilters()) rw->InheritTypeFilters(activewin);

    Mess("Done creating a window");

    // Add the window the editors list

    editor->AddWindow(rw);
    RutilSetActiveWindow(rw);
    rw->DrawStore();
    rw->Show();
    RutilRaiseWindow(rw);

    WindowSetCursor(CURSOR_NORMAL);
    SetIntResult(RutilActiveWindow()->GetWindowNumber(), result);
    return 1;
}

// -------------- case functions ------------------------------

static int OpenEditor(int argc, char *argv[], CharType& result) {
    if (editor->NumSelectedNodes() != 1) {
        RigiError(ERRORFLAG, "Exactly one node must be selected");
        return 1;
    }
    AvlPos *P;
    GraphNode *graphNode = editor->FirstSelectedNode(P);
    WindowTextEdit(graphNode);
    SetIntResult(RutilActiveWindow()->GetWindowNumber(), result);
    return 1;
}

static int Rename(int argc, char *argv[], CharType& result) {
    GraphNode *graphNode;

    if (!NumArgsOK(argc, 4)) return 0;

    if (*argv[NODEID] == 0) {
        // with three or fewer non-null arguments,
        // the command is "__node 1 newname" and a node must be selected
        if (editor->NumSelectedNodes() != 1)
            return RigiError(ERRORFLAG, "Exactly one node must be selected");
        AvlPos *P;
        graphNode = editor->FirstSelectedNode(P);
    } else if ((graphNode = editor->FindNode(atoi(argv[NODEID]))) == NULL) {
        return RigiError(ERRORFLAG, "Node %s not found", argv[NODEID]);
    }
    EventUpdateNodeLabel(graphNode, argv[NODENAME]);
    SetIntResult(RutilActiveWindow()->GetWindowNumber(), result);
    return 1;
}

static int Overview(int argc, char *argv[], CharType& result) {
    int x, y;
    RigiWindow *activewin;
    BaseWindow *basewin;
    
    if (!GetRigiWindow("", activewin)) return 0;
    if (!GetBaseWindow(activewin, basewin)) return 0;
    if (editor->NumSelectedNodes() != 1) {
        RigiError(ERRORFLAG, "Exactly one node must be selected");
        return 1;
    }

    AvlPos *P;
    GraphNode *graphNode = editor->FirstSelectedNode(P);

    RigiDomain *domain = RinitGetDomain();

    WindowSetCursor(CURSOR_WAIT);

    // Do an overview of the current hierarchy that is loaded in
    // at the current time. Do not load any more nodes at this time.
    int depth = editor->GetOverviewDepth();
    basewin->Location(x,y);
    x += XFRAMEOFFSET; y += YFRAMEOFFSET;

    RigiWindow *rw = new RigiWindow(graphNode, WIN_OVERVIEW,
        depth, domain->NumNodeTypes(), domain->NumArcTypes(), x, y);
    if (editor->GetInheritFilters()) rw->InheritTypeFilters(activewin);

    editor->AddWindow(rw);
    RutilSetActiveWindow(rw);
    rw->DrawStore();
    rw->Show();
    RutilRaiseWindow(rw);

    WindowSetCursor(CURSOR_NORMAL);
    SetIntResult(RutilActiveWindow()->GetWindowNumber(), result);
    return 1;
}

static int OpenNode(int argc, char *argv[], CharType& result) {
    if (editor->NumSelectedNodes() != 1) {
        RigiError(ERRORFLAG, "Exactly one node must be selected");
        return 1;
    }
    AvlPos *P;
    GraphNode *graphNode = editor->FirstSelectedNode(P);
    return OpenSingleNode(graphNode, result);
}

static int NodeNeighbors(int argc, char *argv[], CharType& result) {
    GraphNode *graphNode;
    ArcDirection dir;
    ArcFlags arcset;
    int filter;
    RigiWindow *rw = NULL;
    RigiDomain *domain = RinitGetDomain();

    if (!NumArgsOK(argc, 7)) return 0;

    // Get the type of arc to perform the operation on. We do this by parsing
    // the list of nodes given in the TYPEARG arguement. It is a Tcl list
    // that needs to be split, giving the types of the arcs to traverse.

    int arc_argc;
    const char **arc_argv;
    if (Tcl_SplitList(rcl_interp, argv[TYPEARG], &arc_argc, &arc_argv) != TCL_OK ||
        arc_argc <= 0) {
        RigiError(ERRORFLAG, "Invalid arc type set %s", argv[TYPEARG]);
        return 0;
    }
    int success = ConstructArcSet(arc_argc, arc_argv, arcset);
    free((char*) arc_argv);
    if (!success)
      return 0;

    // Get the filter flag that tells whether we should use all nodes or
    // just nodes that are not filtered. This isn't used yet...

    filter = atoi(argv[FILTERARG]);

    // Get the node from the editor's hash list
    if (!GetGraphNode(argv[IDARG], graphNode)) return 0;

    // Get the rigi window

    if (!GetRigiWindow(argv[WINDOWARG], rw)) return 0;

    // Process all the arctypes indicated in the arcset bits
    int ntypes = domain->NumArcTypes();
    for (int t = 0; t < ntypes; t++) {
        if (!arcset.GetBit(t)) continue;

	// Get the in/out flag value
	if ((dir = GetDirection(argv[DIRARG])) == ARC_DIR_UNKNOWN) return 0;

        // Process all the arcs of this node, of this particular type
        GraphArc *a; AvlPos *P;
        for (a = graphNode->FirstArc( P,t,dir ); a; a = graphNode->NextArc( P,t,dir )) {
            RigiArc *ra = NULL;

            // If filtering, skip invisible arcs

            if (filter) ra = rw->FindRigiArc(a);
            if (filter) if (!ra || !ra->isVisible()) continue;

            // Get the appropriate node based on whether we are
            // traversing an incoming or outgoing arc

            GraphNode *other = a->GetOtherNode(graphNode);
            RigiNode *rn = NULL;

            // Add node id to list we have created if not filtered

            if (filter) rn = rw->FindRigiNode(other);
            if (!rn || rn->isVisible())
                SetIntResult(other->GetId(), result);
        }
    }

    return 1;
}

static int NodeArcList(int argc, char *argv[], CharType &result) {
    GraphNode *graphNode;
    ArcDirection dir;
    ArcFlags arcset;
    int filter;
    RigiWindow *rw = NULL;
    RigiDomain *domain = RinitGetDomain();

    if (!NumArgsOK(argc, 7)) return 0;


    // Get the type of arc to perform the operation on. We do this by parsing
    // the list of nodes given in the TYPEARG arguement. It is a Tcl list
    // that needs to be split, giving the types of the arcs to traverse.

    int arc_argc;
    const char **arc_argv;
    if (Tcl_SplitList(rcl_interp, argv[TYPEARG], &arc_argc, &arc_argv) != TCL_OK ||
        arc_argc <= 0) {
        RigiError(ERRORFLAG, "Invalid arc type set %s", argv[TYPEARG]);
        return 0;
    }
    int success = ConstructArcSet(arc_argc, arc_argv, arcset);
    free((char*) arc_argv);
    if (!success)
      return 0;

    // Get the filter flag that tells whether we should use all nodes or
    // just nodes that are not filtered. This isn't used yet...

    filter = atoi(argv[FILTERARG]);

    // Get the node from the editor's hash list
    if (!GetGraphNode(argv[IDARG], graphNode)) return 0;

    // Get the rigi window

    if (!GetRigiWindow(argv[WINDOWARG], rw)) return 0;

    // Process all the arctypes indicated in the arcset bits
    int ntypes = domain->NumArcTypes();
    for (int t = 0; t < ntypes; t++) {
        if (!arcset.GetBit(t)) continue;

	// Get the in/out flag value
	if ((dir = GetDirection(argv[DIRARG])) == ARC_DIR_UNKNOWN) return 0;

        // Process all the arcs of this node, of this particular type
        GraphArc *a; AvlPos *P;
        for (a = graphNode->FirstArc( P,t,dir ); a; a = graphNode->NextArc( P,t,dir )) {
            RigiArc *ra = NULL;

            // If filtering, skip invisible arcs

            if (filter) if (rw) ra = rw->FindRigiArc(a);
            if (filter) if (!ra || !ra->isVisible()) continue;

            // Get the appropriate node based on whether we are
            // traversing an incoming or outgoing arc

            // Add the arc id to the list we are returning

            SetIntResult(a->GetId(), result);
        }
    }
    return 1;
}

static int NodeIdByPosition(int argc, char *argv[], CharType &result) {
    // args: 0:"__node", 1:"9", 2:window, 3:x, 4: y
    Point p;
    RigiWindow *rw;
    RigiNode *rn;

    if (!NumArgsOK(argc, 5)) return 0;

    if (!GetRigiWindow(argv[WIN], rw)) return 0;
    p.x = atoi(argv[X_COORD]);
    p.y = atoi(argv[Y_COORD]);
    p.z = 0;
    p = rw->Canvas2World(rw->Window2Canvas(p));
    rn = rw->FindNode(p);
    if (rn == NULL)
        result.Assign("");
    else
        SetIntResult(rn->GetNode()->GetId(), result);
    return 1;
}

static int Case99(int argc, char *argv[], CharType & /*result*/) {
    int ntypes = RinitGetDomain()->NumNodeTypes();
    ArcDirection dir;
    AvlPos *P;
    GraphNode *graphNode = editor->FirstSelectedNode(P);
    GraphArc *a;
    int i, j, t;

    printf("ARC_OUT\n");
    dir = ARC_OUT;
    for (i = 0, t = 0; t<ntypes; t++)
        for (a = graphNode->FirstArc(P, t, dir); a; i++,a = graphNode->NextArc(P, t, dir))
            printf("%d %d %d: %d %d  %s %s\n",a,a->GetId(),a->GetType(),
                a->GetSrc(),a->GetDst(), a->GetSrc()->GetLabel(),
                a->GetDst()->GetLabel());

    printf("ARC_IN\n");
    dir = ARC_IN;
    for (j = 0, t = 0; t < ntypes; t++)
        for (a = graphNode->FirstArc(P, t, dir); a; j++,a=graphNode->NextArc(P, t, dir))
            printf("%d %d %d: %d %d  %s %s\n",a,a->GetId(),a->GetType(),
                a->GetSrc(),a->GetDst(), a->GetSrc()->GetLabel(),
                a->GetDst()->GetLabel());
#if 0
    int k, m, p, q, r, s;
    printf("ARC_INOUT\n");
    dir = ARC_INOUT;
    for (k = 0, t = 0; t < ntypes; t++)
        for (a = graphNode->FirstArc(P, t, dir); a; k++,a=graphNode->NextArc(P, t, dir))
            printf("%d %d: %s %s\n",a,a->GetType(),
                a->GetSrc()->GetLabel(),a->GetDst()->GetLabel());

    printf("ARC_ANY\n");
    dir = ARC_INOUT;
    for (m = 0, a = graphNode->FirstArc(P, ANY_ARC, dir); a;
        m++,a=graphNode->NextArc(P, ANY_ARC, dir))
        printf("%d %d: %s %s\n",a,a->GetType(),
            a->GetSrc()->GetLabel(),a->GetDst()->GetLabel());

    printf("inArcs\n");
    for (p = 0, a = graphNode->FirstAdj(P,inArcs); a; p++, a = graphNode->NextAdj(P,inArcs))
        printf("%d %d: %s %s\n",a,a->GetType(),
            a->GetSrc()->GetLabel(),a->GetDst()->GetLabel());

    printf("outArcs\n");
    for (q = 0, a = graphNode->FirstAdj(P,outArcs); a; q++, a = graphNode->NextAdj(P,outArcs))
        printf("%d %d: %s %s\n",a,a->GetType(),
            a->GetSrc()->GetLabel(),a->GetDst()->GetLabel());

    printf("inLevArcs\n");
    for (r = 0, a = graphNode->FirstAdj(P,inLevArcs); a; r++, a = graphNode->NextAdj(P,inLevArcs))
        printf("%d %d: %s %s\n",a,a->GetType(),
            a->GetSrc()->GetLabel(),a->GetDst()->GetLabel());
    
    printf("outLevArcs\n");
    for (s = 0, a = graphNode->FirstAdj(P,outLevArcs); a; s++, a = graphNode->NextAdj(P,outLevArcs))
        printf("%d %d: %s %s\n",a,a->GetType(),
            a->GetSrc()->GetLabel(),a->GetDst()->GetLabel());

    printf("%d out arcs, %d in arcs, %d(%d) total\n",i,j,k,m);
    printf("%d  outArcs, %d  inArcs, %d outLev, %d inLev\n",q,p, s,r );

    if (i!=graphNode->ArcNum(ARC_OUT) || j!=graphNode->ArcNum(ARC_IN)
        || k != graphNode->ArcNum() || m!=k || (i+j)!=k)
        mayday("problem with arc counts!");

    if ((q+s) != graphNode->ArcNum(ARC_OUT) || (p+r) != graphNode->ArcNum(ARC_IN))
        mayday("problem with arc adj counts!");
#endif

    P = NULL;
    return 1;
}

static int Case100(int argc, char *argv[], CharType &/*result*/) {
    AvlPos *P;
    RigiWindow *rw;
    GraphNode *graphNode = editor->FirstSelectedNode(P);
    if (!GetRigiWindow("", rw)) return 0;
    RigiNode *rn = rw->FindRigiNode(graphNode);
    RigiArc *ra;

    printf("\nOUT rigiarcs\n");
    for (ra = rn->FirstArc(P,ARC_OUT); ra; ra = rn->NextArc(P,ARC_OUT))
        printf( "src=%d dst=%d\n", ra->GetSrc(), ra->GetDst() );
    printf("IN rigiarcs\n");
    for (ra = rn->FirstArc(P,ARC_IN); ra; ra = rn->NextArc(P,ARC_IN))
        printf( "src=%d dst=%d\n", ra->GetSrc(), ra->GetDst() );
    return 1;
}

int NodeOpOpenNode(GraphNode *graphNode) {
    // used by operationops.C
    CharType result;
    return OpenSingleNode(graphNode, result);
}

int NodeOp(int argc, char *argv[], CharType& result) {
    int op;

    // must have the operation argument to start with
    if (argc < OPARG + 1) {
        op = OPENNODE;
    } else {
        op = atoi(argv[OPARG]);
    }

    // Get the operation type and check to see if it is OK

    switch (op) {
    case OPENEDITOR:
        return OpenEditor(argc, argv, result);
    case RENAME:
        return Rename(argc, argv, result);
    case OVERVIEW:
        return Overview(argc, argv, result);
    case OPENNODE:
        return OpenNode(argc, argv, result);
    case NODENEIGHBORS:
        return NodeNeighbors(argc, argv, result);
    case NODEARCLIST:
        return NodeArcList(argc, argv, result);
    case NODEIDBYPOSITION:
        return NodeIdByPosition(argc, argv, result);
    case 99:
        return Case99(argc, argv, result);
    case 100:
        return Case100(argc, argv, result);
    default:
        return RigiError(ERRORFLAG, "Unknown node operation (%s)", argv[OPARG]);
    }
}
