
//
// rcl_commands.c: RCL commands and interpreter.	[MAD 17Aug93]

// $Author: jmartin $
// $Date: 2002-11-10 22:16:58 $
// $Source: /root/cvs/Rigi/rigiedit/rcl_commands.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: rcl_commands.C,v 1.83 2002-11-10 22:16:58 jmartin Exp $";

//------------------------------------------------------------------------

// All low level rcl commands are prefixed by "__" (double underscore).
// They can be used by the user, but their use is discouraged as
// they are not easy to use.  Higher level commands will be
// available in the rc.rcl file.  A user can customize these
// and add more in their own rc.rcl file.

#ifndef __STDIO_H
#include <stdio.h>
#endif
#include <unistd.h>

#if defined(__WINDOWS__) || defined(__WIN32__)
#include <windows.h>
#include <string.h>
#include <process.h>
#define PATHPARSECHAR '\\'
#define UNIXPATHPARSECHAR '/'
#endif

#include <tcl.h>	// for tcl calls

#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#include "CT.h"

// include the files for all operations possible
#ifndef WINDOWOPS_H
#include "windowops.h"
#endif
#ifndef SELECTOPS_H
#include "selectops.h"
#endif
#ifndef MOVEOPS_H
#include "moveops.h"
#endif
#ifndef SCALEOPS_H
#include "scaleops.h"
#endif
#ifndef NODEOPS_H
#include "nodeops.h"
#endif
#ifndef EDITOPS_H
#include "editops.h"
#endif
#ifndef FILEOPS_H
#include "fileops.h"
#endif
#ifndef STATISTICSOPS_H
#include "statisticsops.h"
#endif
#ifndef OPERATIONOPS_H
#include "operationops.h"
#endif
#ifndef VIEWOPS_H
#include "viewops.h"
#endif
#ifndef ARCOPS_H
#include "arcops.h"
#endif
#ifndef FILTEROPS_H
#include "filterops.h"
#endif
#ifndef RIGIPARAMOPS_H
#include "rigiparamops.h"
#endif
#ifndef EDITOROPS_H
#include "editorops.h"
#endif
#ifndef ARCTHRESHOPS_H
#include "arcthreshops.h"
#endif
#ifndef INFOOPS_H
#include "infoops.h"
#endif
#ifndef SERVEROPS_H
#include "serverops.h"
#endif
#ifndef SPECIALOPS_H
#include "specialops.h"		// weird ops, which use a varargs parm list
#endif
#ifndef EVENTOPS_H
#include "eventops.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef QUIT_H
#include "quit.h"
#endif

// #include "comops.h"

// need this for atoi
extern "C"
{
    int atoi( const char* ) ;
}

// Which set of operations?
static const int SELECTOP = 0;
static const int MOVEOP = 1;
static const int SCALEOP = 2;
static const int NODEOP = 3;
static const int EDITOP = 5;
static const int FILEOP = 6;
static const int STATISTICSOP = 7;
static const int OPERATIONOP = 8;
static const int VIEWOP = 9;
static const int ARCOP = 10;
static const int FILTEROP = 11;
static const int RIGIPARAMOP = 12;
static const int WINDOWOP = 13;
static const int EDITOROP = 14;
static const int INFOOP = 15;

// I am allowing a variable number of parameters to be
// passed in for the script commands.  The following give
// the default values used if some parameters are not
// used: (only some operations can be defaulted)
// which op?
static const int DEFAULT_SELECTOP = SELECTALL;
static const int DEFAULT_NODEOP = OPENNODE;
static const int DEFAULT_SCALEOP = SCALEWIN;
static const int DEFAULT_FILEOP = REFRESH;
static const int DEFAULT_STATISTICSOP = WINSTATS;
static const int DEFAULT_OPERATIONOP = PROJECTSYS;
static const int DEFAULT_VIEWOP = SAVEVIEW;
static const int DEFAULT_RIGIPARAMOP = CURRENTARCTYPEPARAM;
static const int DEFAULT_WINDOWOP = SELECTWINDOW;
static const int DEFAULT_EDITOROP = SELECTEDITOR;
static const int DEFAULT_ARCTHRESHOP = NUMARCSTHRESH;

// other variables needed
static const int DEFAULT_SHIFTDOWN = 0; // default to not having shift key down
static const int DEFAULT_X = 0;	        // used in move operations, default to
static const int DEFAULT_Y = 0;	        // upper left corner
static const int DEFAULT_FLAG = 0;      // for arc and node filters
static char* DEFAULT_STRING = ""; // for grep operations
static const int DEFAULT_NEWVALUE = 1;  // default = 1



// forward declare the rcl command procedures
static int cmdSelect( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdMove( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdEvent( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdScale( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdNode( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdEdit( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdFile( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdFilter( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdStatistics( ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdOperation(  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdServer(     ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdView(       ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdRigiParam(  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdWindow(     ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdEditor(     ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdQuit( 	  ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdCursor(     ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdInfo(       ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdSpecial(    ClientData,  Tcl_Interp*, int, char * argv[]);

static int cmdExecFG(     ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdExecBG(     ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdFork(       ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdExecVP(     ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdSleep(      ClientData,  Tcl_Interp*, int, char * argv[]);
static int cmdUnlink(     ClientData,  Tcl_Interp*, int, char * argv[]);

// int cmdTrap(       ClientData,  Tcl_Interp*, int, char * argv[]);

// this structure is used for defining rcl commands
typedef struct {
        char* name;  	    // name of the command
        Tcl_CmdProc *proc;  // Procedure that executes the command
} CmdInfo;

// initialize the array of feasible commands
static CmdInfo rclCmds[] =  {
    {"__select", 	(Tcl_CmdProc*) cmdSelect},
    {"__move",   	(Tcl_CmdProc*) cmdMove},
    {"__event",   	(Tcl_CmdProc*) cmdEvent},
    {"__scale",  	(Tcl_CmdProc*) cmdScale},
    {"__node",   	(Tcl_CmdProc*) cmdNode},
    {"__edit",   	(Tcl_CmdProc*) cmdEdit},
    {"__file",   	(Tcl_CmdProc*) cmdFile},
    {"__filter", 	(Tcl_CmdProc*) cmdFilter},
    {"__statistics",    (Tcl_CmdProc*) cmdStatistics},
    {"__operation",     (Tcl_CmdProc*) cmdOperation},
    {"__server",        (Tcl_CmdProc*) cmdServer},
    {"__view",	        (Tcl_CmdProc*) cmdView},
    {"__rigiparam",     (Tcl_CmdProc*) cmdRigiParam},
    {"__window",        (Tcl_CmdProc*) cmdWindow},
    {"__editor",        (Tcl_CmdProc*) cmdEditor},
    {"__cursor",        (Tcl_CmdProc*) cmdCursor},
    {"__info",	        (Tcl_CmdProc*) cmdInfo},
    {"__quit",	        (Tcl_CmdProc*) cmdQuit},
    {"__special",       (Tcl_CmdProc*) cmdSpecial},

    {"execfg",          (Tcl_CmdProc*) cmdExecFG},
    {"execbg",          (Tcl_CmdProc*) cmdExecBG},
    {"fork",            (Tcl_CmdProc*) cmdFork},
    {"execvp",          (Tcl_CmdProc*) cmdExecVP},
    {"sleep",           (Tcl_CmdProc*) cmdSleep},
    {"unlink",          (Tcl_CmdProc*) cmdUnlink},

    {NULL,	        (Tcl_CmdProc *)NULL}
}; // rclCmds


// A static buffer that is used to write temporary string results
static char result_buf[80];
static CharType chartype;

Tcl_Interp *rcl_interp = NULL;

static int ArgsError(int argc) {
    RigiError(ERRORFLAG, "Wrong number of arguments (%d)", argc);
    return RigiErrState();
}


// rcl command procedure definitions
// Self explanatory, this format allows there to be a variable
// number of parameters passed in for each command.  When less
// parameters are passed in, default values are selected.

static int cmdSelect( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    int returnvalue;
    if ( argc == 1 )
        returnvalue = SelectOp( DEFAULT_SELECTOP, DEFAULT_SHIFTDOWN, NULL );
    else if ( argc == 2 )
        returnvalue = SelectOp( atoi(argv[1]), DEFAULT_SHIFTDOWN, NULL );
    else if ( argc == 3 )
        returnvalue = SelectOp( atoi(argv[1]), atoi(argv[2]), NULL );
    else if ( argc == 4 )
        returnvalue = SelectOp( atoi(argv[1]), atoi(argv[2]), argv[3] );
    else
         return ArgsError(argc);
    sprintf( result_buf, "%d", returnvalue );
    Tcl_AppendResult(interp, result_buf, 0);
    return RigiErrState();
}

static int cmdScale( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    CharType result;

    if (ScaleOp(argc, argv, result))
        Tcl_AppendResult(interp, result.string(), 0);
    return RigiErrState();
}

static int cmdEdit( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    CharType result;

    if (EditOp(argc, argv, result))
        Tcl_AppendResult(interp, result.string(), 0);
    return RigiErrState();
}

static int cmdFile( ClientData, Tcl_Interp *interp, int argc, char *argv[] ) {
    if ( argc == 1 )
        FileOp( DEFAULT_FILEOP );
    else if ( argc == 2 )
        FileOp( atoi ( argv[1] ) );
    else if ( argc == 3 )
        FileOp( atoi( argv[1] ), argv[2] );
    else
        RigiError(ERRORFLAG, "Wrong # of args for File Operation");
    return RigiErrState();
}

static int cmdFilter( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    CharType result;

    FilterOp(argc, argv, result);
    Tcl_AppendResult(interp, result.string(), 0);
    return RigiErrState();
}

static int cmdNode( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    CharType result;

    if (NodeOp(argc, argv, result))
        Tcl_AppendResult(interp, result.string(), 0);
    return RigiErrState();
}

static int cmdEvent( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    CharType result;

    if (EventOp(argc, argv, result))
        Tcl_AppendResult(interp, result.string(), 0);
    return RigiErrState();
}

static int cmdMove( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    CharType result;

    if (MoveOp(argc, argv, result))
        Tcl_AppendResult(interp, result.string(), 0);
    return RigiErrState();
}

static int cmdStatistics( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    if ( argc == 1 )
        StatisticsOp(DEFAULT_STATISTICSOP);
    else if ( argc == 2 )
        StatisticsOp(atoi(argv[1]));
    else
        return ArgsError(argc);
    return RigiErrState();
}

static int cmdServer( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {

    // Get the operation to perform
    int op = atoi( argv[1] );

    if (argc == 2)
        ServerOp( op, NULL, NULL );
    else if (argc == 3)
        ServerOp( atoi( argv[1] ), argv[2], NULL );
    else if (argc == 4)
        ServerOp( atoi( argv[1] ), argv[2], argv[3] );
    else
        return ArgsError(argc);
    return RigiErrState();
}

static int cmdOperation( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    CharType result;

    // SCRIPTRETURN
    int numnodes; // number of nodes selected after this operation
    int op = atoi(argv[1]);
    if ( op == OPENGENERAL || op == MAKETREE ||
        op == GETCONTRIBS || op == MAKERTREE ) {
        OperationOp( argc, argv, result);
        Tcl_AppendResult( interp, result.string(), 0 );
        return RigiErrState();
    }
    if (argc == 1)
        numnodes = OperationOp( DEFAULT_OPERATIONOP, DEFAULT_STRING, 0);
    else if ( argc == 2 )
        numnodes = OperationOp(atoi( argv[1] ), DEFAULT_STRING, 0);
    else if ( argc == 3 )
        numnodes = OperationOp(atoi( argv[1] ), argv[2], 0);
    else if ( argc == 4 )
        numnodes = OperationOp(atoi( argv[1] ), argv[2], atoi( argv[3] ));
    else
        return ArgsError(argc);

    sprintf( result_buf, "%d", numnodes );
    Tcl_AppendResult(interp, result_buf, 0);
    return RigiErrState();
}

static int cmdView( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    if (argc == 1)
        ViewOp(DEFAULT_VIEWOP, DEFAULT_STRING);
    else if (argc == 2)
        ViewOp(atoi( argv[1] ), DEFAULT_STRING);
    else if (argc == 3)
        ViewOp(atoi( argv[1] ), argv[2]);
    else
        return ArgsError(argc);
    return RigiErrState();
}


static int cmdRigiParam( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    int r, op;
    CharType result;
    const int BUFFSIZE = 128;
    char str[BUFFSIZE];

    // If we are passed no operation, do the default op. Do we ever use the
    // default op? Hmmmm  [BDC May 4/94]

    if (argc == 1) {
        r = RigiParamOp( DEFAULT_RIGIPARAMOP, DEFAULT_NEWVALUE );
        sprintf( str, "%d", r );
        Tcl_AppendResult(interp, str, 0);
    } else if (argc >= 2) {

        // Get the type of operation that we are performing

        op = atoi( argv[1] );

        // If we are not processing a GETPARAMETER operation treat it the old
        // way by assuming that the next argument is an integer

        if (op == ARCSETGET || op == ARCSETCLEAR ||
            op == ARCSETSET || op == ARCSETADD || op == ARCSETDEL) {
            if (RigiParamOp( argc, argv, result ))
                Tcl_AppendResult(interp, result.string(), 0 );
            else
                RigiError(ERRORFLAG, "Parameter operation failed");
            return RigiErrState();

        } else if (op != GETPARAMETER) {
            if (argc == 2) {
                r = RigiParamOp( op, DEFAULT_NEWVALUE );
            } else if ( argc == 3 ) {
                if (op == CURRENTARCTYPEPARAM && !strcmp("any",argv[2]))
                    r = RigiParamOp( op, ANY_ARC);
                else
                    r = RigiParamOp( op, atoi( argv[2])  );
            } else {
                return ArgsError(argc);
            }

            // Save the result string

            sprintf( str, "%d", r );
            Tcl_AppendResult( interp, str, 0 );
        } else {
            str[0] = '\0';
            if (argc<3 || argc>5) return ArgsError(argc);
            if (argc>3) if (strlen(argv[3])>=BUFFSIZE) {
                RigiError(ERRORFLAG, "Argument string %-20.20s... is too long",argv[3]);
                return RigiErrState();
            } else {
                sprintf( str, "%s", argv[3] );
            }

            char *name = (argc>4) ? argv[4] : NULL;
            RigiGetParam( atoi( argv[2] ), str, BUFFSIZE, name );
            Tcl_AppendResult( interp, str, 0 );
        }
    }
    return RigiErrState();
}

static int cmdWindow( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {

    int returnvalue;
    if ( argc == 1 )
        returnvalue = WindowOp( DEFAULT_WINDOWOP, 0, 0, 0, DEFAULT_STRING );
    else if ( argc == 2 )
        returnvalue = WindowOp( atoi(argv[1]), 0, 0, 0, DEFAULT_STRING );
    else if ( argc == 3 )
        returnvalue = WindowOp( atoi(argv[1]), atoi(argv[2]),0,0,DEFAULT_STRING);
    else if ( argc == 4 )
        returnvalue = WindowOp( atoi(argv[1]), atoi(argv[2]), atoi(argv[3]),
            0, DEFAULT_STRING );
    else if ( argc == 5 )
        returnvalue = WindowOp( atoi(argv[1]), atoi(argv[2]), atoi(argv[3]),
            atoi(argv[4]), DEFAULT_STRING );
    else if ( argc == 6 )
        returnvalue = WindowOp( atoi(argv[1]), atoi(argv[2]), atoi(argv[3]),
            atoi(argv[4]), argv[5] );
    else
        return ArgsError(argc);
    sprintf( result_buf, "%d", returnvalue);
    Tcl_AppendResult(interp, result_buf, 0);
    return RigiErrState();
}

// SCRIPT10
// Commands for opening/moving/resizing/closing/iconifying/ editors.
static int cmdEditor( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    RigiError(ERRORFLAG, "Editor RCL commands are no longer supported");
    return RigiErrState();
}


// SCRIPT13
// Commands for setting the x and y coordinates of the cursor
static int cmdCursor( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    if ( argc == 1 )
        CursorOp( DEFAULT_X, DEFAULT_Y );
    else if ( argc == 2 )
        CursorOp( atoi( argv[1] ), atoi( argv[1] ) );
    else if ( argc == 3 )
        CursorOp( atoi( argv[1] ), atoi( argv[2]) );
    else
        return ArgsError(argc);
    return RigiErrState();
}

// This function creates all the rcl commands that will be
// available.
void  CreateRclCommands( Tcl_Interp* interp) {
    CmdInfo *cmdInfoPtr;

    for ( cmdInfoPtr = rclCmds; cmdInfoPtr->name != NULL; cmdInfoPtr++ )
        Tcl_CreateCommand( interp, cmdInfoPtr->name, cmdInfoPtr->proc,
            (ClientData)cmdInfoPtr->name, (Tcl_CmdDeleteProc *)NULL );
}


// cmdQuit: Execute the quit() function.	[SRT 19Sep93]
//	    quit => 	quit()  (quit and verify)
//	    quit 1	=> quit( 1 );	
//
static int cmdQuit( ClientData, Tcl_Interp *interp, int argc, char *[]) {
    if (argc == 1)
        quit( );
    else if (argc == 2)
        quit( 0 );
    else
        return ArgsError(argc);
    return RigiErrState();
}

static int cmdInfo( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    const char* returnvalue;
    int op;

    if (argc > 2) op = atoi(argv[1]);

    // Special case for environment varaible ops
    if (op == SETENVIRON || op == GETENVIRON ||
        op == GETARCCOLOR || op == SETARCCOLOR ||
        op == GETNODECOLOR || op == SETNODECOLOR ||
        op == GETNODEBITMAP)
        // Second op is ignored
        returnvalue = InfoOp( op, op, argv[2], (argc==4)?(argv[3]):(NULL) );
    else if ( argc == 2 )
        returnvalue = InfoOp( atoi(argv[1]),0,DEFAULT_STRING,NULL );
    else if ( argc == 3 )
        returnvalue = InfoOp( atoi(argv[1]),atoi(argv[2]),DEFAULT_STRING,NULL );
    else if ( argc == 4 )
        returnvalue = InfoOp( atoi(argv[1]), atoi(argv[2]), argv[3], NULL );
    else if ( argc == 5 )
        returnvalue = InfoOp( atoi(argv[1]), atoi(argv[2]), argv[3], argv[4] );
    else
        return ArgsError(argc);
    returnvalue ? chartype.Assign( returnvalue ) : chartype.Assign( "" );
    Tcl_AppendResult( interp, chartype.string(), 0);
    return RigiErrState();
}

static int cmdExecFG( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    RigiBool catchOn = RigiFalse;
    if (argc < 1) return ArgsError(argc);

    // check for the catch key word and remove it from the cmd list
    // construct process name followed by arguments
    CharType cmdLine("exec ");
#ifdef __WINDOWS__
    char *lp;
    char filePathName[_MAX_PATH];
    if (!SearchPath( NULL, argv[1], NULL, _MAX_PATH, filePathName, &lp )
        && (strrchr(argv[1], '.') != NULL)
        || (!SearchPath( NULL, argv[1], ".exe", _MAX_PATH, filePathName, &lp )
        && !SearchPath( NULL, argv[1], ".com", _MAX_PATH, filePathName, &lp ))) {
        RigiError(ERRORFLAG, "Unable to find process %s", argv[1]);
        return RigiErrState();
    }
    for (lp = filePathName; (lp = strchr(lp, PATHPARSECHAR)) != NULL; lp++) {
            *lp = UNIXPATHPARSECHAR;
    }
    argv[1] = filePathName;
#endif
    for (int i = 1; i < argc; i++) {
        // extract the catch parameter
        if (!strcmp(argv[i], "-catch")) {
            catchOn = RigiTrue;
        } else {
            cmdLine.Append(argv[i]);
            if (i < argc - 1) cmdLine.Append(" ");
        }
    }
    int result = Tcl_Eval(interp, cmdLine.string());
    if (catchOn) return TCL_OK;
    if ( result != TCL_OK) RigiError(ERRORFLAG, interp->result);
    return RigiErrState();
} /* cmdExecFG */

static int cmdExecBG( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    if(argc < 1) return ArgsError(argc);

    // construct process name followed by arguments
#ifdef __WINDOWS__
    char *lpFileName;
    char filePathName[_MAX_PATH];
    if (!SearchPath( NULL, argv[1], NULL, _MAX_PATH, filePathName, &lpFileName )
        && (strrchr(argv[1], '.') != NULL)
        || (!SearchPath( NULL, argv[1], ".exe", _MAX_PATH, filePathName, &lpFileName )
        && !SearchPath( NULL, argv[1], ".com", _MAX_PATH, filePathName, &lpFileName ))) {
        RigiError(ERRORFLAG, "Unable to find process %s", argv[1]);
        return RigiErrState();
    }
    argv[1] = filePathName;
    CharType cmdLine;
#else
    CharType cmdLine("exec ");
#endif
    for (int i = 1; i < argc; i++) {
        cmdLine.Append(argv[i]);
        if (i < argc - 1) cmdLine.Append(" ");
    }

#ifdef __WINDOWS__
    STARTUPINFO startInfo;
    PROCESS_INFORMATION procInfo;

    // prepare to lauch new process
    memset(&startInfo, 0, sizeof(startInfo));
    startInfo.cb = sizeof(startInfo);
    startInfo.dwX = (unsigned long)CW_USEDEFAULT;
    startInfo.dwY = (unsigned long)CW_USEDEFAULT;
    startInfo.dwXSize = 540;
    startInfo.dwYSize = 480;
    startInfo.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESIZE;
    startInfo.wShowWindow = SW_SHOW;

    // create new, windows process
    if (!CreateProcess(NULL, cmdLine.string(), NULL, NULL, TRUE,
        NORMAL_PRIORITY_CLASS, NULL, NULL, &startInfo, &procInfo) == 1)
        RigiError(ERRORFLAG, "Unable to create new process");
    return RigiErrState();
#else
    cmdLine.Append(" &");
    int result = Tcl_Eval(interp, cmdLine.string());
    if (result != TCL_OK) RigiError(ERRORFLAG, interp->result);
    return RigiErrState();
#endif
}   /*cmdExecBG*/


static int cmdFork( ClientData, Tcl_Interp *interp, int argc, char *[]) {
#if !defined(__WINDOWS__) && !defined(__WIN32__)
    int pid;

    // Check the command structure
    if(argc != 1) return ArgsError(argc);

    // Fork off a new process and check for success
    pid = fork();
    if( pid < 0 ){
        RigiError(ERRORFLAG, "Unable to fork process");
        return RigiErrState();
    }

    sprintf( result_buf, "%d", pid);
    Tcl_AppendResult( interp, result_buf, 0);
    return RigiErrState();
#endif
    RigiError(ERRORFLAG, "\"fork\" is not implemented");
    return RigiErrState();
}

static int cmdExecVP( ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    Tcl_DString buffer;
    const char *execName, **av;
    int ac;

    if(argc != 3) return ArgsError(argc);

    if( Tcl_SplitList(interp, argv[2], &ac, &av) != TCL_OK )
        return TCL_ERROR;
    execName = Tcl_TildeSubst(interp, argv[1], &buffer);

#ifdef _MSC_VER
    execvp(execName, (const char*const*)av);
#else
    execvp(execName, (char*const*)av);
#endif

    free((char *) av);
    Tcl_DStringFree(&buffer);
    RigiError(ERRORFLAG, "Unable to execute %s", argv[1]);
    return RigiErrState();
}

static int cmdSleep( ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    Tcl_DString buffer;

    if(argc != 2) return ArgsError(argc);

#ifdef __WIN32__
    Sleep(atoi(argv[1]) * 1000);
#else
    sleep(atoi(argv[1]));
#endif

    return RigiErrState();
}

static int cmdUnlink( ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
    Tcl_DString buffer;

    if(argc < 2) return ArgsError(argc);
    
    for (int i = 1; i < argc; i++)
      unlink(argv[1]);

    return RigiErrState();
}

static int cmdSpecial( ClientData, Tcl_Interp *interp, int argc, char *argv[]) {
    int args[7];
    if (argc > 7) return ArgsError(argc);
    for ( int i=argc; i<7; i++ )
        args[i] = 0;
    while (--argc>0)
        args[argc] = atoi(argv[argc]);

    SpecialOp( args[1], args[2], args[3], args[4], args[5], args[6] );
    return RigiErrState();
}

char *exec_command( char *cmd ) {
    // Check and make sure the interpreter has been created. If not,
    // print an error message and return.

    if (rcl_interp == NULL) {
        mayday("exec_command - NULL RCL interpreter, command = %s", cmd);
        return 0;
    }

    int result = Tcl_Eval( rcl_interp, cmd );

    if ( result != TCL_OK)
    {
        CharType str, com;
        mayday( "exec_command(%s)", cmd);
        str.Assign( "RCL Error: See standard error for details " );
        mayday( "%s", str.string() );
        mayday( "%s", rcl_interp->result);
        return NULL;
    }
    return rcl_interp->result;
}

