/* RigiEdit mainline
        Dr. H. A. Muller,
        University of Victoria,
        Victoria, BC

        File:		rigi.{C,cpp}
        Purpose:	mainline program for rigiedit
*/


#ifdef __OS2__
#define INCL_DOSPROCESS
#include <os2.h>
#endif

#ifdef __WINDOWS__
#include <windows.h>
#include <io.h>
#else
#include <unistd.h>
#endif
#include <iostream>
#include <stdarg.h>
#include <fcntl.h>
#include <time.h>

#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef CCHEADER_H
#include "CCheader.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef ENVIRON_H
#include "environ.h"
#endif
#ifndef GRAPHERR_H
#include "grapherr.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif
#ifndef VIEW_H
#include "view.h"
#endif
#ifndef EDITUTIL_H
#include "editutil.h"
#endif
#ifndef TK_COLORMAP_H
#include "tk_colormap.h"
#endif
#ifndef WINUTIL_H
#include "win_util.h"
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef WIN_UTIL_H
#include "win_util.h"
#endif
#ifndef RSF_H
#include "rsf.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif

#include <tk.h>
#ifndef _MSC_VER
#ifndef MEMCHK_H
#include "memchk.h"
#endif

#include "rigisignal.h"

#endif

#if HAVE_TIX_H && (((TK_MAJOR_VERSION == 4) && (TK_MINOR_VERSION < 2)) || (TK_MAJOR_VERSION < 4))
#include <tix.h>
#endif

#if HAVE_BLT_H
#include <blt.h>
#endif

// constants --------------------------------------------

static const int MsgBuffSize = 1024;

// prototypes -------------------------------------------

#if HAVE_BLT_H
extern "C" {
int Blt_Init(Tcl_Interp *);		    // (tknt40r1\src\blt18\src\bltinit.c)
extern void RegisterWidgetClass(HINSTANCE); // (tknt40r1\src\include\wndproc.h)
extern void RegisterTerminalClass(HINSTANCE);
extern void RegisterTopLevelClass(HINSTANCE);
extern HWND CreateTerminal( HANDLE );	    // (tknt40r1\src\include\util.h)
extern void DestroyTclTerminal( void );
extern void Tknt_SetHasConsole( int );
extern void Tknt_SetHasWindows( int );
extern void Tknt_SetAppInstance( HINSTANCE );
extern void Tknt_SetPostWindow( HWND );
extern void SetHookDeliverEL( HHOOK );	    // (tknt40r1\src\xlib\callback.c)
}
#endif

// extern variables -------------------------------------

char *progname;                         // program name

// Globals switches to be declared "extern" where needed
int      XMJW		= 0;        	// HACK for MJW experimenting
RigiBool XKW		= RigiFalse;	// HACK for KW experimenting
RigiBool K2_CHECK	= RigiFalse;	// use K2-partite checking
RigiBool POLL       = RigiFalse;    // don't call the rcl_poll_proc
RigiBool RSFMODE	= RigiFalse;	// auto load of rsf file
RigiBool VERBOSE	= RigiFalse;	// display Mess() messages
RigiBool RUNSCRIPT      = RigiFalse;

// global parameters declared in global.h
EditorState  *editor;               // Class that contains the state info (cl_rigistate.cpp)
GraphNode    *basenode = NULL;      // The base node for the graph (cl_graphnode.cpp)
RigiWindow   *basewindow;           // The base window for the graph (cl_rigiwindow.cpp)
RigiColorMap *rigi_colormap = NULL; // declared in global.h

// state variables ---------------------------------------

#ifdef __WINDOWS__
static HWND hConsoleWnd;			// Console (stdout)
static const int MaxTitleSize = 128;
#endif
static char *iRclBuff = NULL;       // RCL file to load at start up
static RigiWindow* clipboard = NULL;// Clipboard window for editing

// public functions --------------------------------------

void        RigiSetBaseNode(GraphNode*n) { basenode = n; }
GraphNode   *RigiBaseNode() { return basenode; }
RigiWindow  *RigiBaseWindow() { return basewindow; }

// The base node for the clipboard. It is created on demand, and if it is
// never used then it is never created.  The clip node is not added to the
// editor state (it hangs in mid-air).
GraphNode* RigiClipNode() {
    // create base node for clipboard
    static GraphNode* clipnode = NULL;
    if ( clipnode == NULL) {
        Mess("Creating clipnode");
        clipnode = new GraphNode( "Clip Board" );
    }
    return clipnode;
}

// Access function for the clipboard RigiWindow. It is created on demand
// and if it is never used it is never created.
RigiWindow* RigiClipBoard() {
    if ( clipboard == NULL ) {
        int x, y;
        RutilGetRootLocation(x, y);
        RigiDomain *domain = RinitGetDomain();
        Mess("Creating clipboard");
        clipboard = new RigiWindow( RigiClipNode(), WIN_SCRATCH,
            domain->NumNodeTypes(), domain->NumArcTypes(), x, y );
        editor->AddWindow(clipboard);
    }
    return clipboard;
}

void RigiUpdateClipBoard() {
    if (clipboard == NULL) return;
    clipboard->Update();
    clipboard->ScaleToView();
    clipboard->RigiDisplay();
}

void RigiDeleteClipBoard() {
    delete clipboard;
    clipboard = NULL;
}

//----------------------------------------------------------------------
// local functions
//

#ifdef __WINDOWS__XX
// Terminates command program
extern "C" {
  static void RGE_Exit(void) {
    fflush(stdout);
    fflush(stderr);
    DestroyWindow( hConsoleWnd );
    DestroyTclTerminal();
    ExitProcess(0);
  }
}
#endif

// Formats and displays fatal application errors
//
static void RigiFatalAppExit( char *format ... ) {
    char msgBuff[MsgBuffSize + 1];

    va_list ap;
    va_start(ap, format);
    vsprintf(msgBuff, format, ap);
    va_end(ap);
#ifdef __WINDOWS__
    // wait for a second...
    time_t t = time(NULL);
    while (time(NULL) < t + 1) ;
    FatalAppExit( 0, msgBuff );
#else
    mayday(msgBuff);
    exit(-1);
#endif
}

static void RigiPrintOptions() {
    cout << "Usage: " << progname << " [filename] [options]\n\n";
    cout << "filename             rcl script to execute\n\n";
    cout << "supported options:\n";
    cout << "  -dm <subdirectory> set subdirectory of domain model\n";
    cout << "  -env <cfg_file>    read specified configuration file\n";
#ifdef __WINDOWS__
    cout << "  -min               minimize stdout/stderr window\n";
#endif
    cout << "  -poll              call rcl_poll_proc every second\n";
    cout << "  -rsf               load file \"rsf\" from DBDIR directory\n";
    cout << "  -v                 verbose mode\n";
    cout << "  --                 pass all remaining options to rcl script\n\n";
    cout << "  all other options will be passed on to tcl/tk\n";
}

// Create base Rigi Window
static void RigiInitBase() {
    int x, y;
    RutilGetRootLocation(x, y);
    Mess( "creating basewindow" );
    basewindow = new RigiWindow(NULL, WIN_GENERAL, 0, 0, x, y);
    editor->AddWindow(basewindow);
    WindowSetCursor(CURSOR_WAIT);
    RutilSetActiveWindow(basewindow);
}

// Loads the RSF file specified
static void RGE_RSFInit() {
    Mess( "loading RSF... " );
    int i = DbLoadRSF();
    Mess( "done" );
    Mess( "generating composite arcs... " );
    edit_check_composite(basenode);
    Mess("updating graph");
    basewindow->Update();
}

static char *findRigi(char *path) {
  char *temp = path;
  char *RIGI = NULL;

  do {
    char* temp1 = strstr(temp, "/bin");
    char* temp2 = strstr(temp, "\\bin");
    char* temp3 = strstr(temp, "\\BIN");
    if ((temp = temp1 ? temp1 : temp2 ? temp2 : temp3)) 
      RIGI = temp++;
  } while (temp);

  if (RIGI) {
    temp = (char*) malloc((RIGI - path) + 6);
    strcpy(temp, "RIGI=");
    strncpy(temp + 5, path, RIGI - path);
    temp[RIGI - path + 5] = '\0';
    putenv(temp);
    RIGI = temp + 5;
  }
  
  return RIGI;
}
// checks and sets the TCL, TK, BLT, and RIGI env variables
static void RigiCheckEnv(char *progpath) {
  char *RIGI = getenv("RIGI");
  if (RIGI == NULL) {
    char cwd[MAXPATHLEN + 9];
    
    if ((RIGI = findRigi(progpath)) ||
	(RIGI = findRigi(getcwd(cwd, sizeof(cwd)))))
      printf("Environment variable RIGI not set, trying %s.\n", RIGI);
    else
      RigiFatalAppExit("Environment variable RIGI not set, "
		       "unable to auto detect");
  }

  int RIGI_len = strlen(RIGI);

  if (getenv("TCL_LIBRARY") == NULL) {
    char *temp = (char*) malloc(12 + RIGI_len + 12);
    strcpy(temp, "TCL_LIBRARY=");
    strcat(temp, RIGI);
    strcat(temp, "/lib/tcl" TCL_VERSION);
    putenv(temp);
  }
  if (getenv("TK_LIBRARY") == NULL) {
    char *temp = (char*) malloc(11 + RIGI_len + 11);
    strcpy(temp, "TK_LIBRARY=");
    strcat(temp, RIGI);
    strcat(temp, "/lib/tk" TK_VERSION);
    putenv(temp);
  }
#ifdef BLT_VERSION
  if (getenv("BLT_LIBRARY") == NULL) {
    char *temp = (char*) malloc(12 + RIGI_len + 12);
    strcpy(temp, "BLT_LIBRARY=");
    strcat(temp, RIGI);
    strcat(temp, "/lib/blt");
    putenv(temp);
  }
#endif
#ifdef TIX_VERSION
  if (getenv("TIX_LIBRARY") == NULL) {
    char *temp = (char*) malloc(12 + RIGI_len + 12);
    strcpy(temp, "TIX_LIBRARY=");
    strcat(temp, RIGI);
    strcat(temp, "/lib/tix" TIX_VERSION);
    putenv(temp);
  }
#endif
}

static void RigiInitialize(char *argv0) {
    // Save the program name so it can be used in error messages etc.
    // trim the executable name from the path name
    // in order to establish the executable's directory
    // as part of the search path
    for (progname = argv0 + strlen(argv0) - 1; progname >= argv0
        && *progname != PATHPARSECHAR
        && *progname != UNIXPATHPARSECHAR
        && *progname != DEVPARSECHAR; progname-- ) ;
    progname++;
}

#ifdef __WINDOWS__XX
// Initializes console window, register Tk window classes
static void RigiInitTk(int argc, char **argv) {
    HANDLE hInstance = NULL, hPrevInstance = NULL;
    COORD bufSize = {80, 200};
    char tmpName[MaxTitleSize];
    char title[MaxTitleSize];

    Tknt_SetHasConsole(1);              // tell Tknt that this is a console application
    Tknt_SetHasWindows(1);              // tell Tknt that this application has a window

    Tknt_SetAppInstance(hInstance);     // tell Tknt the application instance
    if (!hPrevInstance) {               // register the window classes
        RegisterWidgetClass(hInstance);
        RegisterTopLevelClass(hInstance);
        RegisterTerminalClass(hInstance);
    }

    HWND hTerm = CreateTerminal(hInstance);
                                                                                // recreate the console window
    Tknt_SetPostWindow(hTerm);          // tell Tknt the terminal handle
    HINSTANCE hLib = LoadLibrary("lxlib.dll");

    HHOOK hHook = SetWindowsHookEx(WH_MOUSE,
        (HOOKPROC)GetProcAddress(hLib, (LPCSTR)"DeliverEnterLeave"),
        hLib, (DWORD) GetCurrentThreadId() /*NULL*/);
    SetHookDeliverEL(hHook);
    FreeLibrary(hLib);

    HANDLE hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
    SetConsoleScreenBufferSize(hStdout, bufSize);

    // Get a Window handle for our console
    // What follows is a kludge since Windows does not really
    // want to use handles with console windows.
    char *s = "lr&gc";
    strcpy(tmpName, s);
    int prefixlen = strlen(tmpName);
    int match = 0;
    for (int suffix = 0; !match; suffix++) {
        itoa(suffix, &tmpName[prefixlen], 10);

        int len = strlen(tmpName);
        SetConsoleTitle(tmpName);
        for (HWND hThisWnd = GetWindow(hTerm, GW_HWNDFIRST); IsWindow( hThisWnd );
            hThisWnd = GetWindow(hThisWnd, GW_HWNDNEXT)) {
            if (GetWindowText(hThisWnd, title, MaxTitleSize) == len
                && !strcmp(title, tmpName)) {
                match++;
                hConsoleWnd = hThisWnd;
                break;
            }
        }
    }

    SetConsoleTitle(EnvGet("RIGITITLE"));

    for (int i = 1; i < argc; i++)      // iconisizes the window if desired
        if (!strcmp(argv[i], "-min"))
	  ShowWindow( hConsoleWnd, SW_SHOWMINIMIZED );
    atexit(RGE_Exit);                   // Sets up the exit routine
}
#endif

static void RigiPrintWelcome() {
    printf("Welcome to RigiEdit Version %s\n", VERSION);
    printf("Copyright 1986-2003,\n");
    printf("H.A. Muller, University of Victoria, All rights reserved\n");
    printf("Configuration used: %s\n", EnvGet("CFGFILE"));
}

// Parses all Rigi cmd line switches
static void RigiCommandParse(int argc, char **argv,
			     int *tclargc, char ***tclargv) {
  int i = 0;

  *tclargc = 0;
  *tclargv = (char **) malloc((argc + 1) * sizeof(char*));
  (*tclargv)[(*tclargc)++] = argv[0];

  // check whether first parameter is a script name
  // (i.e. does not start with a '-')
  if ((RUNSCRIPT = ((argc > 1) && (argv[1][0] != '-'))))
    (*tclargv)[(*tclargc)++] = argv[++i];

  while (++i < argc) {
    if (!strcmp("--", argv[i])) {
      // script parameters follow
      // transfer remaining arguments to tcl/tk argument list
      while (i < argc)
	(*tclargv)[(*tclargc)++] = argv[i++];
      (*tclargv)[*tclargc] = NULL;
      break;
    }

    if (!strcmp("-dm", argv[i])) {
      if (++i == argc) {
	mayday( "-dm must be followed by a subdirectory name");
      } else {
	Mess("resetting RIGIDOMAIN to %s", argv[i]);
	EnvSet("RIGIDOMAIN", argv[i]);
      }
    } else if (!strcmp("-env", argv[i])) {
      // the configuration file has already been parsed!
      // just skip over the file name.
      i++;
    } else if (!strcmp("-kw", argv[i])) {
      XKW = RigiTrue;
#ifdef _MSC_VER
    } else if (!strcmp("-min", argv[i])) {
      // do nothing
#endif
    } else if (!strncmp("-mjw", argv[i], 4)) {
      XMJW = atoi(argv[i] + 4);
      if (!XMJW) XMJW = 1;
      warn("XMJW flag set to %d", XMJW);
    } else if (!strcmp("-poll", argv[i])) {
      POLL = RigiTrue;
    } else if (!strcmp("-rsf", argv[i])) {
      RSFMODE = RigiTrue;
    } else if (!strcmp( "-v", argv[i])) {
      VERBOSE = RigiTrue;
    } else {
      (*tclargv)[(*tclargc)++] = argv[i];
    }
  }

}

// Creates and initializes rcl interpreter
static void RigiInitInterp() {
    Mess("setting up RCL interpreter");
#ifdef __WINDOWS__
    Tcl_SetVar(rcl_interp, "tcl_msWindows", "1", TCL_GLOBAL_ONLY);
#endif /* _MSC_VER */

#if ((TK_MAJOR_VERSION == 4) && (TK_MINOR_VERSION < 2)) || (TK_MAJOR_VERSION < 4)
    TK_Window mainWindow = Tk_CreateMainWindow(rcl_interp, NULL, "RigiEdit", "Tk");
    if (mainWindow == NULL)
        RigiFatalAppExit( "Tk_CreateMainWindow failed:\n%s",
            rcl_interp->result );
#endif
#if 0  // now done in Tk_Main()
    // save the command line arguments
    char *args = Tcl_Merge(tclargc, tclargv);
    Tcl_SetVar(rcl_interp, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    char buff[MAXPATHLEN + 9];
    sprintf(buff, "%d", tclargc);
    Tcl_SetVar(rcl_interp, "argc", buff, TCL_GLOBAL_ONLY);
    Tcl_SetVar(rcl_interp, "argv0", progname, TCL_GLOBAL_ONLY);
#endif

    // Initialize Tcl, Tk and Blt interpreters
    if (Tcl_Init( rcl_interp ) == TCL_ERROR)
        RigiFatalAppExit( "Tcl_Init failed:\n%s", rcl_interp->result );
    if (Tk_Init( rcl_interp ) == TCL_ERROR)
        RigiFatalAppExit( "Tk_Init failed:\n%s", rcl_interp->result );
    Tcl_StaticPackage(rcl_interp, "Tk", Tk_Init, NULL);

#ifdef TIX_VERSION
    if (Tix_Init( rcl_interp ) == TCL_ERROR)
        RigiFatalAppExit( "Tix_Init failed:\n%s", rcl_interp->result );
    Tcl_StaticPackage(rcl_interp, "Tix", Tix_Init, NULL);
#endif /* TIX_VERSION */

#ifdef BLT_VERSION
    if (Blt_Init( rcl_interp ) == TCL_ERROR)
        RigiFatalAppExit( "Blt_Init failed:\n%s", rcl_interp->result );
    Tcl_StaticPackage(rcl_interp, "Blt", Blt_Init, NULL);
#endif

    // Add RCL commands to the interp
    CreateRclCommands( rcl_interp );

    // Install the rigi icon

    Tk_DefineBitmap(rcl_interp, Tk_GetUid("RigiIcon"), RigiIconBits,
        RIGIICONWIDTH, RIGIICONHEIGHT);

    // source the common procedure lib
    char *rcl = EnvGet("RIGIRCL");
    if (rcl && *rcl) {
        Mess("initializing RIGIRCL, %s", rcl);
        if (Tcl_EvalFile(rcl_interp, rcl) != TCL_OK) {            mayday("failure to evaluate RIGIRCL");
            mayday(rcl_interp->result);
        }
    }
    // reset the windows variable (rc.rcl script sets it to 0)
#ifdef __WINDOWS__
    Tcl_SetVar(rcl_interp, "tcl_msWindows", "1", TCL_GLOBAL_ONLY);
#endif
    // source the user procedure lib
    rcl = EnvGet("RIGIURCL");
    if (rcl && *rcl) {
        Mess("initializing RIGIURCL, %s", rcl);
        if (Tcl_EvalFile(rcl_interp, rcl) != TCL_OK) {
            mayday("failure to evaluate RIGIURCL");
            mayday(rcl_interp->result);
        }
    }
}

// Executes the standard style and user style scripts
// for the menu options.
static void RigiInitStyle() {

    // The styles contain the widgets and menus.
    char *sty = EnvGet( "RIGISTY" );
    if (sty && *sty) {
        Mess("initializing RIGISTY, %s", sty);
        if (Tcl_EvalFile(rcl_interp, sty) != TCL_OK) {
            mayday("failure to evaluate RIGISTY");
            mayday(rcl_interp->result);
        }
    } else {
        RigiFatalAppExit("configuration variable RIGISTY not set");
    }
    sty = EnvGet( "RIGIUSTY" );
    if (sty && *sty) {
        Mess("initializing RIGIUSTY, %s", sty);
        if (Tcl_EvalFile(rcl_interp, sty) != TCL_OK) {
            mayday("failure to evaluate RIGIUSTY");
            mayday(rcl_interp->result);
        }
    }
}

static void RigiAppInit( ) {
    if (basenode == NULL) {                 // Sets up the root or base node of the graph
        char *root = ROOT;
        basenode = new GraphNode(root);
        basenode->SetId(1);
        basenode->SetType(0);
        GutilAddNode(basenode);             // Adds root to hash list and name set
    }
    editor->SetDisplayCommand(RigiFalse);   // Turns off rigi-generated RCL command area
    editor->SetDisplayMenu(RigiFalse);      // Turns off rigi-generated menu

    Mess("creating colormap");
    rigi_colormap = new RigiColorMap();
    rigi_colormap->InitColorMap();

    Mess("displaying basewindow");
    basewindow->SetNode(basenode);          // cl_rigiwindow.cpp
    basewindow->Update();                   // display the base window (cl_rigiwindow.cpp)

    if (iRclBuff && *iRclBuff) {            // Executes the command-line RCL script (-i option)
        Mess("initializing from command line Tcl script");
        CharType srcCmd("source ");
        srcCmd.Append(iRclBuff);
        exec_command(srcCmd.string());
    }

    if (RSFMODE) RGE_RSFInit();             // Loads rsf file if cmd line arg
    basewindow->Show();

    WindowSetCursor(CURSOR_NORMAL);         // tk_stub.cpp
}

static int Tcl_AppInit(Tcl_Interp* interp) {
    rcl_interp = interp;
    RigiInitInterp();               // initializes RCL interpreter
    RigiInitStyle();                // initializes menu styles
#ifndef __WINDOWS__
#if ((TK_MAJOR_VERSION == 4) && (TK_MINOR_VERSION >= 2)) || (TK_MAJOR_VERSION > 4)
    Tk_UnmapWindow(Tk_MainWindow(rcl_interp));
#endif
#endif
    RigiClipNode();                 // initialize base node of clipboard
    RigiInitBase();                 // initializes base window
    editor->Reset();                // may not need this!
    RinitInit(NULL);                // initializes domain
    RigiAppInit();                  // starts the application

    if (POLL) {
        // THIS IS BAD - WE SHOULD USE TIMERS INSTEAD!!!
        // Main Interpreter loop.  This is essentially
        // from Tk_MainLoop but we've added the execution
        // of the rcl_poll_proc as well.  Note that tkwait
        // causes the Tk version of Tk_MainLoop to execute, not
        // this one.  This means that the polling is not performed
        // whenever the tkwait command is being executed.

        static time_t t0 = time(NULL);
        for (; ;) {
            time_t t1 = time(NULL);
	        Tk_DoOneEvent(TK_DONT_WAIT);
            if (t1 > t0) {

                // Once a second, call a tcl proc - it does not
                // actually have to exist.  This is intended
                // primarily for interprogram communication and
                // synchronization.

                t0 = t1;
                Tcl_GlobalEval(rcl_interp, "rcl_poll_proc");
            }
        }
        Tcl_Exit(0);
    }

    // if we are running in an interactive shell, do not exit from Tk_Main
    // to prevent the command prompt from appearing

    if (!RUNSCRIPT) {
      Tk_MainLoop();
      Tcl_Exit(0);
    }

    return TCL_OK;
}

//----------------------------------------------------------------------
// main functions
//

#ifdef __WINDOWS__XX
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nCmdShow) {
  int argc = 1;
  char *argv[] = { "rigiedit.exe", NULL };
#else
int main(int argc, char *argv[] ) {
#endif
#ifdef __OS2__
    TIB *ptib;
    PIB *ppib;
    /* morph into PM application */
    DosGetInfoBlocks(&ptib, &ppib);
    if (ppib->pib_ultype == 0) {
      ULONG written;
      DosWrite(2, "This program must not be run from an OS/2 fullscreen session.\r\n", 63, &written);
      DosExit(EXIT_PROCESS, 1);
    }
    ppib->pib_ultype = 3;
#endif
#ifdef _MSC_VER
    // _MSC_VER uses externs for command line args!
    extern int __argc;
    extern char **__argv;
    argc = __argc;
    argv = __argv;
#else
#ifdef MEMCHK_H
    MemoryDebugInit(argv[0]);       // Initialize memory debugging stuff
#endif
    catchSignal();                  // Trap signals
#endif

    RigiInitialize(argv[0]);        // initialize global variables

    // search for the help cmd line argument first
    for (int i = 1; i < argc; i++) {
        if (!strcmp(argv[i],"-h")) {
            RigiPrintOptions();
            return 0;
        }
    }

    RigiCheckEnv(argv[0]);           // check for critical env vars

    // Read the configuration file
    switch(EnvInit(argc, argv)) {
    case EnvOK:break;
    case EnvDefault:break;
    case EnvError:
        RigiFatalAppExit("Invalid configuration file.");
    }

    editor = new EditorState();
    editor->Init();
    RinitDomain();                  // Initializes the basic domain
#ifdef _MSC_VER
    RigiInitTk(argc, argv);         // init console, register win class
#endif
    RigiPrintWelcome();             // display welcome on stdout
    int tclargc;
    char **tclargv;
    RigiCommandParse(argc, argv, &tclargc, &tclargv); // parse the command line
    Tk_Main(tclargc, tclargv, Tcl_AppInit);
    return 0;
}

