
// $Author: jmartin $
// $Date: 2004-02-18 10:57:04 $
// $Source: /root/cvs/Rigi/rigiedit/rigiutil.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: rigiutil.C,v 1.43 2004-02-18 10:57:04 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <ctype.h>

#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef BASE_WINDOW_H
#include "cl_base_window.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef ACTIVEARC_H
#include "cl_activearc.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif

#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef RIGI_H
#include "rigi.h"		// for RigiBaseWindow() - HACK!
#endif
#ifndef ENVIRON_H
#include "environ.h"
#endif

// Access functions to the currently active window.

static RigiWindow* active_window = NULL;
RigiWindow *RutilActiveWindow() {
    if (active_window == NULL)
        RigiError(ERRORFLAG, "NULL active window");
    return active_window;
}

void RutilRaiseWindow(RigiWindow *rw) {
    rw->Window()->Restack(Above, NULL);
}

void RutilSetActiveWindow(RigiWindow *rw) {
    //Mess("**** Active window = %d", rw->Window()->GetId());

    // MJW Danger! Make sure active_window either NULL or a valid window!
    if (active_window!=rw) {
        if (active_window) active_window->UpdateLabel( NULL, "" );
        rw->UpdateLabel( NULL, "<<ACTIVE>>" );
    }
    active_window = rw;
}

// Check to see if all RigiNodes of the list are contained in window

// Return a textual representation of the name of this type of window.

const int WINDOW_PROJECT  = 0;
const int WINDOW_SCRATCH  = 1;
const int WINDOW_OVERVIEW = 2;
const int WINDOW_GENERAL  = 3;
const int WINDOW_UNKNOWN  = 4;

char *WINNAME[] =
        { "PROJECT","CLIPBOARD","OVERVIEW","GENERAL","UNKNOWN" };

char* RutilWinTypeName( RigiWinType type ) {
    switch( type )
    {
        case WIN_SCRATCH:  return WINNAME[WINDOW_SCRATCH];
        case WIN_PROJECT:  return WINNAME[WINDOW_PROJECT];
        case WIN_OVERVIEW: return WINNAME[WINDOW_OVERVIEW];
        case WIN_GENERAL:  return WINNAME[WINDOW_GENERAL];
        default:           return WINNAME[WINDOW_UNKNOWN];
    }
}

void RutilGetRootLocation( int &x, int &y) {
    x = -1;
    char *loc = EnvGet("ROOTLOCATION");
    if (loc != NULL && sscanf(loc, "%d %d", &x, &y) != 2) {
        warn("Invalid location specified for the root window \"%s\"", loc);
        x = -1;
    }
    if (x == -1) {
        x = 100;
        y = 200;
    }
}

void RutilGetDefFrameSize( int &w, int &h) {
    w = -1;
    char *size = EnvGet("ROOTFRAMEDIM");
    if (size != NULL && sscanf(size, "%d %d", &w, &h) != 2) {
        warn("Invalid frame size specified for the root window \"%s\"", size);
        w = -1;
    }
    if (w == -1) {
        w = 529;
        h = 456;
    }
}

void RutilGetDefWindowSize( int &w, int &h) {
    w = -1; 
    char *size = EnvGet("ROOTWINDOWDIM");
    if (size != NULL && sscanf(size, "%d %d", &w, &h) != 2) {
        warn("Invalid window size specified for the root window \"%s\"", size);
        w = -1;
    }
    if (w == -1) {
        w = 500;
        h = 400;
    }
}

void RutilGetMaxCanvasSize( int &w, int &h) {
    w = -1;
    char *size = EnvGet("MAXCANVASDIM");
    if (size != NULL && sscanf(size, "%d %d", &w, &h) != 2) {
        warn("Invalid maximum canvas size specified \"%s\"", size);
        w = -1;
    }
    if (w == -1) {
        w = 2048;
        h = 2048;
    }
}

//------------------------------------------------------------------------
// File IO utility routines
//------------------------------------------------------------------------

int RutilFscanl( FILE *fp, char *buff, int size )
{
    // Load buff with next line in the file, skipping comments
    // and compressing substrings of space characters
    register int c,r,i,x;

    for ( i=0;;)
    {
        c = getc(fp);
        if (c=='#') i = 1; else if (c=='\n') i = 0;
        else if (isspace(c)) continue; else if (i==0) break;
    }

    r = size-1;
    for ( i=x=0; ; )
    {
        if (c==EOF || c=='\n') break;
        if (isspace(c))
            { if (x) if (i<r) buff[i++] = ' '; x=0; }
        else
            { if (i<r) buff[i++] = c; x=1; }
        c = getc(fp);
    }
    if (i) buff[i] = '\0';
    if (i==r) warn("fscanl(), buffer overflow, contents are\n%s",buff);
    return i;
}

int RutilFscans( FILE *fp, char *buff, int size )
{
    // Load buff with next string in the file, skipping comments.
    // Return 0 if no chars read, 1 if normal, 2 if OK read but
    // also one or more newlines consumed before read.

    register int c,r,i,x;
    int y,z;
    char quote;

    for ( i=0;;)
    {
        c = getc(fp);
        if (c=='#') i = 1;
        else if (c=='\n') i = 2;
        else if (isspace(c)) continue;
        else if (i!=1) break;
    }
    z = i; // 2 or 0

    if ((c=='"') || (c=='\'')) { quote=c; x=1; c=getc(fp); } else x=0;
    r = size-1;
    for ( i=y=0; ; )
    {
        if (c==EOF) break;
        if (x)
        {
            if (!y) {  if (c==quote) break; }
            y = c=='\\';
        }
        else if (isspace(c)) break;
        if (i<r) buff[i++] = c;
        c = getc(fp);
    }
    buff[i] = '\0';
    if (i==r) warn("fscans(), buffer overflow, contents are\n%s",buff);
    z = i ? (z?z:1) : x;
    return z;
}

