// $Author: jmartin $
// $Date: 2003-01-08 21:27:34 $
// $Source: /root/cvs/Rigi/rigiedit/rsf.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#ifndef __CTYPE_H
#include	<ctype.h>
#endif
#ifndef GLOBAL_H
#include	"global.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"     // Global instance of editor's state
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif

#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif
#ifndef ATTRIBUTE_H
#include "cl_attribute.h"     // for sets of strings
#endif

#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef ENVIRON_H
#include "environ.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif
#ifndef RSF_H
#include "rsf.h"
#endif

static const int ARCSIZE = 128;
static const int SRCSIZE = 256;
static const int DSTSIZE = 512;

static char arcname[ARCSIZE], srcnamebuff[SRCSIZE], dstnamebuff[DSTSIZE];
static char *srcname, *dstname;

static GraphNode *new_graphnode(char*, int t, int id);
static GraphArc *new_grapharc(int,GraphNode*, GraphNode*, int arcId);
static int ParseNodeName(char **str, char **name, int *id, int *type);
static int CheckNodeType(GraphNode *n, int type);

static int DumpGraph( GraphNode* );
static int DumpNode( GraphNode*, NodeSet& );
static int DumpArc( GraphArc* );
static int DumpNodeAttr( GraphNode*,char* );
static int DumpArcAttr( GraphArc*,char* );
static void AddArcAttr(char *attrkey, char *arcid, char *attrval);

// locals used in dumping of Rigi graph
static FILE *rsf_fp;
static GraphNode *root;

static char *format	= "%s\t%s\t%s\n";
static char *attrformat	= "%s\t%s\t\"%s\"\n";

// statistics
static int nred;
static int ndup;
static int nrep;

// adds an arc attribute to a arc identified by arcid
// prevents duplicates
static void AddArcAttr(char *attrkey, char *arcid, char *attrval) {
    GraphArc *arc = editor->FindArc( atoi(arcid) );
    if (!arc) {
        mayday("Arc %s-->%s missing - attribute %s = %s omitted\n",
            arcid, attrkey, attrval);
        return;
    }
    if (arc->GetAttr(attrkey) != NULL) {
        mayday("Duplicate attribute %s for arc id %s\n", attrkey, arcid);
        return;
    }

    arc->SetAttr(attrkey, attrval);
}

int SaveRSF( char *fn, RigiWindow *rw )
{
    if (rw==NULL) return -1;

    root = rw->GetNode();
    if (!root) {
        mayday("NULL node of activewin! Correct wintype??");
        return -1;
    }

    char *rootname = root->GetLabel();
    warn( "saving RSF file from node %s into file %s",rootname,fn );

    rsf_fp = fopen( fn,"w" );
    if (!rsf_fp) {
        mayday( "cannot open file %s",fn );
        return -1;
    }

    int r = DumpGraph( root );

    fclose( rsf_fp );
    return r;
}

int LoadRSF( char *fn, RigiWindow *rw )
{
    int init = 0;

    GraphNode *src, *dst;
    char *objectname;
    int i,j;
    int srcid, dstid;
    int numbadarcs=0, numbadarctypes=0;
    AttributeSet nodes, badarctypes;
    ArcId arcid;

    RigiDomain *domain = RinitGetDomain();

    // open file
    FILE *fp = fopen( fn,"r" );
    if (fp==NULL)
    { mayday("couldn't open RSF file %s",fn ); return -1; }

    ndup=nred=nrep=0;

    int numnames = 0;
    for ( j=0;;j++) {
        int srctype=0, dsttype=0, attr=0;

        srcname = srcnamebuff; dstname = dstnamebuff;

        // Scan the triple (allow for names in double quotes)
        if (RutilFscans( fp, arcname, ARCSIZE ) == 0) break;
        if (!strcmp(arcname,"@end")) break;

        if (RutilFscans( fp, srcname, SRCSIZE ) != 1) {
            mayday("RSF parse error on src");
            break;
        }
        if (RutilFscans( fp, dstname, DSTSIZE ) != 1) {
            mayday("RSF parse error on dst");
            break;
        }

        // remove the arc/node prefix from the first parameter
        // if one exists.
        char *pArcname = arcname;
        int arctype = 0;

        if (arcname[1] == '!') {
            if (arcname[0] != 'n' && arcname[0] != 'a') {
                RigiError(ALERT, "Invalid RSF object type in %s", arcname);
                continue;
            }
            pArcname = &arcname[2];
            if (arcname[0] == 'n') arctype = -1;
        }

        if (arctype != -1)
            arctype = domain->ArcTypeFromName( pArcname );

        if (arctype == -1) {
            int badarc = 0;
            // The arcname is not already recognized as a type
            if (pArcname == arcname) {
                if (domain->IsNodeAttr(0, pArcname)) {
                    attr = 1;
                } else if (domain->IsArcAttr(0, pArcname)) {
                    AddArcAttr(arcname, srcname, dstname);
                    continue;
                } else {
                    badarc = 1;
                }
            } else {
                if (arcname[0] == 'n'
                    && domain->IsNodeAttr(0, pArcname)) {
                    attr = 1;
                } else if (arcname[0] == 'a'
                    && domain->IsArcAttr(0, pArcname)) {
                    AddArcAttr(pArcname, srcname, dstname);
                    continue;
                } else {
                    badarc = 1;
                }
            }
            if (badarc) {
                numbadarcs++;
                if (!badarctypes.SetAttr( pArcname,1 )) {
                    numbadarctypes++;
                    RigiError(ALERT, "Unknown arc type (%s)", pArcname );
                }
                continue;
            }
        }

        if (attr == 0) {
            // the line references either an arc or a type
            srctype = domain->ArcSrcType( arctype );
            dsttype = domain->ArcDstType( arctype );
        }

        // decompose compound node names (HACK!)
        srcid = dstid = 0;
        char *rigisrcname = srcname;
        char *rigidstname = dstname;
        ParseNodeName( &srcname, &rigisrcname, &srcid, &srctype );
        ParseNodeName( &dstname, &rigidstname, &dstid, &dsttype );

        // unknown src/dst nodetypes default to known type 0
        if (srctype<0) srctype = 0;
        if (dsttype<0) dsttype = 0;

        // HACK for handling the root node; must be FIRST name
        if (!init) {
            init = 1;
            if (strcmp( rigisrcname,ROOT)) {
                warn( "RSF file is missing \"%s\" as root node.", ROOT);
                warn( "Reading file as unstructured RSF.");
                fclose(fp);
                ReadRSF(fn);
                return -1;
            }
            GraphNode *basenode = RigiBaseNode();
            editor->RemoveNode( basenode );
            if (srcid>0) basenode->SetId( srcid );
            editor->AddNode( basenode );

            objectname = srcid ? srcname : rigisrcname;
            nodes.SetAttr( objectname, (int)basenode );
        }

        objectname = srcid ? srcname : rigisrcname;
        i = nodes.GetIntAttr( objectname );
        if (i==INVALID_ATTR) {
            numnames++;
            src = new_graphnode( rigisrcname, srctype, srcid );
            nodes.SetAttr( objectname, (int)src );
        } else {
            CheckNodeType( src = (GraphNode*)i, srctype );
        }

            // Now, if this an attribute declaration, we just assign
        // the attribute to the object, and continue. Otherwise,
            // the dst name must be an object.

        if (attr) {
            domain->SetNodeAttr( src, pArcname, dstname );
            continue;
        }

        // for arc types, if there is no destination node,
        // one must be created.


        objectname = dstid ? dstname : rigidstname;
        i = nodes.GetIntAttr( objectname );
        if (i==INVALID_ATTR) {
            numnames++;
            dst = new_graphnode( rigidstname, dsttype, dstid );
            nodes.SetAttr( objectname, (int)dst );
        } else {
            CheckNodeType( dst = (GraphNode*)i, dsttype );
        }

        if (src==NULL || dst==NULL) mayday("problem in LoadRSF()");

        arcid = atoi(rigisrcname);
        GraphArc *arc = new_grapharc(arctype, src, dst, arcid);

    }

    // Now, traverse graph to find parentless nodes, and connect
    // them to the node of this window (not done)

    if (nred) warn("%d redundant composite arc%s removed",
        nred,nred==1?"":"s" );
    if (ndup) warn("%d duplicate arc%s removed",
        ndup,ndup==1?"":"s" );

    if (numbadarctypes) warn("%d arctype%s and %d arc%s ignored",
        numbadarctypes,numbadarctypes==1?"":"s",
        numbadarcs,numbadarcs==1?"":"s" );

    //rw->Update();
    fclose(fp);
    return numnames;
}

int ReadRSF( char *fn )
{
    RigiNode *rn, *src, *dst;
    GraphArc *a;
    int i,j,k;
    int numbadarcs=0, numbadarctypes=0;
    AttributeSet badarctypes, nodes;

    RigiWindow *rw = RutilActiveWindow();
    if (rw==NULL) return -1;

    // Get the current domain...

    RigiDomain *domain = RinitGetDomain();

    // open file
    FILE *fp = fopen( fn,"r" );
    if (fp==NULL) { mayday("couldn't open RSF file %s",fn ); return -1; }

    ndup=nrep=j=k=0;

    AvlPos *P;
    for ( rn=rw->FirstNode(P); rn; rn=rw->NextNode(P) ) {
        i = nodes.SetAttr( rn->GetNode()->GetLabel(), (int)rn );
        j++;
        if (i == 0) k++;
    }
    if (j) warn("ReadRSF(), %d node%s already in window",j,j==1?"":"s");
    if (j&&j!=k) warn("ReadRSF(), only %d distinct names!",k );

    int numnames = 0;
    for (;;) {
        int srctype = 0, dsttype = 0, attr = 0;

        srcname = srcnamebuff; dstname = dstnamebuff;

        // Scan the triple (allow for names in double quotes)
        if (RutilFscans( fp, arcname, ARCSIZE ) == 0) break;
        if (!strcmp(arcname,"@end")) break;

        if (RutilFscans( fp, srcname, SRCSIZE ) != 1) {
            mayday("RSF parse error on src");
            break;
        }
        if (RutilFscans( fp, dstname, DSTSIZE ) != 1) {
            mayday("RSF parse error on dst");
            break;
        }

        // determine three types
        int arctype = domain->ArcTypeFromName( arcname );
        if (arctype == -1) {
            // The arcname is not already recognized as a type
            if (domain->IsNodeAttr( 0,arcname )) {
                attr = 1;
            } else if (domain->IsArcAttr( 0, arcname )) {
                AddArcAttr(arcname, srcname, dstname);
                continue;
            } else {
                numbadarcs++;
                if (!badarctypes.SetAttr( arcname,1 )) {
                    numbadarctypes++;
                    warn("ReadRSF(), arctype \"%s\" unknown",arcname );
                }
                continue;
            }
        }

        if (attr == 0) {
            // the line references either an arc or a type
            srctype = domain->ArcSrcType( arctype );
            if (srctype < 0) srctype = 0;
            dsttype = domain->ArcDstType( arctype );
            if (dsttype < 0) dsttype = 0;
        }


        // Keep a list of node attributes/types not permitting duplicates.
        i = nodes.GetIntAttr( srcname );
        if (i == INVALID_ATTR) {
            numnames++;
            src = GutilNewNode( rw, srcname, srctype, 0, 0 );
            nodes.SetAttr( srcname, (int)src );
        } else {
            CheckNodeType( (src = (RigiNode*)i)->GetNode(), srctype );
        }

        // Now, if this an attribute declaration, we just assign
        // the attribute to the object, and continue. Otherwise,
        // the dst name is treated as another object

        if (attr) {
            domain->SetNodeAttr( src->GetNode(), arcname, dstname );
            continue;
        }

        // check to see if the node type has been already declared
        // if it has, check to see if an existing node type is being
        // overridden.
        i = nodes.GetIntAttr( dstname );
        if (i==INVALID_ATTR) {
            numnames++;
            dst = GutilNewNode( rw, dstname, dsttype, 0, 0 );
            nodes.SetAttr( dstname, (int)dst );
        } else {
            CheckNodeType( (dst = (RigiNode*)i)->GetNode(), dsttype );
        }

        if (src==NULL || dst==NULL) mayday("problem in ReadRSF()");

        if ( a = GutilNewArc( rw,arctype,src,dst ) ) {
            ndup++;
            Mess("Multiple arcs %s and %s for %s -> %s",
                domain->ArcNameFromType(arctype),
                domain->ArcNameFromType(a->GetType()),
                srcname, dstname);

            // MJW HACK: set type to that of higher priority (ie. lower num)
            if (arctype < a->GetType()) {
                nrep++;
                GutilChangeArcType( a, arctype );
            }
        }
    }
    if (ndup) warn("ReadRSF(), %d duplicate arc%s read, %d replaced",
        ndup,ndup==1?"":"s",nrep);

    if (numbadarctypes)
        warn("ReadRSF(), %d arctype%s (and %d arc%s) ignored",
            numbadarctypes,numbadarctypes==1?"":"s",
            numbadarcs,numbadarcs==1?"":"s" );

    fclose(fp);
    return numnames;
}

static int ParseNodeName( char **str, char **name, int *id, int *type )
{
    // Decompose compound node names (HACK!).
    // Input str could be (eg.) "4!16575!Root" or "System!16575!Root"
    // or "16575!Root" or "Root".
    // str is the input compound name, and output name or id!name,
    // name is the name by itself

    char *s, *t;
    if (!(str && name && id && type)) mayday( "problem in ParseNodeName()" );

    if ( s = strrchr( *str,'!' ) ) {
        *name = s+1; *s='\0';
        if (t=strrchr( *str,'!')) {
            *t = '\0'; 
            *type = RinitGetDomain()->NodeTypeFromName( *str );
            *str = t+1;
        }
        *id = atoi( *str ); /* *s='!';*/
    }
    return 0;
}

static int CheckNodeType( GraphNode *n, int type )
{
    RigiDomain *domain = RinitGetDomain();
    if ( type>0 && n->GetType()>0 && type!=n->GetType() )
    {
        mayday("nodetype clash (%s vs old %s) for %s, resetting",
            domain->NodeNameFromType(type),
            domain->NodeNameFromType(n->GetType()),n->GetLabel());
    }
    if (type) n->SetType( type );
    return 0;
}

int WriteRSF( char *fn )
{
    RigiNode *rn; RigiArc *ra;
    RigiWindow *rw = RutilActiveWindow();
    if (rw==NULL) return -1;
    
    // Get the current domain

    RigiDomain *domain = RinitGetDomain();

    // open file
    FILE *fp = fopen( fn,"w" );
    if (fp==NULL)
    {
        mayday("couldn't open RSF file %s",fn );
        return -1;
    }

    // write out attributes of visible nodes in the window
    AvlPos *P;
    for ( rn = rw->FirstNode(P); rn; rn = rw->NextNode(P) )
    {
        if (!rn->isVisible()) continue;
        GraphNode *n = rn->GetNode();
        char *attr, *name, *value;
        name = n->GetLabel();

        // The only "special" attr in this context is "type"

        value = domain->NodeNameFromType( n->GetType() );
        fprintf( fp, format, "type",name,value );
        
        AvlPos *AP;
        for ( attr = n->FirstAttr(AP); attr; attr = n->NextAttr(AP) )
        {
            value = n->GetAttr( attr );
            char *f = (value && *value=='\"') ? format : attrformat;
            fprintf( fp, f, attr,name,value );
        }
    }

    // write out the visible arcs in the window
    int i=0; AvlPos *Q;
    for ( rn = rw->FirstNode(P); rn; rn = rw->NextNode(P) )
    for ( ra = rn->FirstArc(Q,ARC_OUT); ra; ra = rn->NextArc(Q,ARC_OUT) )
    {
        if (!ra->isVisible()) continue;

        int arctype = ra->GetArc()->GetType();

        i++;

        char *arcname = domain->ArcNameFromType( arctype );
        if (arcname==NULL) arcname = "unknown";

        fprintf( fp, format, arcname,
                ra->GetSrc()->GetNode()->GetLabel(),
                ra->GetDst()->GetNode()->GetLabel() );
    }
    fclose(fp);
    return i;
}

static GraphNode *new_graphnode( char*n,int type,int id ) {
    GraphNode *node = new GraphNode( n );
    node->SetType( type );
    if (id>0) node->SetId(id);
    GutilAddNode( node );
    return node;
}

static GraphArc *new_grapharc( int type, GraphNode*src, GraphNode*dst, 
    int arcId) {

    GraphArc *a;
    int lt = editor->LevelArc();

    AdjList outtype = (type == lt) ? outLevArcs : outArcs ;

    /*
    if ((a = src->FindArc(dst, outtype)) != NULL) {
        if (a->GetType() == editor->CompositeArc()) {
            // HACK for old GRAS model which had redundant composite arcs
            src->RemoveArc(a);
            if (src!=dst) dst->RemoveArc(a);
            GutilDeleteArc(a);
            nred++;
        } else if (type == editor->CompositeArc()) {
            // HACK for old GRAS model which had redundant composite arcs
            nred++;
            return NULL;
        } else if (type == a->GetType()) {
            // Arcs of same type not permitted
            ndup++;
            return NULL;
        } else {
            //mayday("duplicate arc disallowed, type %s (original %s)", 
                //domain->ArcNameFromType(type),
                //domain->ArcNameFromType(a->GetType()) );
            //return NULL;
        }
    }
    */

    GraphArc *arc = new GraphArc(src, dst, ((type < 0)? 0 : type));
    if (type < 0) {
        mayday("invalid arctype %d in new_grapharc()",type);
        return NULL;
    }

    if (arcId != 0) arc->SetId(arcId);
    src->InsertArc(arc);
    dst->InsertArc(arc);
    if (type == lt) dst->SetLevel();

    GutilAddArc(arc);
    return arc;
}

static int DumpNode( GraphNode *n, NodeSet& visited ) {
    ArcDirection arcout = ARC_OUT;

    // Check to see if we have been to this node, if so return. If not
    // add it to the visited list, and process it.

    int r = 0;
    if ( !n ) mayday( "NULL node in DumpNode" );
    if ( visited.Find( n->GetId() ) ) return 0;
    visited.Add( n );

    // Write all the node attributes

    DumpNodeAttr( n,"type" );

    AvlPos *P;
    for ( char *attr = n->FirstAttr(P); attr; attr = n->NextAttr(P) )
        DumpNodeAttr( n,attr );

    // Write all the outgoing arcs

    GraphArc *arc = n->FirstArc(P, ANY_ARC, arcout);
    while ( arc ) {
        DumpArc(arc);
        arc = n->NextArc(P, ANY_ARC, arcout);
    }

    // Recurse along every single outgoing arc

    arc = n->FirstArc( P,ANY_ARC, arcout);
    while ( arc ) {
        r += DumpNode( arc->GetDst(),visited );
        arc = n->NextArc( P,ANY_ARC, arcout);
    }
    return r;
}

static int DumpGraph( GraphNode* n )
{
    // Try the below as an avl tree sometime, for comparison
    NodeSet visited;
    return DumpNode( n, visited );
}


static char namebuff[SRCSIZE];

static int DumpArc( GraphArc *arc )
{
    // arcs are written out as "type srcid!arcid dstid!dstname"

    if (!arc) {
        mayday("NULL arc in DumpArc()");
        return 0;
    }
    if (!rsf_fp) mayday("null file");
    GraphNode *src = arc->GetSrc(); GraphNode *dst = arc->GetDst();

    char *arcname = RinitGetDomain()->ArcNameFromType( arc->GetType() );

    if (arcname) fprintf( rsf_fp, "a!%s\t", arcname );
        else fprintf( rsf_fp, "a!type-%d\t",arc->GetType() );

    fprintf( rsf_fp, "%d!%d\t", src->GetId(), arc->GetId() );

    fprintf( rsf_fp, "\"%d!%s\"\n", dst->GetId(), dst->GetLabel());

    AvlPos *P;
    for ( char *attr = arc->FirstAttr(P); attr; attr = arc->NextAttr(P) ) {
        int i;
        // only write out those arc attributes that have nothing to do
        // with composite arc contributors (Yuck)...
        if (strcmp(attr, "NumContribs") && sscanf(attr, "C%d", &i) != 1)
                DumpArcAttr( arc, attr );
    }
    return 1;
}

static int DumpNodeAttr( GraphNode *n, char *attr )
{
    char *value;
    if (!strcmp(attr,"type"))
        value = RinitGetDomain()->NodeNameFromType( n->GetType() );
    else value = n->GetAttr( attr );

    if (!value) return 0;
    fprintf( rsf_fp, "n!%s\t", attr );
    fprintf( rsf_fp, "\"%d!%s\"\t", n->GetId(), n->GetLabel() );

    if ( value && *value=='\"' ) fprintf( rsf_fp, "%s\n", value );
        else fprintf( rsf_fp, "\"%s\"\n", value );
    return 1;
}

static int DumpArcAttr( GraphArc *arc, char *attr )
{
    char *value;
    value = arc->GetAttr( attr );

    if (!value) return 0;

    fprintf( rsf_fp, "a!%s\t%d\t", attr, arc->GetId());

    if ( value && *value=='\"' ) fprintf( rsf_fp, "%s\n", value );
        else fprintf( rsf_fp, "\"%s\"\n", value );
    return 1;
}

int DbLoadRSF()
{
    CharType dbfile;

    // Get the directory in which to the databases are stored
    char *s = EnvGet("DBDIR");
    if (!s) dbfile.Assign("."); else dbfile.Assign( s );

    // Tack on the file that the rsf stream is in
    dbfile.Append( "/rsf" );

    // Load the RSF file
    Mess( "DbLoadRSF: file = %s", dbfile.string() );
    int r = LoadRSF( dbfile.string(), RigiBaseWindow() );
    return r;
}

int DbSaveRSF()
{
    CharType dbfile;
    
    // Get the directory in which to the databases are stored
    char *s = EnvGet("DBDIR");
    if (!s) dbfile.Assign("."); else dbfile.Assign( s );

    // Tack on the file that the rsf stream is in
    dbfile.Append( "/rsf" );

    // Save the RSF file
    Mess( "DbSaveRSF: file = %s", dbfile.string() );
    int r = SaveRSF( dbfile.string(), RigiBaseWindow() );
    return r;
}

