//
// $Author: jmcdanie $
// $Date: 1996-04-23 22:30:54 $
// $Source: /root/cvs/Rigi/rigiedit/selectops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: selectops.C,v 1.43 1996-04-23 22:30:54 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"         // Basic graph definitions
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Class definition for graph Nodes
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"       // Class definition for RigiWindows
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"         // Class definition for RigiNodes
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"           // Class definition for RigiDomain
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"       // Window menu state access functions
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef EDIT_H
#include "edit.h"
#endif
#ifndef CONNECT_H
#include "connect.h"
#endif
#ifndef SELECTOPS_H
#include "selectops.h"
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"                // Basic control functions for the editor
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"           // Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"            // Utility control functions for the editor
#endif
#ifndef RIGIERR_H
#include "rigierr.h"             // Error reporting functions
#endif
#ifndef WINUTIL_H
#include "win_util.h"            // for WindowSetCursor()
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"           // for RinitGetDomain()
#endif

// Operations include files
#ifndef CLIENT_H
#include "client.h"              // Client/Supplier operations
#endif
#ifndef SELECTDST_H
#include "selectdst.h"           // Neighbour selection operations
#endif
#ifndef GROUP_H
#include "group.h"               // Selection grouping operations
#endif
#ifndef SELECTTREE_H
#include "selecttree.h"          // Tree selection operation.
#endif

// This function can be called for any of the selection items which are
// present in the Selection Submenu of the background menu in the motif
// version.  All of these operation selection nodes.  This function is
// called by the notify procedure attached to the Selection submenu,
// and it can now be called by RCL (Rigi Command Language).

// SCRIPTRETURN, value returned depends on the operation invoked.

int SelectOp( int op, int select_keep, char *id_or_type ) 
{
    int use_window;
    static AvlPos *SP;

    // Get some initial state information for the windows.
    RigiWindow* activewin = RutilActiveWindow();
    if (activewin==NULL) return 0;

    WindowSetCursor( CURSOR_WAIT ); 
    int arctype = editor->GetArcType();

    // Careful with this list. It is possible that an operation can result in
    // the deletion of RigiNodes. This list contains pointers to RigiNodes, so
    // the list may not point to valid objects after some operations. The ops
    // that you need to worry about are the edit operations, as they update
    // the windows. Most other ops just change the state of the nodes.

    RigiNodeList slist;
    // DO NOT get the slist, for the following operations, it messes
    // up the NextSelectedNode pointer! (PS)
    if ( op != FIRSTSELECTEDNODE )
    if ( op != NEXTSELECTEDNODE )
    if ( op != NUMSELECTEDNODES )
    if ( op != SELECTEDARC ) 
    if ( op != SELECTID ) 
    if ( op != DESELECTID ) 
    if ( op != SELECTTYPE ) 
    {
        editor->GetSelectionList( slist, activewin );
    }

    // See if should deselect the previous selection
    if ( op != SELECTALL )
    if ( op != SELECTINVERT )
    if ( op != DESELECT )
    if ( op != FIRSTSELECTEDNODE )
    if ( op != NEXTSELECTEDNODE )
    if ( op != NUMSELECTEDNODES )
    if ( op != SELECTEDARC )
        if ( !select_keep )
        {
            activewin->DrawSynch( RigiTrue );
            activewin->BunchNode( RigiTrue );
            editor->EmptySelection();
            activewin->BunchNode( RigiFalse );
            activewin->DrawSynch( RigiFalse );
        }

    switch ( op )
    {
        case SELECTALL:
            activewin->DrawSynch( RigiTrue );
            activewin->BunchNode( RigiTrue );
            EditSelectAll() ; 
            activewin->BunchNode( RigiFalse );
            activewin->DrawSynch( RigiFalse );
            break ;
        case SELECTINVERT:
            activewin->DrawSynch( RigiTrue );
            activewin->BunchNode( RigiTrue );
            editor->InvertSelection() ; 
            activewin->BunchNode( RigiFalse );
            activewin->DrawSynch( RigiFalse );
            break ;
        case SELECTDST:
            SelectDst( slist, activewin, arctype ); 
            break ;
        case SELECTSRC:
            SelectSrc( slist, activewin, arctype ); 
            break ;
        case SELECTREC:
            SelectRecursive( activewin, arctype ); 
            break ;
        case SELECTTREE:
            SelectTree( slist, activewin, outArcs, (ArcType)arctype );
            break;
        case SELECTRTREE:
            SelectTree( slist, activewin, inArcs, (ArcType)arctype );
            break;
        case STRONGCONNECT:
            ConnectionSelect( activewin, STRONG, editor->GetHighThreshold() ); 
            break ;
        case STRONGCOMPONENT:
            ComponentSelect( activewin, slist, editor->GetHighThreshold() ); 
            break ;
        case WEAKCONNECT:
            ConnectionSelect( activewin, WEAK, editor->GetLowThreshold() ); 
            break ;
        case ALLCOMMONCS:
            if (arctype == editor->CompositeArc())
            {
                RigiError(ERRORFLAG, "Can't perform client/supplier on composite Arcs");
                break;
            }
            if ( id_or_type == NULL ) use_window = 1;
            else use_window = atoi(id_or_type);
            CommonCSSelect( activewin,
                editor->GetHighThreshold(), editor->GetLowThreshold(),
                editor->GetHighThreshold(), editor->GetLowThreshold(),
                arctype, use_window );
            break ;
        case COMMONCLIENT:
            if ( id_or_type == NULL ) use_window = 1;
            else use_window = atoi(id_or_type);
            CommonCSComponent( activewin, slist,
                editor->GetHighThreshold(), editor->GetLowThreshold(),
                editor->GetHighThreshold(), editor->GetLowThreshold(),
                arctype, RIGI_CLIENT, use_window ); 
            break ;
        case COMMONSUPPLIER:
            if ( id_or_type == NULL ) use_window = 1;
            else use_window = atoi(id_or_type);
            CommonCSComponent( activewin, slist,
                editor->GetHighThreshold(), editor->GetLowThreshold(),
                editor->GetHighThreshold(), editor->GetLowThreshold(),
                arctype, RIGI_SUPPLIER, use_window ) ; 
            break ;
        case DESELECT:
            activewin->DrawSynch( RigiTrue );
            activewin->BunchNode( RigiTrue );
            editor->EmptySelection();
            activewin->BunchNode( RigiFalse );
            activewin->DrawSynch( RigiFalse );
            break;
        case FIRSTSELECTEDNODE:
        {
            WindowSetCursor( CURSOR_NORMAL );
            GraphNode* n = editor->FirstSelectedNode(SP);
            if ( n == NULL ) return 0;
            return n->GetId();
        }
        case NEXTSELECTEDNODE:
        {
            WindowSetCursor( CURSOR_NORMAL );
            GraphNode* n = editor->NextSelectedNode(SP);
            if ( n == NULL ) return 0;
            return n->GetId();
        }
        case NUMSELECTEDNODES:
            WindowSetCursor( CURSOR_NORMAL );
            return editor->NumSelectedNodes();
        case SELECTEDARC:
        {
            WindowSetCursor( CURSOR_NORMAL );
            RigiArc* ra = editor->GetSelectedArc();
            if ( ra == NULL ) return 0;
            return ra->GetArc()->GetId();
        }
        case SELECTID:
        case DESELECTID:
        { 
            GraphNode *node = editor->FindNode( atoi(id_or_type) );
            if (node) {
                if (op==SELECTID) editor->SelectNode( node );
                else              editor->DeselectNode( node );
            } else {
                GraphArc *arc = editor->FindArc( atoi(id_or_type) );
                if (arc) {
                    if (op==SELECTID) editor->SelectArc( arc );
                    else              editor->DeselectArc();
                }
            }
            break;
        }
        case SELECTTYPE:
        {
            int type = RinitGetDomain()->NodeTypeFromName( id_or_type );

            if (type == -1)
                RigiError(ERRORFLAG, "Invalid node type %s",id_or_type );
            else
                SelectType( activewin, type );
            break;
        }
        default:
            RigiError(ERRORFLAG, "Unrecognized select operation %d", op);
            break;
    }
    WindowSetCursor( CURSOR_NORMAL );

    // make a displayed help message indicating number of selected nodes
    char mess[32];
    int i = editor->NumSelectedNodes();
    sprintf( mess, "%d node%s selected",i,i==1?"":"s");
    activewin->Window()->SetMessage( mess );
    // SCRIPTRETURN
    return  i;
}

