//
//

// $Author: jmcdanie $
// $Date: 1996-04-23 22:30:54 $
// $Source: /root/cvs/Rigi/rigiedit/selecttree.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: selecttree.C,v 1.14 1996-04-23 22:30:54 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#ifndef SELECTTREE_H
#include "selecttree.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef HASHNODELIST_H
#include "cl_hashnodelist.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif

#ifndef RIGIERR_H
#include "rigierr.h"
#endif

static HashNodeList vnodelist;   // Hash node list for visited nodes
static HashNodeList wnodelist;   // Hash node list for nodes in the window

// queue_node is called in the tree traversal, once for each node in the
// traversal. It checks to see if this node has been visited already. If
// it has not been visited, it adds it to the visited list, and returns
// TRUE. The retur nvalue of the function tells the traversal algorithm
// whether or not this node should be processed again in the traversal. If
// we have already visited the node once, then we do not need to again. 

static int queue_node(GraphNode* n)
{
    GraphNode* np = vnodelist.Find(n->GetId());
    if (np == NULL) vnodelist.Add(*n);
    return (int)(np == NULL);
}

// Select node is also called once for each node. It performs the operation
// specified at every node, if the node is in the window we care about
// (wnodelist) and if we can find the RigiNode in that window. It returns
// nothing to the traversal algorithm.

static void select_node(GraphNode *n,RigiWindow* win)
{
    if (wnodelist.Find(n->GetId()) == NULL) return;
    RigiNode* rnp = win->FindRigiNode(n);
    if (rnp == NULL) return;
    editor->SelectNode(rnp);
}

void SelectTree(RigiNodeList& slist, RigiWindow* win,
                AdjList type, ArcType subtype)
{
    // Empty the window and the visited node lists.

    wnodelist.Empty();
    vnodelist.Empty();

    // Get the list of Nodes we are interested in, that is, only those that
    // are in the window.

    AvlPos *P;
    RigiNode* winnode = win->FirstNode(P);
    while(winnode != NULL)
    {
        wnodelist.Add( *(winnode->GetNode()) );
        winnode = win->NextNode(P);
    }

    // For each node in the list passed in, perform the breadth first search,
    // marking each node that we reach as selected through the iterator
    // function select_node(). The iterator function queue_node tells the
    // iterator whether or not it should queue the node for further processing
    // in the traversal. This function returns true only if the node has not
    // been processed yet.

    RigiNode* rnp = slist.First();
    while(rnp != NULL)
    {
        GraphNode *np = rnp->GetNode();
        if (np == NULL)
            RigiError(ERRORFLAG, "NULL node in SelectTree.");
        else
            np->TreeBFS(type,subtype,(VarArgFunRef)(&(select_node)),win,
                NULL,NULL,(IntFunction)(&(queue_node)),NULL);
        rnp = slist.Next();
    }
}

