// $Author: jmartin $
// $Date: 2000-02-05 02:18:34 $
// $Source: /root/cvs/Rigi/rigiedit/tk_colormap.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#include <tk.h>
#include "tk_colormap.h"
#include "cl_domain.h"
#include "rigi_init.h"
#include "rcl_commands.h"
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef WIN_UTIL_H
#include "win_util.h"
#endif

RigiColorMap::RigiColorMap() {
    colors = NULL;
    pixels = NULL;
    ncolors = 0;
}

RigiColorMap::~RigiColorMap() {
    register int i;

    if (ncolors > 0) {
        if (colors) {
            for(i=0 ; i<ncolors ; i++) {
                if (colors[i].info) Tk_FreeColor((XColor*)(colors[i].info));
            }
            delete[] colors;
        }
        if (pixels) delete[] pixels;
    }
}

/////// Init Procedure ////////

void RigiColorMap::InitColorMap() {
    register int i;
    RigiColor domcol;

    RigiDomain *domain = RinitGetDomain();
    if (colors) {
        for(i=0 ; i<ncolors ; i++) {
            if (colors[i].info) Tk_FreeColor((XColor*)(colors[i].info));
        }
        delete[] colors;
    }
    if (pixels) delete[] pixels;

    int nnodes = domain->NumNodeTypes();
    int narcs = domain->NumArcTypes();
    ncolors = nnodes + narcs + 1; 
    colors = new RigiColor[ncolors];
    pixels = new unsigned long[ncolors];

    // Compute the string colors based on the canvas color.

    WindowForegroundColor(&fg);
    char *colorName;
    colorName = strdup((!fg.red && !fg.green && !fg.blue)? "#ffffff" : "#000000");
    XColor *xColor;
    xColor = Tk_GetColor(rcl_interp, Tk_MainWindow(rcl_interp), colorName);
    free(colorName);
    fg.red = xColor->red; fg.green = xColor->green; fg.blue = xColor->blue;
    fg.pixel = xColor->pixel;
    
    // Initialize the color map stuff

    for ( i=0; i<ncolors; i++ ) {
        colors[i].red = colors[i].green = colors[i].blue = 0;
        colors[i].pixel = pixels[i] = 0;
        colors[i].info = NULL;
    }

    // Set the string color to be the foreground color

    SetStringColor(0,&fg);

    // Set up the node and arc colors

    for ( i = 0 ; i < nnodes ; i++ ) {
        domain->GetNodeColor( i, &domcol );
        SetNodeColor( i, &domcol );
    }
    for ( i = 0 ; i < narcs ; i++ ) {
        domain->GetArcColor( i, &domcol );
        SetArcColor( i, &domcol );
    }
}

//////// Get Procedures ////////

int RigiColorMap::GetColorIndex( const RigiColor& color ) {
    return color.pixel;
}

RigiColor *RigiColorMap::GetNodeColor( int i ) {
    return colors+i;
}

RigiColor *RigiColorMap::GetArcColor( int i ) {
    return colors+i+RinitGetDomain()->NumNodeTypes();
}

RigiColor *RigiColorMap::GetStringColor( int i ) {
    RigiDomain *domain = RinitGetDomain();
    int j = i+domain->NumNodeTypes()+domain->NumArcTypes();
    return colors+j;
}

//////// Set Procedures ////////


void RigiColorMap::SetColor(int i, RigiColor *col) {
    XColor xcol, *rcol;

    // Check for out of bounds indexing

    if ( i>=ncolors || i<0) { 
        RigiError(ALERT, "Color array overflow");
        return;
    }

    // Free up the Tk color if it has been allocated

    if (colors[i].info) Tk_FreeColor((XColor*)(colors[i].info));

    // Set up the rgb triple and retrieve the color from Tk
    xcol.red = col->red;
    xcol.green = col->green;
    xcol.blue = col->blue;
    rcol = Tk_GetColorByValue( Tk_MainWindow(rcl_interp), &xcol);

    // Copy the color into our Color structure

    colors[i].red = rcol->red;
    colors[i].green = rcol->green;
    colors[i].blue = rcol->blue;
    colors[i].pixel = rcol->pixel;
    colors[i].info = (void*)rcol;
    pixels[i] = rcol->pixel;
}

void RigiColorMap::SetNodeColor( int i, RigiColor *col ) {
    SetColor(i, col);
}

void RigiColorMap::SetArcColor( int i, RigiColor *col ) {
    i += RinitGetDomain()->NumNodeTypes();
    if (i >= ncolors) {
        RigiError(ALERT, "Color array overflow");
        return;
    }
    SetColor(i, col);
}

void RigiColorMap::SetStringColor( int i, RigiColor *col ) {
    RigiDomain *domain = RinitGetDomain();
    i += domain->NumNodeTypes() + domain->NumArcTypes();
    if ( i >= ncolors ) {
        RigiError(ALERT, "Color array overflow");
        return;
    }
    SetColor(i, col);
}

/////// Reset Procedures ///////

void RigiColorMap::ResetNodeColor( int i, RigiColor *col ) {
    SetNodeColor(i, col);
}

void RigiColorMap::ResetArcColor( int i, RigiColor *col ) {
    SetArcColor(i, col);
}

void RigiColorMap::ResetStringColor( int i, RigiColor *col ) {
    SetStringColor(i, col);
}

/*------------------------------------------------------------------*/

RigiBool RigiColorMap::operator==(const RigiColorMap& map) const {
    return this->isEqual( (void*)(&map) );
}

void RigiColorMap::operator=(const RigiColorMap&) {
}

RigiBool RigiColorMap::isEqual(void* d) const {
    if (d == NULL) return RigiFalse;
    return *this==*(RigiColorMap*)d;
}

