
const static char rcsid[] = "$Id: tk_stub.C,v 1.19 2000-02-09 00:18:36 jmartin Exp $";

//------------------------------------------------------------------------

#include <stdio.h>         // for FILE etc.
#include <sys/stat.h>      // for checking files to be opened
#include <tk.h>
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif
#ifndef ENVIRON_H
#include "environ.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"  // Global instance of editor's state
#endif
#include "cl_graphnode.h"
#include "cl_rigiwindow.h"
#include "cl_base_window.h"

#include "tk_textsw.h"
#include "rigierr.h"
#include "global.h"
#include "win_util.h"

// Line thickness
int SMALL     = 1;
int BIG       = 3;
int THICK     = 5;

// Cursor types
int CURSOR_NORMAL = 1;
int CURSOR_WAIT   = 2;


// reads the CANVASCOLOR config parameter and gets the closest
// X color.
RigiBool WindowForegroundColor(RigiColor *col) {
    static XColor *xColor = NULL;

    if (xColor == NULL) {
        char *colorName = EnvGet("CANVASCOLOR");
        if (colorName != NULL && *colorName != 0) {
            xColor = Tk_GetColor(rcl_interp, Tk_MainWindow(rcl_interp), colorName);
            if (xColor == NULL)
                RigiError(ALERT, "Unable to get XColor for canvas color %s", colorName);
        } else {
            xColor = Tk_GetColor(rcl_interp, Tk_MainWindow(rcl_interp), "white");
            if (xColor == NULL || xColor->pixel)
                xColor = Tk_GetColor(rcl_interp, Tk_MainWindow(rcl_interp), "black");
        }
    }
    if (xColor == NULL) {
        memset(col, 0, sizeof(RigiColor));
        return RigiError(ERRORFLAG, "Unable to allocate canvas color");
    }
    col->red = xColor->red; col->green = xColor->green; col->blue = xColor->blue;
    col->pixel = xColor->pixel;
    col->info = (void *)xColor;
    return RigiTrue;
}

// Gets the closest X color to the native background color
// (pixel 0) or "black".
RigiBool WindowBackgroundColor(RigiColor *col) {
    static XColor *xColor = NULL;

    if (xColor == NULL) {
        char *colorName = "black";
        xColor = Tk_GetColor(rcl_interp, Tk_MainWindow(rcl_interp), colorName);
        if (xColor == NULL)
            RigiError(ALERT, "Unable to get XColor for background color %s", colorName);
    }
    if (xColor == NULL) {
        memset(col, 0, sizeof(RigiColor));
        return RigiFalse;
    }
    col->red = xColor->red; col->green = xColor->green; col->blue = xColor->blue;
    col->pixel = xColor->pixel;
    col->info = (void *)xColor;
    return RigiTrue;
}

void WindowSetCursor( int type ) {
    static int cursorType = CURSOR_NORMAL;
    if (type != cursorType) {
        RigiWindow *rw = editor->FirstWindow();
#ifdef __WINDOWS__
        if (rw != NULL)
#else
        for ( ; rw != NULL; rw = editor->NextWindow())
            // with motif, each window has a separate cursor
#endif
            if (type == CURSOR_NORMAL ) rw->Window()->NormalCursor();
            else if (type == CURSOR_WAIT) rw->Window()->WaitCursor();
        cursorType = type;

    }
}

int WindowIsShifted() {
    RigiWindow *rw = editor->FirstWindow();
    if (rw == NULL) return 0;
    return rw->Window()->IsShifted();
}

void WindowSetNormalCursor( RigiWindow* rwp ) {
    rwp->Window()->NormalCursor();
}

int WindowTextEdit(GraphNode* np) {
    warn("Node editor not supported");
    return 0;
}


void create_arcmenu() { }

void create_nodemenu() { }

void WindowAppMainLoop() { }




