
// $Author: jmcdanie $
// $Date: 1996-05-20 18:20:57 $
// $Source: /root/cvs/Rigi/rigiedit/win_archierarchy.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: win_archierarchy.C,v 1.35 1996-05-20 18:20:57 jmcdanie Exp $";

//------------------------------------------------------------------------

#include <stream.h>
#include <math.h>
#include <limits.h>
#include "win_archierarchy.h"
#include "global.h"

#include "cl_list.h"
#include "cl_graphnodeset.h"
#include "cstuff.h"
#include "grapherr.h"

#include "cl_graphnode.h"
#include "cl_grapharc.h"

#include "cl_riginode.h"
#include "cl_rigiwindow.h"

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

static const int MAXLEVELS = 200;    // Maximum number of levels in hierarchy
static const int INVLOC = -1;        // Constant for not finding node at level

static int winwidth;   // Window width/height that remain constant
static int winheight;  // until hierarchy is redrawn

// B19 - adding this to allow user to specify depth for forward/backward trees
static int leveldepth;

static int level;                           // Current level
static int levelsize[MAXLEVELS];            // List of level sizes
static ItemList levellist[MAXLEVELS];       // List of nodes at each level
static ItemList levellist2[MAXLEVELS];      // List of nodes at each level

static NodeSet nodeset;                     // Node set for all nodes
static NodeSet winnodeset;                  // Node set for window nodes

static RigiBool use_windowlist;             // Flag to determine windowlist use
static int num_children;                    // # of children remaining at level
static int sum_children;                    // # of children for the next level
static int max_children;                    // maximum size the levels reached
static AdjList type;                        // type of arc to process
static ArcType subtype;                     // subtype of arc to process

// Given an integer index x,y calculate the window location of the point
// based on the width and height of the window, the number of levels, and
// the size of the window border.

static void getlocation( int x, int y, int winw, int winh,
                int border, int& xloc, int& yloc )
{
    int xstep = (int)(winw/(levelsize[y]+1));
    int ystep = winh/(level+1);
    xloc = (x+1)*xstep + border;
    yloc = y*ystep + border;
}

// Find the index of a given node in a given level. If it does not exist
// return INVLOC

static int GetParentIndex( GraphNode* srcnode,int lev )
{
    // B19 - copied from xview version for this change
    if (lev > level) return INVLOC;
    int id = srcnode->GetId();
    int count = 0;

    GraphNode *np = (GraphNode*)levellist2[lev].First();
    while( np != NULL )
    {
        if ( np->GetId() == id ) return count;
        count++;
        np = (GraphNode*)levellist2[lev].Next();
    }
    return INVLOC;
}

// Add the node n to the nodeset.

static int queue_node( GraphNode* n )
{
    if ( level >= leveldepth-1) return 0;

    GraphNode* np = nodeset.Find( n->GetId() );
    if ( !np ) nodeset.Add( n );
    return !np;
}

// A hack to get the number of arcs of a given type and subtype. I think this
// is a method for nodes now, but I don't feel like changing it right now. Opps
// can't use the method cause we only want to count nodes that are in the
// node set.

static int get_num_arcs( GraphNode* n, AdjList adj, ArcType type )
{
    int count = 0; GraphArc* ap; AvlPos *P;
    for ( ap = n->FirstAdj( P, adj ); ap; ap = n->NextAdj( P, adj ) )
    {
        GraphNode* othernode = ap->GetOtherNode(n);
        GraphNode* np = nodeset.Find(othernode->GetId());

        RigiBool window_check = use_windowlist ?
            winnodeset.Find(othernode->GetId())!=NULL : RigiTrue;

        if (( ap->GetType()==type || ANY_ARC==type ) &&
            ( !nodeset.Find( othernode->GetId() )) &&
            ( window_check )) count++;
    }
    return count;
}

// To do this level thing, I do some fancy counting. This seems more difficult
// than it should be, but I couldn't think of a better way to do it at the
// time. We have three counting variables. num_children is the number of
// nodes at the current level. sum_children is the number of nodes at the
// next level down. max_children is the maximum number of children at any
// level so far. count_children() compiles this info and keeps that info
// in a list called levellist and keeps the size of each level in a list
// called levelsize. The levels are limited to MAXLEVELS, and things will
// probably die if the limit is exceeded.
//
// How does it work. For each node in the BFS iteration it decreases
// num_children.  If num children is greater than 0 then we add the node
// to this level, and add the number of children it has to the total of
// sum_children, which is keeping track of the number of children on the
// next level. If num_children is less than 0 we have finished a level
// and therefore set max_children if necessary, and set up for processing
// the next level. We do this by increasing the level and adding this node
// to the new level.

static void count_children( GraphNode* n )
{
    // If over the limit just return.
    if ( level >= leveldepth ) return;

    // If using the window list and we can't find the node in the list then
    // we don't want to process this node.
    if ( use_windowlist && !winnodeset.Find(n->GetId()) )
        return;

    // Decrease the number of children at this level.

    num_children--;

    // If we have finished this level, to some stuff.

    if ( num_children < 0 )
    {

        // Increase max_children if necessary (largest level size)

        if ( max_children < sum_children )
            max_children = sum_children;

        // Set the number of children in the next level to be the sum of all of
        // the children of the nodes at this level.

        num_children = sum_children;

        // Increase the level and set up the lists and size.

        // B19 - changed MAXLEVELS to leveldepth
        if (level >= leveldepth)
        {
            cerr << "Too many levels\n";
            return;
        }
        level++;
        levellist[level].Empty();
        levellist2[level].Empty();
        levelsize[level] = 0;

        // Process the first node by getting the number of children, decreasing
        // the number of nodes at this level, and adding the node to the list.

        sum_children = get_num_arcs( n,type,subtype );
        num_children--;

        if ( GetParentIndex( n,level ) == INVLOC )
        {
            levellist[level].Insert( n );
            levellist2[level].Insert( n );
            levelsize[level]++;
        }
    }

    // Here we process a node without changing levels.

    else
    {
        // Add to the number of arcs at the next level down.

        sum_children += get_num_arcs( n,type,subtype );

        // Add the node to the list and keep track of this levels size.

        if ( GetParentIndex( n,level ) == INVLOC )
        {
            levellist[level].Insert( n );
            levellist2[level].Insert( n );
            levelsize[level]++;
        }
    }
}

// Draw an arc hierarchy in the given window.

void WindowArcHierarchy(ArcType stype, ArcDirection dir, RigiWindow* rwp,
                        RigiNode* rnp, int xoffset,int yoffset, int depth,
                        double angle)
{
    register int i,j;
    GraphNode* np;
    int xloc,yloc;
    double sina, cosa;

    // Compute the sin and cos of the angle

    sina = sin(angle * M_PI / 180.0 );
    cosa = cos(angle * M_PI / 180.0 );

    // B19 - set the level depth
    leveldepth = depth + 1;
    if ( ( leveldepth <= 1 ) || (leveldepth >= MAXLEVELS))
    {
        Mess("WindowArcHierarchy - Resetting depth from %d to %d",
            depth, MAXLEVELS-1);
        leveldepth = MAXLEVELS-1; 
    }

    // Get the type and subtype of the arcs.

    if ( dir == ARC_IN )
        type = (stype == editor->LevelArc()) ? inLevArcs : inArcs;
    else if ( dir == ARC_OUT )
        type = (stype == editor->LevelArc()) ? outLevArcs : outArcs;

    subtype = stype;

    // Get the node to perform the arc hierarchy for.

    if ( rnp == NULL ) cerr << "NULL RigiNode in WindowArcHierarchy\n";
    np = rnp->GetNode();

    // Initialize the stuff before the traversal.

    num_children = 1;
    max_children = 1;
    sum_children = 0;
    level = 0;
    levellist[level].Empty();
    levellist2[level].Empty();
    levelsize[level] = 0;

    nodeset.Empty();
    winnodeset.Empty();
    use_windowlist = RigiFalse;

    // Get the list of Nodes we are interested in

    AvlPos *P;
    for ( RigiNode* rn = rwp->FirstNode(P); rn; rn = rwp->NextNode(P) )
    {
        use_windowlist = RigiTrue;
        winnodeset.Add( rn->GetNode() );
    }

    // Do the traversal. This puts the nodes in the proper levels lists
    // for the hierarchy.

    np->TreeBFS( stype, dir, (VarArgFunRef)(&(count_children)), NULL,
                NULL, NULL, (IntFunction)(&(queue_node)),NULL );

    // Set up for drawing the hierarchy. Get the width and height of the window
    // (in this case, it is the size of the tree with minimal spacing).

    winheight = (int)((level+1)*4*rnp->NodeSize()*rwp->GetYScale());

    winwidth = (int)(max_children*2*rnp->NodeSize()*rwp->GetXScale());

    // For each level, move the node to its appropriate position

    for( i=0; i<=level; i++ )
    {
        np = (GraphNode*)levellist[i].First();

        j = 0;
        while( np != NULL )
        {

            // Draw the node at its location

            getlocation( j,i,winwidth,winheight,0,xloc,yloc );
            Point pt;
            if (type == inArcs || type == inLevArcs)
            {
                pt.x = xoffset + xloc - winwidth/2;
                pt.y = yoffset - yloc;
                pt.z = 0;
            }
            if (type == outArcs || type == outLevArcs)
            {
                pt.x = xloc + xoffset - winwidth/2;
                pt.y = yloc + yoffset;
                pt.z = 0;
            }
            
            // Perform the transformation based on the angle given

            double tx, ty;
            tx = (double)(pt.x - xoffset);
            ty = (double)(pt.y - xoffset);
            pt.x = xoffset + (int)(tx * cosa - ty * sina);
            pt.y = xoffset + (int)(ty * cosa + tx * sina);

            // Transform to NDC coordinates and set the position

            pt = rwp->Canvas2World(pt);
            NDC_Point2d ndcp = rwp->World2NDC2d( pt );
            np->SetNDCPosition( ndcp );

            RigiNode* rnp = rwp->FindRigiNode(np);
            Point2d p; p.x = pt.x; p.y = pt.y;
            if (rnp == NULL)
                mayday("WindowArcHierarchy: Couldn't find RigiNode");
            else
                rnp->SetPosition(p);
            j++;
            np = (GraphNode*)levellist[i].Next();
        }
    }
}

