
// $Author: jmcdanie $
// $Date: 1996-05-20 18:20:57 $
// $Source: /root/cvs/Rigi/rigiedit/windowops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: windowops.C,v 1.52 1996-05-20 18:20:57 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef GLOBAL_H
#include "global.h"		// for editor and access functions
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"               // Global instance of editor's state
#endif

#ifndef WINDOWOPS_H
#include "windowops.h"		// Includes the definitions for this file
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"		// For RigiBaseWindow()
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"		// Utility control functions for the editor
#endif
#ifndef WINUTIL_H
#include "win_util.h"		// Window state access functions
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif

#ifndef GRAPHNODELIST_H
#include "cl_graphnodelist.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"	//
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"	//
#endif
#ifndef RIGIARC_H
#include "cl_rigiarc.h"		//
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"		//
#endif
#ifndef EVENT_H
#include "event.h"		// for EventDoneProc
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

// SCRIPTRETURN: returned value depends on the operation invoked.
// x==1 doubles as filters-off indicator for FIRST/NEXT/NUM NODE/ARC.
// Default is filters-on (visible stuff only) (x==0).

int WindowOp( int op, int winNum, int x, int y, char* str ) {
    RigiWindow *rw;

    if (winNum==0) {
        if (op == SELECTWINDOW) {
            rw = RigiBaseWindow();
        } else {
            rw = RutilActiveWindow();
        }
        if (rw == NULL) return 0;
    } else if ((rw = editor->GetWindowByNumber(winNum)) == NULL) {
        return RigiError(ERRORFLAG, "Window %d not found",winNum );
    }

    WindowSetCursor( CURSOR_WAIT );
    int r = rw->GetWindowNumber();

    // Create local copies of nlist and alist, so that the next pointer
    // cannot be screwed up by anyone else.

    static NodeList plist;
    if ( op == FIRSTPARENT ) {
        plist.Empty();
        GraphNode *n; AvlPos *P;
        for ( n=rw->FirstParent(P); n; n = rw->NextParent(P) )
            plist.Insert(n);
    }

    static RigiNodeList nlist;
    if ( op == FIRSTNODE ){
        nlist.Empty();
        RigiNode *rn; AvlPos *P;
        for ( rn=rw->FirstNode(P); rn; rn = rw->NextNode(P) )
            if ( x || rn->isVisible() ) nlist.Insert(rn);
    }

    static RigiArcList alist;
    if ( op == FIRSTARC ) {
        alist.Empty();
        RigiNode *rn; RigiArc *ra; AvlPos *P, *Q;
        for ( rn=rw->FirstNode(P); rn; rn = rw->NextNode(P) )
        for ( ra=rn->FirstArc(Q,ARC_OUT); ra; ra=rn->NextArc(Q,ARC_OUT) )
            if ( x || ra->isVisible() ) alist.Insert(ra);
    }

    switch (op) {
    case SELECTWINDOW:
        if (RutilActiveWindow()->GetWindowNumber() != winNum) {
            RutilSetActiveWindow(rw);
            RutilRaiseWindow(rw);
            if (!rw->BackStore()) rw->DrawStore();
        }
        break;
    case MOVEWINDOW:
        rw->Window()->SetLocation( x,y );
        break;
    case RESIZEWINDOW:
        rw->Window()->SetFrameSize( x,y );
        break;
    case CLOSEWINDOW:
        if (rw == RigiBaseWindow()) {
            r = RigiError(ERRORFLAG, "This window may not be closed");
            break;
        }
        if (RutilActiveWindow() == rw)
            RigiBaseWindow()->RigiDisplay();
        editor->CloseWindow(rw);
        break;
    case ICONIFYWINDOW:
        rw->Window()->SetIconified( 1 );
        break;
    case DEICONIFYWINDOW:
        rw->Window()->SetIconified( 0 );
        break;
    case GETWINDOWID:
        break;
    case SETMESSAGE:
        rw->Window()->SetMessage( str );
        break;
    case RESETMESSAGE:
        rw->Window()->ResetMessage();
        break;

        // MJW: the following are based on VISIBLE objects (ie nonfiltered)
        // in the rigiwindow of interest.

    case FIRSTNODE:
    {
        RigiNode* rn = nlist.First();
        if ( rn == NULL ) { r=0; break; }
        if ( rn->GetNode()==NULL ) {
            r = RigiError(ERRORFLAG, "Node does not exist");
            break;
        }
        r = rn->GetNode()->GetId();
        break;
    }
    case NEXTNODE:
    {
        RigiNode* rn = nlist.Next();
        if ( rn == NULL ) { r=0; break; }
        if ( rn->GetNode()==NULL ) {
            r = RigiError(ERRORFLAG, "Node does not exist");
            break;
        }
        r = rn->GetNode()->GetId();
        break;
    }
    case NUMNODES:
        r = x ? rw->NodeNumber() : rw->ShowedNodeNumber();
        break;

    case FIRSTARC:
    {
        RigiArc* ra = alist.First();
        if ( ra == NULL ) { r=0; break; };
        if ( ra->GetArc()==NULL ) {
            r = 0;
            RigiError(ERRORFLAG, "Arc does not exist");
            break;
        }
        r = ra->GetArc()->GetId();
        break;
    }
    case NEXTARC:
    {
        RigiArc* ra = alist.Next();
        if ( ra == NULL ) { r=0; break; };
        if ( ra->GetArc()==NULL ) {
            r = 0;
            RigiError(ERRORFLAG, "Arc does not exist");
            break;
        }
        r = ra->GetArc()->GetId();
        break;
    }
    case NUMARCS:
        r = x ? rw->ArcNumber() : rw->ShowedArcNumber();
        break;
    case FIRSTPARENT:
    {
        GraphNode* n = plist.First();
        if ( n == NULL ) { r=0; break; };
        r = n->GetId();
        break;
    }
    case NEXTPARENT:
    {
        GraphNode* n = plist.Next();
        if ( n == NULL ) { r=0; break; };
        r = n->GetId();
        break;
    }
    case NUMPARENTS:
        r = rw->NumParents();
        break;
    case FINDARC:
    {
        // x is srcid, y is dstid, str is type
        int type = RinitGetDomain()->ArcTypeFromName( str );
        GraphNode *src = editor->FindNode(x);
        if (!src) { r = 0; break; }
        GraphNode *dst = editor->FindNode(y);
        
        GraphArc *a = src->FindArc( dst,outArcs );
        RigiArc *ra = rw->FindRigiArc(a);
        r = (ra && ra->isVisible() && (type==ANY_ARC || type==a->GetType()))
                ? a->GetId() : 0;
        break;
    }

    case SETDRAWING:
        rw->Window()->Drawing( x );
        break;

    case GETCANVASWIDTH:
    case GETCANVASHEIGHT:
    case GETCANVASX:
    case GETCANVASY:
    {
        WindowSetCursor( CURSOR_NORMAL );
        int x,y, w,h;
        rw->Dimensions( &x,&y, &w,&h );
        if (op==GETCANVASWIDTH) return w;
        if (op==GETCANVASHEIGHT) return h;
        if (op==GETCANVASX) return x;
        if (op==GETCANVASY) return y;
        return 0;
    }

    case GETWINDOWX:
    case GETWINDOWY:
    {
        int x, y;
        rw->Window()->Location( x, y );
        if (op == GETWINDOWX) r = x;
        if (op == GETWINDOWY) r = y;
        break;
    }
    case SETWINDOWFONT:
        rw->Window()->SetFont( str );
        rw->RigiDisplay();
        break;

    default:
        RigiError(ERRORFLAG, "Invalid operation (%d)", op);
        break;
    }
    WindowSetCursor( CURSOR_NORMAL );
    return r;
}

